/**********************************************************************
 **                                                                  **
 **                            cbrplanner.cc                         **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This module instantiates DomainManager, MemoryManager, and      **
 **  Planner (sub-)modules within its class to server as a CBR       **
 **  Planner.                                                        **
 **                                                                  **
 **  Copyright 2003 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: cbrplanner.cc,v 1.1 2003/04/06 08:43:44 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <getopt.h>
#include <sys/types.h>

#include "cbrplanner.h"
#include "cbrplanner_domain.h"
#include "domain_manager.h"
#include "memory_manager.h"
#include "planner.h"
#include "windows.h"

//-----------------------------------------------------------------------
// Constatns
//-----------------------------------------------------------------------
const int CBRPlanner::MAINLOOP_USLEEPTIME_ = 10000; // 10 msec.

//-----------------------------------------------------------------------
// Constructor for CBRPlanner class.
//-----------------------------------------------------------------------
CBRPlanner::CBRPlanner(CBRPlanner **cbrplannerInstance) : self_(cbrplannerInstance)
{
    if (cbrplannerInstance != NULL)
    {
	*cbrplannerInstance = this;
    }

    domainManager_ = NULL;
    memoryManager_ = NULL;
    planner_ = NULL;
}

//-----------------------------------------------------------------------
// Distructor for CBRPlanner class.
//-----------------------------------------------------------------------
CBRPlanner::~CBRPlanner(void)
{
    if (planner_ != NULL)
    {
	delete planner_;
    }

    if (memoryManager_ != NULL)
    {
	delete memoryManager_;
    }

    if (domainManager_ != NULL)
    {
	delete domainManager_;
    }

    if (self_ != NULL)
    {
	*self_ = NULL;
    }
}

//---------------------------------------------------------------------
// This function makes the program to just loop until terminated by
// the user.
//---------------------------------------------------------------------
void CBRPlanner::mainLoop(void)
{
    while (true)
    {
	usleep(MAINLOOP_USLEEPTIME_);
    }
}

//---------------------------------------------------------------------
// This function starts all the subsystems within the CBR Planner.
//---------------------------------------------------------------------
void CBRPlanner::startSubsystems(void)
{
    new DomainManager(&domainManager_);
    new MemoryManager(&memoryManager_);
    new Planner(&planner_);
}

//-----------------------------------------------------------------------
// This function will save a new mission plan into its memory.
//-----------------------------------------------------------------------
vector<int> CBRPlanner::saveNewMissionPlan(
    CBRPlanner_ACDLMissionPlan_t acdlPlan,
    MemoryManager::Features_t features)
{
    MemoryManager::MissionPlan_t missionplan;
    vector<int> dataIndexList;
    bool hasStartSubMission;

    // Convert the ACDL represented mision plan into a format that
    // MemoryManager can understand.
    missionplan = domainManager_->acdl2missionplan(acdlPlan);
    hasStartSubMission = domainManager_->hasStartSubMision(acdlPlan);

    // Save the plan into the memory along with the feature.
    dataIndexList = planner_->saveNewMissionPlan(
	missionplan,
	features,
	hasStartSubMission);

    return dataIndexList;
}

//-----------------------------------------------------------------------
// This function updates the FSA list of domain manager.
//-----------------------------------------------------------------------
void CBRPlanner::updateFSANames(CBRPlanner_FSANames_t fsaNames)
{
    domainManager_->updateFSANames(fsaNames);
}

//-----------------------------------------------------------------------
// This function will get a mission plan from Planner, and convert it in
// ACDL.
//-----------------------------------------------------------------------
vector<CBRPlanner_ACDLMissionPlan_t> CBRPlanner::getACDLMissionPlans(
    MemoryManager::Features_t features,
    int maxRating)
{
    CBRPlanner_ACDLMissionPlan_t acdlPlan;
    vector<CBRPlanner_ACDLMissionPlan_t> acdlPlans;
    vector<MemoryManager::MissionPlan_t> missionplans;
    int i;

    // Get a mision plan from the Planner module, based on the feature.
    missionplans = planner_->getMissionPlans(features, maxRating);
    
    for (i = 0; i < ((int)(missionplans.size())); i++)
    {
	// Convert the mission plan into an ACDL mission plan.
	acdlPlan = domainManager_->missionplan2acdl(missionplans[i]);
	acdlPlans.push_back(acdlPlan);
    }

    return acdlPlans;
}

/**********************************************************************
 * $Log: cbrplanner.cc,v $
 * Revision 1.1  2003/04/06 08:43:44  endo
 * Initial revision
 *
 *
 **********************************************************************/
