/**********************************************************************
 **                                                                  **
 **                           communicator.cc                        **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This module serves as an interface to the CfgEdit (cbrclient),  **
 **  by instantiating a socket-based communication with it.          **
 **                                                                  **
 **  Copyright 2003 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: communicator.cc,v 1.1 2003/04/06 08:43:44 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <getopt.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <linux/un.h>
#include <string>

#include "communicator.h"
#include "cbrplanner_protocol.h"
#include "cbrplanner_domain.h"
#include "cbrplanner.h"
#include "domain_manager.h"
#include "memory_manager.h"
#include "windows.h"
#include "debugger.h"

//-----------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------
extern void gQuitProgram(int exitStatus); // Defined in main.cc

//-----------------------------------------------------------------------
// Constanns
//-----------------------------------------------------------------------
const int Communicator::SOCKET_BACKLOG_SIZE_ = 5;
const int Communicator::SELECT_WAIT_SEC_ = 10;
const int Communicator::MAX_SYNC_ERROR_COUNT_ = 5;

//-----------------------------------------------------------------------
// Mutex initialization
//-----------------------------------------------------------------------
pthread_mutex_t Communicator::privateConnectionInfoListMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Constructor for CBRPlanner class.
//-----------------------------------------------------------------------
Communicator::Communicator(Communicator **communicatorInstance) : self_(communicatorInstance)
{
    if (communicatorInstance != NULL)
    {
	*communicatorInstance = this;
    }

    socketName_ = "";
    socketFd_ = -1;
    daemonThreadIsUp_ = false;
}
//-----------------------------------------------------------------------
// Distructor for Communicator class.
//-----------------------------------------------------------------------
Communicator::~Communicator(void)
{
    // Close all the private connections.
    disconnectAllPrivateConnections_();

    // Make sure to kill the deamon thread.
    if (daemonThreadIsUp_)
    {
	pthread_cancel(daemonThread_);
	pthread_join(daemonThread_, NULL );
	daemonThreadIsUp_ = false;
    }

    if (self_ != NULL)
    {
	*self_ = NULL;
    }
}

//-----------------------------------------------------------------------
// This function starts up the daemon to initiate the communication with
// CfgEdit / mlab.
//-----------------------------------------------------------------------
void Communicator::startDaemon(void)
{
    char buf[UNIX_PATH_MAX];

    if (socketName_ == "")
    {
	// The name of the socket is not yet assigned. Create it using
	// the user name.
	sprintf(buf, "/tmp/%s-cbrplanner.socket", getenv("USER"));
	socketName_ = buf;
    }

    // Delete the old socket.
    unlink(socketName_.c_str());

    // Open the socket and bind the name to it.
    setupSocket_();

    // Start the communication deamon thread.
    pthread_create(&daemonThread_, NULL, &startDaemonThread_, (void *)this );
}

//-----------------------------------------------------------------------
// This function starts up a thread to run the communication daemon.
//-----------------------------------------------------------------------
void *Communicator::startDaemonThread_(void *communicatorInstance)
{
    ((Communicator *)communicatorInstance)->daemonMainLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function starts up the daemon to initiate the communication with
// CfgEdit / mlab.
//-----------------------------------------------------------------------
void Communicator::daemonMainLoop_(void)
{
    int count = 0;
    int fd = -1;

    daemonThreadIsUp_ = true;

    while(true)
    {
	// Make sure termination was not requested.
        pthread_testcancel();

	// Wait for a client connection at the socket.
	fd = waitForNewConnection_(count);

	// Make sure termination was not requested.
        pthread_testcancel();

	// Start another thread specific to this connection.
	startPrivateConnection_(fd);

	count++;
    }
}

//-----------------------------------------------------------------------
// This function opens a socket and bind the given name to it. It returns
// the file descriptor of the socket.
//-----------------------------------------------------------------------
void Communicator::setupSocket_(void)
{
    int status;
    sockaddr_un localAddress;

    // Open the socket.
    socketFd_ = socket(AF_UNIX, SOCK_STREAM, 0);

    if (socketFd_ < 0)
    {
	gWindows->printfCommWindow("Error: Socket could not be opened.\n");
	gQuitProgram(-1);
    }

    // Bind the name to a socket.
    localAddress.sun_family = AF_UNIX;
    strcpy(localAddress.sun_path, socketName_.c_str());
    status = bind(socketFd_, (sockaddr *)&localAddress, sizeof(localAddress));

    if (status < 0)
    {
	gWindows->printfCommWindow("Error: \"%s\" could not be binded to the socket.\n",
				  socketName_.c_str());
	gQuitProgram(-1);
    }
}

//-----------------------------------------------------------------------
// This function listens the socket for a client connection. After the
// communication is established and it returns a new file descriptor to
// talks to the cilent privately.
//-----------------------------------------------------------------------
int Communicator::waitForNewConnection_(int count)
{
    int newFd = -1, status = -1;
    sockaddr_un remoteAddress;
    socklen_t len;

    // Listen the socket for connection.
    status = listen(socketFd_, SOCKET_BACKLOG_SIZE_);

    if (status < 0)
    {
	gWindows->printCommWindow("Error: Listening socket failed.\n");
	gQuitProgram(-1);
    }

    gWindows->printfCommWindow("Waiting for %s client connection.\n",
			      (count > 0)? "another" : "the first");

    // Wait for the connection with CfgEdit
    newFd = accept(socketFd_, (sockaddr *)&remoteAddress, &len);

    if (newFd < 0)
    {
	gWindows->printCommWindow("Error: Connection to the client failed.\n");
	gQuitProgram(-1);
    }

    gWindows->clearCommWindow();
    gWindows->printCommWindow("Connection to the client established.\n");

    return newFd;
}

//-----------------------------------------------------------------------
// This function starts another thread to talk only to this connection.
//-----------------------------------------------------------------------
void Communicator::startPrivateConnection_(const int fd)
{
    PrivateConnectionThreadInput_t *data = NULL;
    PrivateConnectionInfo_t connectionInfo;
    pthread_t newThread;

    // First, add this fd to the connection info list.
    connectionInfo.fd = fd;
    connectionInfo.connectionThreadIsUp = false;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);
    shPrivateConnectionInfoList_.push_back(connectionInfo);
    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    // Set up the input for the thread.
    data = new PrivateConnectionThreadInput_t;
    data->fd = fd;
    data->communicatorInstance = this;

    // Start the new thread to talk to this connection.
    pthread_create(&newThread, NULL, &startPrivateConnectionThread_, (void *)data);

    saveConnectionThread_(fd, newThread);
}

//-----------------------------------------------------------------------
// This function starts up a thread to communicate with this connection
// privately.
//-----------------------------------------------------------------------
void *Communicator::startPrivateConnectionThread_(void *input)
{
    PrivateConnectionThreadInput_t *data = NULL;
    int fd;
    Communicator *communicatorInstance;

    // Unpack the data
    data = (PrivateConnectionThreadInput_t *)input;
    fd = data->fd;
    communicatorInstance = data->communicatorInstance;
    
    // Start the main loop for the connection
    communicatorInstance->privateConnectionMainLoop_(fd);

    return NULL;
}

//-----------------------------------------------------------------------
// This function loops around to read commands from the client.
//-----------------------------------------------------------------------
void Communicator::privateConnectionMainLoop_(const int fd)
{
    setConnectionThreadIsUp_(fd, true);
    
    while(true)
    {
	// Make sure termination was not requested.
        pthread_testcancel();

	// Read the arrived data, and process it based on the instruction.
	readAndProcessPrivateConnectionData_(fd);
    }
}

//-----------------------------------------------------------------------
// This function reads the data sent by the client, and process it
// according to the instruction.
//-----------------------------------------------------------------------
bool Communicator::readAndProcessPrivateConnectionData_(const int fd)
{
    unsigned char oneByteBuf[1];
    bool synced = false, received = false;
    CBRPlanner_ClientInfo_t clientInfo;
    CBRPlanner_ACDLMissionPlan_t acdlPlan;
    vector<CBRPlanner_ACDLMissionPlan_t> acdlPlans;
    vector<int> dataIndexList;
    MemoryManager::Features_t features;
    CBRPlanner_FSANames_t fsaNames;
    int maxRating, index;

    // Make sure that it is synced to the client.
    synced = syncToClient_(fd);

    if (!synced)
    {
	// Not synced. Something went wrong. Disconnect it.
	disconnectPrivateConnection_(fd);
	return false;
    }

    // Now it is synced. Get a command byte.
    received = readNBytes_(fd, (unsigned char *)&oneByteBuf, 1);

    switch (oneByteBuf[0]) {

    case CBRPLANNER_COMM_CMD_CLIENTINIT:
	clientInfo = readClientInfo_(fd);
	updateClientInfo_(fd, clientInfo);
	break;

    case CBRPLANNER_COMM_CMD_CLIENTEXIT:
	disconnectPrivateConnection_(fd);
	break;

    case CBRPLANNER_COMM_CMD_FSANAMES:
	fsaNames = readFSANames_(fd);
	gCBRPlanner->updateFSANames(fsaNames);
	break;

    case CBRPLANNER_COMM_CMD_MISSIONPLAN_SAVE:
	acdlPlan = readACDLMissionPlan_(fd);
	features = readFeatures_(fd);
	dataIndexList = gCBRPlanner->saveNewMissionPlan(
	    acdlPlan,
	    features);
	writeDataIndexList_(fd, dataIndexList);
	break;

    case CBRPLANNER_COMM_CMD_MISSIONPLAN_RETRIEVE:
	features = readFeatures_(fd);
	maxRating = readInteger_(fd);
	acdlPlans = gCBRPlanner->getACDLMissionPlans(features, maxRating);
 	writeACDLMissionPlans_(fd, acdlPlans);
	break;

    case CBRPLANNER_COMM_CMD_SELECTED_MISSION:
	index = readInteger_(fd);
	break;

    case CBRPLANNER_COMM_CMD_POSITIVE_FEEDBACK:
	dataIndexList = readDataIndexList_(fd);
	gCBRPlanner->givePositiveFeedback(dataIndexList);
	break;

    case CBRPLANNER_COMM_CMD_NEGATIVE_FEEDBACK:
	dataIndexList = readDataIndexList_(fd);
	gCBRPlanner->giveNegativeFeedback(dataIndexList);
	break;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function makes sure that the server and the client are syncronized.
// Two keys are used to check the syncronization.
//-----------------------------------------------------------------------
bool Communicator::syncToClient_(const int fd)
{
    int errorCount = 0;
    unsigned char oneByteBuf[1];
    bool synced = false, received = false;


    while (!synced)
    {
	received = readNBytes_(fd, (unsigned char *)&oneByteBuf, 1);

	if (received)
	{
	    if (oneByteBuf[0] == CBRPLANNER_COMM_SYNC_BYTE1)
	    {
		// The first sync byte is OK. Check the second one.
		received = readNBytes_(fd, (unsigned char *)&oneByteBuf, 1);

		if (received)
		{
		    if (oneByteBuf[0] == CBRPLANNER_COMM_SYNC_BYTE2)
		    {
			// The second sync byte matched as well. Assuming
			// this is a right data.
			synced = true;
			break;
		    }
		    else
		    {
			// It seems, it was not the header. Continue to read.
			continue;
		    }
		}
		else
		{
		    // Read error occured.
		    errorCount++;
		}
	    }
	    else
	    {
		// It seems, it was not the header. Continue to read.
		continue;
	    }
	}
	else
	{
	    // Read error occured.
	    errorCount++;
	}

	// Check if the errors were occured too often.
	if (errorCount > MAX_SYNC_ERROR_COUNT_)
	{
	    gWindows->printCommWindow("WARNING - Max Sync count exceeded!\n");
	    break;
	}
    }

    return synced;
}

//-----------------------------------------------------------------------
// This function reads n bytes from the file descriptor.
//-----------------------------------------------------------------------
bool Communicator::readNBytes_(const int fd, unsigned char *buf, const int n)
{
    int msgLength = -1, index = 0, remaining = 0;

    remaining = n;

    while (remaining > 0)
    {
        msgLength = read(fd, &buf[index], remaining);

        if (msgLength == -1)
        {
	    gWindows->printfCommWindow("WARNING - Read error %d!\n", errno);

            return false;
        }

        index += msgLength;
        remaining -= msgLength;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function updates the status of the thread in the info list.
//-----------------------------------------------------------------------
void Communicator::setConnectionThreadIsUp_(const int fd, const bool isUp)
{
    int i, listSize;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);

    listSize = shPrivateConnectionInfoList_.size();

    for (i = 0; i < listSize; i++)
    {
	if (shPrivateConnectionInfoList_[i].fd == fd)
	{
	    shPrivateConnectionInfoList_[i].connectionThreadIsUp = isUp;
	    break;
	}
    }

    pthread_mutex_unlock(&privateConnectionInfoListMutex_);
}

//-----------------------------------------------------------------------
// This function updates the status of the thread in the info list.
//-----------------------------------------------------------------------
void Communicator::saveConnectionThread_(const int fd, const pthread_t connectionThread)
{
    int i, listSize;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);

    listSize = shPrivateConnectionInfoList_.size();

    for (i = 0; i < listSize; i++)
    {
	if (shPrivateConnectionInfoList_[i].fd == fd)
	{
	    shPrivateConnectionInfoList_[i].connectionThread = connectionThread;
	    break;
	}
    }

    pthread_mutex_unlock(&privateConnectionInfoListMutex_);
}

//-----------------------------------------------------------------------
// This function gets the connectionThread stored in the info list.
//-----------------------------------------------------------------------
pthread_t Communicator::getConnectionThread_(const int fd)
{
    pthread_t connectionThread;
    int i, listSize;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);

    listSize = shPrivateConnectionInfoList_.size();

    for (i = 0; i < listSize; i++)
    {
	if (shPrivateConnectionInfoList_[i].fd == fd)
	{
	    connectionThread = shPrivateConnectionInfoList_[i].connectionThread;
	    break;
	}
    }

    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    return connectionThread;
}

//-----------------------------------------------------------------------
// This function gets the connectionThread stored in the info list.
//-----------------------------------------------------------------------
CBRPlanner_ClientInfo_t Communicator::getClientInfo_(const int fd)
{
    CBRPlanner_ClientInfo_t clientInfo;
    int i, listSize;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);

    listSize = shPrivateConnectionInfoList_.size();

    for (i = 0; i < listSize; i++)
    {
	if (shPrivateConnectionInfoList_[i].fd == fd)
	{
	    clientInfo = shPrivateConnectionInfoList_[i].clientInfo;
	    break;
	}
    }

    pthread_mutex_unlock(&privateConnectionInfoListMutex_);

    return clientInfo;
}

//-----------------------------------------------------------------------
// This function disconnects the private connection.
//-----------------------------------------------------------------------
void Communicator::disconnectPrivateConnection_(const int fd)
{
    CBRPlanner_ClientInfo_t clientInfo;
    pthread_t connectionThread;

    // First, report the disconnection to the user
    clientInfo = getClientInfo_(fd);
    gWindows->printfCommWindow(
	"Client %s (PID %d) disconnected.\n",
	clientInfo.name, clientInfo.pid);

    // Update the info list.
    setConnectionThreadIsUp_(fd, false);

    // Finally, quit the thread.
    connectionThread = getConnectionThread_(fd);
    pthread_cancel(connectionThread);
    pthread_join(connectionThread, NULL);
}

//-----------------------------------------------------------------------
// This function disconnects all of the private connections.
//-----------------------------------------------------------------------
void Communicator::disconnectAllPrivateConnections_(void)
{
    int i, listSize;
    
    pthread_mutex_lock(&privateConnectionInfoListMutex_);

    listSize = shPrivateConnectionInfoList_.size();

    for (i = 0; i < listSize; i++)
    {
	if (shPrivateConnectionInfoList_[i].connectionThreadIsUp)
	{
	    pthread_cancel(shPrivateConnectionInfoList_[i].connectionThread);
	    pthread_join(shPrivateConnectionInfoList_[i].connectionThread, NULL);

	    shPrivateConnectionInfoList_[i].connectionThreadIsUp = false;
	}
    }

    pthread_mutex_unlock(&privateConnectionInfoListMutex_);
}

//-----------------------------------------------------------------------
// This function reads the client info being received
//-----------------------------------------------------------------------
CBRPlanner_ClientInfo_t Communicator::readClientInfo_(const int fd)
{
    CBRPlanner_ClientInfo_t clientInfo;
    bool received = false;

    received = readNBytes_(fd, (unsigned char *)&clientInfo, sizeof(CBRPlanner_ClientInfo_t));

    if (received)
    {
	gWindows->printfCommWindow(
	    "Client is %s (PID %d).\n",
	    clientInfo.name, clientInfo.pid);
	fflush(stdout);
    }
    else
    {
	// Something went wrong during the reading of the data. Warn the user.
	gWindows->printfCommWindow("WARNING - Reading the client info failed!\n");
    }

    return clientInfo;
}

//-----------------------------------------------------------------------
// This function updates the client list.
//-----------------------------------------------------------------------
void Communicator::updateClientInfo_(const int fd, const CBRPlanner_ClientInfo_t clientInfo)
{
    int i, listSize;

    pthread_mutex_lock(&privateConnectionInfoListMutex_);

    listSize = shPrivateConnectionInfoList_.size();

    for (i = 0; i < listSize; i++)
    {
	if (shPrivateConnectionInfoList_[i].fd == fd)
	{
	    shPrivateConnectionInfoList_[i].clientInfo = clientInfo;
	    break;
	}
    }

    pthread_mutex_unlock(&privateConnectionInfoListMutex_);
}

//-----------------------------------------------------------------------
// This function reads the mission plan being received.
//-----------------------------------------------------------------------
CBRPlanner_ACDLMissionPlan_t Communicator::readACDLMissionPlan_(const int fd)
{
    CBRPlanner_ACDLMissionPlan_t acdlPlan;

    // Get the mission plan type.
    readNBytes_(fd, (unsigned char *)&(acdlPlan.type), sizeof(int));

    // Get the mission plan rating.
    readNBytes_(fd, (unsigned char *)&(acdlPlan.rating), sizeof(int));

    // Get the data size.
    readNBytes_(fd, (unsigned char *)&(acdlPlan.dataSize), sizeof(int));

    // Get the data itself.
    acdlPlan.data = new char[acdlPlan.dataSize+1];
    readNBytes_(fd, (unsigned char *)(acdlPlan.data), acdlPlan.dataSize);
    acdlPlan.data[acdlPlan.dataSize] = '\0';

    return acdlPlan;
}

//-----------------------------------------------------------------------
// This function writes the mission plan to the server.
//-----------------------------------------------------------------------
void Communicator::writeACDLMissionPlans_(
    const int fd,
    vector<CBRPlanner_ACDLMissionPlan_t> acdlPlans)
{
    int i, j, numPlans, dataIndexListSize;

    // First, send the number of the plans.
    numPlans = acdlPlans.size();
    write(fd, &(numPlans), sizeof(int));

    // Next, send the mission plan. The mission plan data is composed of the type flag,
    // data size, and the data itselt.
    for (i = 0; i < numPlans; i++)
    {
	// Send the mission type.
	write(fd, &(acdlPlans[i].type), sizeof(int));

	// Send the rating.
	write(fd, &(acdlPlans[i].rating), sizeof(int));

	// Send the list of the data indexes.
	dataIndexListSize = acdlPlans[i].dataIndexList.size();
	write(fd, &(dataIndexListSize), sizeof(int));
	for (j = 0; j < dataIndexListSize; j++)
	{
	    write(fd, &(acdlPlans[i].dataIndexList[j]), sizeof(int));
	}

	// Send the data.
	write(fd, &(acdlPlans[i].dataSize), sizeof(int));
	write(fd, acdlPlans[i].data, acdlPlans[i].dataSize);
    }

    return;
}

//-----------------------------------------------------------------------
// This function reads an integer number.
//-----------------------------------------------------------------------
int Communicator::readInteger_(const int fd)
{
    int number;

    readNBytes_(fd, (unsigned char *)&number, sizeof(int));

    return number;
}

//-----------------------------------------------------------------------
// This function reads a boolean value.
//-----------------------------------------------------------------------
bool Communicator::readBoolean_(const int fd)
{
    bool value;

    readNBytes_(fd, (unsigned char *)&value, sizeof(bool));

    return value;
}

//-----------------------------------------------------------------------
// This function reads the features being received.
//-----------------------------------------------------------------------
MemoryManager::Features_t Communicator::readFeatures_(const int fd)
{
    MemoryManager::Feature_t feature;
    MemoryManager::Features_t features;
    bool nonIndex;
    bool received = false;
    double weight;
    int i, numFeatures, nameSize, valueSize;
    char *name = NULL, *value = NULL;
    const bool DEBUG = true;

    // Get the number of features.
    received = readNBytes_(fd, (unsigned char *)&(numFeatures), sizeof(int));

    if (DEBUG)
    {
	gDebugger->printfLine();
	gDebugger->printfDebug("RECEIVED FEATURES:\n");
    }

    for (i = 0; i < numFeatures; i++)
    {
	// Get the length of the name.
	received = readNBytes_(fd, (unsigned char *)&(nameSize), sizeof(int));

	// Get the name.
	name = new char[nameSize+1];
	received = readNBytes_(fd, (unsigned char *)name, nameSize);
	name[nameSize] = '\0';
	feature.name = name;
	delete [] name;
	name = NULL;
	
	// Get the length of the value.
	received = readNBytes_(fd, (unsigned char *)&(valueSize), sizeof(int));

	// Get the name.
	value = new char[valueSize+1];
	received = readNBytes_(fd, (unsigned char *)value, valueSize);
	value[valueSize] = '\0';
	feature.value = value;
	delete [] value;
	value = NULL;

	// Get the nonIndex flag.
	received = readNBytes_(fd, (unsigned char *)&(nonIndex), sizeof(bool));
	feature.nonIndex = nonIndex;

	// Get the weight.
	received = readNBytes_(fd, (unsigned char *)&(weight), sizeof(double));
	feature.weight = weight;

	// Put in the list.
	features.push_back(feature);

	if (DEBUG)
	{
	    gDebugger->printfDebug("%s  ", feature.name.c_str());
	    gDebugger->printfDebug("%s  ", feature.value.c_str());
	    gDebugger->printfDebug("%d  ", feature.nonIndex);
	    gDebugger->printfDebug("%.2f\n", feature.weight);
	}
    }

    if (DEBUG)
    {
	gDebugger->printfLine();
    }

    return features;
}

//-----------------------------------------------------------------------
// This function reads the FSA list info being received.
//-----------------------------------------------------------------------
CBRPlanner_FSANames_t Communicator::readFSANames_(const int fd)
{
    CBRPlanner_FSANames_t fsaNames;
    string taskName, triggerName;
    bool received = false;
    int i, numTasks, numTriggers, nameSize;
    char *name = NULL;

    // Get the number of tasks.
    received = readNBytes_(fd, (unsigned char *)&(numTasks), sizeof(int));

    for (i = 0; i < numTasks; i++)
    {
	// Get the length of the task name;
	received = readNBytes_(fd, (unsigned char *)&(nameSize), sizeof(int));

	// Get the task name;
	name = new char[nameSize+1];
	received = readNBytes_(fd, (unsigned char *)name, nameSize);
	name[nameSize] = '\0';
	taskName = name;
	fsaNames.taskNames.push_back(taskName);
	delete [] name;
	name = NULL;
    }

    // Get the number of triggers.
    received = readNBytes_(fd, (unsigned char *)&(numTriggers), sizeof(int));

    for (i = 0; i < numTriggers; i++)
    {
	// Get the length of the trigger name;
	received = readNBytes_(fd, (unsigned char *)&(nameSize), sizeof(int));

	// Get the trigger name;
	name = new char[nameSize+1];
	received = readNBytes_(fd, (unsigned char *)name, nameSize);
	name[nameSize] = '\0';
	triggerName = name;
	fsaNames.triggerNames.push_back(triggerName);
	delete [] name;
	name = NULL;
    }

    return fsaNames;
}

//-----------------------------------------------------------------------
// This function reads the list of data indexes.
//-----------------------------------------------------------------------
vector<int> Communicator::readDataIndexList_(const int fd)
{
    vector<int> dataIndexList;
    int i, dataIndexListSize;
    int dataIndex;
    bool received = false;
    
    // Get the size of the list.
    received = readNBytes_(fd, (unsigned char *)&(dataIndexListSize), sizeof(int));

    for (i = 0; i < dataIndexListSize; i++)
    {
	received = readNBytes_(fd, (unsigned char *)&(dataIndex), sizeof(int));
	dataIndexList.push_back(dataIndex);
    }

    return dataIndexList;
}

//-----------------------------------------------------------------------
// This function sends the list of data indexes.
//-----------------------------------------------------------------------
void Communicator::writeDataIndexList_(const int fd, vector<int> dataIndexList)
{
    int i, dataIndexListSize;
    
    dataIndexListSize = dataIndexList.size();
    write(fd, &dataIndexListSize, sizeof(int));

    for (i = 0; i < dataIndexListSize; i++)
    {
	write(fd, &(dataIndexList[i]), sizeof(int));
    }
}

/**********************************************************************
 * $Log: communicator.cc,v $
 * Revision 1.1  2003/04/06 08:43:44  endo
 * Initial revision
 *
 *
 **********************************************************************/
