/**********************************************************************
 **                                                                  **
 **                          keyboard_reader.cc                      **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This module deals with processing of the user inputs via        **
 **  key entries.                                                    **
 **                                                                  **
 **  Copyright 2003 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: keyboard_reader.cc,v 1.1 2003/04/06 08:43:44 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <getopt.h>
#include <sys/types.h>
#include <pthread.h>
#include <curses.h>

#include "keyboard_reader.h"
#include "windows.h"
#include "cbrplanner.h"

//-----------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------
extern void gQuitProgram(int exitStatus); // Defined in main.cc

//-----------------------------------------------------------------------
// Constatns
//-----------------------------------------------------------------------
const int KeyboardReader::DAEMON_USLEEPTIME_ = 10000; // 10 msec.
const string KeyboardReader::KEYBIND_TITLE_ = "Enter a key:";

//-----------------------------------------------------------------------
// Mutex initialization
//-----------------------------------------------------------------------
pthread_mutex_t KeyboardReader::readerMutex_ = PTHREAD_MUTEX_INITIALIZER;

//-----------------------------------------------------------------------
// Constructor for KeyboardReader class.
//-----------------------------------------------------------------------
KeyboardReader::KeyboardReader(KeyboardReader **keyboardReaderInstance) : self_(keyboardReaderInstance)
{
    if (keyboardReaderInstance != NULL)
    {
	*keyboardReaderInstance = this;
    }

    daemonThreadIsUp_ = false;
    displayDefaultKeybind();
}

//-----------------------------------------------------------------------
// Distructor for KeyboardReader class.
//-----------------------------------------------------------------------
KeyboardReader::~KeyboardReader(void)
{
    // Make sure to kill the deamon thread.
    if (daemonThreadIsUp_)
    {
	daemonThreadIsUp_ = false;
    }

    if (self_ != NULL)
    {
	*self_ = NULL;
    }
}

//-----------------------------------------------------------------------
// This function starts up a thread to run the keyboard reader daemon.
//-----------------------------------------------------------------------
void *KeyboardReader::startDaemonThread_(void *keyboardReaderInstance)
{
    ((KeyboardReader *)keyboardReaderInstance)->daemonMainLoop_();

    return NULL;
}

//-----------------------------------------------------------------------
// This function is a daemon to start reading user inputs
//-----------------------------------------------------------------------
void KeyboardReader::daemonMainLoop_(void)
{
    daemonThreadIsUp_ = true;

    while(true)
    {
	// Make sure termination was not requested.
        pthread_testcancel();

	// Catch and proces the user input.
	catchAndProcessInput_();

	// Make sure termination was not requested.
        pthread_testcancel();

	// Sleep a little to reduce the CPU.
	usleep(DAEMON_USLEEPTIME_);
    }
}

//-----------------------------------------------------------------------
// This function starts up the daemon to start reading user inputs
//-----------------------------------------------------------------------
void KeyboardReader::startDaemon(void)
{
    // Start the communication deamon thread.
    pthread_create(&daemonThread_, NULL, &startDaemonThread_, (void *)this );
}

//-----------------------------------------------------------------------
// This function reads the user input.
//-----------------------------------------------------------------------
void KeyboardReader::catchAndProcessInput_(void)
{
    char c;

    c = getch();

    switch(c) {

    case 'l':
    case 'L':
	gCBRPlanner->loadCBRLibrary();
	displayDefaultKeybind();
	break;

    case 's':
    case 'S':
	gCBRPlanner->saveCBRLibrary();
	displayDefaultKeybind();
	break;

    case 'q':
    case 'Q':
	if (gCBRPlanner->memoryChanged())
	{
	    if (!(gCBRPlanner->confirmClear()))
	    {
		displayDefaultKeybind();
		break;
	    }
	}
	gWindows->printfPlannerWindow("Program terminted by the user.\n");
	gQuitProgram(0);
	break;

    case 'x':
    case 'X':
	gWindows->printfPlannerWindow("Program terminted by the user.\n");
	gQuitProgram(0);
	break;
    }
}

//-----------------------------------------------------------------------
// This function display keyboard for the default keyboarding.
//-----------------------------------------------------------------------
void KeyboardReader::displayDefaultKeybind(void)
{
    gWindows->clearKeyboardWindow();
    gWindows->printfKeyboardWindow("l)oad library,  ");
    gWindows->printfKeyboardWindow("s)ave library,  ");
    gWindows->printfKeyboardWindow("q)uit");
    gWindows->setKeyboardWindowTitle(KEYBIND_TITLE_);
}

/**********************************************************************
 * $Log: keyboard_reader.cc,v $
 * Revision 1.1  2003/04/06 08:43:44  endo
 * Initial revision
 *
 *
 **********************************************************************/
