/**********************************************************************
 **                                                                  **
 **                               main.cc                            **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This file contains the main function for CBRServer.             **
 **                                                                  **
 **  Copyright 2003 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: main.cc,v 1.1 2003/04/06 08:43:44 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <getopt.h>
#include <sys/types.h>
#include <string>
#include "cbrplanner.h"
#include "communicator.h"
#include "windows.h"
#include "keyboard_reader.h"
#include "debugger.h"

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const bool DEBUG = false;
const string EMPTY_STRING = "";

//-----------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------
CBRPlanner *gCBRPlanner = NULL;
Communicator *gCommunicator = NULL;
Windows *gWindows = NULL;
KeyboardReader *gKeyboardReader = NULL;
Debugger *gDebugger = NULL;

//---------------------------------------------------------------------
// This function kills the program.
//---------------------------------------------------------------------
void gQuitProgram(int exitStatus)
{
    if (gKeyboardReader != NULL)
    {
	delete gKeyboardReader;
    }

    if (gWindows != NULL)
    {
	if (exitStatus == -1)
	{
	    gWindows->printfPlannerWindow("Abnormal termination.\n");
	}
	delete gWindows;
    }

    if (gCommunicator != NULL)
    {
	delete gCommunicator;
    }

    if (gCBRPlanner != NULL)
    {
	delete gCBRPlanner;
    }

    if (gDebugger != NULL)
    {
	delete gDebugger;
    }

    exit(exitStatus);
}

//---------------------------------------------------------------------
// This function displays usage.
//---------------------------------------------------------------------
void showUsageAndExit(char *binaryName)
{
    fprintf(stderr, "\nUsage: %s <option>\n\n", binaryName);
    fprintf(stderr, "Where <option> is:\n");
    fprintf(stderr, "  -s <socket name>  Use the specified socket name to initiate the\n");
    fprintf(stderr, "                    communication with CfgEdit/mlab.\n");
    fprintf(stderr, "  -l <library name> Load the specified CBR library.\n");
    fprintf(stderr, "                -d  Enable debug.\n");
    fprintf(stderr, "                -h  Show this usage.\n");
    fprintf(stderr, "\n");

    gQuitProgram(0);
}

//---------------------------------------------------------------------
// This function catches the signal.
//---------------------------------------------------------------------
void signalHandler(int sig)
{
    switch (sig) {

    case SIGABRT:
    case SIGSEGV:
    case SIGBUS:
    case SIGHUP:
	gQuitProgram(-1);
	break;

    default:
	gQuitProgram(0);
	break;

    }
}

//-----------------------------------------------------------------------
// Main function for cbrpserver
//-----------------------------------------------------------------------
int main(int argc, char **argv)
{
    int c;
    bool validArgument = true;
    bool debug = DEBUG;
    string specifiedSocketName = EMPTY_STRING;
    string cbrlibraryName = EMPTY_STRING;

    // Set the signal handler.
    signal(SIGINT, signalHandler);
    signal(SIGABRT, signalHandler);
    signal(SIGTERM, signalHandler);
    signal(SIGBUS, signalHandler);
    signal(SIGHUP, signalHandler);
    signal(SIGSEGV, signalHandler);

    // Get the command line arguments.
    if (argc > 1)
    {
        while ((c = getopt(argc, argv, "s:l:hd")) != -1)
        {
            switch (c) {
            case 'h':
                validArgument = false;
                break;
            case 's':
		specifiedSocketName = optarg;
                break;
            case 'l':
		cbrlibraryName = optarg;
                break;
            case 'd':
		debug = true;
                break;
            }
        }

        if (!validArgument)
	{
	    showUsageAndExit(argv[0]);
	}
    }

    // Instantiate the Debugger class.
    new Debugger(&gDebugger);

    // Instantiate the CBRPlanner class.
    new CBRPlanner(&gCBRPlanner);

    // Instantiate the Communicator class.
    new Communicator(&gCommunicator);

    // Instantiate the Windows class, and start windows.
    new Windows(&gWindows);
    gWindows->startWindows();

    // Instantiate the KeyboardReader class, and start the deamon.
    new KeyboardReader(&gKeyboardReader);
    gKeyboardReader->startDaemon();

    if (debug)
    {
	// Enable debugger.
	gDebugger->enableDebug();
    }

    // Start all the subsystems of the CBR Planner.
    gCBRPlanner->startSubsystems();

    // If user specified, load the CBR library.
    if (cbrlibraryName != EMPTY_STRING)
    {
	gCBRPlanner->loadCBRLibrary(cbrlibraryName);
    }

    // If user specified, use that socket name instead.
    if (specifiedSocketName != EMPTY_STRING)
    {
	gWindows->printfCommWindow("CBR Planner: Using the socket \"%s\".", specifiedSocketName.c_str());
	gCommunicator->setSocketName(specifiedSocketName);
    }

    // Start the communication daeman.
    gCommunicator->startDaemon();

    // Start a main loop until the program is terminated by the user.
    gCBRPlanner->mainLoop();

    return 0;
}

/**********************************************************************
 * $Log: main.cc,v $
 * Revision 1.1  2003/04/06 08:43:44  endo
 * Initial revision
 *
 *
 **********************************************************************/
