/**********************************************************************
 **                                                                  **
 **                           memory_manager.cc                      **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This modules deals with storing and retrieving of missions.     **
 **                                                                  **
 **  Copyright 2003 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: memory_manager.cc,v 1.1 2003/04/06 08:43:44 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <getopt.h>
#include <sys/types.h>

#include "cbrplanner_domain.h"
#include "memory_manager.h"
#include "windows.h"
#include "debugger.h"

//-----------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------
extern void gQuitProgram(int exitStatus); // Defined in main.cc

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const string MemoryManager::EMPTY_STRING_ = "";
const string MemoryManager::CBRLIBRARY_EXTENSION_ = "cbl";
const string MemoryManager::STRING_PERIOD_ = ".";
const string MemoryManager::CBL_STRING_DATA_ = "DATA";
const string MemoryManager::CBL_STRING_TYPE_ = "TYPE";
const string MemoryManager::CBL_STRING_RATING_ = "RATING";
const string MemoryManager::CBL_STRING_ARCHTYPE_ = "ARCHTYPE";
const string MemoryManager::CBL_STRING_PLAN_ = "PLAN";
const string MemoryManager::CBL_STRING_OPERATOR_ = "OPERATOR";
const string MemoryManager::CBL_STRING_OPERATOR_NAME_ = "OPERATOR-NAME";
const string MemoryManager::CBL_STRING_OPERATOR_INDEX_ = "OPERATOR-INDEX";
const string MemoryManager::CBL_STRING_OPERATOR_PARM_ = "OPERATOR-PARM";
const string MemoryManager::CBL_STRING_OPERATOR_PARM_NAME_ = "OPERATOR-PARM-NAME";
const string MemoryManager::CBL_STRING_OPERATOR_PARM_VALUE_ = "OPERATOR-PARM-VALUE";
const string MemoryManager::CBL_STRING_OPERATOR_PRECONDITION_ = "OPERATOR-PRECONDITION";
const string MemoryManager::CBL_STRING_OPERATOR_PRECONDITION_NAME_ = "OPERATOR-PRECONDITION-NAME";
const string MemoryManager::CBL_STRING_OPERATOR_EFFECT_ = "OPERATOR-EFFECT";
const string MemoryManager::CBL_STRING_OPERATOR_EFFECT_NAME_ = "OPERATOR-EFFECT-NAME";
const string MemoryManager::CBL_STRING_OPERATOR_EFFECT_INDEX_ = "OPERATOR-EFFECT-INDEX";
const string MemoryManager::CBL_STRING_OPERATOR_EFFECT_PARM_ = "OPERATOR-EFFECT-PARM";
const string MemoryManager::CBL_STRING_OPERATOR_EFFECT_PARM_NAME_ = "OPERATOR-EFFECT-PARM-NAME";
const string MemoryManager::CBL_STRING_OPERATOR_EFFECT_PARM_VALUE_ = "OPERATOR-EFFECT-PARM-VALUE";
const string MemoryManager::CBL_STRING_DECISIONTREE_ = "DECISIONTREE";
const string MemoryManager::CBL_STRING_NODE_ = "NODE";
const string MemoryManager::CBL_STRING_NODE_NAME_ = "NODE-NAME";
const string MemoryManager::CBL_STRING_NODE_LEAF_ = "NODE-LEAF";
const string MemoryManager::CBL_STRING_NODE_EDGE_ = "NODE-EDGE";
const string MemoryManager::CBL_STRING_NODE_EDGE_PARENT_ = "NODE-EDGE-PARENT";
const string MemoryManager::CBL_STRING_NODE_EDGE_VALUE_ = "NODE-EDGE-VALUE";
const string MemoryManager::CBL_STRING_NODE_EDGE_NEXTNODE_ = "NODE-EDGE-NEXTNODE";
const string MemoryManager::CBL_STRING_NODE_DATA_ = "NODE-DATA";
const string MemoryManager::CBL_STRING_NODE_DATA_INDEXES_ = "NODE-DATA-INDEXES";
const string MemoryManager::CBL_STRING_NODE_DATA_INDEX_ = "NODE-DATA-INDEX";
const string MemoryManager::CBL_STRING_NODE_DATA_STATUSES_ = "NODE-DATA-STATUSES";
const string MemoryManager::CBL_STRING_NODE_DATA_STATUS_ = "NODE-DATA-STATUS";
const string MemoryManager::CBL_STRING_NODE_DATA_OTHERFEATURE_ = "NODE-DATA-OTHERFEATURE";
const string MemoryManager::CBL_STRING_NODE_DATA_OTHERFEATURE_NAME_ = "NODE-DATA-OTHERFEATURE-NAME";
const string MemoryManager::CBL_STRING_NODE_DATA_OTHERFEATURE_VALUE_ = "NODE-DATA-OTHERFEATURE-VALUE";
const string MemoryManager::CBL_STRING_START_ = "START";
const string MemoryManager::CBL_STRING_END_ = "END";
const int UNSPECIFIED_DATA_INDEX_ = -1;

//-----------------------------------------------------------------------
// Constructor for MemoryManager class.
//-----------------------------------------------------------------------
MemoryManager::MemoryManager(MemoryManager **memoryManagerInstance) : self_(memoryManagerInstance)
{
    if (memoryManagerInstance != NULL)
    {
	*memoryManagerInstance = this;
    }

    decisionTree_ = NULL;
    memoryChanged_ = false;
}

//-----------------------------------------------------------------------
// Distructor for MemoryManager class.
//-----------------------------------------------------------------------
MemoryManager::~MemoryManager(void)
{
    if (self_ != NULL)
    {
	*self_ = NULL;
    }
}

//-----------------------------------------------------------------------
// This function will save a new mission plan into its memory (decision
// tree).
//-----------------------------------------------------------------------
int MemoryManager::saveNewMissionPlan(MissionPlan_t missionplan, Features_t features)
{
    Feature_t feature;
    DecisionTreeNode_t *node = NULL, *childNode = NULL;
    vector<DecisionTreeNode_t *>nodeStack;
    DecisionTreeData_t data;
    string value;
    int index;
    int i, numEdges;
    bool dataSaved = false;
    const bool DEBUG = true;

    // Save the case.
    index = missionPlanList_.size();
    missionplan.dataIndexList.push_back(index);
    missionPlanList_.push_back(missionplan);
    memoryChanged_ = true;

    // Provide the feedback to the user.
    gWindows->printfPlannerWindow(
	"New case (ID %d) added to the memory.\n",
	index);

    if (decisionTree_ == NULL)
    {
	decisionTree_ = new DecisionTree_t;

	// Get the feature of the highest priority.
	feature = features.front();

	// Update the list (i.e., remove the feature from the list).
	removeFeatureFromList_(&features, feature.name);

	// Construct the data for the leaf node.
	data.missionPlanListIndexes.push_back(index);
	data.statuses.push_back(NEUTRAL_DATA);
	data.otherFeatures = features;

	// Create a new node.
	addNewNodeToDecisionTree_(decisionTree_, feature.name, feature.value, data);
    }
    else
    {
	nodeStack.push_back(decisionTree_);
	dataSaved = false;
	
	while (!dataSaved)
	{
	    // Get the node from the stack.
	    node = nodeStack.back();
	    nodeStack.pop_back();

	    // Get the value of the feature.
	    value = findValueOfFeature_(features, node->name);

	    // Update the list (i.e., remove the feature from the list).
	    removeFeatureFromList_(&features, node->name);

	    numEdges = node->edges.size();

	    // Look for edges to see if the value exists.
	    for (i = 0; i < numEdges; i++)
	    {
		if (value == node->edges[i].value)
		{
		    childNode = (DecisionTreeNode_t *)(node->edges[i].nextNode);

		    if (childNode->isLeafNode)
		    {
			// The child is a leaf node. It means that the child contains data.
			if ((int)(childNode->data.otherFeatures.size()) > 0)
			{
			    // Other feature(s) is still available to expand the tree.
			    // Create a new grandchild node, using the next priority of the
			    // feature. Copy the data to the grandchild as well.
			    feature = childNode->data.otherFeatures.front();
			    removeFeatureFromList_(&(childNode->data.otherFeatures), feature.name);

			    addNewNodeToDecisionTree_(
				childNode,
				feature.name,
				feature.value,
				childNode->data);
			}
			else
			{
			    // There is no other feature to distinguish this data
			    // and the existing one. Append this data in the child node.
			    childNode->data.missionPlanListIndexes.push_back(index);
			    childNode->data.statuses.push_back(NEUTRAL_DATA);
			    dataSaved = true;
			    break;
			}
		    }
		    else
		    {
			// The child is not a leaf node. It means there is a grandchild node
			// being already created. Just rememeber it to explore it later.
		    }

		    nodeStack.push_back(childNode);
		    break;
		}
	    }

	    if (((int)(nodeStack.size()) == 0) && (!dataSaved))
	    {
		// It is a new value. Add new edge and node to the tree.
		// First, construct the data for the leaf node.
		data.missionPlanListIndexes.push_back(index);
		data.statuses.push_back(NEUTRAL_DATA);
		data.otherFeatures = features;

		// Create a new node.
		addNewNodeToDecisionTree_(node, node->name, value, data);
		dataSaved = true;
	    }
	}
    }

    if (DEBUG)
    {
	dumpDecisionTree_();
    }

    return index;
}

//-----------------------------------------------------------------------
// This function creates a new node to the decision tree.
//-----------------------------------------------------------------------
void MemoryManager::addNewNodeToDecisionTree_(
    DecisionTree_t *decisionTree,
    string nodeName,
    string edgeValue,
    DecisionTreeData_t data)
{
    DecisionTreeEdge_t edge;
    DecisionTreeNode_t *childNode = NULL;

    // Copy the node name.
    decisionTree->name = nodeName;

    // This is no longer a leaf node.
    decisionTree->isLeafNode = false;

    // Construct the child node.
    childNode = new DecisionTreeNode_t;
    childNode->isLeafNode = true;
    childNode->data = data;

    // Construct an edge based on the feature value, and connect
    // the parent and child nodes with it.
    edge.parentName = nodeName;
    edge.value = edgeValue;
    edge.nextNode = (void *)childNode;
    decisionTree->edges.push_back(edge);
}

//-----------------------------------------------------------------------
// This function removes a feature from the list.
//-----------------------------------------------------------------------
void MemoryManager::removeFeatureFromList_(Features_t *features, string featureName)
{
    int i, j, listSize;

    listSize = features->size();

    for (i = 0; i < listSize; i++)
    {
	if ((*features)[i].name == featureName)
	{
	    for (j = i; j < (listSize-1); j++)
	    {
		(*features)[j] = (*features)[j+1];
	    }

	    features->pop_back();

	    return;
	}
    }
}

//-----------------------------------------------------------------------
// This function finds a value of the feature from the list.
//-----------------------------------------------------------------------
string MemoryManager::findValueOfFeature_(
    Features_t features,
    string featureName)
{
    string value = EMPTY_STRING_;
    int i, listSize;

    listSize = features.size();

    for (i = 0; i < listSize; i++)
    {
	if (features[i].name == featureName)
	{
	    value = features[i].value;
	    break;
	}
    }

    return value;
}

//-----------------------------------------------------------------------
// This function will get an saved mission plan into its memory (decision
// tree).
//-----------------------------------------------------------------------
vector<MemoryManager::MissionPlan_t> MemoryManager::getSavedMissionPlans(
    Features_t features,
    int maxRating)
{
    MissionPlan_t missionplan;
    vector<MissionPlan_t> missionplans;
    const vector<MissionPlan_t> EMPTY_MISSION_PLANS;
    DecisionTreeNode_t *node, *childNode;
    vector<DecisionTreeNode_t *> evaluationNodeStack, backtrackNodeStack;
    vector<string> visitedNodeNames;
    string value;
    int index;
    int i, j, numEdges, rating;
    bool savedBefore = false;
    const bool DEBUG = true;

    rating = maxRating;

    if (decisionTree_ == NULL)
    {
	// No tree, no solution.
	return EMPTY_MISSION_PLANS;
    }
    else
    {
	if (DEBUG)
	{
	    dumpDecisionTree_();
	}

	evaluationNodeStack.push_back(decisionTree_);
    }

    while(((int)(evaluationNodeStack.size())) > 0)
    {
	// Get the node in the stack.
	node = evaluationNodeStack.back();
	evaluationNodeStack.pop_back();

	// Remember the node for backtracking.
	backtrackNodeStack.push_back(node);

	// Get the value of the feature.
	value = findValueOfFeature_(features, node->name);

	numEdges = node->edges.size();

	// Look for edges to see if the value exits.
	for (i = 0; i < numEdges; i++)
	{
	    if (value == node->edges[i].value)
	    {
		childNode = (DecisionTreeNode_t *)(node->edges[i].nextNode);

		if (childNode->isLeafNode)
		{
		    for (j = 0; j < ((int)(childNode->data.missionPlanListIndexes.size())); j++)
		    {
			// This child node is a leaf node. It means that it contains the
			// data. Return the mission plan in the data.
			index = childNode->data.missionPlanListIndexes[j];
			missionplan = missionPlanList_[index];
			missionplan.type = CBRPLANNER_MISSIONPLAN_OLD;
			
			// If it had a negative feedback rate it less.
			// If it had a positive feedback rate it more.
			missionplan.rating = rating;

			switch (childNode->data.statuses[j]) {

			case POSITIVE_DATA:
			    missionplan.rating++;
			    break;

			case NEGATIVE_DATA:
			    missionplan.rating--;
			    break;
			}

			if (missionplan.rating > 0)
			{
			    missionplans.push_back(missionplan);
			}
		    }
		    return missionplans;
		}
		else
		{
		    // If the child is not a leaf node. Remember this node to explore
		    // it later.
		    evaluationNodeStack.push_back(childNode);
		}
	    }
	}
    }


    // The perfect match not found. Backtrack to see if compromised version exists.
    while (((int)(backtrackNodeStack.size())) > 0)
    {
	if (rating <= 1)
	{
	    //break;
	}
	else
	{
	    rating--;
	}

	node = backtrackNodeStack.back();
	backtrackNodeStack.pop_back();

	visitedNodeNames.push_back(node->name);

	numEdges = node->edges.size();

	// Look for all the edges to see if data exits.
	for (i = 0; i < numEdges; i++)
	{
	    childNode = (DecisionTreeNode_t *)(node->edges[i].nextNode);
	    
	    if (childNode->isLeafNode)
	    {
		for (j = 0; j < ((int)(childNode->data.missionPlanListIndexes.size())); j++)
		{
		    // This child node is a leaf node. It means that it contains the
		    // data. Return the mission plan in the data.
		    index = childNode->data.missionPlanListIndexes[j];
		    missionplan = missionPlanList_[index];
		    missionplan.type = CBRPLANNER_MISSIONPLAN_OLD;
		    missionplan.rating = rating;

		    // If it had a negative feedback rate it less.
		    // If it had a positive feedback rate it more.
		    switch (childNode->data.statuses[j]) {
			
		    case POSITIVE_DATA:
			missionplan.rating++;
			break;
			
		    case NEGATIVE_DATA:
			missionplan.rating--;
			break;
		    }

		    if (missionplan.rating > 0)
		    {
			missionplans.push_back(missionplan);
		    }
		}
	    }
	    else
	    {
		// If this node is saved before, save it for further backtracking.
		savedBefore = false;
		for (j = 0; j < ((int)(visitedNodeNames.size())); j++)
		{
		    if (visitedNodeNames[j] == childNode->name)
		    {
			savedBefore = true;
			break;
		    }
		}

		if (!savedBefore)
		{
		    backtrackNodeStack.push_back(childNode);
		}
	    }
	}
    }

    return missionplans;
}

//-----------------------------------------------------------------------
// This function prints out the entire decision tree.
//-----------------------------------------------------------------------
void MemoryManager::dumpDecisionTree_(void)
{
    int i, numEdges;
    DecisionTreeNode_t *node;
    DecisionTreeEdge_t edge;
    vector<DecisionTreeEdge_t> edgeStack;

    gDebugger->printfLine();
    gDebugger->printfDebug("ROOT NODE:\n");
    gDebugger->printfDebug("NAME %s\n", decisionTree_->name.c_str());
    gDebugger->printfDebug("LEAF %d\n", decisionTree_->isLeafNode);
    gDebugger->printfDebug("EDGES %d\n", decisionTree_->edges.size());

    edgeStack = decisionTree_->edges;

    while(!(edgeStack.empty()))
    {
	edge = edgeStack.back();
	edgeStack.pop_back();
	gDebugger->printfLine();
	gDebugger->printfDebug("EDGE PARENT %s\n", edge.parentName.c_str());
	gDebugger->printfDebug("EDGE VALUE %s\n", edge.value.c_str());

	node = (DecisionTreeNode_t *)edge.nextNode;
	gDebugger->printfDebug("NEW NODE:\n");
	gDebugger->printfDebug("NAME %s\n", node->name.c_str());
	gDebugger->printfDebug("LEAF %d\n", node->isLeafNode);
	gDebugger->printfDebug("EDGES %d\n", node->edges.size());

	if (node->isLeafNode)
	{
	    for (i = 0; i < ((int)(node->data.missionPlanListIndexes.size())); i++)
	    {
		gDebugger->printfDebug("DATA %d\n", node->data.missionPlanListIndexes[i]);
		gDebugger->printfDebug("STATUS %d\n", node->data.statuses[i]);
	    }
	}

	numEdges = node->edges.size();
	for (i = 0; i < numEdges; i++)
	{
	    edgeStack.push_back(node->edges[i]);
	}
    }
}

//-----------------------------------------------------------------------
// This function asks the CBR library name from the user.
//-----------------------------------------------------------------------
string MemoryManager::askCBRLibraryName_(void)
{
    string cbrlibraryName;
    string title;

    title = "Enter the CBR library name.";

    gWindows->clearKeyboardWindow();
    gWindows->setKeyboardWindowTitle(title);
    gWindows->printfKeyboardWindow(": ");
    cbrlibraryName = gWindows->scanfKeyboardWindow();

    if (!hasCBRLibraryExtension_(cbrlibraryName))
    {
	cbrlibraryName = addCBRLibraryExtension_(cbrlibraryName);
    }

    return cbrlibraryName;
}

//-----------------------------------------------------------------------
// This function loads a CBR library.
//-----------------------------------------------------------------------
void MemoryManager::loadCBRLibrary(string cbrlibraryName)
{
    FILE *in = NULL;

    in = fopen(cbrlibraryName.c_str(), "r");

    if (in == NULL)
    {
	// Error opening the file.
	gWindows->printfPlannerWindow(
	    "Warning: CBR library \"%s\" could not be loaded.\n",
	    cbrlibraryName.c_str());
	return;
    }

    if (memoryChanged_)
    {
	if (!confirmClear_())
	{
	    gWindows->printfPlannerWindow("CBR library not loaded.\n");
	    return;
	}
    }


    // Clear the memory first.
    missionPlanList_.clear();
    delete decisionTree_;
    decisionTree_ = NULL;

    // Load the library.
    loadCBRLibrary_(in);

    // Close the file.
    fclose(in);

    // Provide the feedbacks to the memory.
    gWindows->printfPlannerWindow(
	"CBR library \"%s\" loaded.\n",
	cbrlibraryName.c_str());

    gWindows->printfPlannerWindow(
	"%d case%s in the current memory.\n",
	(int)(missionPlanList_.size()),
	((int)(missionPlanList_.size())>1)? "s" : "");
};

//-----------------------------------------------------------------------
// This function loads a CBR library.
//-----------------------------------------------------------------------
void MemoryManager::loadCBRLibrary(void)
{
    string cbrlibraryName;

    cbrlibraryName = askCBRLibraryName_();

    if (!loadVerify_(cbrlibraryName))
    {
	// The user changed the mind.
	gWindows->printfPlannerWindow("Cases not loaded.\n");
	return;
    }

    loadCBRLibrary(cbrlibraryName);
}

//-----------------------------------------------------------------------
// This function loads a CBR library.
//-----------------------------------------------------------------------
void MemoryManager::loadCBRLibrary_(FILE *in)
{
    char buf1[1024], buf2[1024];
    string bufString1, bufString2;
    MissionPlan_t missionPlan;
    int index;

    while (fscanf(in, "%s %s", buf1, buf2) != EOF)
    {
	bufString1 = buf1;
	bufString2 = buf2;

	if ((bufString1 == CBL_STRING_DATA_) &&
	    (bufString2 == CBL_STRING_START_))
	{
	    missionPlan = loadMissionPlan_(in);
	    index = missionPlanList_.size();
	    missionPlan.dataIndexList.push_back(index);
	    missionPlanList_.push_back(missionPlan);
	}
	else if ((bufString1 == CBL_STRING_DECISIONTREE_) &&
		 (bufString2 == CBL_STRING_START_))
	{
	    decisionTree_ = loadDecisionTree_(in);
	}	
    }
}

//-----------------------------------------------------------------------
// This function loads a CBR library.
//-----------------------------------------------------------------------
MemoryManager::MissionPlan_t MemoryManager::loadMissionPlan_(FILE *in)
{
    char buf1[1024], buf2[1024], buf3[1024], buf4[1024];
    char buf5[1024], buf6[1024], buf7[1024], buf8[1024];
    char buf9[1024], buf10[1024];
    string bufString1, bufString2, bufString3, bufString4;
    string bufString5, bufString6, bufString7, bufString8;
    string bufString9, bufString10;
    MissionPlan_t missionplan;
    CBRPlanner_PlanOperator_t planOperator;
    CBRPlanner_Parameter_t parm, effectParm;
    CBRPlanner_Precondition_t precondition;
    CBRPlanner_Effect_t effect;
    CBRPlanner_PlanSequence_t plan;

    while (fscanf(in, "%s %s", buf1, buf2) != EOF)
    {
	bufString1 = buf1;
	bufString2 = buf2;
		
	if (bufString1 == CBL_STRING_TYPE_)
	{
	    missionplan.type = atoi(buf2);
	}
	else if (bufString1 == CBL_STRING_ARCHTYPE_)
	{
	    missionplan.archType = bufString2;
	}
	else if ((bufString1 == CBL_STRING_PLAN_) &&
		 (bufString2 == CBL_STRING_START_))
	{
	    plan.clear();

	    while (fscanf(in, "%s %s", buf3, buf4) != EOF)
	    {
		bufString3 = buf3;
		bufString4 = buf4;
		
		if ((bufString3 == CBL_STRING_OPERATOR_) &&
		    (bufString4 == CBL_STRING_START_))
		{
		    planOperator.parameters.clear();
		    planOperator.preconditions.clear();
		    planOperator.effects.clear();

		    while (fscanf(in, "%s %s", buf5, buf6) != EOF)
		    {
			bufString5 = buf5;
			bufString6 = buf6;

			if (bufString5 == CBL_STRING_OPERATOR_NAME_)
			{
			    planOperator.name = bufString6;
			}
			else if (bufString5 == CBL_STRING_OPERATOR_INDEX_)
			{
			    planOperator.index = bufString6;
			}
			else if ((bufString5 == CBL_STRING_OPERATOR_PARM_) &&
				 (bufString6 == CBL_STRING_START_))
			{
			    while (fscanf(in, "%s %s", buf7, buf8) != EOF)
			    {
				bufString7 = buf7;	
				bufString8 = buf8;
				
				if (bufString7 == CBL_STRING_OPERATOR_PARM_NAME_)
				{
				    parm.name = bufString8; 
				}
				else if (bufString7 == CBL_STRING_OPERATOR_PARM_VALUE_)
				{
				    parm.value = bufString8; 
				}
				else if ((bufString7 == CBL_STRING_OPERATOR_PARM_) &&
					 (bufString8 == CBL_STRING_END_))
				{
				    planOperator.parameters.push_back(parm);
				    break;
				}
			    }
			}
			else if ((bufString5 == CBL_STRING_OPERATOR_PRECONDITION_) &&
				 (bufString6 == CBL_STRING_START_))
			{
			    while (fscanf(in, "%s %s", buf7, buf8) != EOF)
			    {
				bufString7 = buf7;	
				bufString8 = buf8;

				if (bufString7 == CBL_STRING_OPERATOR_PRECONDITION_NAME_)
				{
				    precondition.name = bufString8; 
				}
				else if ((bufString7 == CBL_STRING_OPERATOR_PRECONDITION_) &&
					 (bufString8 == CBL_STRING_END_))
				{
				    planOperator.preconditions.push_back(precondition);
				    break;
				}
			    }
			}
			else if ((bufString5 == CBL_STRING_OPERATOR_EFFECT_) &&
				 (bufString6 == CBL_STRING_START_))
			{
			    effect.parameters.clear();

			    while (fscanf(in, "%s %s", buf7, buf8) != EOF)
			    {
				bufString7 = buf7;	
				bufString8 = buf8;

				if (bufString7 == CBL_STRING_OPERATOR_EFFECT_NAME_)
				{
				    effect.name = bufString8; 
				}
				else if (bufString7 == CBL_STRING_OPERATOR_EFFECT_INDEX_)
				{
				    effect.index = bufString8; 
				}
				else if ((bufString7 == CBL_STRING_OPERATOR_EFFECT_PARM_) &&
					 (bufString8 == CBL_STRING_START_))
				{
				    while (fscanf(in, "%s %s", buf9, buf10) != EOF)
				    {
					bufString9 = buf9;
					bufString10 = buf10;

					if (bufString9 == CBL_STRING_OPERATOR_EFFECT_PARM_NAME_)
					{
					    effectParm.name = bufString10; 
					}
					else if (bufString9 == CBL_STRING_OPERATOR_EFFECT_PARM_VALUE_)
					{
					    effectParm.value = bufString10; 
					}
					else if ((bufString9 == CBL_STRING_OPERATOR_EFFECT_PARM_) &&
						 (bufString10 == CBL_STRING_END_))
					{
					    effect.parameters.push_back(effectParm);
					    break;
					}
				    }
				}
				else if ((bufString7 == CBL_STRING_OPERATOR_EFFECT_) &&
					 (bufString8 == CBL_STRING_END_))
				{
				    planOperator.effects.push_back(effect);
				    break;
				}
			    }
			}
			else if ((bufString5 == CBL_STRING_OPERATOR_) &&
				 (bufString6 == CBL_STRING_END_))
			{
			    plan.push_back(planOperator);
			    break;
			}
		    }
		}
		else if ((bufString3 == CBL_STRING_PLAN_) &&
			 (bufString4 == CBL_STRING_END_))
		{
		    missionplan.plans.push_back(plan);
		    break;
		}
	    }
	}
	else if ((bufString1 == CBL_STRING_DATA_) &&
		 (bufString2 == CBL_STRING_END_))
	{
	    break;
	}
    }

    return missionplan;
}

//-----------------------------------------------------------------------
// This function loads a decision tree from a file.
//-----------------------------------------------------------------------
MemoryManager::DecisionTree_t *MemoryManager::loadDecisionTree_(FILE *in)
{
    DecisionTree_t *decisionTree = NULL;

    decisionTree = loadNode_(in);

    return decisionTree;
}

//-----------------------------------------------------------------------
// This function loads a node from a file.
//-----------------------------------------------------------------------
MemoryManager::DecisionTreeData_t MemoryManager::loadNodeData_(FILE *in)
{
    char buf1[1024], buf2[1024], buf3[1024], buf4[1024];
    char buf5[1024], buf6[1024];
    string bufString1, bufString2, bufString3, bufString4;
    string bufString5, bufString6;
    DecisionTreeData_t data;
    Feature_t otherFeature;
    int index, status;
    bool dataFound = false;

    while ((fscanf(in, "%s %s", buf1, buf2) != EOF) && (!dataFound))
    {
	bufString1 = buf1;
	bufString2 = buf2;

	if ((bufString1 == CBL_STRING_NODE_DATA_) &&
	    (bufString2 == CBL_STRING_START_))
	{

	    while (fscanf(in, "%s %s", buf3, buf4) != EOF)
	    {
		bufString3 = buf3;
		bufString4 = buf4;

		if ((bufString3 == CBL_STRING_NODE_DATA_INDEXES_) &&
		    (bufString4 == CBL_STRING_START_))
		{
		    while (fscanf(in, "%s %s", buf5, buf6) != EOF)
		    {
			bufString5 = buf5;
			bufString6 = buf6;

			if (bufString5 == CBL_STRING_NODE_DATA_INDEX_)
			{
			    index = atoi(buf6);
			    data.missionPlanListIndexes.push_back(index);
			}
			else if ((bufString5 == CBL_STRING_NODE_DATA_INDEXES_) &&
				 (bufString6 == CBL_STRING_END_))
			{
			    break;
			}
		    }  
		}
		if ((bufString3 == CBL_STRING_NODE_DATA_STATUSES_) &&
		    (bufString4 == CBL_STRING_START_))
		{
		    while (fscanf(in, "%s %s", buf5, buf6) != EOF)
		    {
			bufString5 = buf5;
			bufString6 = buf6;

			if (bufString5 == CBL_STRING_NODE_DATA_STATUS_)
			{
			    status = atoi(buf6);
			    data.statuses.push_back(status);
			}
			else if ((bufString5 == CBL_STRING_NODE_DATA_STATUSES_) &&
				 (bufString6 == CBL_STRING_END_))
			{
			    break;
			}
		    }  
		}
		else if ((bufString3 == CBL_STRING_NODE_DATA_OTHERFEATURE_) &&
			 (bufString4 == CBL_STRING_START_))
		{
		    while (fscanf(in, "%s %s", buf5, buf6) != EOF)
		    {
			bufString5 = buf5;
			bufString6 = buf6;

			if (bufString5 == CBL_STRING_NODE_DATA_OTHERFEATURE_NAME_)
			{
			    otherFeature.name = bufString6;
			}
			else if (bufString5 == CBL_STRING_NODE_DATA_OTHERFEATURE_VALUE_)
			{
			    otherFeature.value = bufString6;
			}
			else if ((bufString5 == CBL_STRING_NODE_DATA_OTHERFEATURE_) &&
				 (bufString6 == CBL_STRING_END_))
			{
			    data.otherFeatures.push_back(otherFeature);
			    break;
			}
		    }

		}
		else if ((bufString3 == CBL_STRING_NODE_DATA_) &&
			 (bufString4 == CBL_STRING_END_))
		{
		    dataFound = true;
		    break;
		}
	    }
	}
    }    

    return data;
}

//-----------------------------------------------------------------------
// This function loads a node from a file.
//-----------------------------------------------------------------------
MemoryManager::DecisionTreeNode_t *MemoryManager::loadNode_(FILE *in)
{
    char buf1[1024], buf2[1024], buf3[1024], buf4[1024];
    char buf5[1024], buf6[1024];
    string bufString1, bufString2, bufString3, bufString4;
    string bufString5, bufString6;
    DecisionTreeNode_t *node = NULL;
    DecisionTreeEdge_t edge;

    while (fscanf(in, "%s %s", buf1, buf2) != EOF)
    {
	bufString1 = buf1;
	bufString2 = buf2;

	if ((bufString1 == CBL_STRING_NODE_) &&
	    (bufString2 == CBL_STRING_START_))
	{
	    node = new DecisionTreeNode_t;

	    while (fscanf(in, "%s %s", buf3, buf4) != EOF)
	    {
		bufString3 = buf3;
		bufString4 = buf4;

		if (bufString3 == CBL_STRING_NODE_NAME_)
		{
		    node->name = bufString4;
		}
		else if (bufString3 == CBL_STRING_NODE_LEAF_)
		{
		    node->isLeafNode = atoi(buf4);

		    if (node->isLeafNode)
		    {
			node->data = loadNodeData_(in);
			return node;
		    }

		}
		else if ((bufString3 == CBL_STRING_NODE_EDGE_) &&
			 (bufString4 == CBL_STRING_START_))
		{
		    while (fscanf(in, "%s %s", buf5, buf6) != EOF)
		    {
			bufString5 = buf5;
			bufString6 = buf6;
			
			if (bufString5 == CBL_STRING_NODE_EDGE_PARENT_)
			{
			    edge.parentName = bufString6;
			}
			else if (bufString5 == CBL_STRING_NODE_EDGE_VALUE_)
			{
			    edge.value = bufString6;
			}
			else if ((bufString5 == CBL_STRING_NODE_EDGE_NEXTNODE_) &&
				 (bufString6 == CBL_STRING_START_))
			{
			    (DecisionTreeNode_t *)(edge.nextNode) = loadNode_(in);
			}
			else if ((bufString5 == CBL_STRING_NODE_EDGE_) &&
				 (bufString6 == CBL_STRING_END_))
			{
			    node->edges.push_back(edge);
			    break;
			}
		    }
		}
		else if ((bufString3 == CBL_STRING_NODE_) &&
			 (bufString4 == CBL_STRING_END_))
		{
		    return node;
		}
	    }
	}
	else if ((bufString1 == CBL_STRING_DECISIONTREE_) &&
		 (bufString2 == CBL_STRING_END_))
	{
	    break;
	}
	else if ((bufString1 == CBL_STRING_NODE_EDGE_NEXTNODE_) &&
		 (bufString2 == CBL_STRING_END_))
	{
	    break;
	}
    }

    return node;
}

//-----------------------------------------------------------------------
// This function saves the current cases into a library.
//-----------------------------------------------------------------------
void MemoryManager::saveCBRLibrary(void)
{
    string cbrlibraryName;

    cbrlibraryName = askCBRLibraryName_();

    if (!saveVerify_(cbrlibraryName))
    {
	// The user changed the mind.
	gWindows->printfPlannerWindow("Cases not saved.\n");
	return;
    }

    saveCBRLibrary_(cbrlibraryName);
}

//-----------------------------------------------------------------------
// This function saves the current cases into a library.
//-----------------------------------------------------------------------
void MemoryManager::saveCBRLibrary_(string cbrlibraryName)
{
    FILE *out = NULL;

    out = fopen(cbrlibraryName.c_str(), "w");

    if (out == NULL)
    {
	// Error opening the file.
	gWindows->printfPlannerWindow(
	    "Warning: \"%s\" could not be opend. Cases not saved.\n",
	    cbrlibraryName.c_str());

	return;
    }

    // Write the data (i.e., mission plans).
    writeMissionPlans_(out);

    // Write the decision tree.
    writeDecisionTree_(out);

    // Close the file.
    fclose(out);

    // Report the status.
    gWindows->printfPlannerWindow("Cases saved in \"%s\".\n", cbrlibraryName.c_str());

    memoryChanged_ = false;
}

//-----------------------------------------------------------------------
// This function saves the entire decision tree in a file.
//-----------------------------------------------------------------------
void MemoryManager::writeDecisionTree_(FILE *out)
{
    fprintf(
	out,
	"%s %s\n",
	CBL_STRING_DECISIONTREE_.c_str(),
	CBL_STRING_START_.c_str());

    writeNode_(out, decisionTree_);

    fprintf(
	out,
	"%s %s\n",
	CBL_STRING_DECISIONTREE_.c_str(),
	CBL_STRING_END_.c_str());

    fflush(out);
}

//-----------------------------------------------------------------------
// This function save data in a file. 
//-----------------------------------------------------------------------
void MemoryManager::writeNodeData_(FILE *out, DecisionTreeData_t data)
{
    Feature_t otherFeature;
    int i, index, status;

    fprintf(
	out,
	"%s %s\n",
	CBL_STRING_NODE_DATA_.c_str(),
	CBL_STRING_START_.c_str());

    // Write indexes.
    fprintf(
	out,
	"%s %s\n",
	CBL_STRING_NODE_DATA_INDEXES_.c_str(),
	CBL_STRING_START_.c_str());

    for (i = 0; i < (int)(data.missionPlanListIndexes.size()); i++)
    {
	index = data.missionPlanListIndexes[i];

	fprintf(
	    out,
	    "%s %d\n",
	    CBL_STRING_NODE_DATA_INDEX_.c_str(),
	    index);
    }

    fprintf(
	out,
	"%s %s\n",
	CBL_STRING_NODE_DATA_INDEXES_.c_str(),
	CBL_STRING_END_.c_str());

    // Write statuses.
    fprintf(
	out,
	"%s %s\n",
	CBL_STRING_NODE_DATA_STATUSES_.c_str(),
	CBL_STRING_START_.c_str());

    for (i = 0; i < (int)(data.statuses.size()); i++)
    {
	status = data.statuses[i];

	fprintf(
	    out,
	    "%s %d\n",
	    CBL_STRING_NODE_DATA_STATUS_.c_str(),
	    status);
    }

    fprintf(
	out,
	"%s %s\n",
	CBL_STRING_NODE_DATA_STATUSES_.c_str(),
	CBL_STRING_END_.c_str());

    // Write features.
    for (i = 0; i < (int)(data.otherFeatures.size()); i++)
    {
	otherFeature = data.otherFeatures[i];

	fprintf(
	    out,
	    "%s %s\n",
	    CBL_STRING_NODE_DATA_OTHERFEATURE_.c_str(),
	    CBL_STRING_START_.c_str());

	fprintf(
	    out,
	    "%s %s\n",
	    CBL_STRING_NODE_DATA_OTHERFEATURE_NAME_.c_str(),
	    otherFeature.name.c_str());

	fprintf(
	    out,
	    "%s %s\n",
	    CBL_STRING_NODE_DATA_OTHERFEATURE_VALUE_.c_str(),
	    otherFeature.value.c_str());

	fprintf(
	    out,
	    "%s %s\n",
	    CBL_STRING_NODE_DATA_OTHERFEATURE_.c_str(),
	    CBL_STRING_END_.c_str());
    }

    fprintf(
	out,
	"%s %s\n",
	CBL_STRING_NODE_DATA_.c_str(),
	CBL_STRING_END_.c_str());
}

//-----------------------------------------------------------------------
// This function save a node in a file. 
//-----------------------------------------------------------------------
void MemoryManager::writeNode_(FILE *out, DecisionTreeNode_t *node)
{
    int i;
    DecisionTreeEdge_t edge;
    DecisionTreeNode_t *nextNode;

    fprintf(
	out,
	"%s %s\n",
	CBL_STRING_NODE_.c_str(),
	CBL_STRING_START_.c_str());

    if (node->name != EMPTY_STRING_)
    {
	fprintf(
	    out,
	    "%s %s\n",
	    CBL_STRING_NODE_NAME_.c_str(),
	    node->name.c_str());
    }

    fprintf(
	out,
	"%s %d\n",
	CBL_STRING_NODE_LEAF_.c_str(),
	node->isLeafNode);

    if (node->isLeafNode)
    {
	writeNodeData_(out, node->data);
	return;
    }

    for (i = 0; i < (int)(node->edges.size()); i ++)
    {
	edge = node->edges[i];
	fprintf(
	    out,
	    "%s %s\n",
	    CBL_STRING_NODE_EDGE_.c_str(),
	    CBL_STRING_START_.c_str());

	if (edge.parentName != EMPTY_STRING_)
	{
	    fprintf(
		out,
		"%s %s\n",
		CBL_STRING_NODE_EDGE_PARENT_.c_str(),
		edge.parentName.c_str());
	}

	if (edge.value != EMPTY_STRING_)
	{
	    fprintf(
		out,
		"%s %s\n",
		CBL_STRING_NODE_EDGE_VALUE_.c_str(),
		edge.value.c_str());
	}

	nextNode = (DecisionTreeNode_t *)(edge.nextNode);

	fprintf(
	    out,
	    "%s %s\n",
	    CBL_STRING_NODE_EDGE_NEXTNODE_.c_str(),
	    CBL_STRING_START_.c_str());

	writeNode_(out, nextNode);

	fprintf(
	    out,
	    "%s %s\n",
	    CBL_STRING_NODE_EDGE_NEXTNODE_.c_str(),
	    CBL_STRING_END_.c_str());

	fprintf(
	    out,
	    "%s %s\n",
	    CBL_STRING_NODE_EDGE_.c_str(),
	    CBL_STRING_END_.c_str());
    }

    fprintf(
	out,
	"%s %s\n",
	CBL_STRING_NODE_.c_str(),
	CBL_STRING_END_.c_str());

    fflush(out);
}

//-----------------------------------------------------------------------
// This function save the entire decision tree.
//-----------------------------------------------------------------------
void MemoryManager::writeMissionPlans_(FILE *out)
{
    int i, j, k, l, p;
    MissionPlan_t missionplan;
    CBRPlanner_PlanOperator_t planOperator;
    CBRPlanner_Parameter_t parm, effectParm;
    CBRPlanner_Precondition_t precondition;
    CBRPlanner_Effect_t effect;
    CBRPlanner_PlanSequence_t plan;

    for (i = 0; i < (int)(missionPlanList_.size()); i++)
    {
	missionplan = missionPlanList_[i];
	fprintf(
	    out,
	    "%s %s\n",
	    CBL_STRING_DATA_.c_str(),
	    CBL_STRING_START_.c_str());

	fprintf(
	    out,
	    "%s %d\n",
	    CBL_STRING_TYPE_.c_str(),
	    missionplan.type);

	fprintf(
	    out,
	    "%s %s\n",
	    CBL_STRING_ARCHTYPE_.c_str(),
	    missionplan.archType.c_str());

	for (p = 0; p < (int)(missionplan.plans.size()); p++)
	{
	    fprintf(
		out,
		"%s %s\n",
		CBL_STRING_PLAN_.c_str(),
		CBL_STRING_START_.c_str());

	    plan = missionplan.plans[p];

	    for (j = 0; j < (int)(plan.size()); j++)
	    {
		planOperator = plan[j];
		fprintf(
		    out,
		    "%s %s\n",
		    CBL_STRING_OPERATOR_.c_str(),
		    CBL_STRING_START_.c_str());
		fprintf(
		    out,
		    "%s %s\n",
		    CBL_STRING_OPERATOR_NAME_.c_str(),
		    planOperator.name.c_str());
		fprintf(
		    out,
		    "%s %s\n",
		    CBL_STRING_OPERATOR_INDEX_.c_str(),
		    planOperator.index.c_str());

		for (k = 0; k < (int)(planOperator.parameters.size()); k++)
		{
		    parm = planOperator.parameters[k];
		    fprintf(
			out,
			"%s %s\n",
			CBL_STRING_OPERATOR_PARM_.c_str(),
			CBL_STRING_START_.c_str());
		    fprintf(
			out,
			"%s %s\n",
			CBL_STRING_OPERATOR_PARM_NAME_.c_str(),
			parm.name.c_str());
		    fprintf(
			out,
			"%s %s\n",
			CBL_STRING_OPERATOR_PARM_VALUE_.c_str(),
			parm.value.c_str());
		    fprintf(
			out,
			"%s %s\n",
			CBL_STRING_OPERATOR_PARM_.c_str(),
			CBL_STRING_END_.c_str());
		}

		for (k = 0; k < (int)(planOperator.preconditions.size()); k++)
		{
		    precondition = planOperator.preconditions[k];
		    fprintf(
			out,
			"%s %s\n",
			CBL_STRING_OPERATOR_PRECONDITION_.c_str(),
			CBL_STRING_START_.c_str());
		    fprintf(
			out,
			"%s %s\n",
			CBL_STRING_OPERATOR_PRECONDITION_NAME_.c_str(),
			precondition.name.c_str());
		    fprintf(
			out,
			"%s %s\n",
			CBL_STRING_OPERATOR_PRECONDITION_.c_str(),
			CBL_STRING_END_.c_str());
		}

		for (k = 0; k < (int)(planOperator.effects.size()); k++)
		{
		    effect = planOperator.effects[k];
		    fprintf(
			out,
			"%s %s\n",
			CBL_STRING_OPERATOR_EFFECT_.c_str(),
			CBL_STRING_START_.c_str());
		    fprintf(
			out,
			"%s %s\n",
			CBL_STRING_OPERATOR_EFFECT_NAME_.c_str(),
			effect.name.c_str());

		    fprintf(
			out,
			"%s %s\n",
			CBL_STRING_OPERATOR_EFFECT_INDEX_.c_str(),
			effect.index.c_str());

		    for (l = 0; l < (int)(effect.parameters.size()); l++)
		    {
			effectParm = effect.parameters[l];
			fprintf(
			    out,
			    "%s %s\n",
			    CBL_STRING_OPERATOR_EFFECT_PARM_.c_str(),
			    CBL_STRING_START_.c_str());
			fprintf(
			    out,
			    "%s %s\n",
			    CBL_STRING_OPERATOR_EFFECT_PARM_NAME_.c_str(),
			    effectParm.name.c_str());
			fprintf(
			    out,
			    "%s %s\n",
			    CBL_STRING_OPERATOR_EFFECT_PARM_VALUE_.c_str(),
			    effectParm.value.c_str());
			fprintf(
			    out,
			    "%s %s\n",
			    CBL_STRING_OPERATOR_EFFECT_PARM_.c_str(),
			    CBL_STRING_END_.c_str());
		    }
		    fprintf(
			out,
			"%s %s\n",
			CBL_STRING_OPERATOR_EFFECT_.c_str(),
			CBL_STRING_END_.c_str());
		}
		fprintf(
		    out,
		    "%s %s\n",
		    CBL_STRING_OPERATOR_.c_str(),
		    CBL_STRING_END_.c_str());
	    }

	    fprintf(
		out,
		"%s %s\n",
		CBL_STRING_PLAN_.c_str(),
		CBL_STRING_END_.c_str());
	}

	fprintf(
	    out,
	    "%s %s\n",
	    CBL_STRING_DATA_.c_str(),
	    CBL_STRING_END_.c_str());
    }

    fflush(out);
}

//-----------------------------------------------------------------------
// This function asks the user to verify the file name.
//-----------------------------------------------------------------------
bool MemoryManager::saveVerify_(string cbrlibraryName)
{
    string title;
    char c;
    bool save = false;

    title = "Save to \"";
    title += cbrlibraryName;
    title += "\"?";

    gWindows->clearKeyboardWindow();
    gWindows->printfKeyboardWindow("s)ave,  ");
    gWindows->printfKeyboardWindow("c)ancel");
    gWindows->setKeyboardWindowTitle(title);

    c = getch();

    switch(c) {

    case 's':
    case 'S':
	save = true;
	break;
    }

    return save;
}

//-----------------------------------------------------------------------
// This function asks the user to verify the file name.
//-----------------------------------------------------------------------
bool MemoryManager::loadVerify_(string cbrlibraryName)
{
    string title;
    char c;
    bool load = false;

    title = "Load \"";
    title += cbrlibraryName;
    title += "\"?";

    gWindows->clearKeyboardWindow();
    gWindows->printfKeyboardWindow("l)oad,  ");
    gWindows->printfKeyboardWindow("c)ancel");
    gWindows->setKeyboardWindowTitle(title);

    c = getch();

    switch(c) {

    case 'l':
    case 'L':
	load = true;
	break;
    }

    return load;
}

//-----------------------------------------------------------------------
// This function asks the user to confirm deletion of the current cases.
//-----------------------------------------------------------------------
bool MemoryManager::confirmClear_(void)
{
    string title;
    char c;
    bool clear = false;

    title = "Discard all cases in the current memory?";

    gWindows->clearKeyboardWindow();
    gWindows->printfKeyboardWindow("d)iscard,  ");
    gWindows->printfKeyboardWindow("c)ancel");
    gWindows->setKeyboardWindowTitle(title);

    c = getch();

    switch(c) {

    case 'd':
    case 'D':
	clear = true;
	break;
    }

    return clear;
}

//-----------------------------------------------------------------------
// This function checks to see it has an appropriate extension for the
// library.
//-----------------------------------------------------------------------
bool MemoryManager::hasCBRLibraryExtension_(string cbrlibraryName)
{
    int index, size, extensionSize;
    string extension, possibleExtension;

    extension =  STRING_PERIOD_ + CBRLIBRARY_EXTENSION_;

    size = cbrlibraryName.size();

    extensionSize = extension.size();

    index = size - extensionSize;

    if (index < 1)
    {
	return false;
    }

    possibleExtension = cbrlibraryName.substr(index, extensionSize);

    if (possibleExtension != extension)
    {
	return false;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function adds an appropriate extension for the library.
//-----------------------------------------------------------------------
string MemoryManager::addCBRLibraryExtension_(string cbrlibraryName)
{
    int size;
    string extension;

    size = cbrlibraryName.size();

    if (size == 0)
    {
	cbrlibraryName = createCBRLibraryNamePrefix_();
    }

    extension =  STRING_PERIOD_ + CBRLIBRARY_EXTENSION_;

    cbrlibraryName += extension;

    return cbrlibraryName;
}

//-----------------------------------------------------------------------
// This function sets the default CBR library name prefix.
//-----------------------------------------------------------------------
string MemoryManager::createCBRLibraryNamePrefix_(void)
{
    time_t cur_time;
    tm local_time;
    string prefix;
    char date[100], month[100], day[100], year[100], hour[100], min[100], sec[100];
    char *userName = NULL, buf[1024];

    // Get the user name.
    userName = getenv("USER");

    // Get the current date and time.
    cur_time = time(NULL);
    strncpy(date,ctime(&cur_time),sizeof(date));
    local_time = *localtime(&cur_time);
    strftime(month, sizeof(month), "%m", &local_time);
    strftime(day, sizeof(day), "%d", &local_time);
    strftime(year, sizeof(year), "%Y", &local_time);
    strftime(hour, sizeof(hour), "%H", &local_time);
    strftime(min, sizeof(min), "%M", &local_time);
    strftime(sec, sizeof(sec), "%S", &local_time);

    // Create the prefix.
    sprintf(buf, "cbrserver-%s-%s%s%s-%s%s%s",
	    userName,
	    month,
	    day,
	    year,
	    hour,
	    min,
	    sec);
    prefix = buf;

    return prefix;
}

//-----------------------------------------------------------------------
// This function deals with the user's positive feedback.
//-----------------------------------------------------------------------
void MemoryManager::givePositiveFeedback(vector<int> dataIndexList)
{
    int i;

    for (i = 0; i < (int)(dataIndexList.size()); i++)
    {
	giveFeedback_(decisionTree_, dataIndexList[i], POSITIVE_DATA);

	// Report the status.
	gWindows->printfPlannerWindow("Positive feedback recorded (ID %d).\n");
    }
}

//-----------------------------------------------------------------------
// This function deals with the user's negative feedback.
//-----------------------------------------------------------------------
void MemoryManager::giveNegativeFeedback(vector<int> dataIndexList)
{
    int i;

    for (i = 0; i < (int)(dataIndexList.size()); i++)
    {
	giveFeedback_(decisionTree_, dataIndexList[i], NEGATIVE_DATA);

	// Report the status.
	gWindows->printfPlannerWindow("Negative feedback recorded (ID %d).\n");
    }
}

//-----------------------------------------------------------------------
// This function deals with the user's positive feedback.
//-----------------------------------------------------------------------
void MemoryManager::giveFeedback_(DecisionTreeNode_t *node, int index, int status)
{
    int i;
    DecisionTreeEdge_t edge;
    DecisionTreeNode_t *nextNode;

    if (node->isLeafNode)
    {
	for (i = 0; i < (int)(node->data.missionPlanListIndexes.size()); i++)
	{
	    if (node->data.missionPlanListIndexes[i] == index)
	    {
		node->data.statuses[i] = status;
		memoryChanged_ = true;
		return;
	    }
	}
	return;
    }

    for (i = 0; i < (int)(node->edges.size()); i ++)
    {
	edge = node->edges[i];

	nextNode = (DecisionTreeNode_t *)(edge.nextNode);

	giveFeedback_(nextNode, index, status);
    }
}

/**********************************************************************
 * $Log: memory_manager.cc,v $
 * Revision 1.1  2003/04/06 08:43:44  endo
 * Initial revision
 *
 *
 **********************************************************************/
