/**********************************************************************
 **                                                                  **
 **                           memory_manager.h                       **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  Copyright 2003 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/
#ifndef MEMORY_MANAGER_H
#define MEMORY_MANAGER_H

/* $Id: memory_manager.h,v 1.1 2003/04/06 08:43:44 endo Exp $ */

#include <vector>
#include <string>

#include "cbrplanner_protocol.h"
#include "cbrplanner_domain.h"

class MemoryManager {

public:
    typedef struct Coordinates_t
    {
	double x;
	double y;
    };

    typedef struct MissionPlan_t
    {
	int type;
	int rating;
	string archType;
	string maxVelValue;
	string baseVelValue;
	vector<CBRPlanner_PlanSequence_t> plans;
	vector<int> dataIndexList;
    };

    typedef vector<MissionPlan_t> MissionPlanList_t;

    typedef struct Feature_t
    {
	string name;
	string value;
	bool nonIndex;
	double weight;
    };

    typedef vector<Feature_t> Features_t;

    typedef enum DecisionTreeDataStatus_t
    {
	NEUTRAL_DATA,
	POSITIVE_DATA,
	NEGATIVE_DATA
    };

    typedef struct DecisionTreeData_t
    {
	vector<int> missionPlanListIndexes;
	vector<int> statuses;
	Features_t otherFeatures;
    };

    typedef struct DecisionTreeEdge_t
    {
	string parentName;
	string value;
	void *nextNode;
    };

    typedef struct DecisionTreeNode_t
    {
	string name;
	bool isLeafNode;
	vector<DecisionTreeEdge_t> edges;
	DecisionTreeData_t data;
    };

    typedef DecisionTreeNode_t DecisionTree_t;

protected:
    MemoryManager **self_;
    MissionPlanList_t missionPlanList_;
    DecisionTree_t *decisionTree_;
    bool memoryChanged_;
    
    static const string EMPTY_STRING_;
    static const string CBRLIBRARY_EXTENSION_;
    static const string STRING_PERIOD_;
    static const string CBL_STRING_DATA_;
    static const string CBL_STRING_TYPE_;
    static const string CBL_STRING_RATING_;
    static const string CBL_STRING_ARCHTYPE_;
    static const string CBL_STRING_PLAN_;
    static const string CBL_STRING_OPERATOR_;
    static const string CBL_STRING_OPERATOR_NAME_;
    static const string CBL_STRING_OPERATOR_INDEX_;
    static const string CBL_STRING_OPERATOR_PARM_;
    static const string CBL_STRING_OPERATOR_PARM_NAME_;
    static const string CBL_STRING_OPERATOR_PARM_VALUE_;
    static const string CBL_STRING_OPERATOR_PRECONDITION_;
    static const string CBL_STRING_OPERATOR_PRECONDITION_NAME_;
    static const string CBL_STRING_OPERATOR_EFFECT_;
    static const string CBL_STRING_OPERATOR_EFFECT_NAME_;
    static const string CBL_STRING_OPERATOR_EFFECT_INDEX_;
    static const string CBL_STRING_OPERATOR_EFFECT_PARM_;
    static const string CBL_STRING_OPERATOR_EFFECT_PARM_NAME_;
    static const string CBL_STRING_OPERATOR_EFFECT_PARM_VALUE_;
    static const string CBL_STRING_DECISIONTREE_;
    static const string CBL_STRING_NODE_;
    static const string CBL_STRING_NODE_NAME_;
    static const string CBL_STRING_NODE_LEAF_;
    static const string CBL_STRING_NODE_EDGE_;
    static const string CBL_STRING_NODE_EDGE_PARENT_;
    static const string CBL_STRING_NODE_EDGE_VALUE_;
    static const string CBL_STRING_NODE_EDGE_NEXTNODE_;
    static const string CBL_STRING_NODE_DATA_;
    static const string CBL_STRING_NODE_DATA_INDEXES_;
    static const string CBL_STRING_NODE_DATA_INDEX_;
    static const string CBL_STRING_NODE_DATA_STATUSES_;
    static const string CBL_STRING_NODE_DATA_STATUS_;
    static const string CBL_STRING_NODE_DATA_OTHERFEATURE_;
    static const string CBL_STRING_NODE_DATA_OTHERFEATURE_NAME_;
    static const string CBL_STRING_NODE_DATA_OTHERFEATURE_VALUE_;
    static const string CBL_STRING_START_;
    static const string CBL_STRING_END_;

    void addNewNodeToDecisionTree_(
	DecisionTree_t *decisionTree,
	string nodeName,
	string edgeValue,
	DecisionTreeData_t data);
    string findValueOfFeature_(Features_t features, string featureName);
    void removeFeatureFromList_(Features_t *features, string featureName);
    void dumpDecisionTree_(void);
    string askCBRLibraryName_(void);
    bool saveVerify_(string cbrlibraryName);
    bool loadVerify_(string cbrlibraryName);
    bool hasCBRLibraryExtension_(string cbrlibraryName);
    string addCBRLibraryExtension_(string cbrlibraryName);
    string createCBRLibraryNamePrefix_(void);
    void saveCBRLibrary_(string cbrlibraryName);
    void loadCBRLibrary_(FILE *in);
    void checkSaveCBRLibrary(void);
    MissionPlan_t loadMissionPlan_(FILE *in);
    DecisionTree_t *loadDecisionTree_(FILE *in);
    DecisionTreeData_t loadNodeData_(FILE *in);
    DecisionTreeNode_t *loadNode_(FILE *in);
    void writeNodeData_(FILE *out, DecisionTreeData_t data);
    void writeNode_(FILE *out, DecisionTreeNode_t *node);
    void writeDecisionTree_(FILE *out);
    void writeMissionPlans_(FILE *out);
    bool confirmClear_(void);
    void giveFeedback_(DecisionTreeNode_t *node, int index, int status);

public:
    MemoryManager(void);
    MemoryManager(MemoryManager **self);
    ~MemoryManager(void);

    int saveNewMissionPlan(MissionPlan_t missionplan, Features_t features);
    vector<MissionPlan_t> getSavedMissionPlans(Features_t features, int maxRating);
    void loadCBRLibrary(void);
    void loadCBRLibrary(string cbrlibraryName);
    void saveCBRLibrary(void);
    void confirmSaveCBRLibrary(void);
    bool confirmClear(void);
    bool memoryChanged(void);
    void givePositiveFeedback(vector<int> dataIndexList);
    void giveNegativeFeedback(vector<int> dataIndexList);
};

inline bool MemoryManager::confirmClear(void)
{
    return (confirmClear_());
};

inline bool MemoryManager::memoryChanged(void)
{
    return (memoryChanged_);
};

#endif
/**********************************************************************
 * $Log: memory_manager.h,v $
 * Revision 1.1  2003/04/06 08:43:44  endo
 * Initial revision
 *
 *
 **********************************************************************/
