/**********************************************************************
 **                                                                  **
 **                              planner.cc                          **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This module builds a mission based by adjusting ballpark        **
 **  solutions suggested by MemoryManager.                           **
 **                                                                  **
 **  Copyright 2003 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: planner.cc,v 1.1 2003/04/06 08:43:44 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include <getopt.h>
#include <sys/types.h>
#include <string>

#include "cbrplanner_domain.h"
#include "planner.h"
#include "cbrplanner.h"
#include "memory_manager.h"
#include "windows.h"
#include "debugger.h"

//-----------------------------------------------------------------------
// Constatns
//-----------------------------------------------------------------------
const string Planner::EMPTY_STRING_ = "";
const string Planner::STRING_EMPTY_ARCHITECTURE_ = "free";
const string Planner::STRING_START_STATE_ = "Stop";
const string Planner::STRING_START_STATE_INDEX_ = "Start";
const string Planner::STRING_IMMEDIATE_ = "Immediate";
const string Planner::STRING_START_SUB_MISSION_ = "StartSubMission";
const string Planner::STRING_START_SUB_MISSION_INDEX_ = "$StartSubMission";
const string Planner::STRING_SUB_MISSION_NAME_ = "%sub_mission_name";
const string Planner::STRING_GOAL_LOCATION_ = "%Goal_Location";
const string Planner::STRING_MOVE_TO_LOCATION_GAIN_ = "%move_to_location_gain";
const string Planner::STRING_WANDER_GAIN_ = "%wander_gain";
const string Planner::STRING_AVOID_OBSTACLE_GAIN_ = "%avoid_obstacle_gain";
const string Planner::STRING_AVOID_OBSTACLE_SPHERE_ = "%avoid_obstacle_sphere";
const string Planner::STRING_AVOID_OBSTACLE_SAFETY_MARGIN_ = "%avoid_obstacle_safety_margin";
const string Planner::STRING_SUB_MISSION_READY_ = "SubMissionReady";
const string Planner::STRING_SUB_MISSION_DEPLOYMENT_ = "%sub_mission_deployment";
const string Planner::STRING_DEPLOY_FIRST_ = "{DEPLOY_FIRST}";
const string Planner::STRING_EXECUTE_IMMEDIATELY_ = "{EXECUTE_IMMEDIATELY}";
const string Planner::STRING_GOAL_TOLERANCE_ = "%Goal_Tolerance";
const string Planner::FEATURE_NAME_MAX_VELOCITY_ = "MaxVelocity";
const string Planner::FEATURE_NAME_AGGRESSIVENESS_ = "Aggressiveness";
const string Planner::FEATURE_NAME_NUM_ROBOTS_ = "NumberOfRobots";
const string Planner::FEATURE_NAME_START_X_ = "StartXCoord";
const string Planner::FEATURE_NAME_START_Y_ = "StartYCoord";
const string Planner::FEATURE_NAME_TASK_NAME_ = "Task";
const string Planner::FEATURE_NAME_LOCALIZATION_ = "Localization";
const string Planner::FEATURE_NAME_ENVIRONMENT_ = "Environment";
const string Planner::FEATURE_VALUE_USEMAPS_ = "UseMaps";
const string Planner::FEATURE_VALUE_INDOOR_ = "Indoor";
const double Planner::DEFAULT_MOVE_TO_LOCATION_GAIN_ = 1.0;
const double Planner::DEFAULT_WANDER_GAIN_ = 0.0;
const double Planner::DEFAULT_INDOOR_AVOID_OBSTACLE_GAIN_ = 0.3;
const double Planner::DEFAULT_OUTDOOR_AVOID_OBSTACLE_GAIN_ = 1.0;
const double Planner::DEFAULT_INDOOR_AVOID_OBSTACLE_SPHERE_ = 0.7;
const double Planner::DEFAULT_OUTDOOR_AVOID_OBSTACLE_SPHERE_ = 2.2;
const double Planner::DEFAULT_AVOID_OBSTACLE_SAFETY_MARGIN_ = 0.3;
const double Planner::DEFAULT_GOAL_TOLERANCE_ = 0.5;

//-----------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Constructor for Planner class.
//-----------------------------------------------------------------------
Planner::Planner(Planner **plannerInstance) : self_(plannerInstance)
{
    if (plannerInstance != NULL)
    {
	*plannerInstance = this;
    }
}

//-----------------------------------------------------------------------
// Distructor for Planner class.
//-----------------------------------------------------------------------
Planner::~Planner(void)
{
    if (self_ != NULL)
    {
	*self_ = NULL;
    }
}

//-----------------------------------------------------------------------
//  This function creates a misssion plan.
//-----------------------------------------------------------------------
vector<MemoryManager::MissionPlan_t> Planner::getMissionPlans(
    MemoryManager::Features_t features,
    int maxRating)
{
    vector<MemoryManager::MissionPlan_t> missionPlans, bigMissionPlans;
    string maxVelValue, baseVelValue, aggressivenessValue, taskName;
    MemoryManager::Coordinates_t coordinates;
    vector<MemoryManager::Features_t> subFeaturesList;
    char buf[1024];
    double maxVel, baseVel, aggressiveness;
    bool useMaps, indoor;
    int i, j, numRobots, startSubMissionOperatorIndex;

    // First, get the max velocity from the global features.
    maxVelValue = extractMaxVelocityValue_(features);
    maxVel = atof(maxVelValue.c_str());

    // Next, get the base velocity from the global features through
    // "aggressiveness".
    aggressivenessValue = extractAggressivenessValue_(features);
    aggressiveness = atof(aggressivenessValue.c_str());
    baseVel = maxVel * aggressiveness;
    sprintf(buf, "%.2f", baseVel);
    baseVelValue = buf;

    // Get the number of robots from the global features.
    numRobots = extractNumRobots_(features);

    // Decompose the features with respect to sub-missions.
    subFeaturesList = extractSubFeatures_(features);

    // Reset the SubMissionOperator index.
    startSubMissionOperatorIndex = 0;

    for (i = 0; i < (int)(subFeaturesList.size()); i ++)
    {
	// Get the coordinates and other info.
	coordinates = extractCoordinates_(subFeaturesList[i]);
	taskName = extractTaskName_(subFeaturesList[i]);
	useMaps = extractUseMaps_(subFeaturesList[i]);
	indoor = extractIndoor_(subFeaturesList[i]);

	// Remove all non-index features.
	subFeaturesList[i] = filterNonIndexFeatures_(subFeaturesList[i]);

	// Get the mission plan for these "subfeatures".
	missionPlans = gCBRPlanner->memoryManager()->getSavedMissionPlans(
	    subFeaturesList[i], maxRating);

	for (j = 0; j < ((int)(missionPlans.size())); j++)
	{
	    // Adjust the number of robots.
	    missionPlans[j] = adjustNumRobots_(missionPlans[j], numRobots);

	    // Put the "StartSubMisson" state in the beginning.
	    missionPlans[j] = addStartSubMissionState_(
		missionPlans[j],
		taskName,
		coordinates,
		useMaps,
		indoor,
		startSubMissionOperatorIndex++);

	    missionPlans[j].maxVelValue = maxVelValue;
	    missionPlans[j].baseVelValue = baseVelValue;
	}

	// Append the plans to create a sequence of sub-missions.
	bigMissionPlans = appendMissionPlans_(bigMissionPlans, missionPlans, i);
    }

    // Straighten up some parameters.
    for (i = 0; i < (int)(bigMissionPlans.size()); i++)
    {
	bigMissionPlans[i] = addStartState_(bigMissionPlans[i]);

	if (bigMissionPlans[i].type == CBRPLANNER_MISSIONPLAN_EMPTY)
	{
	    bigMissionPlans[i].archType = STRING_EMPTY_ARCHITECTURE_;
	}
	else
	{
	    bigMissionPlans[i].type = CBRPLANNER_MISSIONPLAN_CONSTRUCTED;
	}
    }

    bigMissionPlans = sortMissionPlans_(bigMissionPlans);

    return bigMissionPlans;
}

//-----------------------------------------------------------------------
// This function sorts the mission plans based on their ratings.
//-----------------------------------------------------------------------
vector<MemoryManager::MissionPlan_t> Planner::sortMissionPlans_(
    vector<MemoryManager::MissionPlan_t> missionPlans)
{
    vector<MemoryManager::MissionPlan_t> newMissionPlans, tmpMissionPlans;
    int i, heighestRating, heighestMissionIndex;

    while ((int)(missionPlans.size()) > 0)
    {
	heighestRating = -1;
	heighestMissionIndex = -1;

	for (i = 0; i < (int)(missionPlans.size()); i++)
	{
	    if (missionPlans[i].rating > heighestRating)
	    {
		heighestMissionIndex = i;
		heighestRating = missionPlans[i].rating;
	    }
	}

	if (heighestMissionIndex != -1)
	{
	    tmpMissionPlans.clear();

	    for (i = 0; i < (int)(missionPlans.size()); i++)
	    {
		if (i == heighestMissionIndex)
		{
		    newMissionPlans.push_back(missionPlans[i]);
		}
		else
		{
		    tmpMissionPlans.push_back(missionPlans[i]);
		}
	    }

	    missionPlans = tmpMissionPlans;
	}
	else
	{
	    break;
	}
    }

    return newMissionPlans;
}

//-----------------------------------------------------------------------
// This function adds one mission to another to create a seqeubce of
// sub-missions.
//-----------------------------------------------------------------------
vector<MemoryManager::MissionPlan_t> Planner::appendMissionPlans_(
    vector<MemoryManager::MissionPlan_t> bigMissionPlans,
    vector<MemoryManager::MissionPlan_t> missionPlans,
    int trackingNumber)
{
    MemoryManager::MissionPlan_t newBigMissionPlan;
    vector<MemoryManager::MissionPlan_t> newBigMissionPlans;
    CBRPlanner_Precondition_t precondition;
    CBRPlanner_Effect_t effect;
    char bufIndex[1024];
    int i, j, k, l, m, numRobots, lastIndex;

    if ((int)(bigMissionPlans.size()) == 0)
    {
	return missionPlans;
    }

    if ((int)(missionPlans.size()) == 0)
    {
	return bigMissionPlans;
    }

    // bufIndex will be added to any index that would possibly conflict
    // with any existing index.
    sprintf(bufIndex, "%d", trackingNumber);

    for (i = 0; i < (int)(bigMissionPlans.size()); i++)
    {
	for (j = 0; j < (int)(missionPlans.size()); j++)
	{
	    newBigMissionPlan = bigMissionPlans[i];

	    // Take the average of two ratings.
	    newBigMissionPlan.rating = (newBigMissionPlan.rating + missionPlans[j].rating)/2;

	    // Append the list of the data index as well.
	    newBigMissionPlan.dataIndexList = mergeDataIndexLists_(
		newBigMissionPlan.dataIndexList,
		missionPlans[j].dataIndexList);

	    numRobots = newBigMissionPlan.plans.size();

	    if (numRobots != (int)(missionPlans[j].plans.size()))
	    {
		// Number of robots does not match. Do not append
		// this mission.
		continue;
	    }

	    for (k = 0; k < numRobots; k++)
	    {
		lastIndex = newBigMissionPlan.plans[k].size() - 1;
		
		if ((lastIndex > 0) && ((int)(missionPlans[j].plans[k].size()) > 0))
		{
		    effect.name = STRING_IMMEDIATE_;
		    effect.index = missionPlans[j].plans[k][0].index;
		    effect.index += bufIndex;
		    newBigMissionPlan.plans[k][lastIndex].effects.push_back(effect);

		    precondition.name = STRING_IMMEDIATE_;
		    missionPlans[j].plans[k][0].preconditions.push_back(precondition);
		}
		
		for (l = 0; l < (int)(missionPlans[j].plans[k].size()); l++)
		{
		    missionPlans[j].plans[k][l].index += bufIndex;

		    for (m = 0; m < (int)(missionPlans[j].plans[k][l].effects.size()); m++)
		    {
			missionPlans[j].plans[k][l].effects[m].index += bufIndex;
		    }
		    
		    newBigMissionPlan.plans[k].push_back(missionPlans[j].plans[k][l]);
		}
	    }

	    newBigMissionPlans.push_back(newBigMissionPlan);
	}
    }

    return newBigMissionPlans;
}

//-----------------------------------------------------------------------
// This function merges two lists of data indexes.
//-----------------------------------------------------------------------
vector<int> Planner::mergeDataIndexLists_(
    vector<int> dataIndexListA,
    vector<int> dataIndexListB)
{
    int i;

    for (i = 0; i < (int)(dataIndexListB.size()); i++)
    {
	dataIndexListA.push_back(dataIndexListB[i]);
    }

    dataIndexListA = trimRedundantDataIndex_(dataIndexListA);

    return dataIndexListA;
}

//-----------------------------------------------------------------------
// This function gets rid of redundant indexes in the list.
//-----------------------------------------------------------------------
vector<int> Planner::trimRedundantDataIndex_(vector<int> dataIndexList)
{
    int i, j;
    vector<int> newDataIndexList;
    bool dataExists;

    for (i = 0; i < (int)(dataIndexList.size()); i++)
    {
	dataExists = false;

	for (j = 0; j < (int)(newDataIndexList.size()); j++)
	{
	    if (dataIndexList[i] == newDataIndexList[j])
	    {
		dataExists = true;
		break;
	    }
	}

	if (!dataExists)
	{
	    newDataIndexList.push_back(dataIndexList[i]);
	}
    }

    return newDataIndexList;
}

//-----------------------------------------------------------------------
// This function extracts the maximum velocity from the features list.
//-----------------------------------------------------------------------
string Planner::extractMaxVelocityValue_(MemoryManager::Features_t features)
{
    string maxVelValue = EMPTY_STRING_;
    int i;

    for (i = 0; i < (int)(features.size()); i++)
    {
	if (features[i].name == FEATURE_NAME_MAX_VELOCITY_)
	{
	    maxVelValue = features[i].value;
	    break;
	}
    }

    return maxVelValue;
}

//-----------------------------------------------------------------------
// This function extracts the aggressiveness from the features list.
//-----------------------------------------------------------------------
string Planner::extractAggressivenessValue_(MemoryManager::Features_t features)
{
    string aggressivenessValue = EMPTY_STRING_;
    int i;

    for (i = 0; i < (int)(features.size()); i++)
    {
	if (features[i].name == FEATURE_NAME_AGGRESSIVENESS_)
	{
	    aggressivenessValue = features[i].value;
	    break;
	}
    }

    return aggressivenessValue;
}

//-----------------------------------------------------------------------
// This function extracts the number of robots from the features list.
//-----------------------------------------------------------------------
int Planner::extractNumRobots_(MemoryManager::Features_t features)
{
    int i, numRobots;

    for (i = 0; i < (int)(features.size()); i++)
    {
	if (features[i].name == FEATURE_NAME_NUM_ROBOTS_)
	{
	    numRobots = atoi(features[i].value.c_str());
	    break;
	}
    }

    return numRobots;
}

//-----------------------------------------------------------------------
// This function extracts sub-features (features that correspond to
// sub-missions).
//-----------------------------------------------------------------------
vector<MemoryManager::Features_t> Planner::extractSubFeatures_(MemoryManager::Features_t features)
{
    vector<MemoryManager::Features_t> subFeaturesList;
    MemoryManager::Features_t subFeatures, globalFeatures;
    int i, j;
    bool firstTaskFound = false;
    const bool DEBUG = true;

    for (i = 0; i < (int)(features.size()); i++)
    {
	if (features[i].name == FEATURE_NAME_TASK_NAME_)
	{
	    if (firstTaskFound)
	    {
		subFeaturesList.push_back(subFeatures);
	    }
	    else
	    {
		firstTaskFound = true;
	    }

	    subFeatures = globalFeatures;
	    subFeatures.push_back(features[i]);
	}
	else
	{
	    if (firstTaskFound)
	    {
		subFeatures.push_back(features[i]);
	    }
	    else
	    {
		globalFeatures.push_back(features[i]);
	    }
	}
    }

    if ((int)(subFeatures.size()) > 0)
    {
	subFeaturesList.push_back(subFeatures);
    }

    if (DEBUG)
    {
	gDebugger->printfLine();
	gDebugger->printfDebug("EXTRACTED SUB-FEATURES:\n");

	for (i = 0; i < (int)(subFeaturesList.size()); i++)
	{
	    gDebugger->printfLine();
	    for (j = 0; j < (int)(subFeaturesList[i].size()); j++)
	    {
		gDebugger->printfDebug("%s  ", subFeaturesList[i][j].name.c_str());
		gDebugger->printfDebug("%s  ", subFeaturesList[i][j].value.c_str());
		gDebugger->printfDebug("%d ", subFeaturesList[i][j].nonIndex);
		gDebugger->printfDebug("%.2f\n", subFeaturesList[i][j].weight);
	    }
	}
	gDebugger->printfLine();
    }

    return subFeaturesList;
}

//-----------------------------------------------------------------------
// This function extracts the coordinates from the features list.
//-----------------------------------------------------------------------
MemoryManager::Coordinates_t Planner::extractCoordinates_(MemoryManager::Features_t features)
{
    MemoryManager::Coordinates_t coordinates;
    bool xFound = false, yFound = false;
    int i;

    for (i = 0; i < (int)(features.size()); i++)
    {
	if (features[i].name == FEATURE_NAME_START_X_)
	{
	    coordinates.x = atof(features[i].value.c_str());
	    xFound = true;
	}
	else if (features[i].name == FEATURE_NAME_START_Y_)
	{
	    coordinates.y = atof(features[i].value.c_str());
	    yFound = true;
	}

	if ((xFound) && (yFound))
	{
	    break;
	}
    }

    return coordinates;
}

//-----------------------------------------------------------------------
// This function extracts the task name from the features list.
//-----------------------------------------------------------------------
string Planner::extractTaskName_(MemoryManager::Features_t features)
{
    string taskName = EMPTY_STRING_;
    int i;

    for (i = 0; i < (int)(features.size()); i++)
    {
	if (features[i].name == FEATURE_NAME_TASK_NAME_)
	{
	    taskName = features[i].value;
	    break;
	}
    }

    return taskName;
}

//-----------------------------------------------------------------------
// This function extracts the localization value from the features list,
// and checks whether maps should be used or not.
//-----------------------------------------------------------------------
bool Planner::extractUseMaps_(MemoryManager::Features_t features)
{
    bool useMaps = false;
    int i;

    for (i = 0; i < (int)(features.size()); i++)
    {
	if (features[i].name == FEATURE_NAME_LOCALIZATION_)
	{
	    if (features[i].value == FEATURE_VALUE_USEMAPS_)
	    {
		useMaps = true;
	    }

	    break;
	}
    }

    return useMaps;
}

//-----------------------------------------------------------------------
// This function extracts the environment value from the features list,
// and checks whether the should be running indoor or not.
//-----------------------------------------------------------------------
bool Planner::extractIndoor_(MemoryManager::Features_t features)
{
    bool indoor = false;
    int i;

    for (i = 0; i < (int)(features.size()); i++)
    {
	if (features[i].name == FEATURE_NAME_ENVIRONMENT_)
	{
	    if (features[i].value == FEATURE_VALUE_INDOOR_)
	    {
		indoor = true;
	    }

	    break;
	}
    }

    return indoor;
}

//-----------------------------------------------------------------------
// This function will sort the feature list based on the weights.
//-----------------------------------------------------------------------
MemoryManager::Features_t Planner::sortFeatures_(MemoryManager::Features_t features)
{
    MemoryManager::Features_t tmpFeatures, newFeatures;
    double heighestWeight = -1.0;
    int i, heighestFeatureIndex = -1;

    while ((int)(features.size()) > 0)
    {
	heighestWeight = -1.0;
	heighestFeatureIndex = -1;

	for (i = 0; i < (int)(features.size()); i++)
	{
	    if (features[i].weight > heighestWeight)
	    {
		heighestFeatureIndex = i;
		heighestWeight = features[i].weight;
	    }
	}

	if (heighestFeatureIndex != -1)
	{
	    tmpFeatures.clear();

	    for (i = 0; i < (int)(features.size()); i++)
	    {
		if (i == heighestFeatureIndex)
		{
		    newFeatures.push_back(features[i]);
		}
		else
		{
		    tmpFeatures.push_back(features[i]);
		}
	    }

	    features = tmpFeatures;
	}
	else
	{
	    break;
	}
    }

    return newFeatures;
}

//-----------------------------------------------------------------------
// This function will remove all the nonIndex features from the list.
//-----------------------------------------------------------------------
MemoryManager::Features_t Planner::filterNonIndexFeatures_(MemoryManager::Features_t features)
{
    MemoryManager::Features_t filteredFeatures;
    int i;

    for (i = 0; i < (int)(features.size()); i++)
    {
	if (!(features[i].nonIndex))
	{
	    filteredFeatures.push_back(features[i]);
	}
    }

    return filteredFeatures;
}

//-----------------------------------------------------------------------
//  This function adds StartSubMission state in the beginning
//-----------------------------------------------------------------------
MemoryManager::MissionPlan_t Planner::addStartSubMissionState_(
    MemoryManager::MissionPlan_t missionPlan,
    string taskName,
    MemoryManager::Coordinates_t coordinates,
    bool useMaps,
    bool indoor,
    int startSubMissionOperatorIndex)
{
    vector<CBRPlanner_PlanSequence_t> newPlans;
    CBRPlanner_PlanSequence_t plan;
    CBRPlanner_PlanOperator_t startSubMissionOperator;
    int i, j;

    startSubMissionOperator = createStartSubMissionOperator_(
	missionPlan,
	taskName,
	coordinates,
	useMaps,
	indoor,
	startSubMissionOperatorIndex);

    for (i = 0; i < (int)(missionPlan.plans.size()); i++)
    {
	plan.clear();
	startSubMissionOperator.effects[0].index = missionPlan.plans[i][0].index;
	plan.push_back(startSubMissionOperator);

	for (j = 0; j < (int)(missionPlan.plans[i].size()); j++)
	{
	    plan.push_back(missionPlan.plans[i][j]);
	}

	newPlans.push_back(plan);
    }

    missionPlan.plans = newPlans;

    return missionPlan;
}

//-----------------------------------------------------------------------
//  This function adds Start state in the beginning
//-----------------------------------------------------------------------
MemoryManager::MissionPlan_t Planner::addStartState_(
    MemoryManager::MissionPlan_t missionPlan)
{
    vector<CBRPlanner_PlanSequence_t> newPlans;
    CBRPlanner_PlanSequence_t plan;
    CBRPlanner_PlanOperator_t startOperator;
    CBRPlanner_Effect_t effect;
    CBRPlanner_Precondition_t precondition;
    int i, j;

    startOperator = createStartOperator_();
    effect.name = STRING_IMMEDIATE_;
    precondition.name = STRING_IMMEDIATE_;

    for (i = 0; i < (int)(missionPlan.plans.size()); i++)
    {
	plan.clear();
	plan.push_back(startOperator);

	if ((int)(missionPlan.plans[i].size()) > 0)
	{
	    effect.index = missionPlan.plans[i][0].index;
	    plan[0].effects.push_back(effect);

	    missionPlan.plans[i][0].preconditions.push_back(precondition);

	    for (j = 0; j < (int)(missionPlan.plans[i].size()); j++)
	    {
		plan.push_back(missionPlan.plans[i][j]);
	    }
	}
	newPlans.push_back(plan);
    }

    missionPlan.plans = newPlans;

    return missionPlan;
}

//-----------------------------------------------------------------------
//  This function creates Start operator.
//-----------------------------------------------------------------------
CBRPlanner_PlanOperator_t Planner::createStartOperator_(void)
{
    CBRPlanner_PlanOperator_t startOperator;

    startOperator.name = STRING_START_STATE_;
    startOperator.index = STRING_START_STATE_INDEX_;

    return startOperator;}

//-----------------------------------------------------------------------
//  This function creates StartSubMission operator.
//-----------------------------------------------------------------------
CBRPlanner_PlanOperator_t Planner::createStartSubMissionOperator_(
    MemoryManager::MissionPlan_t missionPlan,
    string taskName,
    MemoryManager::Coordinates_t coordinates,
    bool useMaps,
    bool indoor,
    int startSubMissionOperatorIndex)
{
    CBRPlanner_PlanOperator_t startSubMissionOperator;
    CBRPlanner_Parameter_t parameter;
    CBRPlanner_Effect_t effect;
    CBRPlanner_Precondition_t precondition;
    char buf[1024];

    startSubMissionOperator.name = STRING_START_SUB_MISSION_;
    sprintf(
	buf,
	"%s%d",
	STRING_START_SUB_MISSION_INDEX_.c_str(),
	startSubMissionOperatorIndex);
    startSubMissionOperator.index = buf;

    parameter.name = STRING_SUB_MISSION_NAME_;
    sprintf(buf, "{\"%s\"}", taskName.c_str());
    parameter.value = buf;
    startSubMissionOperator.parameters.push_back(parameter);
    
    parameter.name = STRING_GOAL_LOCATION_;
    sprintf(buf, "{%.2f,+%.2f}", coordinates.x, coordinates.y);
    parameter.value = buf;
    startSubMissionOperator.parameters.push_back(parameter);

    parameter.name = STRING_MOVE_TO_LOCATION_GAIN_;
    sprintf(buf, "{%.2f}", DEFAULT_MOVE_TO_LOCATION_GAIN_);
    parameter.value = buf;
    startSubMissionOperator.parameters.push_back(parameter);

    parameter.name = STRING_WANDER_GAIN_;
    sprintf(buf, "{%.2f}", DEFAULT_WANDER_GAIN_);
    parameter.value = buf;
    startSubMissionOperator.parameters.push_back(parameter);

    parameter.name = STRING_AVOID_OBSTACLE_GAIN_;
    if (indoor)
    {
	sprintf(buf, "{%.2f}", DEFAULT_INDOOR_AVOID_OBSTACLE_GAIN_);
    }
    else
    {
	sprintf(buf, "{%.2f}", DEFAULT_OUTDOOR_AVOID_OBSTACLE_GAIN_);
    }
    parameter.value = buf;
    startSubMissionOperator.parameters.push_back(parameter);

    parameter.name = STRING_AVOID_OBSTACLE_SPHERE_;
    if (indoor)
    {
	sprintf(buf, "{%.2f}", DEFAULT_INDOOR_AVOID_OBSTACLE_SPHERE_);
    }
    else
    {
	sprintf(buf, "{%.2f}", DEFAULT_OUTDOOR_AVOID_OBSTACLE_SPHERE_);
    }
    parameter.value = buf;
    startSubMissionOperator.parameters.push_back(parameter);

    parameter.name = STRING_AVOID_OBSTACLE_SAFETY_MARGIN_;
    sprintf(buf, "{%.2f}", DEFAULT_AVOID_OBSTACLE_SAFETY_MARGIN_);
    parameter.value = buf;
    startSubMissionOperator.parameters.push_back(parameter);

    precondition.name = STRING_IMMEDIATE_;
    startSubMissionOperator.preconditions.push_back(precondition);

    effect.name = STRING_SUB_MISSION_READY_;
    parameter.name = STRING_SUB_MISSION_DEPLOYMENT_;
    if (useMaps)
    {
	parameter.value = STRING_DEPLOY_FIRST_;
    }
    else
    {
	parameter.value = STRING_EXECUTE_IMMEDIATELY_;
    }
    effect.parameters.push_back(parameter);

    parameter.name = STRING_GOAL_TOLERANCE_;
    sprintf(buf, "{%.2f}", DEFAULT_GOAL_TOLERANCE_);
    parameter.value = buf;
    effect.parameters.push_back(parameter);

    parameter.name = STRING_GOAL_LOCATION_;
    sprintf(buf, "{%.2f,+%.2f}", coordinates.x, coordinates.y);
    parameter.value = buf;
    effect.parameters.push_back(parameter);

    startSubMissionOperator.effects.push_back(effect);

    return startSubMissionOperator;
}

//-----------------------------------------------------------------------
// This function adjusts the number of robots in the mission plan.
//-----------------------------------------------------------------------
MemoryManager::MissionPlan_t Planner::adjustNumRobots_(MemoryManager::MissionPlan_t missionPlan, int numRobots)
{
    CBRPlanner_PlanSequence_t plan;
    int i, j, k;
    char bufIndex[1024];

    // Duplicate existing robot if it has less than specified number of
    // robots.
    for (i = 0; i < numRobots; i++)
    {
	if ((int)(missionPlan.plans.size()) < (i+1))
	{
	    plan = missionPlan.plans[0];
	    sprintf(bufIndex, "%d", i);

	    for (j = 0; j < (int)(plan.size()); j++)
	    {
		plan[j].index += bufIndex;

		for (k = 0; k < (int)(plan[j].effects.size()); k++)
		{
		    plan[j].effects[k].index += bufIndex;
		}
	    }

	    missionPlan.plans.push_back(plan);
	}
    }

    // Delete excessive robot plans.
    while ((int)(missionPlan.plans.size()) > numRobots)
    {
	missionPlan.plans.pop_back();
    }

    return missionPlan;
}

//-----------------------------------------------------------------------
// This function saves new mission plan in the memory.
//-----------------------------------------------------------------------
vector<int> Planner::saveNewMissionPlan(
    MemoryManager::MissionPlan_t missionPlan,
    MemoryManager::Features_t features,
    bool hasStartSubMission)
{
    MemoryManager::MissionPlanList_t missionPlans;
    vector<MemoryManager::Features_t> featuresList;
    vector<int> dataIndexList;
    const vector<int> EMPTY_DATA_INDEX_LIST;
    vector<string> featureNameList;
    int i, dataIndex;

    if (hasStartSubMission)
    {
	missionPlans = separateSubMissions_(
	    missionPlan,
	    features,
	    &featuresList);
    }
    else
    {
	missionPlans.push_back(missionPlan);
	featuresList.push_back(features);
    }

    if (missionPlans.size() != featuresList.size())
    {
	gWindows->printfPlannerWindow(
	    "Warning: Inadequate mission plan. Not saved.\n");
	return EMPTY_DATA_INDEX_LIST;
    }

    for (i = 0; i < (int)(missionPlans.size()); i++)
    {
	features = filterNonIndexFeatures_(featuresList[i]);
	features = sortFeatures_(features);
	dataIndex = gCBRPlanner->memoryManager()->saveNewMissionPlan(
	    missionPlans[i],
	    features);
	dataIndexList.push_back(dataIndex);
    }

    return dataIndexList;
}

//-----------------------------------------------------------------------
// This function saves new mission plan in the memory.
//-----------------------------------------------------------------------
MemoryManager::MissionPlanList_t Planner::separateSubMissions_(
    MemoryManager::MissionPlan_t missionPlan,
    MemoryManager::Features_t features,
    vector<MemoryManager::Features_t> *featuresList)
{
    MemoryManager::MissionPlanList_t missionPlans;
    const MemoryManager::MissionPlanList_t EMPTY_MISSION_PLANS;
    MemoryManager::MissionPlan_t tmpMissionPlan;
    CBRPlanner_PlanSequence_t plan;
    string taskName;
    int i, j, lastIndex, missionCount;
    bool firstStartSubMissionFound = false;
    
    tmpMissionPlan = missionPlan;
    featuresList->clear();

    for (i = 0; i < (int)(missionPlan.plans.size()); i++)
    {
	// Check for all the robots.
	firstStartSubMissionFound = false;
	tmpMissionPlan.plans.clear();
	plan.clear();
	missionCount = 0;

	for (j = 0; j < (int)(missionPlan.plans[i].size()); j++)
	{
	    // Check the plan sequence for each robot.
	    if (missionPlan.plans[i][j].name == STRING_START_SUB_MISSION_)
	    {
		if (firstStartSubMissionFound)
		{
		    plan[0].preconditions.clear();
		    lastIndex = (int)(plan.size()) - 1;
		    plan[lastIndex].effects.clear();

		    if (i == 0)
		    {
			// Case when this is for the first robot.
			tmpMissionPlan.plans.push_back(plan);
			missionPlans.push_back(tmpMissionPlan);

			features = renameTaskInFeatures_(features, taskName);
			featuresList->push_back(features);
		    }
		    else
		    {
			// Case when this is for the second or higher robot.
			if (((int)(missionPlans.size()) > missionCount) &&
			    ((int)(featuresList->size()) > missionCount))
			{
			    if (getTaskNameInFeatures_((*featuresList)[missionCount]) != taskName)
			    {
				gWindows->printfPlannerWindow(
				    "Warning: Inadequate mission plan. Not saved.\n");
				return EMPTY_MISSION_PLANS;
			    }

			    missionPlans[missionCount].plans.push_back(plan);
			}
			else
			{
			    gWindows->printfPlannerWindow(
				"Warning: Inadequate mission plan. Not saved.\n");
			    return EMPTY_MISSION_PLANS;
			}
		    }

		    missionCount++;
		    plan.clear();
		    tmpMissionPlan.plans.clear();
		}
		else
		{
		    firstStartSubMissionFound = true;
		}

		taskName = extractTaskNameFromStartSubMission_(missionPlan.plans[i][j]);
	    }
	    else
	    {
		if (firstStartSubMissionFound)
		{
		    plan.push_back(missionPlan.plans[i][j]);
		}
	    }
	}

	if (firstStartSubMissionFound)
	{
	    if (i == 0)
	    {
		if ((int)(plan.size()) > 0)
		{
		    // Case when this is for the first robot.
		    tmpMissionPlan.plans.push_back(plan);
		    missionPlans.push_back(tmpMissionPlan);

		    features = renameTaskInFeatures_(features, taskName);
		    featuresList->push_back(features);
		}
	    }
	    else
	    {
		// Case when this is for the second or higher robot.
		if (((int)(missionPlans.size()) > missionCount) &&
		    ((int)(featuresList->size()) > missionCount))
		{
		    if (getTaskNameInFeatures_((*featuresList)[missionCount]) != taskName)
		    {
			gWindows->printfPlannerWindow(
			    "Warning: Inadequate mission plan. Not saved.\n");
			return EMPTY_MISSION_PLANS;
		    }

		    missionPlans[missionCount].plans.push_back(plan);
		}
		else
		{
		    gWindows->printfPlannerWindow(
			"Warning: Inadequate mission plan. Not saved.\n");
		    return EMPTY_MISSION_PLANS;
		}
	    }
	}
    }

    return missionPlans;
}

//-----------------------------------------------------------------------
// This function changes the name of the task in the feature.
//-----------------------------------------------------------------------
string Planner::getTaskNameInFeatures_(MemoryManager::Features_t features)
{
    string taskName = EMPTY_STRING_;
    int i;

    for (i = 0; i < (int)(features.size()); i++)
    {
	if (features[i].name == FEATURE_NAME_TASK_NAME_)
	{
	    taskName = features[i].value;
	}
    }

    return taskName;
}

//-----------------------------------------------------------------------
// This function changes the name of the task in the feature.
//-----------------------------------------------------------------------
MemoryManager::Features_t Planner::renameTaskInFeatures_(
    MemoryManager::Features_t features,
    string taskName)
{
    int i;

    for (i = 0; i < (int)(features.size()); i++)
    {
	if (features[i].name == FEATURE_NAME_TASK_NAME_)
	{
	    features[i].value = taskName;
	}
    }

    return features;
}

//-----------------------------------------------------------------------
// This function extracts the task name from StartSubMission state.
//-----------------------------------------------------------------------
string Planner::extractTaskNameFromStartSubMission_(
    CBRPlanner_PlanOperator_t startSubMissionOperator)
{
    int i;
    string taskName = EMPTY_STRING_;
    CBRPlanner_Parameters_t parameters;

    parameters = startSubMissionOperator.parameters;

    for (i = 0; i < (int)(parameters.size()); i++)
    {
	if (parameters[i].name == STRING_SUB_MISSION_NAME_)
	{
	    taskName = extractValueContent_(parameters[i].value, true);
	}
    }

    return taskName;
}

//-----------------------------------------------------------------------
// This function extracts the task name from StartSubMission state.
//-----------------------------------------------------------------------
string Planner::extractValueContent_(string value, bool withoutQuote)
{
    string content;
    string::size_type index1, index2;


    index1 = value.find("{");
    index2 = value.find("}");

    if ((index1 != string::npos) && (index2 != string::npos))
    {
	index1++;
	content = value.substr(index1, index2-index1);

	if (withoutQuote)
	{
	    index1 = content.find("\"");
	    index2 = content.find("\"", index1+1);

	    if ((index1 != string::npos) && (index1 != string::npos))
	    {
		index1++;
		content = content.substr(index1, index2-index1);
	    }
	}
    }

    return content;
}

/**********************************************************************
 * $Log: planner.cc,v $
 * Revision 1.1  2003/04/06 08:43:44  endo
 * Initial revision
 *
 *
 **********************************************************************/
