/**********************************************************************
 **                                                                  **
 **                              cbrclient.cc                        **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This file contains functions for the CBRClient class, whose     **
 **  job is to act as an interface between CfgEdit and CBRServer.    **
 **                                                                  **
 **  Copyright 2002 - 2003 Georgia Tech Research Corporation         **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: cbrclient.cc,v 1.2 2003/04/06 08:48:14 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <fstream.h>
#include <iostream.h>
#include <string>
#include <sys/types.h>
#include <sys/socket.h>
#include <linux/un.h>

#include "cbrclient.h"
#include "cbrplanner_protocol.h"
#include "mission_expert.h"
#include "globals.h"
#include "EventLogging.h"

//-----------------------------------------------------------------------
// Constatns
//-----------------------------------------------------------------------
// Feature file related constants.
const string CBRClient::STRING_TASK_ = "Task";
const string CBRClient::STRING_START_X_ = "StartXCoord";
const string CBRClient::STRING_START_Y_ = "StartYCoord";

// Others
const string CBRClient::RCTABLE_SERVERSOCKET_STRING_= "CBRServerSocketName";
const string CBRClient::EMPTY_STRING_ = "";
//-----------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Constructor for CBRClient class.
//-----------------------------------------------------------------------
CBRClient::CBRClient(void)
{
    socketFd_ = -1;

    clientInfo_.type = CBRPLANNER_CLIENT_CFGEDIT;
    clientInfo_.pid = getpid();
    sprintf(clientInfo_.name, "CfgEdit");
}

//-----------------------------------------------------------------------
// Distructor for CBRClient class.
//-----------------------------------------------------------------------
CBRClient::~CBRClient(void)
{
    sendCommandToServer_(CBRPLANNER_COMM_CMD_CLIENTEXIT);
}

//-----------------------------------------------------------------------
// This function initializes the CBRClient.
//-----------------------------------------------------------------------
bool CBRClient::initialize(void)
{
    bool status;

    // Connect to server.
    status = connectToServer_();

    if (!status)
    {
	fprintf(stderr, "Error(cfgedit): CBRClient::initialize(). Failed to connect with the server.\n");
	return false;
    }
    
    return true;
}

//-----------------------------------------------------------------------
// This function uploads the FSA (tasks and triggers) names to the server.
//-----------------------------------------------------------------------
void CBRClient::saveFSANames(config_fsa_lists_t fsaLists)
{
    sendFSANamesToServer_(fsaLists);
}

//-----------------------------------------------------------------------
// This function connects to the server through a socket.
//-----------------------------------------------------------------------
bool CBRClient::connectToServer_(void)
{
    rc_chain *serverSocketNameList;
    char *serverSocketName = NULL;
    int status;
    char buf[UNIX_PATH_MAX];
    sockaddr_un localAddress;

    // Get the user defined name for the socket.
    serverSocketNameList = (rc_chain *)rc_table.get(RCTABLE_SERVERSOCKET_STRING_.c_str());

    if (serverSocketNameList != NULL)
    {
	serverSocketNameList->first(&serverSocketName);
    }

    if (serverSocketName == NULL)
    {
	// Create a default one using the user name.
	sprintf(buf, "/tmp/%s-cbrplanner.socket", getenv("USER"));
	serverSocketName = strdup(buf);
    }

    // Open the socket.
    socketFd_ = socket(AF_UNIX, SOCK_STREAM, 0);

    if (socketFd_ < 0)
    {
	fprintf(stderr, "Error(cfgedit): CBRClient::connectToServer(). Socket could not be opened.\n");
	return false;
    }

    // Connect to the socket.
    localAddress.sun_family = AF_UNIX;
    strcpy(localAddress.sun_path, serverSocketName);
    status = connect(socketFd_, (sockaddr *)&localAddress, sizeof(localAddress));

    if (status < 0)
    {
	fprintf(stderr, "Error(cfgedit): CBRClient::initialize(). Connection to the server failed.\n");
	return false;
    }

    fprintf(stdout, "CBR Planner connected.\n");
    fflush(stdout);

    sendCommandToServer_(CBRPLANNER_COMM_CMD_CLIENTINIT);
    sendDataToServer_((void *)&clientInfo_, sizeof(CBRPlanner_ClientInfo_t));

    return true;
}

//-----------------------------------------------------------------------
// This function saves the new mission plan into the CBRServer.
//-----------------------------------------------------------------------
vector<int> CBRClient::saveMissionPlan(
    vector<MExpFeature_t> globalFeatures_,
    MExpMissionTask_t task,
    double taskWeight,
    string solutionFileName)
{
    int i, j, k;
    char buf[1024];
    MExpFeature_t globalFeature;
    MExpFeature_t localFeature;
    CBRPlanner_MissionPlan_t missionplan;
    CBRPlanner_Feature_t feature;
    vector<CBRPlanner_Feature_t> features;
    vector<int> dataIndexList;

    // Describe the mission plan type.
    missionplan.type = CBRPLANNER_MISSIONPLAN_NEW;

    // Describe the mission plan rating.
    missionplan.rating = 0;

    // Open the solution file to read.
    ifstream istr(solutionFileName.c_str());

    if (istr.bad()) 
    {
	fprintf
	    (stderr,
	     "Error(cfgedit): CBRClient::saveMissionPlan(). file %s not readable",
	     solutionFileName.c_str());
	return dataIndexList;
    }

    istr.seekg (0, ios::end);
    missionplan.dataSize = istr.tellg();
    istr.seekg (0, ios::beg);

    // To append NULL at the end.
    missionplan.dataSize++; 

    // Read in the solution.
    missionplan.data = new char[missionplan.dataSize];
    istr.read (missionplan.data, missionplan.dataSize-1);

    // Append a null, to make it a char string.
    missionplan.data[missionplan.dataSize-1] = '\0';

    // Close the solution file.
    istr.close();

    // Number of features.
    i = 0;

    // Save the global features.
    for (j = 0; j<(int)globalFeatures_.size(); j++)
    {
	globalFeature = globalFeatures_[j];
	feature.name = globalFeature.name;
	feature.nonIndex = globalFeature.nonIndex;
	feature.weight = globalFeature.weight;

	switch (globalFeature.optionType) {
	    
	case MEXP_FEATURE_OPTION_TOGGLE:
	    for (k = 0; k < ((int)(globalFeature.options.size())); k++)
	    {
		if (globalFeature.selectedOption == k)
		{
		    feature.value = globalFeature.options[k].value;
		    features.push_back(feature);
		    break;
		}
	    }
	    break;

	case MEXP_FEATURE_OPTION_SLIDER1:
	    sprintf(buf, "%.2f", globalFeature.selectedOption);
	    feature.value = buf;
	    features.push_back(feature);
	    break;

	case MEXP_FEATURE_OPTION_SLIDER10:
	    sprintf(buf, "%d", (int)(globalFeature.selectedOption));
	    feature.value = buf;
	    features.push_back(feature);
	    break;
	}
    }

    // Save the task name.
    feature.name = STRING_TASK_;
    feature.value = task.name;
    feature.nonIndex = false;
    feature.weight = taskWeight;
    features.push_back(feature);

    // Save other local features.
    for (j = 0; j < ((int)(task.localFeatures.size())); j++)
    {
	localFeature = task.localFeatures[j];
	feature.name = localFeature.name;
	feature.nonIndex = localFeature.nonIndex;
	feature.weight = localFeature.weight;

	switch (localFeature.optionType) {
	    
	case MEXP_FEATURE_OPTION_TOGGLE:
	    for (k = 0; k < ((int)(localFeature.options.size())); k++)
	    {
		if (localFeature.selectedOption == k)
		{
		    feature.value = localFeature.options[k].value;
		    features.push_back(feature);
		    break;
		}
	    }
	    break;

	case MEXP_FEATURE_OPTION_SLIDER1:
	    sprintf(buf, "%.2f", localFeature.selectedOption);
	    feature.value = buf;
	    features.push_back(feature);
	    break;

	case MEXP_FEATURE_OPTION_SLIDER10:
	    sprintf(buf, "%d", (int)(localFeature.selectedOption));
	    feature.value = buf;
	    features.push_back(feature);
	    break;
	}
    }

    dataIndexList = sendMissionPlanToServer_(missionplan, features);

    return dataIndexList;
}

//-----------------------------------------------------------------------
// This function retrieve a mission planfrom the CBRServer based on the
// features defined in the specified feature file.
//-----------------------------------------------------------------------
vector<MExpRetrievedMission_t> CBRClient::retrieveMissionPlan(
    string featureFileName,
    vector<MExpMissionTask_t> defaultTaskList,
    vector<MExpFeature_t> defaultGlobalFeatures,
    vector<MExpFeature_t> defaultLocalFeatures,
    double taskWeight,
    int maxRating,
    string *overlayFileName)
{
    MExpRetrievedMission_t retrievedMission;
    vector<MExpRetrievedMission_t> retrievedMissions;
    const vector<MExpRetrievedMission_t> EMPTY_RETRIEVED_MISSIONS;
    const MExpRetrievedMission_t EMPTY_RETRIEVED_MISSION;
    FILE *featureFile;
    char buf[1024], buf1[1024], buf2[1024], buf3[1024];
    string bufString1;
    int i, taskNum, taskId, localFeatureId, globalFeatureId;
    float option;
    bool dataExists = false, hadError = false;
    vector<CBRPlanner_MissionPlan_t> missionplans;
    CBRPlanner_Feature_t feature, globalFeature;
    vector<CBRPlanner_Feature_t> features, globalFeatures;
    
    featureFile = fopen(featureFileName.c_str(), "r");
    if (featureFile == NULL)
    {
	return EMPTY_RETRIEVED_MISSIONS;
    }

    while(!feof(featureFile))
    {
	if (fscanf(featureFile, "%s", buf1) == EOF)
	{
	    break;
	}

	bufString1 = buf1;

	if (bufString1 ==  MEXP_STRING_OVERLAY)
	{
	    // Save the overlay name specified.
	    fscanf(featureFile, "%s", buf2);
	    *overlayFileName = buf2;
	}
	else if (bufString1 == MEXP_STRING_GLOBALFEATURE)
	{
	    // Build a list of the global preferences.  
	    // The task entries will append them to each feature list.

	    // Get the preference ID (e.g., NumberOfRobots, Time, etc.)
	    if (fscanf(featureFile, "%s", buf2) == EOF)
	    {
		hadError = true;
		break;
	    }
	    globalFeatureId = atoi(buf2);

	    // Get the preference value
	    if (fscanf(featureFile, "%s", buf3) == EOF)
	    {
		hadError = true;
		break;
	    }
	    option = atof(buf3);

	    for (i = 0; i < ((int)(defaultGlobalFeatures.size())); i++)
	    {
		if (defaultGlobalFeatures[i].id == globalFeatureId)
		{
		    globalFeature.name = defaultGlobalFeatures[i].name;
		    globalFeature.nonIndex = defaultGlobalFeatures[i].nonIndex;
		    globalFeature.weight = defaultGlobalFeatures[i].weight;

		    switch (defaultGlobalFeatures[i].optionType) {

		    case MEXP_FEATURE_OPTION_TOGGLE:
			globalFeature.value = defaultGlobalFeatures[i].options[(int)option].value;
			break;
			
		    case MEXP_FEATURE_OPTION_SLIDER1:
			sprintf(buf, "%.2f", option);
			globalFeature.value = buf;
			break;

		    case MEXP_FEATURE_OPTION_SLIDER10:
			sprintf(buf, "%d", (int)option);
			globalFeature.value = buf;
			break;
		    }

		    break;
		}
	    }

	    features.push_back(globalFeature);
	}
	else if (bufString1 == MEXP_STRING_TASK)
	{
	    feature.name = STRING_TASK_;
	    feature.nonIndex = false;
	    feature.weight = taskWeight;

	    // Get the task number (i.e., how many tasks?).
	    if (fscanf(featureFile, "%s", buf2) == EOF)
	    {
		hadError = true;
		break;
	    }
	    taskNum = atoi(buf2);

	    // Get the task ID (e.g., Biohazrd, EOD, etc.)
	    if (fscanf(featureFile, "%s", buf3) == EOF)
	    {
		hadError = true;
		break;
	    }
	    taskId = atoi(buf3);

	    for (i = 0; i < ((int)(defaultTaskList.size())); i++)
	    {
		if (defaultTaskList[i].id == taskId)
		{
		    feature.value = defaultTaskList[i].name;
		    break;
		}
	    }
	    
	    features.push_back(feature);

	    dataExists = true;
	}
	else if (bufString1 == MEXP_STRING_COORDINATE)
	{
	    feature.nonIndex = true;
	    feature.weight = 0.0;

	    // Get the coordinate.
	    if (fscanf(featureFile, "%s %s", buf2, buf3) == EOF)
	    {
		hadError = true;
		break;
	    }

	    feature.name = STRING_START_X_;
	    feature.value = buf2;
	    features.push_back(feature);

	    feature.name = STRING_START_Y_;
	    feature.value = buf3;
	    features.push_back(feature);

	    dataExists = true;
	}
	else if (bufString1 == MEXP_STRING_FEATURE)
	{
	    // Get the feature ID.
	    if (fscanf(featureFile, "%s", buf2) == EOF)
	    {
		hadError = true;
		break;
	    }
	    localFeatureId = atoi(buf2);

	    // Get the feature option being selected.
	    if (fscanf(featureFile, "%s", buf3) == EOF)
	    {
		hadError = true;
		break;
	    }
	    option = atof(buf3);

	    for (i = 0; i < ((int)(defaultLocalFeatures.size())); i++)
	    {
		if (defaultLocalFeatures[i].id == localFeatureId)
		{
		    feature.name = defaultLocalFeatures[i].name;
		    feature.nonIndex = defaultLocalFeatures[i].nonIndex;
		    feature.weight = defaultLocalFeatures[i].weight;

		    switch (defaultLocalFeatures[i].optionType) {
			
		    case MEXP_FEATURE_OPTION_TOGGLE:
			feature.value = defaultLocalFeatures[i].options[(int)option].value;
			break;

		    case MEXP_FEATURE_OPTION_SLIDER1:
			sprintf(buf, "%.2f", option);
			feature.value = buf;
			break;

		    case MEXP_FEATURE_OPTION_SLIDER10:
			sprintf(buf, "%d", (int)option);
			feature.value = buf;
			break;
		    }

		    break;
		}
	    }

	    features.push_back(feature);

	    dataExists = true;
	}
    }

    fclose(featureFile);

    if ((!dataExists)||(hadError))
    {
	return EMPTY_RETRIEVED_MISSIONS;
    }

    // Get the mission plan data from the server.
    gEventLogging->start("CBRClient Retrieving mission plan from CBRServer");
    missionplans = retrieveMissionPlansFromServer_(features, maxRating);
    gEventLogging->end("CBRClient Retrieving mission plan from CBRServer");

    for (i = 0; i < ((int)(missionplans.size())); i++)
    {
	retrievedMission = EMPTY_RETRIEVED_MISSION;
	retrievedMission.rating = missionplans[i].rating;
	retrievedMission.dataIndexList = missionplans[i].dataIndexList;
	retrievedMission.acdlSolution = missionplans[i].data;
	retrievedMissions.push_back(retrievedMission);
    }
    return retrievedMissions;
}

//-----------------------------------------------------------------------
// This function reads n bytes from the file descriptor.
//-----------------------------------------------------------------------
bool CBRClient::readNBytes_(const int fd, unsigned char *buf, const int n)
{
    int msgLength = -1, index = 0, remaining = 0;

    remaining = n;

    while (remaining > 0)
    {
        msgLength = read(fd, &buf[index], remaining);

        if (msgLength == -1)
        {
	    fprintf(stderr, "Warning(cfgedit): Read error %d!\n", errno);

            return false;
        }

        index += msgLength;
        remaining -= msgLength;
    }

    return true;
}

//-----------------------------------------------------------------------
// This function sends a command to the server
//-----------------------------------------------------------------------
void CBRClient::sendCommandToServer_(const unsigned char cmd)
{
    unsigned char threeBytesBuf[3];

    threeBytesBuf[0] = CBRPLANNER_COMM_SYNC_BYTE1;
    threeBytesBuf[1] = CBRPLANNER_COMM_SYNC_BYTE2;
    threeBytesBuf[2] = cmd;

    write(socketFd_, (unsigned char *)&threeBytesBuf, 3);
}

//-----------------------------------------------------------------------
// This function sends a general data to the server.
//-----------------------------------------------------------------------
void CBRClient::sendDataToServer_(void *data, int dataSize)
{
    write(socketFd_, data, dataSize);
}

//-----------------------------------------------------------------------
// This function sends a mission plan to the server.
//-----------------------------------------------------------------------
vector<int> CBRClient::sendMissionPlanToServer_(
    CBRPlanner_MissionPlan_t missionplan,
    vector<CBRPlanner_Feature_t> features)
{
    int i, numFeatures, nameSize, valueSize, dataIndexListSize;
    int dataIndex;
    bool nonIndex;
    double weight;
    vector<int> dataIndexList;

    // First, send a flag to notify the server that the mission plan will be sent.
    sendCommandToServer_(CBRPLANNER_COMM_CMD_MISSIONPLAN_SAVE);

    // Next, send the mission plan. The mission plan data is composed of the type flag,
    // rating, data size, and the data itselt.
    write(socketFd_, &(missionplan.type), sizeof(int));
    write(socketFd_, &(missionplan.rating), sizeof(int));
    write(socketFd_, &(missionplan.dataSize), sizeof(int));
    write(socketFd_, missionplan.data, missionplan.dataSize);

    // Send the number of features
    numFeatures = features.size();
    write(socketFd_, &numFeatures, sizeof(int));

    for (i = 0; i < (int)features.size(); i++)
    {
	// Send the length of the name
	nameSize = features[i].name.size();
	write(socketFd_, &nameSize, sizeof(int));

	// Send the name
	write(socketFd_, features[i].name.c_str(), nameSize);
	
	// Send the length of the value
	valueSize = features[i].value.size();
	write(socketFd_, &valueSize, sizeof(int));

	// Send the value
	write(socketFd_, features[i].value.c_str(), valueSize);

	// Send the nonIndex flag.
	nonIndex = features[i].nonIndex;
	write(socketFd_, &nonIndex, sizeof(bool));

	// Send the weight.
	weight = features[i].weight;
	write(socketFd_, &weight, sizeof(double));
    }

    // Get the new list of data indexes.
    readNBytes_(socketFd_, (unsigned char *)&(dataIndexListSize), sizeof(int));
    for (i = 0; i < dataIndexListSize; i++)
    {
	readNBytes_(socketFd_, (unsigned char *)&(dataIndex), sizeof(int));
	dataIndexList.push_back(dataIndex);
    }

    return dataIndexList;
}

//-----------------------------------------------------------------------
// This function retriives a mission plan afer sending features to the server.
//-----------------------------------------------------------------------
vector<CBRPlanner_MissionPlan_t> CBRClient::retrieveMissionPlansFromServer_(
    vector<CBRPlanner_Feature_t> features,
    int maxRating)
{
    CBRPlanner_MissionPlan_t missionplan;
    vector<CBRPlanner_MissionPlan_t> missionplans;
    int i, j, numFeatures, nameSize, valueSize, numPlans, dataIndexListSize;
    int dataIndex;
    double weight;
    bool nonIndex;
    const bool DEBUG = false;

    // First, send a flag to notify the server that the mission plan is going to be retrieved.
    sendCommandToServer_(CBRPLANNER_COMM_CMD_MISSIONPLAN_RETRIEVE);

    // Send the number of features
    numFeatures = features.size();
    write(socketFd_, &numFeatures, sizeof(int));

    for (i = 0; i < (int)features.size(); i++)
    {
	// Send the length of the name
	nameSize = features[i].name.size();
	write(socketFd_, &nameSize, sizeof(int));

	// Send the name
	write(socketFd_, features[i].name.c_str(), nameSize);
	
	// Send the length of the value
	valueSize = features[i].value.size();
	write(socketFd_, &valueSize, sizeof(int));

	// Send the value
	write(socketFd_, features[i].value.c_str(), valueSize);

	// Send the nonIndex flag.
	nonIndex = features[i].nonIndex;
	write(socketFd_, &nonIndex, sizeof(bool));

	// Send the weight.
	weight = features[i].weight;
	write(socketFd_, &weight, sizeof(double));
    }

    // Send the maximum suitability rating
    write(socketFd_, &maxRating, sizeof(int));

    // Now retrieve the mission plan from the server. Get the number of the plans.
    readNBytes_(socketFd_, (unsigned char *)&(numPlans), sizeof(int));

    for (i = 0; i < numPlans; i++)
    {
	// Receive the mission type.
	readNBytes_(socketFd_, (unsigned char *)&(missionplan.type), sizeof(int));

	// Receive the rating.
	readNBytes_(socketFd_, (unsigned char *)&(missionplan.rating), sizeof(int));

	// Receive the list of the data indexes.
	missionplan.dataIndexList.clear();
	readNBytes_(socketFd_, (unsigned char *)&dataIndexListSize, sizeof(int));
	for (j = 0; j < dataIndexListSize; j++)
	{
	    readNBytes_(socketFd_, (unsigned char *)&dataIndex, sizeof(int));
	    missionplan.dataIndexList.push_back(dataIndex);
	}

	// Receive the data.
	readNBytes_(socketFd_, (unsigned char *)&(missionplan.dataSize), sizeof(int));
	missionplan.data = new char[missionplan.dataSize+1];
	readNBytes_(socketFd_, (unsigned char *)(missionplan.data), missionplan.dataSize);
	missionplan.data[missionplan.dataSize] = '\0';

	if (DEBUG)
	{
	    fprintf(stderr, "Retrieved Plan (%d)\n", i);
	    fprintf(stderr, "%s\n\n", missionplan.data);
	}

	missionplans.push_back(missionplan);
    }

    return missionplans;
}

//-----------------------------------------------------------------------
// This function sends the FSA (tasks and triggers) names to the server.
//-----------------------------------------------------------------------
void CBRClient::sendFSANamesToServer_(config_fsa_lists_t fsaLists)
{
    int i, numTasks, numTriggers, nameSize;

    // First, send a flag notify the server that fsa lists will be sent.
    sendCommandToServer_(CBRPLANNER_COMM_CMD_FSANAMES);

    // Send the number of tasks.
    numTasks = fsaLists.taskList.size();
    write(socketFd_, &numTasks, sizeof(int));

    for (i = 0; i < numTasks; i++)
    {
	// Send the length of the task name.
	nameSize = fsaLists.taskList[i].name.size();
	write(socketFd_, &nameSize, sizeof(int));

	// Send the name of the task
	write(socketFd_, fsaLists.taskList[i].name.c_str(), nameSize);
    }

    // Send the number of triggers.
    numTriggers = fsaLists.triggerList.size();
    write(socketFd_, &numTriggers, sizeof(int));

    for (i = 0; i < numTriggers; i++)
    {
	// Send the length of the trigger name.
	nameSize = fsaLists.triggerList[i].name.size();
	write(socketFd_, &nameSize, sizeof(int));

	// Send the name of the trigger
	write(socketFd_, fsaLists.triggerList[i].name.c_str(), nameSize);
    }
}

//-----------------------------------------------------------------------
// This function sends the index of the selected mission.
//-----------------------------------------------------------------------
void CBRClient::sendSelectedMissionIndexToServer(int index)
{
    sendCommandToServer_(CBRPLANNER_COMM_CMD_SELECTED_MISSION);

    write(socketFd_, &index, sizeof(int));
}

//-----------------------------------------------------------------------
// This function sends a positive feedback (regarding the latest mission)
// to the server.
//-----------------------------------------------------------------------
void CBRClient::sendPositiveFeedbackToServer(vector<int> dataIndexList)
{
    sendFeedbackToServer_(CBRPLANNER_COMM_CMD_POSITIVE_FEEDBACK, dataIndexList);
}

//-----------------------------------------------------------------------
// This function sends a negative feedback (regarding the latest mission)
// to the server.
//-----------------------------------------------------------------------
void CBRClient::sendNegativeFeedbackToServer(vector<int> dataIndexList)
{
    sendFeedbackToServer_(CBRPLANNER_COMM_CMD_NEGATIVE_FEEDBACK, dataIndexList);
}

//-----------------------------------------------------------------------
// This function sends a feedback (regarding the latest mission) to the
// server.
//-----------------------------------------------------------------------
void CBRClient::sendFeedbackToServer_(const unsigned char cmd, vector<int> dataIndexList)
{
    int i, dataIndexListSize;

    sendCommandToServer_(cmd);

    dataIndexListSize = dataIndexList.size();
    write(socketFd_, &dataIndexListSize, sizeof(int));

    for (i = 0; i < dataIndexListSize; i++)
    {
	write(socketFd_, &(dataIndexList[i]), sizeof(int));
    }
}

/**********************************************************************
 * $Log: cbrclient.cc,v $
 * Revision 1.2  2003/04/06 08:48:14  endo
 * Updated for CBR Wizard Prototype II.
 *
 * Revision 1.1  2002/01/13 01:08:36  endo
 * Initial revision
 *
 **********************************************************************/
