/**********************************************************************
 **                                                                  **
 **                      PositionCalculator.c                        **
 **                                                                  **
 **                                                                  **
 **  Written by:  James Brian Lee                                    **
 **                                                                  **
 **  Copyright 2002 - 2003, Georgia Tech Research Corporation        **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: PositionCalculator.c,v 1.3 2003/04/06 13:32:37 endo Exp $ */

#include <math.h>
#include <unistd.h>

#include "PositionCalculator.h"

const float ClPositionCalculator::m_fSTATUSBAR_UPDATE_TIME_WAIT    = 0.25; // time between updates to the status bar
const float ClPositionCalculator::m_fCOMPASS_RELIABILITY_THRESHOLD = 4.0;  // +/- degree range that the compass readings
                                                                           // must stay within before we'll trust them
const float ClPositionCalculator::m_fCOMPASS_RELIABILITY_INTERVAL  = 1.0;  // interval (in sec) that the compass readings
                                                                           // must stay within the reliability threshold
                                                                           // before we'll trust them

void* ClPositionCalculator::RunReaderThread( void* pvData )
{
    ( (ClPositionCalculator*) pvData )->CalculatePosition();
    return NULL;
}

ClPositionCalculator::ClPositionCalculator( bool bWriteLog, bool watchDogEnable ) : 
    Module( NULL, "Position" )
{
    pthread_mutex_init( &m_mutexXytLock, NULL );
    pthread_mutex_init( &m_mutexRobotLock, NULL );
    pthread_mutex_init( &m_mutexGpsLock, NULL );
    pthread_mutex_init( &m_mutexGyroLock, NULL );
    pthread_mutex_init( &m_mutexCompassLock, NULL );
    pthread_mutex_init( &m_mutexCancelMutex, NULL );
    
    this->watchDogEnable = watchDogEnable; // initialize local watchDogEnable flag 
    if ( this->watchDogEnable == true)
    {
	// initialize mutex & put initial time in timestamp
	pthread_mutex_init( &position_watchdog_mutex, NULL);
	gettimeofday( &tv_watchdog,NULL);
	if ( WatchDog != NULL)
	{
	    // tell watchdog to watch PositionCalculator
	    WatchDog->AddPositionCalculator( &tv_watchdog, &position_watchdog_mutex);
	}
    }

    m_suLastStatusbarUpdateTime.tv_sec = 0;
    m_suLastStatusbarUpdateTime.tv_usec = 0;

    m_suLastWriteTime.tv_sec = 0;
    m_suLastWriteTime.tv_usec = 0;
    m_lFirstWriteSecond = 0;
    m_pfLog = NULL;
    m_bWriteLog = bWriteLog;

    SetCancelThread( false );
    SetRobot( NULL );
    SetGps( NULL );
    SetXyt( 0, 0, 0 );
#ifndef NO_MOBILITY
    SetGyro( NULL );
    SetCompass( NULL );
#endif

    m_pclSensor = new Sensor( SENSOR_XYT, 0, NULL, 0, NULL, 3, m_afXyt, NULL, &m_mutexXytLock );

    pthread_create( &m_threadCalculator, NULL, &RunReaderThread, (void*) this );
}

ClPositionCalculator::~ClPositionCalculator()
{
    SetCancelThread( true );
    pthread_join( m_threadCalculator, NULL );
    if ( m_pclSensor != NULL )
    {
        delete m_pclSensor;
    }
}

void ClPositionCalculator::UpdateStatusBar()
{
    static int c = 0;
    char buf[ 100 ];

    statusStr[ 0 ] = '\0';

    strcat( statusStr, "Position: " );
    switch( c )
    {
    case 0:
        strcat( statusStr, "-" );
        break;
    case 1:
        strcat( statusStr, "\\" );
        break;
    case 2:
        strcat( statusStr, "|" );
        break;
    case 3:
        strcat( statusStr, "/" );
        break;
    }
    c = ( c + 1 ) % 4;
    float fX, fY, fT;
    GetXyt( fX, fY, fT );
    sprintf( buf, " x: %.2f  y: %.2f  t: %.2f", fX, fY, fT );
    strcat( statusStr, buf );
    statusbar->update( statusLine );

    if ( watchDogEnable == true)
    {
	// update timestamp so watchdog will know we are still here
	pthread_cleanup_push((void(*)(void*)) pthread_mutex_unlock, (void *) &position_watchdog_mutex);
	pthread_mutex_lock(&position_watchdog_mutex);
	gettimeofday( &tv_watchdog,NULL);
	pthread_cleanup_pop(1);
    }
}

void ClPositionCalculator::SetRobot( Robot* pclRobot )
{
    pthread_mutex_lock( &m_mutexRobotLock );

    m_pclRobot = pclRobot;
    if ( m_pclRobot != NULL )
    {
	if ( m_pclRobot-> m_bposition_set == true)
	    m_pclRobot->getXYT( m_afLastRobotXyt[ 0 ], m_afLastRobotXyt[ 1 ], m_afLastRobotXyt[ 2 ] );
    }

    pthread_mutex_unlock( &m_mutexRobotLock );
}

void ClPositionCalculator::SetGps( Gps* pclGps )
{
    pthread_mutex_lock( &m_mutexGpsLock );

    m_pclGps = pclGps;

    pthread_mutex_unlock( &m_mutexGpsLock );
}

#ifndef NO_MOBILITY
void ClPositionCalculator::SetGyro( GyroDMUVGX* pclGyro )
{
    pthread_mutex_lock( &m_mutexGyroLock );

    m_pclGyro = pclGyro;
    if ( m_pclGyro != NULL )
    {
        m_pclGyro->getXYT( &(m_afLastGyroReading[ 0 ]), &(m_afLastGyroReading[ 1 ]), &(m_afLastGyroReading[ 2 ]) );
    }

    pthread_mutex_unlock( &m_mutexGyroLock );
}

void ClPositionCalculator::SetCompass( CompassKVHC100* pclCompass )
{
    int i;
    pthread_mutex_lock( &m_mutexCompassLock );

    m_pclCompass = pclCompass;
    if ( m_pclCompass != NULL )
    {
        m_pclCompass->getHeading( m_fStartCompassIntervalReading );
        gettimeofday( &m_suStartCompassIntervalTime, NULL );
    }

    pthread_mutex_unlock( &m_mutexCompassLock );
}
#endif

void ClPositionCalculator::GetXyt( float& fX, float& fY, float& fTheta )
{
    pthread_mutex_lock( &m_mutexXytLock );

    fX = m_afXyt[ 0 ];
    fY = m_afXyt[ 1 ];
    fTheta = m_afXyt[ 2 ];

    pthread_mutex_unlock( &m_mutexXytLock );
}

void ClPositionCalculator::SetXyt( float fX, float fY, float fTheta )
{
    pthread_mutex_lock( &m_mutexXytLock );

    m_afXyt[ 0 ] = fX;
    m_afXyt[ 1 ] = fY;
    m_afXyt[ 2 ] = fTheta;

    pthread_mutex_unlock( &m_mutexXytLock );
}

// this writes log info
void ClPositionCalculator::WriteToLog( long lSec, 
				       float fX, float fY, float fTheta, 
				       float fRobotX, float fRobotY, float fRobotTheta, bool bGotRobot, 
				       float fGyroX, float fGyroY, float fGyroTheta, bool bGotGyro, 
                                       float fCompassHeading, bool bGotCompass,
				       float fGpsX, float fGpsY, float fGpsTheta, bool bGotGps )
{
    char szFileName[ 100 ];

    // if the file hasn't been opened yet, then open it
    if ( m_pfLog == NULL )
    {
        sprintf( szFileName, "hserver_position_%ld.log", lSec );
        m_pfLog = fopen( szFileName, "w" );
        char* szHeader = "Hserver Position Log\n\n";
        if ( m_pfLog != NULL )
        {
            m_lFirstWriteSecond = lSec;
            fwrite( szHeader, strlen( szHeader ), 1, m_pfLog );
            fflush( m_pfLog );
        }
    }

    // write the info
    if ( m_pfLog != NULL )
    {
        fprintf( m_pfLog, "-------------------------------------------------------------------------------\n" );
        fprintf( m_pfLog, "Time: %ld\n", lSec - m_lFirstWriteSecond );
        fprintf( m_pfLog, "Position: \t%.2f \t%.2f \t%.2f\n", fX, fY, fTheta );
        fprintf( m_pfLog, "Robot:    " );
        if ( bGotRobot )
        {
            fprintf( m_pfLog, "\t%.2f \t%.2f \t%.2f\n", fRobotX, fRobotY, fRobotTheta );
        }
        else
        {
            fprintf( m_pfLog, "not available\n" );
        }

        fprintf( m_pfLog, "GPS:      " );
        if ( bGotGps )
        {
            fprintf( m_pfLog, "\t%.2f \t%.2f \t%.2f\n", fGpsX, fGpsY, fGpsTheta );
        }
        else
        {
            fprintf( m_pfLog, "not available\n" );
        }

        fprintf( m_pfLog, "Gyro:     " );
        if ( bGotGyro )
        {
            fprintf( m_pfLog, "\t%.2f \t%.2f \t%.2f\n", fGyroX, fGyroY, fGyroTheta );
        }
        else
        {
            fprintf( m_pfLog, "not available\n" );
        }

        fprintf( m_pfLog, "Compass:     " );
        if ( bGotCompass )
        {
            fprintf( m_pfLog, "\t%.2f \n", fCompassHeading );
        }
        else
        {
            fprintf( m_pfLog, "not available\n" );
        }

        fflush( m_pfLog );
    }
}

// get the average of 2 angles
float ClPositionCalculator::AverageTwoAngles( float fAngle1, float fAngle2 )
{
    float fMax, fMin, fDelta, fAverage;

    // get the max and min angles
    if ( fAngle1 > fAngle2 )
    {
        fMax = fAngle1;
        fMin = fAngle2;
    }
    else
    {
        fMax = fAngle2;
        fMin = fAngle1;
    }

    fDelta = fMax - fMin; // get the difference

    // if the 2 angles don't cross the 0 degree boundary, then we can just average them
    if ( fDelta <= 180.0 )
    {
        fAverage = ( fAngle1 + fAngle2 ) / 2.0;
    }
    // otherwise we have to find half the difference and subtract it from the min angle
    // (you could also add it to the max angle)
    else
    {
        fDelta = 360.0 - fDelta;
        fAverage = fMin - ( fDelta / 2.0 );
        if ( fAverage < 0.0 )
        {
            fAverage += 360.0;
        }
    }

    return fAverage;
}

// calculate a heading based on two points
float ClPositionCalculator::CalculateHeading( float fXOld, float fXNew, 
                                              float fYOld, float fYNew )
{
    float fXDiff = fXNew - fXOld;
    float fYDiff = fYNew - fYOld;
    float fHeading = 0.0;
    bool bRotate = ( fXDiff < 0 ); // if the difference vector is in the 2nd or 3rd quadrant,
                                   // we'll have to rotate it 180 degrees later.

    // guard against a divide-by-zero error
    if ( fXDiff == 0 )
    {
        if ( fYDiff > 0 )
        {
            fHeading = 90.0;
        }
        else if ( fYDiff < 0 )
        {
            fHeading = 270.0;
        }
    }
    else
    {
        // make sure the vector is in the 1st or 4th quadrant by mirroring it if necessary
        if ( bRotate )
        {
            fXDiff *= -1.0;
            fYDiff *= -1.0;
        }
        // get the heading in degrees
        fHeading = atan( fYDiff / fXDiff );
        fHeading = fHeading * ( 180.0 / M_PI );
        if ( fHeading < 0 )
        {
            fHeading += 360.0;
        }

        // if we had to rotate the vector, rotate it around by 180 degrees
        if ( bRotate )
        {
            fHeading += 180.0;
            if ( fHeading >= 360 )
            {
                fHeading -= 360.0;
            }
        }
    }

    return fHeading;
}

// use the robot's X, Y, and theta info to determine whether or not the robot is moving forward
bool ClPositionCalculator::MovingForward( float fXOld, float fXNew, 
                                          float fYOld, float fYNew, 
                                          float fThetaOld, float fThetaNew )
{
    // get the heading based on the current and last positions
    float fHeading = CalculateHeading( fXOld, fXNew, fYOld, fYNew );

    // get the robot's average reported heading
    float fAvgHeading = AverageTwoAngles( fThetaOld, fThetaNew );

    // determine the max and min angle so we can get the angle difference
    float fMaxAngle, fMinAngle;
    if ( fHeading >= fAvgHeading )
    {
        fMaxAngle = fHeading;
        fMinAngle = fAvgHeading;
    }
    else
    {
        fMaxAngle = fAvgHeading;
        fMinAngle = fHeading;
    }

    // get the angle difference
    float fAngleDiff = fMaxAngle - fMinAngle;
    if ( fAngleDiff > 180.0 )
    {
        fAngleDiff = 360.0 - fAngleDiff;
    }

    // If reported and calculated angles are less than 90 degrees off, then the 
    // robot must be moving forward.
    return ( fAngleDiff < 90.0 );
}

// calculate the difference in time between the two timevals.
float ClPositionCalculator::TimeDifferenceInSeconds( timeval suEndTime, timeval suStartTime )
{
    float fDiff;

    // make sure that the "end time" is greater than the "start time"
    if ( ( suEndTime.tv_sec < suStartTime.tv_sec ) || 
         ( ( suEndTime.tv_sec == suStartTime.tv_sec ) && 
           ( suEndTime.tv_usec < suStartTime.tv_usec ) ) )
    {
        fDiff = TimeDifferenceInSeconds( suStartTime, suEndTime );
    }
    else
    {
        // calculate the time difference
        long lSecDiff = suEndTime.tv_sec - suStartTime.tv_sec;
        long lUSecDiff = suEndTime.tv_usec - suStartTime.tv_usec;
        if ( lUSecDiff < 0 )
        {
            lSecDiff--;
            lUSecDiff = 1000000 - suStartTime.tv_usec + suEndTime.tv_usec;
        }

        float fUSecDiff = (float) lUSecDiff;
        fUSecDiff /= 1000000.0;
        fDiff = (float) lSecDiff;
        fDiff += fUSecDiff;
    }

    return fDiff;
}

// This is the main loop. This looks at all of the hardware currently connected and 
// gets the best estimate for the current position.
void ClPositionCalculator::CalculatePosition()
{
    while ( !GetCancelThread() )
    {
        float fStartingX, fStartingY, fStartingTheta;
        float fX, fY, fTheta;
        float fRobotX, fRobotY, fRobotTheta;
        float fGyroX, fGyroY, fGyroTheta;
        float fCompassHeading;
        float fCompassInterval = 0.0;
        bool bGotRobot = false;
        bool bGotGps = false;
        bool bGotGyro = false;
        bool bGotCompass = false;
        bool bPositionSet = false;
        bool bHeadingSet = false;
        SuGpsStats suGpsStats;
        bool bMovingForward = true;
        timeval suCurrentTime;

        // initialize position and heading
        GetXyt( fStartingX, fStartingY, fStartingTheta );
        fX = 0;
        fY = 0;
        fTheta = fStartingTheta;

        bPositionSet = false;
        bHeadingSet  = false;

        // if we have a robot, get the estimated position and heading from the robot
        pthread_mutex_lock( &m_mutexRobotLock );

	if ( m_pclRobot != NULL )
	{
	    bGotRobot = true;
	}
	else
	{
	    bGotRobot = false; 
	}

        if ( bGotRobot )
        {
	    // Ernest - make sure that position was set at least once, so that we don't
	    // have uninitialized values in there. Eventually, we need to verify with time
	    // stamp or other means that position recorded is reasonably current.
	    if ( m_pclRobot->m_bposition_set == true) 
		m_pclRobot->getXYT( fRobotX, fRobotY, fRobotTheta );
        }
        pthread_mutex_unlock( &m_mutexRobotLock );

        if ( bGotRobot )
        {
            bMovingForward = MovingForward( m_afLastRobotXyt[ 0 ], fRobotX, 
                                            m_afLastRobotXyt[ 1 ], fRobotY, 
                                            m_afLastRobotXyt[ 2 ], fRobotTheta );
        }

        // if we have GPS, get the readings
        pthread_mutex_lock( &m_mutexGpsLock );
        bGotGps = ( m_pclGps != NULL );
        if ( bGotGps )
        {
            suGpsStats = m_pclGps->GetStats();
        }
        pthread_mutex_unlock( &m_mutexGpsLock );

        // see if we have a good GPS fix
        if ( bGotGps && ( suGpsStats.iRT20Status < 2 ) ) // this threshold will be a member var of the GPS object later -- jbl
        {
            // set X and Y
            fX = suGpsStats.afXyt[ 0 ];
            fY = suGpsStats.afXyt[ 1 ];
            bPositionSet = true;

            // if we have a reliable GPS heading
            if ( suGpsStats.dHorizontalSpeed >= 0.2 ) // this threshold will be a member var of the GPS object later -- jbl
            {
                // set the heading
                fTheta = suGpsStats.afXyt[ 2 ];

                // reverse the angle if we're going backwards
                if ( !bMovingForward )
                {
                    fTheta = NormalizeAngle( fTheta + 180.0 );
                }
                bHeadingSet = true;
            }
        }

#ifndef NO_MOBILITY
        // if we have a gyro, get the heading
        pthread_mutex_lock( &m_mutexGyroLock );
        bGotGyro = ( m_pclGyro != NULL );
        if ( bGotGyro )
        {
            m_pclGyro->getXYT( &fGyroX, &fGyroY, &fGyroTheta );
        }
        pthread_mutex_unlock( &m_mutexGyroLock );

        // if we have a compass, get the heading
        pthread_mutex_lock( &m_mutexCompassLock );
        bGotCompass = ( m_pclCompass != NULL );
        if ( bGotCompass )
        {
            m_pclCompass->getHeading( fCompassHeading); 
	    if ( fabs( fCompassHeading - m_fStartCompassIntervalReading ) > m_fCOMPASS_RELIABILITY_THRESHOLD )
	    {
		m_fStartCompassIntervalReading = fCompassHeading;
		gettimeofday( &m_suStartCompassIntervalTime, NULL );
	    }
	    gettimeofday( &suCurrentTime, NULL );
	    fCompassInterval = TimeDifferenceInSeconds( suCurrentTime, m_suStartCompassIntervalTime );
	}
        pthread_mutex_unlock( &m_mutexCompassLock );
#endif

        // if the compass is there and we have a stable heading, use it. This will currently overwrite
        // the GPS's heading, but that's OK for now.
        if ( bGotCompass && ( fCompassInterval >= m_fCOMPASS_RELIABILITY_INTERVAL ) )
        {
            fTheta = fCompassHeading;
            bHeadingSet = true;
        }

        // if the heading wasn't set by GPS, use the gyro to change the heading
        if ( bGotGyro && !bHeadingSet )
        {
            float fHeadingDiff = m_afLastGyroReading[ 2 ] - fGyroTheta;
            fTheta = NormalizeAngle( fStartingTheta - fHeadingDiff );
            bHeadingSet = true;
        }

        // save the gyro readings for the next go-round
        if ( bGotGyro )
        {
            pthread_mutex_lock( &m_mutexGyroLock );
            m_afLastGyroReading[ 0 ] = fGyroX;
            m_afLastGyroReading[ 1 ] = fGyroY;
            m_afLastGyroReading[ 2 ] = fGyroTheta;
            pthread_mutex_unlock( &m_mutexGyroLock );
        }

        // if the heading hasn't been set yet, we can only rely on what the robot says, so 
        // use the difference between the robot's last reported heading and it's current heading
        // to calculate the offset to add to the old accepted heading
        if ( bGotRobot && !bHeadingSet )
        {
            fTheta = NormalizeAngle( fStartingTheta + ( fRobotTheta - m_afLastRobotXyt[ 2 ] ) );
            bHeadingSet = true;
        }

        // if the GPS didn't set the position, do the best we can w/ the robot's readings
        if ( bGotRobot && !bPositionSet )
        {
            // use the robot's current and last position reading along with the current heading
            //  to find dX and dY and add them to the current X and Y position  
            double dXDiff = fRobotX - m_afLastRobotXyt[ 0 ];
            double dYDiff = fRobotY - m_afLastRobotXyt[ 1 ];
            double dDistance = sqrt( ( dXDiff * dXDiff ) + ( dYDiff * dYDiff ) );

            // reverse the angle if we're going backwards
            float fTempHeading = fTheta;
            if ( !bMovingForward )
            {
                fTempHeading = NormalizeAngle( fTempHeading + 180.0 );
            }

            double dHeadingInRad = fTempHeading * M_PI / 180.0;

            dXDiff = dDistance * cos( dHeadingInRad );
            dYDiff = dDistance * sin( dHeadingInRad );

            fX = fStartingX + (float) dXDiff;
            fY = fStartingY + (float) dYDiff;
            bPositionSet = true;
        }

        // save the robot's readings for use on the next go-round
        if ( bGotRobot )
        {
            pthread_mutex_lock( &m_mutexRobotLock );
            m_afLastRobotXyt[ 0 ] = fRobotX;
            m_afLastRobotXyt[ 1 ] = fRobotY;
            m_afLastRobotXyt[ 2 ] = fRobotTheta;
            pthread_mutex_unlock( &m_mutexRobotLock );
        }

        // set the finalized XYT values
        SetXyt( fX, fY, fTheta );

        gettimeofday( &suCurrentTime, NULL );
        if ( TimeDifferenceInSeconds( suCurrentTime, m_suLastStatusbarUpdateTime ) > 
             m_fSTATUSBAR_UPDATE_TIME_WAIT )
        {
            UpdateStatusBar();
            m_suLastStatusbarUpdateTime = suCurrentTime;
        }

        // write logging info
        if ( m_bWriteLog )
        {
            timeval suCurrentTime;
            gettimeofday( &suCurrentTime, NULL );

            // only write once a second
            if ( suCurrentTime.tv_sec > m_suLastWriteTime.tv_sec )
            if ( suCurrentTime.tv_sec > m_suLastWriteTime.tv_sec && bGotGps )
            {
		/* Ernest - the step below prevents uninitialized variables from being
		   sent to WriteToLog(). But even if we have "bGotGps" how do we know that the
		   suGpsStats info is valid, especially the first time through the loop? */
		if ( !bGotGps ) 
		{
		    suGpsStats.afXyt[ 0 ] = 0;
		    suGpsStats.afXyt[ 1 ] = 0;
		    suGpsStats.afXyt[ 2 ] = 0;
		}
                WriteToLog( suCurrentTime.tv_sec, 
                            fX, fY, fTheta, 
                            fRobotX, fRobotY, fRobotTheta, bGotRobot, 
                            fGyroX, fGyroY, fGyroTheta, bGotGyro, 
                            fCompassHeading, bGotCompass,
                            suGpsStats.afXyt[ 0 ], suGpsStats.afXyt[ 1 ], suGpsStats.afXyt[ 2 ], bGotGps );
                m_suLastWriteTime = suCurrentTime;
            }
        }

        usleep( 5000 ); // update 200 times / sec
    }
}

/**********************************************************************
# $Log: PositionCalculator.c,v $
# Revision 1.3  2003/04/06 13:32:37  endo
# gcc 3.1.1
#
#**********************************************************************/
