#ifndef _POSITIONCALCULATOR_H
#define _POSITIONCALCULATOR_H

#include <pthread.h>
#include <unistd.h>
#include <sys/time.h>
#include "sensors.h"
#include "robot.h"
#include "gps.h"

#ifndef NO_MOBILITY
#include "gyro_dmu-vgx.h"
#include "compass_kvh-c100.h"
#endif

class ClPositionCalculator : public Module
{
public:
    ClPositionCalculator( bool bWriteLog = false,bool watchDogEnable = true ); // Ernest
    ~ClPositionCalculator();

    void SetXyt( float fX, float fY, float fTheta );

    void SetRobot( Robot* pclRobot );
    void SetGps( Gps* pclGps );
#ifndef NO_MOBILITY
    void SetGyro( GyroDMUVGX* pclGyro );
    void SetCompass( CompassKVHC100* pclCompass );
#endif

protected:
    static const float m_fSTATUSBAR_UPDATE_TIME_WAIT;
    static const float m_fCOMPASS_RELIABILITY_THRESHOLD;
    static const float m_fCOMPASS_RELIABILITY_INTERVAL;

    static void* RunReaderThread( void* pvData );
    static bool MovingForward( float fXOld, float fXNew, 
                               float fYOld, float fYNew, 
                               float fThetaOld, float fThetaNew );
    static float NormalizeAngle( float fAngle );
    static float CalculateHeading( float fXOld, float fXNew, float fYOld, float fYNew );
    static float AverageTwoAngles( float fAngle1, float fAngle2 );

    void CalculatePosition();
    void GetXyt( float& fX, float& fY, float& fTheta );

    bool GetCancelThread();
    void SetCancelThread( bool bCancel );

    static float TimeDifferenceInSeconds( timeval suEndTime, timeval suStartTime );

    void UpdateStatusBar();
    void WriteToLog( long lSec, 
		     float fX, float fY, float fTheta, 
		     float fRobotX, float fRobotY, float fRobotTheta, bool bGotRobot, 
		     float fGyroX, float fGyroY, float fGyroTheta, bool bGotGyro, 
                     float fCompassHeading, bool bGotCompass,
		     float fGpsX, float fGpsY, float fGpsTheta, bool bGotGps );

    pthread_t m_threadCalculator;
    pthread_mutex_t m_mutexXytLock;
    pthread_mutex_t m_mutexRobotLock;
    pthread_mutex_t m_mutexGpsLock;
    pthread_mutex_t m_mutexGyroLock;
    pthread_mutex_t m_mutexCompassLock;
    pthread_mutex_t m_mutexCancelMutex;
    Robot* m_pclRobot;
    Gps* m_pclGps;
#ifndef NO_MOBILITY
    GyroDMUVGX* m_pclGyro;
    CompassKVHC100* m_pclCompass;
#endif
    bool m_bCancelThread;
    float m_afXyt[ 3 ];
    Sensor* m_pclSensor;
    float m_afLastRobotXyt[ 3 ];
    float m_afLastGyroReading[ 3 ];
    float m_fStartCompassIntervalReading;
    timeval m_suStartCompassIntervalTime;

    timeval m_suLastStatusbarUpdateTime;
    timeval m_suLastWriteTime;
    long m_lFirstWriteSecond;
    FILE* m_pfLog;
    bool m_bWriteLog;
    struct timeval tv_watchdog;  // Ernest
    pthread_mutex_t position_watchdog_mutex; // Ernest
    bool watchDogEnable; // Ernest

};

inline bool ClPositionCalculator::GetCancelThread()
{
    bool bCancel;

    pthread_mutex_lock( &m_mutexCancelMutex );
    bCancel = m_bCancelThread;
    pthread_mutex_unlock( &m_mutexCancelMutex );

    return bCancel;
}

inline void ClPositionCalculator::SetCancelThread( bool bCancel )
{
    pthread_mutex_lock( &m_mutexCancelMutex );
    m_bCancelThread = bCancel;
    pthread_mutex_unlock( &m_mutexCancelMutex );
}

inline float ClPositionCalculator::NormalizeAngle( float fAngle )
{
    while ( fAngle < 0 )
    {
        fAngle += 360.0;
    }
    while ( fAngle >= 360.0 )
    {
        fAngle -= 360.0;
    }

    return fAngle;
}

#endif // ndef _POSITIONCALCULATOR_H
