/**********************************************************************
 **                                                                  **
 **                             WatchDog.c                           **
 **                                                                  **
 **  Written by: Ernest P. Merrill                                   **
 **                                                                  **
 **  The WatchDog class monitors hserver's threads by periodically   **
 **  checking to see if each thread has updated a timestamp. If any  **
 **  thread is found not to have updated it's time stamp count (=6)  **
 **  number of times, then hserver is shut down, causing the robot   **
 **  to stop. A message is printed on the console, indicating which  **
 **  thread caused the problem. Additional information can be gotton **
 **  from the "WatchDogdebug.txt" file. The watchdog function can be **
 **  disabled by using -d as a command line option.
 **                                                                  **
 **  Copyright 2002 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: WatchDog.c,v 1.2 2003/04/06 13:32:37 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <pthread.h>
#include <semaphore.h>
#include <stdarg.h>

#include <fstream>
#include <iostream>
#include <sys/types.h>
#include <signal.h>
#include "hserver.h"
#include "WatchDog.h"
#include "nomad.h"
#include "robot.h"

using std::ofstream;

//-----------------------------------------------------------------------
// Constanns
//-----------------------------------------------------------------------
const int ClWatchDog::MAX_PRINT_BUFFER_SIZE = 1024;

//-----------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------

// store the relevant timestamps, mutexes, and enable flags in these structures
SuWDData SuNomad;
SuWDData SuPositionCalculator;
SuWDData SuLaser;
SuWDData SuIPTHandler;
SuWDData SuATRVJr;

ClWatchDog::ClWatchDog ()
{  

    pthread_mutex_init( &debug_mutex, NULL);
    wddebugout = new ofstream("WatchDogdebug.txt");
    pthread_mutex_init( &keeprunning_mutex, NULL );
    pthread_mutex_lock( &keeprunning_mutex );
    keeprunning = false;
    pthread_mutex_unlock( &keeprunning_mutex );
    position_calculator = false;
    ipt_handler_check = false;
    ipt_handler_present = false;
}

ClWatchDog::~ClWatchDog ()
{
    StopWatchdog();
    delete wddebugout;
}


// accept a variable number of arguments in "printf" style and output them to a 
// debugging file. Outputs are protected by mutex to avoid race conditions.
void ClWatchDog::PrintDebug (const char* fmt, ...)
{  
    char *buf;
    va_list args;

    pthread_mutex_lock( &debug_mutex );
    buf = new char[MAX_PRINT_BUFFER_SIZE];
    va_start(args,fmt);
    vsprintf(buf, fmt,args);
    va_end(args);
    *wddebugout << buf;
    delete [] buf;
    buf = NULL;
    pthread_mutex_unlock ( &debug_mutex );
}

void ClWatchDog::AddNomad(struct timeval* nomad_tv, pthread_mutex_t* nomad_mutex)
{
    SuNomad.tv_wd = nomad_tv;
    SuNomad.wd_mutex = nomad_mutex;
    SuNomad.enable_wd = true;
    g_clNomadList.push_back( SuNomad);
}

/* When the nomad does it's zeroing process, it doesn't execute the normal reader_loop()
   often enough to keep the WatchDog from killing hserver. Therefore, the nomad watchdog
   (as well as the ipt watchdog) must be disabled while the nomad is zeroing */
void ClWatchDog::DisableNomadWD()
{
    pthread_mutex_lock( &nomadenable_mutex );
    g_clNomadList[ 0 ].enable_wd = false;
    ipt_handler_check = false; //TODO: fix this (see below)
    pthread_mutex_unlock( &nomadenable_mutex );  
    g_clNomadList[ 0 ].enable_wd = false;
}

// Enable the nomad (and ipt, if applicable) watchdog function after nomad zeroing is complete
void ClWatchDog::EnableNomadWD()
{
    pthread_mutex_lock( &nomadenable_mutex );
    g_clNomadList[ 0 ].enable_wd = true;
    if (ipt_handler_present == true)
    {
	// only enable the watchdog for ipt if ipt was enabled before the DisableNomad()
	// function was called.
	ipt_handler_check = true; //TODO: fix this so it works with and without ipthandler
    }
    pthread_mutex_unlock( &nomadenable_mutex );  
}

void ClWatchDog::AddATRVJr(struct timeval* watchdog_tv, pthread_mutex_t* watchdog_mutex)
{
    SuATRVJr.tv_wd = watchdog_tv;
    SuATRVJr.wd_mutex = watchdog_mutex;
    SuATRVJr.enable_wd = true;
    g_clATRVJrList.push_back( SuATRVJr );
}

void ClWatchDog::AddLaser(struct timeval* watchdog_tv, pthread_mutex_t* watchdog_mutex)
{
    SuLaser.tv_wd = watchdog_tv;
    SuLaser.wd_mutex = watchdog_mutex;
    SuLaser.enable_wd = true;
    g_clLaserList.push_back( SuLaser);
}


void ClWatchDog::AddPositionCalculator(struct timeval* watchdog_tv, pthread_mutex_t* watchdog_mutex)
{
    SuPositionCalculator.tv_wd = watchdog_tv;
    SuPositionCalculator.wd_mutex = watchdog_mutex;
    SuPositionCalculator.enable_wd = true;
    position_calculator = true;
}

void ClWatchDog::AddIPTHandler(struct timeval* watchdog_tv, pthread_mutex_t* watchdog_mutex)
{
    SuIPTHandler.tv_wd = watchdog_tv;
    SuIPTHandler.wd_mutex = watchdog_mutex;
    SuIPTHandler.enable_wd = true;
    ipt_handler_check = true;
    ipt_handler_present = true;
}

// check each (enabled) thread every second shutdown hserver if any thread is not responsive
// "count" number of times in a row.    
void ClWatchDog::WatchdogLoop()
{

    while(keeprunning)
    {
	if ( g_clNomadList.size()> 0 && g_clNomadList[ 0 ].enable_wd == true)
	    CheckNomad();

	if ( g_clATRVJrList.size()> 0 && g_clATRVJrList[ 0 ].enable_wd == true)
	    CheckATRVJr();

	if ( g_clLaserList.size()> 0)
	    CheckLaser();

	if ( position_calculator == true)
	    CheckPosCalculator();
	
	if (ipt_handler_check == true)
	    CheckIPTHandler();

	sleep(2);
    }
}

void ClWatchDog::CheckPosCalculator()
{
        static int count = 0;
	static long sec_prev = 0;
	static long usec_prev = 0;
	timeval poscal_tv_current = {0,0}; // assume only one robot for now

	pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, 
				  (void *) & (*SuPositionCalculator.wd_mutex));
	pthread_mutex_lock(& (*SuPositionCalculator.wd_mutex));
	poscal_tv_current.tv_sec = SuPositionCalculator.tv_wd->tv_sec;
	poscal_tv_current.tv_usec = SuPositionCalculator.tv_wd->tv_usec;
	pthread_cleanup_pop(1);
      
	double dTimeDiff = ( ( poscal_tv_current.tv_sec + ( poscal_tv_current.tv_usec * 1e-6 ) ) - 
				     ( sec_prev  + ( usec_prev  * 1e-6 ) ) );
    
	sec_prev = poscal_tv_current.tv_sec;
	usec_prev = poscal_tv_current.tv_usec;

	if ( static_cast<int>(dTimeDiff) == 0 )
	{
	    count++;
	    PrintDebug( "\t\tWatchdogLoop() PositionCalculator: dTimeDiff = 0 , count = %d \n",count );
	    if (count >= 3)
	    {
		PrintDebug("\t\tWatchdogLoop() PositionCalculator: dTimeDiff = 0 for 3rd time. Exiting...\n");
		printTextWindow("WatchDog.c: PositionCalculator thread died. Exiting...\n");
		kill (0,9);
	    }
	}
	else
	    count = 0; 
}

void ClWatchDog::CheckIPTHandler()
{
        static int count = 0;
	static long sec_prev = 0;
	static long usec_prev = 0;
	timeval tv_current = {0,0}; // assume only one ipt connection 

	pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, 
				  (void *) & (*SuIPTHandler.wd_mutex));
	pthread_mutex_lock(& (*SuIPTHandler.wd_mutex));
	tv_current.tv_sec = SuIPTHandler.tv_wd->tv_sec;
	tv_current.tv_usec = SuIPTHandler.tv_wd->tv_usec;
	pthread_cleanup_pop(1);
      
	double dTimeDiff = ( ( tv_current.tv_sec + ( tv_current.tv_usec * 1e-6 ) ) - 
				     ( sec_prev  + ( usec_prev  * 1e-6 ) ) );

	sec_prev = tv_current.tv_sec;
	usec_prev = tv_current.tv_usec;

	if (static_cast<int>(dTimeDiff) == 0)
	{
	    count++;
	    PrintDebug( "\t\tWatchdogLoop() IPTHandler: dTimeDiff = 0 , count = %d\n", count ); 

	    if (count >= 3)
	    {
		PrintDebug ("\t\tWatchdogLoop() IPTHandler: dTimeDiff = 0 for 3rd time. Exiting...\n");
		printTextWindow("WatchDog.c: IPTHandler thread died. Exiting...\n");
		kill (0,9);
	    }
	}
	else
	    count = 0; 
}


void ClWatchDog::CheckNomad()
{
    static int count = 0;
    timeval nomad_tv_current = {0,0}; // assume only one nomad for now
    for (unsigned int i=1; i <= g_clNomadList.size(); i++)
    {
	static long sec_prev = 0;
	static long usec_prev = 0;

	pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, 
				  (void *) & (*g_clNomadList[ 0 ].wd_mutex));
	pthread_mutex_lock(& (*g_clNomadList[ 0 ].wd_mutex));
	nomad_tv_current.tv_sec = g_clNomadList[ 0 ].tv_wd->tv_sec;
	nomad_tv_current.tv_usec = g_clNomadList[ 0 ].tv_wd->tv_usec;
	pthread_cleanup_pop(1);
      
	double dTimeDiff = ( ( nomad_tv_current.tv_sec + ( nomad_tv_current.tv_usec * 1e-6 ) ) - 
				     ( sec_prev  + ( usec_prev  * 1e-6 ) ) );

	sec_prev = nomad_tv_current.tv_sec;
	usec_prev = nomad_tv_current.tv_usec;

	if (static_cast<int>(dTimeDiff) == 0)
	{
	    count++;
	    PrintDebug( "\t\tWatchdogLoop() Nomad: dTimeDiff = 0 , count = %d:\n",count); 

	    if (count >= 3)
	    {
		PrintDebug("\t\tWatchdogLoop() Nomad: dTimeDiff = 0 for 3rd time. Exiting...\n");
		printTextWindow("WatchDog.c: Nomad thread died. Exiting...\n");
		kill (0,9);
	    }
	}
	else
	    count = 0;	    
    }
}

void ClWatchDog::CheckLaser()
{
    static int count = 0;
    timeval laser_tv_current = {0,0}; // assume only one laser for now
    for (unsigned int i=1; i <= g_clLaserList.size(); i++)
    {

	static long sec_prev = 0;
	static long usec_prev = 0;

	pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, 
				  (void *) & (*g_clLaserList[ 0 ].wd_mutex));
	pthread_mutex_lock(& (*g_clLaserList[ 0 ].wd_mutex));
	laser_tv_current.tv_sec = g_clLaserList[ 0 ].tv_wd->tv_sec;
	laser_tv_current.tv_usec = g_clLaserList[ 0 ].tv_wd->tv_usec;
	pthread_cleanup_pop(1);
      
	double dTimeDiff = ( ( laser_tv_current.tv_sec + ( laser_tv_current.tv_usec * 1e-6 ) ) - 
				     ( sec_prev  + ( usec_prev  * 1e-6 ) ) );

	sec_prev = laser_tv_current.tv_sec;
	usec_prev = laser_tv_current.tv_usec;

	if (static_cast<int>(dTimeDiff) == 0)
	{
	    count++;
	    PrintDebug("\t\tWatchdogLoop() Laser: dTimeDiff = 0 , count = %d\n",count);

	    if (count >= 3)
	    {
		PrintDebug("\t\tWatchdogLoop() Laser: dTimeDiff = 0 for 3rd time. Exiting...\n");
		printTextWindow("WatchDog.c: Laser thread died. Exiting...\n");
		kill (0,9);
	    }
	}
	else
	    count = 0;	    
    }
}

void ClWatchDog::CheckATRVJr()
{
    static int count = 0;
    timeval ATRVJr_tv_current = {0,0}; // assume only one ATRVJr for now
    for (unsigned int i=1; i <= g_clATRVJrList.size(); i++)
    {

	static long sec_prev = 0;
	static long usec_prev = 0;

	pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, 
				  (void *) & (*g_clATRVJrList[ 0 ].wd_mutex));
	pthread_mutex_lock(& (*g_clATRVJrList[ 0 ].wd_mutex));
	ATRVJr_tv_current.tv_sec = g_clATRVJrList[ 0 ].tv_wd->tv_sec;
	ATRVJr_tv_current.tv_usec = g_clATRVJrList[ 0 ].tv_wd->tv_usec;
	pthread_cleanup_pop(1);
      
	double dTimeDiff = ( ( ATRVJr_tv_current.tv_sec + ( ATRVJr_tv_current.tv_usec * 1e-6 ) ) - 
				     ( sec_prev  + ( usec_prev  * 1e-6 ) ) );

	sec_prev = ATRVJr_tv_current.tv_sec;
	usec_prev = ATRVJr_tv_current.tv_usec;

	if (static_cast<int>(dTimeDiff) == 0)
	{
	    count++;
	    PrintDebug("\t\tWatchdogLoop() ATRVJr: dTimeDiff = 0 , count = %d\n",count);

	    if (count >= 3)
	    {
		PrintDebug("\t\tWatchdogLoop() ATRVJr: dTimeDiff = 0 for 3rd time. Exiting...\n");
		printTextWindow("WatchDog.c: ATRVJr thread died. Exiting...\n");
		kill (0,9);
	    }
	}
	else
	    count = 0;	    
    }
}

void ClWatchDog::StartThread(void* wd_instance)
{
    ( (ClWatchDog*) wd_instance)->WatchdogLoop();
}

void ClWatchDog::StartWatchdog()
{
    pthread_mutex_lock( &keeprunning_mutex );
    keeprunning = true;
    pthread_mutex_unlock( &keeprunning_mutex );       
    int res = pthread_create( &watchdog_th, NULL, (void*(*)(void*)) &StartThread, (void*)this);
    if (res != 0)
    {
	perror("StartWatchdog");
	PrintDebug("In startWatchDog(). Error: watchdog thread could not be created.\n");
    }
    else 
	PrintDebug("In StartWatchDog().Watchdog thread was created...\n");
}

void ClWatchDog::StopWatchdog()
{
    PrintDebug("In StopThread(). Waiting for thread join..\n");
    pthread_mutex_lock( &keeprunning_mutex );
    keeprunning = false;
    pthread_mutex_unlock( &keeprunning_mutex );
    pthread_join(watchdog_th, NULL );
    PrintDebug ("In StopThread(). Thread joined.\n");
}

/**********************************************************************
 * $Log: WatchDog.c,v $
 * Revision 1.2  2003/04/06 13:32:37  endo
 * gcc 3.1.1
 *
 *
 **********************************************************************/
