/**********************************************************************
 **                                                                  **
 **                           compass_kvh-c100.c                     **
 **                                                                  **
 **                                                                  **
 **  Written by:  Alexander Stoytchev                                **
 **                                                                  **
 **  KVH C100 Compass driver for HServer                             **
 **                                                                  **
 **  Note: HServer reads compass data from mobility. mobility has to **
 **        be installed in the system.                               **
 **                                                                  **
 **  Copyright 2001, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

// Compass modifications
// src/hardware_drivers/hserver/makefile
// src/hardware_drivers/hserver/hserver.c
// src/hardware_drivers/hserver/message.c
// src/hardware_drivers/hserver/message.h
// src/hardware_drivers/hserver/gyro_dmu-vgx.c
// src/hardware_drivers/hserver/gyro_dmu-vgx.h
// src/hardware_drivers/hserver/ControlInterfaces.c
// src/hardware_drivers/hserver/ControlInterfaces.h
// src/hardware_drivers/hserver/PositionCalculator.c 
// src/hardware_drivers/hserver/PositionCalculator.h

/* $Id: compass_kvh-c100.c,v 1.1 2001/05/18 17:55:02 blee Exp $ */

#include "compass_kvh-c100.h"


// This macro takes a degree (d) and change its value to be inside
// the lower bound (l) and the higher bound (h).
#define COMPASSKVHC100_CRAMPDEG(d,l,h) {while((d)<(l))(d)+=360.0; while((d)>=(h))(d)-=360.0;}

// Name of the COMPASS server.
const char *COMPASSKVHC100_COMPASSSERVER_NAME = "KVH C100";


// How often the data from mobility should be updated (in microsecond).
const int COMPASSKVHC100_READER_THREAD_USLEEP = 1000;

// How often the status bar updated with respect to the sampling time
// in the reader thread.
const int COMPASSKVHC100_SKIP_STATUSBAR_UPDATE = 10;

// Initial values for the parameters to convert from mobility
// Compass heading to HServer angles in degree. *calib_factor* multiplies
// and *offset* adds values to the compass data. The multiplication
// is done first.
const double COMPASSKVHC100_INIT_CALIB_FACTOR_HEADING = -1.0;
const double COMPASSKVHC100_INIT_OFFSET_HEADING = 90.0;


// Pthread mutex to for HServer angles.
pthread_mutex_t compassKVHC100_headingMutex_ = PTHREAD_MUTEX_INITIALIZER;

// This function calls the reader thread.
void* CompassKVHC100::CompassKVHC100_startMainThread( void* compass_kvh_c100_instance )
{
    ((CompassKVHC100*)compass_kvh_c100_instance)->readerLoop();
    return NULL;
}

// Constructor for the CompassKVHC100 class.
CompassKVHC100::CompassKVHC100(CompassKVHC100** a) : Module( (Module**) a, "CompassKVHC100")
{
    int rtn;

    if (report_level > HS_REP_LEV_NONE) printTextWindow("Connecting to Compass KVH C100... ", false);

    firstDataReceived_ = FALSE;
    compassKVHC100_heading_=0.0;
    //memset((void *)(&compassKVHC100_heading_), 0x0, sizeof(float)*1);

    compassKVHC100HeadingCalibFactor_ = COMPASSKVHC100_INIT_CALIB_FACTOR_HEADING;

    compassKVHC100HeadingOffset_ = COMPASSKVHC100_INIT_OFFSET_HEADING;


    rtn = initializeMobility_();

    refreshScreen();


    if (rtn)
    {
        if (report_level > HS_REP_LEV_NONE) printTextWindow("failed", true);
        return;
    }

    // Start reading the COMPASS KVH C100 data.
    pthread_create( &readerThread_, NULL, &CompassKVHC100_startMainThread, (void*) this );

    while (!firstDataReceived_) usleep(COMPASSKVHC100_READER_THREAD_USLEEP);

    if (report_level > HS_REP_LEV_NONE) printTextWindow( "connected", true );

    statusbarUpdate_();
}

// Destructor for the CompassKVHC100 class.
CompassKVHC100::~CompassKVHC100()
{
    pthread_cancel(readerThread_);
    pthread_join(readerThread_, NULL );

    printTextWindow( "Compass KVH C100 disconnected" );
}

// This function updates the status bar.
void CompassKVHC100::statusbarUpdate_()
{
    static int c = 0;
    char buf[100];
  
    statusStr[0] = 0;

    strcat(statusStr, "Compass KVH C100: ");

    switch(c) {
    case 0:
        strcat(statusStr, "-");
        break;
    case 1:
        strcat(statusStr, "\\");
        break;
    case 2:
        strcat(statusStr, "|");
        break;
    case 3:
        strcat(statusStr, "/");
        break;
    }
    c = (c + 1) % 4;

    pthread_mutex_lock(&compassKVHC100_headingMutex_);
  
    sprintf(buf, " heading: %.1f [deg]", compassKVHC100_heading_);
                                        
    pthread_mutex_unlock(&compassKVHC100_headingMutex_);

    strcat(statusStr, buf);
    statusbar->update(statusLine);
}

// This function sets up the interface between HServer and mobility. 
int CompassKVHC100::initializeMobility_(void)
{
    int argc;
    char* argv[4];
    char* compassName = strdup(COMPASSKVHC100_COMPASSSERVER_NAME);
    char pathName[255];

    argv[0] = "HServer";
    argv[1] = "COMPASS";
    argv[2] = "KVHC100";
    argv[3] = 0;
    argc = 3;

    pHelper_ = new mbyClientHelper(argc, argv);

    sprintf(pathName, "%s/Compass", compassName);

    ptempObj_ = pHelper_->find_object(pathName);

    pMbyCompassState_ = FVectorState_i::_duplicate(
        FVectorState_i::_narrow(ptempObj_));

    return 0;
}  

// This function loops around to get the compass KVH C100 data.
void CompassKVHC100::readerLoop(void)
{
    int count = 0;

    while(TRUE)
    {
	// Make sure termination was not requested.
        pthread_testcancel();

	// Sleep a bit to reduce the CPU load.
	usleep(COMPASSKVHC100_READER_THREAD_USLEEP);

	// Check for the pthread termination again.
        pthread_testcancel();

	// Get the data from COMPASS KVH C100.
        updateCompassData_();


	// Show updated data to the screen.
	if (count > COMPASSKVHC100_SKIP_STATUSBAR_UPDATE)
	{
	    statusbarUpdate_();
	    count = 0;
	}

	count ++;
    }
}

// This function updates the compass data by reading it
// from mobility.
void CompassKVHC100::updateCompassData_(void)
{
    // Read the raw data from mobility.
    rawCompassData_ = readMbyCompassData_();

    // Computer HServer version of heading.
    pthread_mutex_lock(&compassKVHC100_headingMutex_);
 
    compassKVHC100_heading_ = correctHeading_(rawCompassData_.heading);

    pthread_mutex_unlock(&compassKVHC100_headingMutex_);

    if (!firstDataReceived_) firstDataReceived_ = TRUE;
}


// This function grabs the Compass KVH C100 data from mobility
CompassKVHC100_CompassData_t CompassKVHC100::readMbyCompassData_(void)
{
    CompassKVHC100_CompassData_t compassData;
    struct timeval timeVal;
    struct timezone timeZone;

    // Put the time stamp.
    gettimeofday(&timeVal, &timeZone);
    compassData.timeStamp = timeVal;

    // Grab the latest data set from mobility.
    pMbyCompassData_ = pMbyCompassState_->get_sample(0); 

    // Retrieve the heading data.
    compassData.heading = pMbyCompassData_->data[0];
 
    return compassData;
}

// This function converts the Compass KVH C100 data value into HServer
// heading
//correctAngularPosition_

float CompassKVHC100::correctHeading_(float rawHeading)
{
    float heading;
 
    heading = rawHeading;

    heading *= compassKVHC100HeadingCalibFactor_;
    heading += compassKVHC100HeadingOffset_;
    COMPASSKVHC100_CRAMPDEG(heading,0.0,360.0);

    return heading;
}


// This function returns the current compass heading.
void CompassKVHC100::getHeading( float& heading )
{
  pthread_mutex_lock(&compassKVHC100_headingMutex_);
  heading = compassKVHC100_heading_;
  pthread_mutex_unlock(&compassKVHC100_headingMutex_);
}

/* jbl
// Control menus.
void CompassKVHC100::control(void)
{
    int c;
    int done = 0;

    messageDrawWindow( EnMessageType_COMPASS_DMU_VGX_CONTROL, EnMessageErrType_NONE );
    redrawWindows();
    do
    {
        c = getch();
        switch( c )
        {
        case 'd':
            messageHide();
            delete this;
            return;
            break;
        case 'r':
	    resetAngles();
            break;
        case 'x':
        case 'Q':
        case KEY_ESC:
            done = true;
            break;
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey( c );
            break;
        }
    } while ( !done );
    messageHide();
}
*/


/**********************************************************************
# $Log: compass_kvh-c100.c,v $
# Revision 1.1  2001/05/18 17:55:02  blee
# Initial revision
#
#**********************************************************************/





