/**********************************************************************
 **                                                                  **
 **                            console.c                             **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Communication interface to robot-executables                    **
 **                                                                  **
 **  Copyright 2000, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: console.c,v 1.7 2003/04/02 22:29:21 zkira Exp $ */

#include <pthread.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <math.h>

#include "hserver.h"
#include "camera.h"
#include "statusbar.h"
#include "robot_config.h"
#include "sensors.h"
#include "message.h"
#include "robot.h"
#include "ipt_handler.h"
#include "video.h"

#include "ipt/connection.h"
#include "hserver_ipt.h"

bool consoleConnected,useConsole;

bool disableConsoleRobotControl = false;

pthread_t console_reader;

IPConnection* console; 

char consoleRstr[ 100 ];

IPCommunicator* communicator;

int console_total_recieve = 0;

char hclient_ipt_name[ 100 ];

// Query for robot setup
void consoleReturnRobotSetup( IPCommunicator* comm, IPMessage* msg )
{
    REC_RobotSetupInfo rec;
    console_total_recieve += msg->SizeData();
    robot_config->fillIptMessage( &rec );
    comm->Reply( msg, MSG_RobotSetupInfo, &rec );
    if ( rec.num_parts > 0 ) delete [] rec.part;
}

// Query for robot battery readings (if available)  // trw zk
void consoleReturnBatteryReadings( IPCommunicator* comm, IPMessage* msg)
{
    REC_BatteryUpdate rec;

    if (robot != NULL)
      rec.battery_level = robot->getBattery();
    else
      rec.battery_level = -2;

    comm->Reply( msg, MSG_BatteryUpdate, &rec );

}

// Query for robot sensor readings
void consoleReturnSensorReadings( IPCommunicator* comm, IPMessage* msg )
{
    int i;
    REC_SensorUpdate rec;

    console_total_recieve += msg->SizeData();
    rec.num_sensors = sensors->numSensors();
    rec.sensor = (REC_SensorReadingT*) malloc( sizeof( REC_SensorReadingT ) * rec.num_sensors );
    for( i = 0; i < rec.num_sensors; i++ )
    {
        rec.sensor[ i ].type = sensors->latest_sensor[ i ]->getType();
        rec.sensor[ i ].len_extra = sensors->latest_sensor[ i ]->getLenExtra();
        rec.sensor[ i ].extra = sensors->latest_sensor[ i ]->getExtra();
        sensors->latest_sensor[ i ]->getAngleInfo( rec.sensor[ i ].angle, rec.sensor[ i ].num_angles );
        sensors->latest_sensor[ i ]->getReadingInfo( rec.sensor[ i ].reading, rec.sensor[ i ].num_readings );

        // if space for angles or readings couldn't be allocated, set their counts to 0
        if ( rec.sensor[ i ].angle == NULL )
        {
            rec.sensor[ i ].num_angles = 0;
        }
        if ( rec.sensor[ i ].reading == NULL )
        {
            rec.sensor[ i ].num_readings = 0;
        }
    }

    comm->Reply( msg, MSG_SensorUpdate, &rec );

    // clean up the copies
    for ( i = 0; i < rec.num_sensors; i++ )
    {
        if ( rec.sensor[ i ].angle != NULL )
        {
            delete [] rec.sensor[ i ].angle;
        }
        if ( rec.sensor[ i ].reading != NULL )
        {
            delete [] rec.sensor[ i ].reading;
        }
    }
}

void consoleRangeStop()
{
    if ( robot != NULL )
    {
        robot->sonarStop();
    }
    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printTextWindow( "Console: range stop" );
    }
}

void consoleRangeStart()
{
    if ( robot != NULL )
    {
        robot->sonarStart();
    }
    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printTextWindow( "Console: range start" );
    }
}

void consoleInitXYT()
{
    if ( robot != NULL )
    {
        robot->initXYT();
    }
    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printTextWindow( "Console: init xyt" );
    }
}
    
void consoleCloseRobot()
{
    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printTextWindow( "Console: close robot" );
    }
}

void consoleStopRobot()
{
    if ( !disableConsoleRobotControl )
    {
        if ( robot != NULL )
        {
            robot->stop();
        }
    }
    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printTextWindow( "Console: stop robot" );
    }
}    

void consoleDriveCommand( IPCommunicator* comm, IPMessage* msg, void* hndData )
{
    console_total_recieve += msg->SizeData();
    if ( disableConsoleRobotControl )
    {
        return;
    }
    REC_DriveCommand* rec = (REC_DriveCommand*) msg->FormattedData();
    
    if ( robot != NULL )
    {
        robot->drive( rec->vel );
    }
    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Console: drive command %f", rec->vel );
    }
}

void consoleSteerCommand( IPCommunicator* comm, IPMessage* msg, void* hndData )
{
    console_total_recieve += msg->SizeData();
    if ( disableConsoleRobotControl )
    {
        return;
    }
    REC_SteerCommand* rec = (REC_SteerCommand*) msg->FormattedData();

    if ( robot != NULL )
    {
        robot->steer( rec->avel );
    }

    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Console: steer command %f", rec->avel );
    }
}

void consoleSteerTowardDriveCommand( IPCommunicator *comm, IPMessage *msg, void *hndData )
{
    float angle;
    float speed;
    console_total_recieve += msg->SizeData();
    if ( disableConsoleRobotControl )
    {
        return;
    }
    REC_SteerTowardDriveCommand* rec = (REC_SteerTowardDriveCommand*) msg->FormattedData();

    angle = rec->theta;
    speed = rec->speed;
    if ( rec->use_reverse )
    {
        if ( fabs( angle ) > 90 )
        {
            angle += 180.0;
            speed *= -1;
        }
    }
    if ( angle < -180 )
    {
        angle += 360;
    }
    if ( angle > 180 )
    {
        angle -= 360; 
    }

    if ( fabs( angle ) > rec->drive_wait_angle )
    {
        speed = 0.0;
    }

    if ( robot != NULL )
    {
        robot->steerToward( int( angle ) );
        robot->drive( speed );
    }
    
    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Console: steer toward drive command %f %f %d", 
                          rec->theta, rec->speed, rec->use_reverse );
    }

}


void consoleFrameUpload( IPCommunicator* comm, IPMessage* msg, void* hndData )
{
    REC_FrameUpload* rec = (REC_FrameUpload*) msg->FormattedData(); 
    REC_SimpleCommand reply;

    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Console: frame upload name: %s  channel: %d ", 
                          rec->name, rec->channel );
    }

    if ( video == NULL )
    {
        reply.command = -1;
        comm->Reply( msg, MSG_SimpleCommand, &reply );
    }
    
    video->frameUpload( rec->channel, rec->name, rec->send, comm, msg );
    if ( !rec->send )
    {
        reply.command = 0;
        comm->Reply( msg, MSG_SimpleCommand, &reply );
    }
}

void consoleSetXYT( IPCommunicator* comm, IPMessage* msg, void* hndData )
{
    console_total_recieve += msg->SizeData();
    REC_SetXYT* rec = (REC_SetXYT*) msg->FormattedData();

    HserverSetXyt( rec->x, rec->y, rec->t );

    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Console: Set xyt command x: %f y:%f t: %f", rec->x, rec->y, rec->t );
    }
}

void consoleMoveCamera( IPCommunicator* comm, IPMessage* msg, void* hndData )
{
    console_total_recieve += msg->SizeData();
    REC_MoveCamera *rec = (REC_MoveCamera*) msg->FormattedData();
    if ( camera!=NULL )
    {
        camera->moveDeg( rec->pan, rec->tilt );
    }
}

void consoleSetCameraTrackerMode( IPCommunicator* comm, IPMessage* msg, void* hndData )
{
    console_total_recieve += msg->SizeData();
    REC_SetCameraTrackerMode* rec = (REC_SetCameraTrackerMode*) msg->FormattedData();
    if ( camera != NULL )
    {
        camera->setTrackerMode( rec->mode );
    }

    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Console: Set camera tracker mode %d", rec->mode );
    }
}

char* consoleSimpleName( int command )
{
    switch( command )
    {
    case HCLIENT_GET_ROBOT_SETUP_CMD:
        return "HCLIENT_GET_ROBOT_SETUP_CMD";
        break;

    case HCLIENT_RANGE_STOP_CMD:
        return "HCLIENT_RANGE_STOP_CMD";
        break;

    case HCLIENT_RANGE_START_CMD:
        return "HCLIENT_RANGE_START_CMD";
        break;

    case HCLIENT_INIT_XYT_CMD:
        return "HCLIENT_INIT_XYT_CMD";
        break;

    case HCLIENT_CLOSE_ROBOT_CMD:
        return "HCLIENT_CLOSE_ROBOT_CMD";
        break;

    case HCLIENT_STOP_CMD:
        return "HCLIENT_STOP_CMD";
        break;

    case HCLIENT_GET_SENSOR_READINGS_CMD:
        return "HCLIENT_GET_SENSOR_READINGS_CMD";
        break;

    case HCLIENT_GET_BATTERY_READINGS_CMD:         // trw zk
        return "HCLIENT_GET_BATTERY_READINGS_CMD";
        break;

    default:
        return "unknown";
    }
}

void consoleHandleSimpleCommand( IPCommunicator* comm, IPMessage* msg, void* hndData )
{
    REC_SimpleCommand reply;

    console_total_recieve += msg->SizeData();
    REC_SimpleCommand* rec = (REC_SimpleCommand*) msg->FormattedData();
    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Console: simple command %x: %s", 
                          rec->command, consoleSimpleName( rec->command ) );
    }
    switch( rec->command )
    {
    case HCLIENT_GET_ROBOT_SETUP_CMD:
        consoleReturnRobotSetup( comm, msg );
        break;

    case HCLIENT_RANGE_STOP_CMD:
        consoleRangeStop();
        break;

    case HCLIENT_RANGE_START_CMD:
        consoleRangeStart();
        break;

    case HCLIENT_INIT_XYT_CMD:
        consoleInitXYT();
        reply.command = 0;
        comm->Reply( msg, MSG_SimpleCommand, &reply );
        break;

    case HCLIENT_CLOSE_ROBOT_CMD:
        consoleCloseRobot();
        break;

    case HCLIENT_STOP_CMD:
        consoleStopRobot();
        break;

    case HCLIENT_GET_SENSOR_READINGS_CMD:
        consoleReturnSensorReadings( comm, msg );
        break;

    case HCLIENT_GET_BATTERY_READINGS_CMD:      // trw zk
        consoleReturnBatteryReadings(comm, msg);
        break;
   
    default:
        if ( iptHandler->reportLevel() >= HS_REP_LEV_ERR )
        {
            printfTextWindow( "Unknown simple command %x", rec->command );
        }
        break;
    }
    //statusbarConsoleUpdate();
}

void consoleHandleInitConnection( IPCommunicator* comm, IPMessage* msg, void* hndData )
{
    console_total_recieve += msg->SizeData();
    REC_InitConnection* rec = (REC_InitConnection*) msg->FormattedData();
    strcpy( hclient_ipt_name, rec->name );
    console = communicator->Connect( hclient_ipt_name, IPT_REQUIRED );
    if ( iptHandler->reportLevel() >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Initialized ipt connection with %s on %s", 
                 rec->name, console->Host() );
    }
}

void consoleRegisterHandlers()
{
    int num_handlers;
    IPMsgHandlerSpec hndArray[] = {
        { MSG_SimpleCommand ,          consoleHandleSimpleCommand,     IPT_HNDL_STD, NULL }, 
        { MSG_InitConnection,          consoleHandleInitConnection,    IPT_HNDL_STD, NULL }, 
        { MSG_DriveCommand,            consoleDriveCommand,            IPT_HNDL_STD, NULL }, 
        { MSG_SteerCommand,            consoleSteerCommand,            IPT_HNDL_STD, NULL }, 
        { MSG_SteerTowardDriveCommand, consoleSteerTowardDriveCommand, IPT_HNDL_STD, NULL }, 
        { MSG_SetXYT,                  consoleSetXYT,                  IPT_HNDL_STD, NULL }, 
        { MSG_MoveCamera,              consoleMoveCamera,              IPT_HNDL_STD, NULL }, 
        { MSG_SetCameraTrackerMode,    consoleSetCameraTrackerMode,    IPT_HNDL_STD, NULL }, 
        { MSG_FrameUpload,             consoleFrameUpload,             IPT_HNDL_STD, NULL }
    };
    num_handlers = sizeof( hndArray )/sizeof( IPMsgHandlerSpec );
    iptHandler->registerHandlers( num_handlers, hndArray );
}


void consoleDisableHandlers()
{
    int num_handlers;
    char* hndNames[] = {
        MSG_SimpleCommand, 
        MSG_InitConnection, 
        MSG_DriveCommand, 
        MSG_SteerCommand, 
        MSG_SteerTowardDriveCommand, 
        MSG_SetXYT, 
        MSG_MoveCamera, 
        MSG_SetCameraTrackerMode
    };
    num_handlers = sizeof( hndNames ) / sizeof( IPMessageType* );
    iptHandler->disableHandlers( num_handlers, hndNames );
}
	

void consoleConnect()
{
    communicator = iptHandler->communicator();
    iptHandler->registerNamedFormatters( iptHclientFormatArray );
    iptHandler->registerMessages( iptHclientMessageArray );
    consoleRegisterHandlers();
    iptHandler->set_state( IPT_MOD_CONSOLE, true, false );
    printTextWindow( "Hclient listening connected" );
}

void consoleDisconnect()
{
    consoleDisableHandlers();
}

/**********************************************************************
# $Log: console.c,v $
# Revision 1.7  2003/04/02 22:29:21  zkira
# Added consoleGetBatteryReadings and battery message cases
#
# Revision 1.6  2001/05/29 22:37:50  blee
# Made changes to incorporate gps data w/ other position info.
#
# Revision 1.5  2000/12/12 22:52:29  blee
# Changed consoleReturnSensorReadings() to copy sensor readings before
# writing them.
#
# Revision 1.4  2000/09/19 03:11:10  endo
# RCS log added.
#
#
#**********************************************************************/
