/**********************************************************************
 **                                                                  **
 **                              fred.c                              **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Robot emulation for HServer                                     **
 **                                                                  **
 **  Copyright 2000, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: fred.c,v 1.7 2001/12/25 22:54:53 endo Exp $ */

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <pthread.h>
#include <math.h>
#include <string.h>

#include "hserver.h"
#include "sensors.h"
#include "robot_config.h"
#include "message.h"
#include "statusbar.h"
#include "ipm/IPM_timer.h"
#include "fred.h"

void Fred::statusbarUpdate()
{
    char buf[ 100 ];
    float fX, fY, fT;

    statusStr[ 0 ] = 0;
    strcat( statusStr, "Fred: " );
    getXYT( fX, fY, fT );
    sprintf( buf, "x: %.2f  y: %.2f  t: %.2f", fX, fY, fT );
    strcat( statusStr, buf );
    statusbar->update( statusLine );
}

static void start_thread( void* fred_instance )
{
    ( (Fred*) fred_instance )->mover_thread();
}

void Fred::reset()
{
    Robot::setXYT( 0.0, 0.0, 0.0 );
}

void Fred::mover_thread()
{
    float fX, fY, fXDiff, fYDiff, time, fTheta, fThetaInRad, dist, adist;
    IPM_timer t;
    
    IPM_timer_clear( &t );
    IPM_timer_start( &t );

    while( true )
    {
        pthread_testcancel();
        usleep( 10000 );
        getXYT( fX, fY, fTheta );
        fThetaInRad = fTheta * M_PI / 180.0;
        IPM_timer_stop( &t );
        time = IPM_timer_read( &t );
        dist  = time * velocity;
        adist = time * avelocity;
        fYDiff = sin( fThetaInRad ) * dist;
        fXDiff = cos( fThetaInRad ) * dist;
        fX += fXDiff;
        fY += fYDiff;
        fTheta += adist;
        if ( fTheta >= 360.0 )
        {
            fTheta -= 360.0;
        }
        if ( fTheta <= -360.0 )
        {
            fTheta += 360.0;
        }
        Robot::setXYT( fX, fY, fTheta );
        statusbarUpdate();
        IPM_timer_clear( &t );
        IPM_timer_start( &t );
    }
}

void Fred::steer( float avel )
{
    avelocity = avel * angular_speed_factor / 10.0;
    if ( report_level >= HS_REP_LEV_DEBUG )
    { 
        printfTextWindow( "Fred steer: %f", avelocity );
    }
}


void Fred::drive( float vel )
{
    velocity = vel * speed_factor / 10.0;
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Fred drive: %f", velocity );
    }
}

void Fred::stop()
{
    velocity  = 0;
    avelocity = 0;
}

void Fred::steerToward( int angle )
{
    float avel; // deg/sec
    int absang;
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Fred steer towards: %d", angle );
    }
    absang = abs( angle );
    if ( absang > 90 )
    {
        avel = 160;
    }
    else if ( absang > 60 )
    {
        avel = 140;
    }
    else if ( absang > 30 )
    {
        avel = 120;
    }
    else
    {
        avel = 80;
    }
    
    if ( angle < 0 )
    {
        avel*=-1;
    }
    if ( fabs( angle ) < 4 )
    {
        avel = 0;
    }
    steer( avel );
}

void Fred::adjustXYT( float xinc, float yinc, float tinc )
{
    float fX, fY, fTheta;

    getXYT( fX, fY, fTheta );
    fX += xinc / 1000.0;
    fY += yinc / 1000.0;
    fTheta -= tinc / 10.0;

    while ( fTheta < 0 )
    {
        fTheta += 360;
    }
    while ( fTheta >= 360 )
    {
        fTheta -= 360;
    }
    Robot::setXYT( fX, fY, fTheta );
    statusbarUpdate();
}

void Fred::initXYT()
{
    Robot::setXYT( 0, 0, 0 );
    statusbarUpdate();
}

void Fred::setXYT( float xi, float yi, float theta )
{
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Fred setxyt  %f %f %f", xi, yi, theta );
    }
    Robot::setXYT( xi, yi, theta );
    statusbarUpdate();
}

void Fred::addSensors()
{
    partXyt = new RobotPartXyt();
}

void Fred::control()
{
    bool bDone = false;
    int msgData[ 2 ];

    msgData[ 0 ] = (int) speed_factor;
    msgData[ 1 ] = (int) angular_speed_factor;

    messageDrawWindow( EnMessageType_FRED_CONTROL, EnMessageErrType_NONE, msgData );
    do
    {
        int c = getch();
        switch( c )
        {
        case 'd':
            messageHide();
            delete this;
            return; // never reached ??
            break;

        case 'i':
            setXYT(5,5,0);
            break;

        case '+':
            angular_speed_factor++;
            msgData[ 1 ] = (int) angular_speed_factor;
            messageDrawWindow( EnMessageType_FRED_CONTROL, EnMessageErrType_NONE, msgData );
            break;

        case '=':
            speed_factor++;
            msgData[0] = (int) speed_factor;
            messageDrawWindow( EnMessageType_FRED_CONTROL, EnMessageErrType_NONE,msgData );
            break;

        case '_':
            angular_speed_factor--;
            if ( angular_speed_factor < 0 )
            {
                angular_speed_factor = 0;
            }
            msgData[ 1 ] = (int) angular_speed_factor;
            messageDrawWindow( EnMessageType_FRED_CONTROL, EnMessageErrType_NONE, msgData );
            break;

        case '-':
            speed_factor--;
            if ( speed_factor < 0 )
            {
                speed_factor=0;
            }
            msgData[ 0 ] = (int) speed_factor;
            messageDrawWindow( EnMessageType_FRED_CONTROL, EnMessageErrType_NONE, msgData );
            break;

        case 'x':
        case 'Q':
        case KEY_ESC:
            bDone = true;
            break;

        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey( c );
            break;

        }
    } while ( !bDone );
    messageHide();
}

Fred::~Fred()
{
    pthread_cancel( mover );
    pthread_join( mover, NULL );
    delete partXyt;
}

Fred::Fred( Robot** a ) : Robot( a, rt_fred, "Fred" )
{
    module_type = MODULE_ROBOT_FRED;
    velocity = 0;
    avelocity = 0;
    report_level = HS_REP_LEV_NORMAL;
    addSensors();
    Robot::setXYT( 0, 0, 0 );
    statusbarUpdate();
    pthread_create( &mover, NULL, (void*(*)(void*)) &start_thread, (void*) this );
}

/**********************************************************************
# $Log: fred.c,v $
# Revision 1.7  2001/12/25 22:54:53  endo
# *** empty log message ***
#
# Revision 1.6  2001/05/29 22:35:41  blee
# Made changes to incorporate gps data w/ other position info.
#
# Revision 1.5  2000/12/12 23:23:38  blee
# Changed #defines for the user interface to enumerated types.
#
# Revision 1.4  2000/09/19 03:11:10  endo
# RCS log added.
#
#
#**********************************************************************/
