/**********************************************************************
 **                                                                  **
 **                           gyro_dmu-vgx.c                         **
 **                                                                  **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  Corssbow DMU-VGX Gyroscope driver for HServer                   **
 **                                                                  **
 **  Note: HServer reads DMU data from mobility. mobility has to be  **
 **        installed in the system.                                  **
 **                                                                  **
 **  Copyright 2001, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: gyro_dmu-vgx.c,v 1.2 2001/12/22 16:35:40 endo Exp $ */

#include "gyro_dmu-vgx.h"

// This macro takes a degree (d) and change its value to be inside
// the lower bound (l) and the higher bound (h).
#define GYRODMUVGX_CRAMPDEG(d,l,h) {while((d)<(l))(d)+=360.0; while((d)>=(h))(d)-=360.0;}

// The pointer to the gyro class.
// jbl GyroDMUVGX *gGyroDMUVGX = NULL;

// Name of the DMU server.
const char *GYRODMUVGX_DMUSERVER_NAME = "crossbow";

// How often the data from mobility should be updated (in microsecond).
const int GYRODMUVGX_READER_THREAD_USLEEP = 1000;

// How often the status bar updated with respect to the sampling time
// in the reader thread.
const int GYRODMUVGX_SKIP_STATUSBAR_UPDATE = 10;

// Initial values for the parameters to convert from mobility
// DMU data to HServer angles in degree. *calib_factor* multiplies
// and *offset* adds values to the DMU data. The multiplication
// is done first.
const double GYRODMUVGX_INIT_CALIB_FACTOR_YAW = -1.0;
const double GYRODMUVGX_INIT_CALIB_FACTOR_PITCH = 1.0;
const double GYRODMUVGX_INIT_CALIB_FACTOR_ROLL = 1.0;
const double GYRODMUVGX_INIT_OFFSET_YAW = 0.0;
const double GYRODMUVGX_INIT_OFFSET_PITCH = 0.0;
const double GYRODMUVGX_INIT_OFFSET_ROLL = 0.0;

// Pthread mutex to for HServer angles.
pthread_mutex_t gyroDMUVGX_positionMutex_ = PTHREAD_MUTEX_INITIALIZER;

// This function calls the reader thread.
void* GyroDMUVGX::GyroDMUVGX_startMainThread( void* gyro_dmu_vgx_instance )
{
    ((GyroDMUVGX*)gyro_dmu_vgx_instance)->readerLoop();
    return NULL;
}

// Constructor for the Gyro DMU-VGX class.
GyroDMUVGX::GyroDMUVGX(GyroDMUVGX** a) : Module( (Module**) a, "GyroDMUVGX")
{
    int rtn;

    if (report_level > HS_REP_LEV_NONE) printTextWindow("Connecting to Gyro DMU-VGX... ", false);

    firstDataReceived_ = FALSE;
    memset(gyroDMUVGX_xyt_, 0x0, sizeof(float)*3);
    gyroDMUVGXAngPosCalibFactor_.yaw = GYRODMUVGX_INIT_CALIB_FACTOR_YAW;
    gyroDMUVGXAngPosCalibFactor_.pitch = GYRODMUVGX_INIT_CALIB_FACTOR_PITCH;
    gyroDMUVGXAngPosCalibFactor_.roll = GYRODMUVGX_INIT_CALIB_FACTOR_ROLL;
    gyroDMUVGXAngPosOffset_.yaw = GYRODMUVGX_INIT_OFFSET_YAW;
    gyroDMUVGXAngPosOffset_.pitch = GYRODMUVGX_INIT_OFFSET_PITCH;
    gyroDMUVGXAngPosOffset_.roll = GYRODMUVGX_INIT_OFFSET_ROLL;

    rtn = initializeMobility_();
    refreshScreen();

    if (rtn)
    {
        if (report_level > HS_REP_LEV_NONE) printTextWindow("failed", true);
        return;
    }

    // Start reading the DMU-VGX data.
    pthread_create( &readerThread_, NULL, &GyroDMUVGX_startMainThread, (void*) this );

    while (!firstDataReceived_) usleep(GYRODMUVGX_READER_THREAD_USLEEP);

    if (report_level > HS_REP_LEV_NONE) printTextWindow( "connected", true );
    resetAngles();
    statusbarUpdate_();
}

// Distructor for the Gyro DMU-VGX class.
GyroDMUVGX::~GyroDMUVGX()
{
    pthread_cancel(readerThread_);
    pthread_join(readerThread_, NULL );

    printTextWindow( "Gyro DMU-VGX disconnected" );
}

// This function updates the status bar.
void GyroDMUVGX::statusbarUpdate_()
{
    static int c = 0;
    char buf[100];
  
    statusStr[0] = 0;

    strcat(statusStr, "Gyro DMU-VGX: ");

    switch(c) {
    case 0:
        strcat(statusStr, "-");
        break;
    case 1:
        strcat(statusStr, "\\");
        break;
    case 2:
        strcat(statusStr, "|");
        break;
    case 3:
        strcat(statusStr, "/");
        break;
    }
    c = (c + 1) % 4;

    sprintf(buf, " angles yaw: %.2f pitch: %.2f roll: %.2f [deg]",
	    gyroDMUVGXAngPos_.yaw, gyroDMUVGXAngPos_.pitch, gyroDMUVGXAngPos_.roll);
    strcat(statusStr, buf);
    statusbar->update(statusLine);
}

// This function sets up the interface between HServer and mobility. 
int GyroDMUVGX::initializeMobility_(void)
{
    int argc;
    char* argv[4];
    char* gyroName = strdup(GYRODMUVGX_DMUSERVER_NAME);
    char pathName[255];

    argv[0] = "HServer";
    argv[1] = "DMU";
    argv[2] = "VGX";
    argv[3] = 0;
    argc = 3;

    pHelper_ = new mbyClientHelper(argc, argv);

    sprintf(pathName, "%s/Drive/State", gyroName);

    ptempObj_ = pHelper_->find_object(pathName);

    pMbyGyroState_ = MobilityActuator::ActuatorState::_duplicate(
        MobilityActuator::ActuatorState::_narrow(ptempObj_));

    return 0;
}  

// This function loops around to get the gyroscopic data DMU-VGX.
void GyroDMUVGX::readerLoop(void)
{
    int count = 0;

    while(TRUE)
    {
	// Make sure termination was not requested.
        pthread_testcancel();

	// Sleep a bit to reduce the CPU load.
	usleep(GYRODMUVGX_READER_THREAD_USLEEP);

	// Check for the pthread termination again.
        pthread_testcancel();

	// Get the data from DMU-VGX.
        updateGyroData_();


	// Show updated data to the screen.
	if (count > GYRODMUVGX_SKIP_STATUSBAR_UPDATE)
	{
	    statusbarUpdate_();
	    count = 0;
	}

	count ++;
    }
}

// This function updates the gyroscopic data by reading it
// from mobility.
void GyroDMUVGX::updateGyroData_(void)
{
    // Read the raw data from mobility.
    rawGyroData_ = readMbyGyroData_();

    // Computer HServer version of angular position.
    pthread_mutex_lock(&gyroDMUVGX_positionMutex_);
    gyroDMUVGXAngPos_ = correctAngularPosition_(rawGyroData_.angPos);
    gyroDMUVGX_xyt_[2] = gyroDMUVGXAngPos_.yaw;
    pthread_mutex_unlock(&gyroDMUVGX_positionMutex_);

    if (!firstDataReceived_) firstDataReceived_ = TRUE;
}

// This function grabs the DMU-VGX data from mobility
GyroDMUVGX_GyroData_t GyroDMUVGX::readMbyGyroData_(void)
{
    GyroDMUVGX_GyroData_t gyroData;
    struct timeval timeVal;
    struct timezone timeZone;

    // Put the time stamp.
    gettimeofday(&timeVal, &timeZone);
    gyroData.timeStamp = timeVal;

    // Grab the latest data set from mobility.
    pMbyGyroData_ = pMbyGyroState_->get_sample(0); 

    // Retrieve the positional data.
    gyroData.angPos.roll = pMbyGyroData_->position[3];
    gyroData.angPos.pitch = pMbyGyroData_->position[4];
    gyroData.angPos.yaw = pMbyGyroData_->position[5];

    // Retrieve the veloicty data.
    gyroData.angVel.roll = pMbyGyroData_->velocity[3];
    gyroData.angVel.pitch = pMbyGyroData_->velocity[4];
    gyroData.angVel.yaw = pMbyGyroData_->velocity[5];

    // Obtain X, Y, Z acceleration in terms of G.
    gyroData.accel.x = pMbyGyroData_->acceleration[0];
    gyroData.accel.y = pMbyGyroData_->acceleration[1];
    gyroData.accel.z = pMbyGyroData_->acceleration[2];

    return gyroData;
}

// This function converts the DMU-VGX data value into HServer
// yaw, pitch, and roll angles in degree.
GyroDMUVGX_YPR_t GyroDMUVGX::correctAngularPosition_(GyroDMUVGX_YPR_t rawAngularPosition)
{
    GyroDMUVGX_YPR_t angPos;

    angPos = rawAngularPosition;

    angPos.yaw *= gyroDMUVGXAngPosCalibFactor_.yaw;
    angPos.yaw += gyroDMUVGXAngPosOffset_.yaw;
    GYRODMUVGX_CRAMPDEG(angPos.yaw,0.0,360.0);

    angPos.pitch *= gyroDMUVGXAngPosCalibFactor_.pitch;
    angPos.pitch += gyroDMUVGXAngPosOffset_.pitch;
    GYRODMUVGX_CRAMPDEG(angPos.pitch,-180.0,180.0);

    angPos.roll *= gyroDMUVGXAngPosCalibFactor_.roll;
    angPos.roll += gyroDMUVGXAngPosOffset_.roll;
    GYRODMUVGX_CRAMPDEG(angPos.roll,-180.0,180.0);

    return angPos;
}

// This function sets the Gyro angles to be the specified value.
void GyroDMUVGX::setAngleYaw_(double yaw)
{
    pthread_mutex_lock(&gyroDMUVGX_positionMutex_);

    gyroDMUVGXAngPosOffset_.yaw += yaw - gyroDMUVGXAngPos_.yaw;
    gyroDMUVGXAngPos_.yaw = yaw;
    GYRODMUVGX_CRAMPDEG(gyroDMUVGXAngPos_.yaw,0.0,360.0);

    pthread_mutex_unlock(&gyroDMUVGX_positionMutex_);
}

// This function sets the Gyro angles to be the specified value.
void GyroDMUVGX::setAnglePitch_(double pitch)
{
    pthread_mutex_lock(&gyroDMUVGX_positionMutex_);

    gyroDMUVGXAngPosOffset_.pitch += pitch - gyroDMUVGXAngPos_.pitch;
    gyroDMUVGXAngPos_.pitch = pitch;
    GYRODMUVGX_CRAMPDEG(gyroDMUVGXAngPos_.pitch,-180.0,180.0);

    pthread_mutex_unlock(&gyroDMUVGX_positionMutex_);
}

// This function sets the Gyro angles to be the specified value.
void GyroDMUVGX::setAngleRoll_(double roll)
{
    pthread_mutex_lock(&gyroDMUVGX_positionMutex_);

    gyroDMUVGXAngPosOffset_.roll += roll - gyroDMUVGXAngPos_.roll;
    gyroDMUVGXAngPos_.roll = roll;
    GYRODMUVGX_CRAMPDEG(gyroDMUVGXAngPos_.roll,-180.0,180.0);

    pthread_mutex_unlock(&gyroDMUVGX_positionMutex_);
}

// This function resets the gyro angles to be zeros.
void GyroDMUVGX::resetAngles(void)
{
    setAngleYaw_(0.0);
    setAnglePitch_(0.0);
    setAngleRoll_(0.0);
}

// This function sets the position of the gyro to be specified
// values. The theta (t) will be set by setting its angle yaw.
void GyroDMUVGX::setXYT(float x, float y, float t)
{
    pthread_mutex_lock(&gyroDMUVGX_positionMutex_);
    gyroDMUVGX_xyt_[0] = x;
    gyroDMUVGX_xyt_[1] = y;
    pthread_mutex_unlock(&gyroDMUVGX_positionMutex_);

    setAngleYaw_(t);
}

// This function returns position of the gyro.
void GyroDMUVGX::getXYT(float *x, float *y, float *t)
{
    pthread_mutex_lock(&gyroDMUVGX_positionMutex_);
    *x = gyroDMUVGX_xyt_[0];
    *y = gyroDMUVGX_xyt_[1];
    *t = gyroDMUVGX_xyt_[2];
    pthread_mutex_unlock(&gyroDMUVGX_positionMutex_);
}
/* jbl
// Control menus.
void GyroDMUVGX::control(void)
{
    int c;
    int done = 0;

    messageDrawWindow( EnMessageType_GYRO_DMU_VGX_CONTROL, EnMessageErrType_NONE );
    redrawWindows();
    do
    {
        c = getch();
        switch( c )
        {
        case 'd':
            messageHide();
            delete this;
            return;
            break;
        case 'r':
	    resetAngles();
            break;
        case 'x':
        case 'Q':
        case KEY_ESC:
            done = true;
            break;
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey( c );
            break;
        }
    } while ( !done );
    messageHide();
}
*/
/* jbl
// Instantiate Gyro DMU-VGX class.
void gyroDMUVGXConnect(void)
{
    new GyroDMUVGX(&gGyroDMUVGX);
}
*/
/**********************************************************************
# $Log: gyro_dmu-vgx.c,v $
# Revision 1.2  2001/12/22 16:35:40  endo
# CALIB_FACTOR changed due to the change in the DMU server.
#
# Revision 1.1  2001/05/18 17:54:47  blee
# Initial revision
#
#**********************************************************************/
