/**********************************************************************
 **                                                                  **
 **                               nomad.c                            **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Nomad 150/200 robot driver for HServer                          **
 **                                                                  **
 **  Copyright 2000, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: nomad.c,v 1.9 2003/04/06 13:32:37 endo Exp $ */

#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include "hserver.h"
#include "message.h"
#include "statusbar.h"
#include "ndirect_hs.h"
#include "nomad.h"
#include "robot_config.h"
#include "sensors.h"
#include "WatchDog.h"


int translation_accel = 100; // default accel .1 inch/s/s
int translation_speed = 100; // default speed .1 inch/s
int steer_accel       = 450; // default steer accel .1 deg/s/s
int steer_speed       = 450; // default steer speed .1 deg/s
int turret_accel      = 450; // default accel .1 deg/s/s
int turret_speed      = 450; // default speed .1 deg/s


pthread_mutex_t nomadMutex;

pthread_cond_t nomad_cond;
pthread_mutex_t nomad_cond_mutex;

float watchdog_speed;

void Nomad::statusbarUpdate()
{

    char buf[ 100 ];
    statusStr[ 0 ] = 0;

    strcat( statusStr, "Nomad: " );
    
    if ( mode == NOMAD_MODE_NORMAL ){
        spinner = ( spinner +1 )%4;
        switch ( spinner )
        {
        case 0:
            strcat( statusStr, "-" );
            break;
        case 1:
            strcat( statusStr, "\\" );
            break;
        case 2:
            strcat( statusStr, "|" );
            break;
        case 3:
            strcat( statusStr, "/" );
            break;
        }
    } 

    float fX, fY, fTheta;
    getXYT( fX, fY, fTheta );
    sprintf( buf, " x: %.2f y: %.2f t: %.2f ", fX, fY, fTheta );
    strcat( statusStr, buf );

    if ( is200 )
    {
        sprintf( buf, " cpu: %.2fV motor: %.2fV ", voltages[ 0 ], voltages[ 1 ] );
        strcat( statusStr, buf );
    }	

    if ( mode == NOMAD_MODE_ZEROING )
    {
        strcat( statusStr, "... zeroing" );
    }
    else if ( mode == NOMAD_MODE_COMPASS )
    {
        strcat( statusStr, "... calibrating compass" );
    }    

    statusbar->update( statusLine );

    /* Ernest start*/
    /* if watchdog function is enabled, then periodically update a timestamp for the 
       WatchDog class to check. */
    if ( watchDogEnable == true)
    {
	 pthread_cleanup_push((void(*)(void*)) pthread_mutex_unlock, (void *) &nomad_watchdog_mutex);
	 pthread_mutex_lock(&nomad_watchdog_mutex);
	 gettimeofday( &tv_watchdog,NULL);
	 pthread_cleanup_pop(1);
    }
    /* Ernest end */
}


int Nomad::compassActivate()
{
    int rtn;
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printTextWindow( "Nomad compass activate" );
    }
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    rtn = conf_cp( 1 );
    pthread_cleanup_pop( 1 );
    if ( !rtn )
    {
        if ( report_level >= HS_REP_LEV_ERR )
        {
            printTextWindow( "Error activating compass" );
        }
        return -1;
    }
    return 0;
}

int Nomad::compassCalibrate()
{
    int rtn;
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printTextWindow( "Nomad compass calibrate" );
    }
    if ( !is200 )
    {
        return -1;
    }
    mode = NOMAD_MODE_COMPASS;
    statusbarUpdate();    
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );

    rtn = conf_cp( 2 );
    pthread_cleanup_pop( 1 );
    if ( !rtn )
    {
        if ( report_level >= HS_REP_LEV_ERR )
        {
            printTextWindow( "Error calibrating compass" );
        } 
        rtn = -1;
    }
    else
    {
        rtn = 0;
    }
    mode = NOMAD_MODE_NORMAL;
    statusbarUpdate();    
    return rtn;
}

void Nomad::setSonarSpeed( int speed )
{
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Nomad set sonar speed: %d", speed );
    }
    sonar_speed = speed;
    conf_sn( sonar_speed, sonar_order );
}

void Nomad::sonarStart()
{
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printTextWindow( "Nomad sonar start" );
    }
    conf_sn( sonar_speed, sonar_order );
}

void Nomad::sonarStop()
{
    int	order[ 16 ] = { 255 };
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printTextWindow( "Nomad sonar stop" );
    }
    conf_sn( 0, order );
}

void Nomad::drive( float vel )	// m per sec 
{
    int vel_i;
    int rtn;

    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Nomad drive: %f", vel );
    }
    watchdog_speed = vel;
    vel = vel * 393.70079;  /* convert to .1 inches per sec */

    vel_i = (int) vel;

    if ( vel_i > NOMAD_MAX_VEL )
    {
        vel_i = NOMAD_MAX_VEL;
    }
    if ( vel_i < -NOMAD_MAX_VEL )
    {
        vel_i = -NOMAD_MAX_VEL;
    }
    translation_speed = vel_i;
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    rtn = mv( MV_VM, translation_speed, MV_IGNORE, MV_IGNORE, MV_IGNORE, MV_IGNORE );
    pthread_cleanup_pop( 1 );

}

/* Test robot connection */
int Nomad::ping( bool output )
{
    int rtn;
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        if ( output )
        {
            printTextWindow( "Nomad ping" );
        }
    }
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    rtn = mv( MV_IGNORE, MV_IGNORE, MV_IGNORE, MV_IGNORE, MV_IGNORE, MV_IGNORE );
    pthread_cleanup_pop( 1 );
    if ( output )
    {
        if ( rtn )
        {
            printTextWindow( "Nomad ping sucessful" );
        }
        else
        {
            printTextWindow( "Nomad ping unsucessful" );
        }
    }
    return rtn;
}

void Nomad::steer( float rvel ) // deg per sec
{
    int irvel;

    irvel = int( rvel );
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Nomad steer: %i", irvel );
    }
    
    if ( irvel > NOMAD_MAX_RVEL )
    {
        irvel = NOMAD_MAX_RVEL;
    }
    if ( irvel < -NOMAD_MAX_RVEL )
    {
        irvel = -NOMAD_MAX_RVEL;
    }
    steer_speed = irvel * 10;
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    mv( MV_IGNORE, MV_IGNORE, MV_VM, steer_speed, MV_VM, steer_speed );
    pthread_cleanup_pop( 1 );
}

void Nomad::stop()
{
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printTextWindow( "Nomad stop" );
    }
    drive( 0 );
    steer( 0 );
}


int Nomad::updateCompass()
{
    int rtn;
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    rtn = get_cp();
    pthread_cleanup_pop( 1 );
    if ( !rtn )
    {
        if ( report_level >= HS_REP_LEV_ERR )
        {
            printTextWindow( "Error updating compass" );
        }
        return !rtn;
    }
    compass[ 0 ] = ( (float) State[ STATE_COMPASS ] ) / 10.0;
    return !rtn;
}

int Nomad::updateSonar()
{
    int i, rtn;
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    rtn = get_sn();
    pthread_cleanup_pop( 1 );
    for( i = 0; i < 16; i++ )
    {
        sonar[ i ] = ( (float) State[ i + STATE_SONAR_0 ] ) * .0254 + radius; 
    }
    return !rtn;
}

int Nomad::updateIR()
{
    int i, rtn;
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    rtn = get_ir();
    pthread_cleanup_pop( 1 );
    for( i = 0; i < 16; i++ )
    {
        if ( State[ i + STATE_IR_0 ] == 15 )
        {
            ir[ i ] = 50.0;
        }
        else
        {
            ir[ i ] = ( (float) State[ i + STATE_IR_0 ] ) * .0254 + radius; 
        }
    }
    return !rtn;
}

int Nomad::zero()
{
    int rtn;
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printTextWindow( "Nomad zero" );
    }
    mode = NOMAD_MODE_ZEROING;
    statusbarUpdate();
    WatchDog->DisableNomadWD();
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    rtn = zr();
    pthread_cleanup_pop( 1 );
    WatchDog->EnableNomadWD();
    if ( !rtn )
    {
        if ( report_level >= HS_REP_LEV_ERR )
        {
            printTextWindow( "Error zeroing nomad" );
        }
        rtn = -1;
    }
    else
    {
        rtn = 0;
    }
    mode = NOMAD_MODE_NORMAL;
    statusbarUpdate();
    return rtn;
}

void Nomad::initXYT()
{
    zero();
}

void Nomad::setXYT( float x, float y, float t )
{
    long lx, ly;
    int steering, turret;
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Nomad set xyt: %f %f %f", x, y, t );
    }
    lx = (long) ( x * 393.70079 );  // meters to .1 inches
    ly = (long) ( y * 393.70079 );  // meters to .1 inches
    steering = turret = (int) ( t * 10 );	// degrees to .1 degrees
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    dp( lx, ly );
    da( steering, turret );
    pthread_cleanup_pop( 1 );
}

void Nomad::adjustXYT( float xinc, float yinc, float tinc )
{
    float fX, fY, fTheta;

    getXYT( fX, fY, fTheta );
    fX += xinc / 1000.0;
    fY += yinc / 1000.0;
    fTheta -= tinc / 10.0;
    while ( fTheta < 0 )
    {
        fTheta += 360;
    }
    while ( fTheta > 360 )
    {
        fTheta -= 360;
    }
    Robot::setXYT( fX, fY, fTheta );
    statusbarUpdate();
}

void Nomad::waitNext()
{
    pthread_mutex_lock( &nomad_cond_mutex );
    pthread_cond_wait( &nomad_cond, &nomad_cond_mutex );
    pthread_mutex_unlock( &nomad_cond_mutex );   
}


void Nomad::setTimeout( unsigned char timeout )
{
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Nomad timeout: %d", timeout );
    }
    this->timeout = timeout;
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    conf_tm( timeout );
    pthread_cleanup_pop( 1 );
}

void Nomad::turn( float degrees, bool wait )
{
    int deg;

    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printfTextWindow( "Nomad turn: %f %d", degrees, wait );
    }
    deg = (int) ( degrees * 10.0 );
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    mv( MV_IGNORE, MV_IGNORE, MV_PR, deg, MV_PR, deg );
    if ( wait )
    {
        ws( false, true, true, NOMAD_WAIT );
    }
    pthread_cleanup_pop( 1 );
}

void Nomad::steerToward( int angle )
{
    turn( (float) angle, false );
}

void Nomad::turnTo( float direction, bool wait )
{
    int deg;
    float fX, fY, fTheta;

    getXYT( fX, fY, fTheta );
    direction -= fTheta;
    if ( direction < -180 )
    {
        direction += 360;
    }
    if ( direction > 180 )
    {
        direction -= 360;
    }
    deg = (int) ( direction * 10.0 );
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    mv( MV_IGNORE, MV_IGNORE, MV_PR, deg, MV_PR, deg );
    pthread_cleanup_pop( 1 );
    if ( wait )
    {
        waitNext();
        //ws(false, true, true, NOMAD_WAIT );
        while( true )
        {
            if ( State[ STATE_VEL_STEER ] == 0 )
            {
                break;
            }
            usleep( 1 );
        }
    }
}

void Nomad::moveDistance( float distance, bool wait )
{
    // The input distance in "feet".

    int dist;

    dist = int( distance * 120.0 );
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    mv( MV_PR, dist, MV_IGNORE, MV_IGNORE, MV_IGNORE, MV_IGNORE );
    pthread_cleanup_pop( 1 );
    if ( wait )
    {
        waitNext();
        //ws( true, false, false, NOMAD_WAIT );
        while( true )
        {
            if ( State[ STATE_VEL_TRANS ] == 0 )
            {
                break;
            }
            usleep( 1 );
        }
    }
}

void Nomad::updateVoltages()
{
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    voltages[ 0 ] = voltCpuGet();
    voltages[ 1 ] = voltMotorGet();
    battery_level  = voltages[ 1 ];
    pthread_cleanup_pop( 1 );
}	    

int Nomad::updateLocation()
{
    int rtn;
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    rtn = get_rc();
    rtn = get_rv();
    pthread_cleanup_pop( 1 );
    Robot::setXYT( State[ STATE_CONF_X ] * 0.00254, 
                   State[ STATE_CONF_Y ] * 0.00254, 
                   State[ STATE_CONF_STEER ] * 0.1 );
    arm[ 0 ] = State[ STATE_CONF_TURRET ] * 0.1;
    return !rtn;
}

int Nomad::setup()
{
    int irorder[ 16 ], firerate, i;
    int	order[ 16 ] = { 0, 2, 15, 1, 14, 3, 13, 4, 12, 5, 11, 6, 10, 7, 9, 8 };

    mode = NOMAD_MODE_NORMAL;
    statusbarUpdate();

    for( i = 0; i < 16; i++ )
    {
        irorder[ i ] = order[ i ];
        sonar_order[ i ] = order[ i ];
    }
    firerate = 0;
    pthread_cleanup_push( (void(*)(void*)) pthread_mutex_unlock, (void*) &nomadMutex );
    pthread_mutex_lock( &nomadMutex );
    conf_sn( firerate, sonar_order );
    sonarStop();
    if ( is200 )
    {
        conf_ir( 2, irorder );
    }
    ac( translation_accel, steer_accel, turret_accel );
    sp( translation_speed, steer_speed, turret_speed );
    conf_tm( timeout ); 
    pthread_cleanup_pop( 1 );
    if ( is200 )
    {
        compassActivate();
    }
    updateVoltages();
    translation_speed = steer_speed = turret_speed = 0;
    mode = NOMAD_MODE_NORMAL;
    statusbarUpdate();
    return 0;
}

static void start_thread( void* nomad_instance )
{
    ( (Nomad*) nomad_instance )->reader_thread();
}

void Nomad::reader_thread()
{
    int rtn;
    int cnt = 0;
    int batteryUpdate = 0;
    rtn = setup();
    if ( rtn )
    {
        pthread_exit( (void*) -1 );
    }
    reader_running = true;
    while( true )
    {
        pthread_testcancel();
        if ( timeout && ( ++cnt > 10 ) )
        {
            drive( watchdog_speed );
            cnt = 0;
        }
        if ( batteryUpdate++ > 100 )
        {
            updateVoltages();
        }	    
        rtn = updateSonar();
        if ( is200 )
        {
            rtn = updateIR();
            rtn = updateCompass();
        }
        rtn = updateLocation();
        pthread_cond_broadcast( &nomad_cond );
        statusbarUpdate();
        usleep( 1 );
    }
}

void Nomad::addSensors()
{
    int i;
    partXyt = new RobotPartXyt();

    partSonar = new RobotPartSonar( NOMAD_NUM_SONAR, sonar_angle );
    sensorSonar = new Sensor( SENSOR_SONAR, 0, NULL, NOMAD_NUM_SONAR, sonar_angle, NOMAD_NUM_SONAR, sonar );
    for( i = 0; i < NOMAD_NUM_SONAR; i++ )
    {
        sonar[ i ] = 0.0;
    }

    partArm = new RobotPartArm();
    sensorArm = new Sensor( SENSOR_ARM, 0, NULL, 1, arm );
    arm[ 0 ] = 0.0;

    if ( is200 )
    {
        partIr = new RobotPartIr( NOMAD_NUM_IR, sonar_angle );
        sensorIr = new Sensor( SENSOR_IR, 0, NULL, NOMAD_NUM_IR, sonar_angle, NOMAD_NUM_IR, ir );
        for( i = 0; i < NOMAD_NUM_IR; i++ )
        {
            ir[ i ] = 0.0;
        }
        partCompass = new RobotPartCompass();
        sensorCompass = new Sensor( SENSOR_COMPASS, 0, NULL, 1, compass );
        compass[ 0 ] = 0.0;
    }
}


void Nomad::reset()
{
    sonarStop();
    drive( 0 );
    steer( 0 );
    if ( report_level >= HS_REP_LEV_DEBUG )
    {
        printTextWindow( "Nomad reset" );
    }
}

int Nomad::connect()
{
    int rtn;

    if ( enType == EnNomadType_150 )
    {
        is200 = false;

        sprintf( rstr, "Connecting to nomad 150 over %s...", strPortString.c_str() );
        printTextWindow( rstr, false );
        nomad_open_serial( strPortString.c_str(), 38400 );
        rtn = ping( false );
        if ( !rtn )
        {
            printTextWindow( "failed" );
            if ( report_level>=HS_REP_LEV_ERR )
            {
                printTextWindow( "error: Nomad open serial port" );
            }
	    mode = NOMAD_MODE_NORMAL; // Ernest- have to set mode before calling statusbarUpdate();
            statusbarUpdate();
            redrawWindows();
            return -1;
        }
    }
    else
    {
        is200 = true;

        printfTextWindow( "Connecting to nomad 200 on %s...", strHost.c_str() );
        printTextWindow( rstr, false );
        rtn = connect_robot( strHost.c_str() );
        if ( rtn )
        {
            printTextWindow( "failed" );
            if ( report_level>=HS_REP_LEV_ERR )
            {
                printfTextWindow( "error: Nomad connect robot %d", rtn );
            }
            return -1;
        }
    }
    printTextWindow( "connected" );
    mode = NOMAD_MODE_ZEROING;
    statusbarUpdate();
    addSensors();
    pthread_cond_init( &nomad_cond, NULL );
    pthread_mutex_init( &nomad_cond_mutex, NULL );
    pthread_mutex_init( &nomadMutex, NULL );

    /* Ernest start */
    /* if watchdog function is enabled, then initialize the mutex for this object, 
       and call AddNomad to add this instance of Nomad to the list of threads that
       the WatchDog class will check. */
    if (watchDogEnable == true)
    {
	pthread_mutex_init( &nomad_watchdog_mutex, NULL );
	int result;
	if( (result = gettimeofday( &tv_watchdog,NULL)) != 0)
	    printTextWindow( "error: Nomad watchdog failed to gettimeofday()" );
	if ( WatchDog != NULL)
	{
	    WatchDog->AddNomad( &tv_watchdog, &nomad_watchdog_mutex);
	    WatchDog->EnableNomadWD(); // Enable the watchdog for the nomad
	}
    }

    /* Ernest end */
    pthread_create( &reader, NULL, (void*(*)(void*)) &start_thread, (void*) this );
    readerRunning = true;
    return 0;
}


Nomad::~Nomad()
{
    if ( readerRunning )
    {
        pthread_cancel( reader );
        pthread_join( reader, NULL );
        reset();
        disconnect_robot();
        delete sensorSonar;
        delete sensorArm;
        delete partArm;
        delete partXyt;
        delete partSonar;
        if ( is200 )
        {
            delete sensorCompass;
            delete sensorIr;
            delete partCompass;
            delete partIr;
        }
        printTextWindow( "Nomad disconnected" );
    }
}


// Ernest - added watchDogEnable flag
Nomad::Nomad( Robot** a, EnNomadType enType, const string& strPortString, 
              const string& strHost, bool watchDogEnable = true) : 
    Robot( a, rt_nomad, "Nomad" )
{
    int rtn, i;

    Robot::setXYT( 0, 0, 0 );
    spinner = 0;
    sonar_on = false;
    timeout = 2;
    readerRunning = false;
    this->enType = enType;
    this->strHost = strHost;
    this->strPortString = strPortString;
    this->watchDogEnable = watchDogEnable; // Ernest- initialize local watchDogEnable flag
    sonar_speed = 0;
    module_type = MODULE_ROBOT_NOMAD;
    radius = 0.24;
    for( i = 0; i < NOMAD_NUM_SONAR; i++ )
    {
        sonar_angle[ i ] = float( i ) * ( 360.0 / NOMAD_NUM_SONAR );
    }
    rtn = connect();
    if ( rtn )
    {
        delete this;
    }
}


void Nomad::control()
{
    bool bDone = false;
    int msgData[ 2 ];

    msgData[ 0 ] = sonar_on;
    msgData[ 1 ] = sonar_speed;

    messageDrawWindow( EnMessageType_NOMAD_CONTROL, EnMessageErrType_NONE, msgData );
    redrawWindows();
    do
    {
        int c = getch();
        switch( c )
        {
        case 'd':
            messageHide();
            delete this;
            return;
            break;
        case 'r':
            refreshScreen();
            break;
        case 'z':
            zero();
            break;
        case 'c':
            compassCalibrate();
            break;
        case 's':
            sonar_on = !sonar_on;
            msgData[ 0 ] = sonar_on;
            messageDrawWindow( EnMessageType_NOMAD_CONTROL, EnMessageErrType_NONE, msgData );
            if ( sonar_on )
            {
                sonarStart();
            }
            else
            {
                sonarStop();
            }
            break;
        case '-':
        case '_':
            sonar_speed--;
            if ( sonar_speed < 0 )
            {
                sonar_speed = 0;
            }
            msgData[ 1 ] = sonar_speed;
            messageDrawWindow( EnMessageType_NOMAD_CONTROL, EnMessageErrType_NONE, msgData );
            setSonarSpeed( sonar_speed );
     	    break;
        case '+':
        case '=':
            sonar_speed++;
            msgData[ 1 ] = sonar_speed;
            messageDrawWindow( EnMessageType_NOMAD_CONTROL, EnMessageErrType_NONE, msgData );
            setSonarSpeed( sonar_speed );
            break;
        case 'p':
            ping();
            break;
        case 'x':
        case 'Q':
        case KEY_ESC:
            bDone = true;
            break;
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case KEY_UP:
        case KEY_DOWN:
        case KEY_LEFT:
        case KEY_RIGHT:
            messageMovePanelKey( c );
            break;
        }
    } while ( !bDone );
    messageHide();
}

/**********************************************************************
# $Log: nomad.c,v $
# Revision 1.9  2003/04/06 13:32:37  endo
# gcc 3.1.1
#
# Revision 1.8  2003/04/02 22:23:22  zkira
# Battery level for telop GUI
#
# Revision 1.7  2002/01/16 22:17:17  ernest
# Added WatchDog function; fixed other misc. bugs.
#
# Revision 1.6  2001/05/29 22:35:57  blee
# Made changes to incorporate gps data w/ other position info.
#
# Revision 1.5  2001/03/23 21:31:44  blee
# altered to use a config file
#
# Revision 1.4  2000/12/12 23:22:16  blee
# Changed #defines for the user interface to enumerated types.
#
# Revision 1.3  2000/09/19 03:47:14  endo
# RCS log added.
#
#
#**********************************************************************/
