/**********************************************************************
 **                                                                  **
 **                               nomad.h                            **
 **                                                                  **
 **                                                                  **
 **  Written by:  William Halliburton                                **
 **                                                                  **
 **  Nomad 150/200 robot driver for HServer                          **
 **                                                                  **
 **  Copyright 2000, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: nomad.h,v 1.7 2003/04/06 13:32:37 endo Exp $ */

#ifndef NOMAD_H_
#define NOMAD_H_

#include <pthread.h>
#include <string>
#include <sys/time.h>
#include "robot.h"
#include "sensors.h"
#include "robot_config.h"

using std::string;

#define NOMAD_MAX_VEL  100
#define NOMAD_MAX_RVEL 450
#define NOMAD_MIN_ACC  40
#define NOMAD_MAX_ACC  320

#define NOMAD_NUM_SONAR 16
#define NOMAD_NUM_IR    16

#define NOMAD_MODE_NORMAL  0
#define NOMAD_MODE_ZEROING 1
#define NOMAD_MODE_COMPASS 2

#define NOMAD_MAX_IR 15 // inches

#define NOMAD_WAIT 20   // seconds

class Nomad : public Robot
{
public:
    enum EnNomadType
    {
        EnNomadType_150,
        EnNomadType_200
    };

private:
    Sensor* sensorSonar;
    Sensor* sensorIr;
    Sensor* sensorArm;
    Sensor* sensorCompass;
    RobotPartSonar* partSonar;
    RobotPartIr* partIr;
    RobotPartXyt* partXyt;
    RobotPartArm* partArm;
    RobotPartCompass* partCompass;
    bool sonar_on;
    int spinner;
    string strPortString;
    int timeout;
    string strHost;
    float sonar_angle[ NOMAD_NUM_SONAR ];
    float sonar[ NOMAD_NUM_SONAR ];
    int sonar_order[ NOMAD_NUM_SONAR ];
    int sonar_speed;
    float ir[ NOMAD_NUM_IR ];
    float arm[ 1 ];
    float compass[ 1 ];
    float voltages[ 2 ];
    int mode;
    int reader_running;
    bool is200;
    float radius;
    pthread_mutex_t robMutex;
    void addSensors();
    void statusbarUpdate();
    int connect();
    int setup();
    bool readerRunning;
    pthread_t reader;
    struct timeval tv_watchdog;  // Ernest - timestamp for watchdog
    pthread_mutex_t nomad_watchdog_mutex; // Ernest
    bool watchDogEnable; // Ernest
    EnNomadType enType;

public:
    void waitNext();
    void reader_thread();
    int updateCompass();
    int updateSonar();
    int updateIR();
    void updateVoltages();
    int updateLocation();

    int compassActivate();
    int compassCalibrate();
    void setSonarSpeed( int speed );
    void sonarStart();
    void sonarStop();
    int ping( bool output = true );
    void setTimeout( unsigned char timeout );
    void moveDistance( float distance, bool wait = true );
    float theta();
    void XY( float* x, float* y );
    void reset();

    int zero();
    void initXYT();
    void setXYT( float xi, float yi, float theta );
    void adjustXYT( float xinc, float yinc, float tinc );
    void control();
    void steer( float avel );  
    void drive( float vel );   
    void steerToward( int angle );
    void stop();
    void turn( float degrees, bool wait = true );
    void turnTo( float direction, bool wait = true );
    Nomad( Robot** a, EnNomadType enType, const string& strPortString, 
           const string& strHost, bool watchDogEnable);
    // Ernest- added watchDogEnable flag to constructor
    ~Nomad();
};    

#endif

/**********************************************************************
# $Log: nomad.h,v $
# Revision 1.7  2003/04/06 13:32:37  endo
# gcc 3.1.1
#
# Revision 1.6  2002/01/16 22:19:11  ernest
# Added WatchDog function
#
# Revision 1.5  2001/05/29 22:36:02  blee
# Made changes to incorporate gps data w/ other position info.
#
# Revision 1.4  2001/03/23 21:31:53  blee
# altered to use a config file
#
# Revision 1.3  2000/09/19 03:47:14  endo
# RCS log added.
#
#
#**********************************************************************/
