/**********************************************************************
 **                                                                  **
 **                         CheckBoxType.hpp                         **
 **                                                                  **
 **                                                                  **
 **  Written by:  Douglas C. MacKenzie                               **
 **                                                                  **
 **  Copyright 1995, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **********************************************************************/

/* $Id: CheckBoxType.hpp,v 1.5 2003/04/06 12:58:48 endo Exp $ */


#ifndef CheckBoxType_HPP
#define CheckBoxType_HPP

#include <Xm/Xm.h>
#include <Xm/LabelG.h>
#include <Xm/RowColumn.h>
#include <Xm/ToggleB.h>

#include "utilities.h"
#include "TypeInfo.hpp"

/*-----------------------------------------------------------------------*/
// Define the CheckBoxType class
class CheckBoxType : public TypeInfo
{
   char  **values;
   int    num_values;

   static void
   toggled(Widget w, bool *included)
   {
      // Toggle the boolean marking if this item is included 
      *included ^= true;
   }

public:
   CheckBoxType(const char *TypeName, char *valuelist[], int value_cnt) : 
	TypeInfo(TypeName)
   {
      num_values = value_cnt;
      if( num_values > 0 )
      {
         values = new (char *)[num_values];
         for(int i=0; i<num_values; i++)
	    values[i] = strdup(valuelist[i]);
      }
      else
      {
	 values = NULL;
      }
   }

   ~CheckBoxType() 
   { 
      if( values )
      {
	 delete values;
	 values = NULL;
      }
   }

   Widget 
   CreateDisplayWidget(Widget parent, Symbol *parm)
   {
      // Create name 
      XmString str = XmStringCreateSimple("Select any of the following");

      Widget w = XtVaCreateManagedWidget("",
	       		xmLabelGadgetClass, parent,
	       		XmNlabelString, str,
	       		NULL);
      XmStringFree(str);

      char *label = "";
      if( parm->name )
      {
         if( parm->name[0] == '%' )
            label = &parm->name[1];
         else
            label = parm->name;
      }

      // Make a list of booleans to keep track of the selections
      // and hang it off of the widget
      bool *current_state = new bool[num_values];

      w = XtVaCreateWidget("", xmRowColumnWidgetClass, parent,
		XtVaTypedArg, XmNtitleString, XmRString, label, strlen(label)+1,
		XmNpacking,	XmPACK_COLUMN,
		XmNnumColumns,	1,
		XmNradioBehavior, false,
		XmNradioAlwaysOne, false,
                XmNuserData, (int)current_state,
		NULL);

      // Default is not selected
      for(int i=0; i<num_values; i++)
         current_state[i] = false;

      // Get the list of currently selected items
      if( parm->input_generator && parm->input_generator->name )
      {
         char *val = parm->input_generator->name;

         // Check each of our values in turn to see if they are specified.
         for(int i=0; i<num_values; i++)
	 {
	    char *p = strstr(val, values[i]);
	    if( p )
	    {
	       // Found a match, make sure we aren't matching a prefix.
	       int len = strlen(values[i]);
	       if( p[len] == '\0' || p[len] == '|' || isspace(p[len]) )
	       {
		  current_state[i] = true;
	       }
	    }
         }
      }
      else
      {
         for(int i=0; i<num_values; i++)
            current_state[i] = false;
      }


      for(int i=0; i<num_values; i++)
      {
	 Widget btn = XtVaCreateManagedWidget(values[i],
	  	xmToggleButtonWidgetClass, w, 
	  	XmNset,		current_state[i],
	  	NULL);

	 XtAddCallback(btn, XmNvalueChangedCallback, 
		    (XtCallbackProc)toggled, (XtPointer)&current_state[i]);
      }
      XtManageChild(w);
      return w;
   }

   void
   UpdateFromDisplayWidget(Widget w, char **value)
   {
      if( value == NULL )
      {
	 fprintf(stderr,"INTERNAL ERROR: Null value to UpdateFromDisplayWidget\nUnable to save modifed values\n");
	 return;
      }

      bool *current_state;
      XtVaGetValues(w, 
		XmNuserData, (int *)&current_state,
		NULL);

      char *val = NULL;

      // Add each specified value to the string (bitwise or'd)
      for(int i=0; i<num_values; i++)
      {
	 if( current_state[i] )
	 {
	    if( val )
	    {
	       val = strdupcat(val," | ", values[i]);
	    }
	    else
	    {
	       val = strdup(values[i]);
	    }

	 }
      }

      // An empty list is "" and not a NULL pointer
      if( val == NULL )
	 val = strdup("");

      // Return to sender
      free(*value);
      *value = val;
   }


   char *
   PrintValue(struct Symbol *parm, const char *separator)
   {
      char buf[2048];
      buf[0] = '\0';  

      // Get the list of currently selected items
      if( parm && parm->input_generator && parm->input_generator->name )
      {
         char *val = parm->input_generator->name;
         char *tok_end = val;
         char *tok_start = NULL;

         while( tok_start != tok_end)
	 {
            // Find the start of the next token.
	    tok_start = tok_end;
   	    while(*tok_start && !isalnum( *tok_start ) && *tok_start != '_')
	       tok_start++;

	    // Find the end of the token
	    tok_end = tok_start;
	    while( isalnum( *tok_end ) || *tok_end == '_' )
	       tok_end++;

            if( tok_start != tok_end )
	    {
	       // Mark end of token so strcat will work.
	       char tmp = *tok_end;
	       *tok_end = '\0';
   
	       // If this isn't the first value, then separate them
	       if( buf[0] )
	          strcat(buf, separator);

	       // Add this value
	       strcat(buf, tok_start);

               // Repair
	       *tok_end = tmp;
            }
         }

         // If empty, then return "None" instead of leaving it blank
         if( buf[0] == '\0' )
	 {
	    strcpy(buf, "None");
	 }
      }
      else
      {
	 // Didn't get a valid parameter list
	 strcpy(buf,"???");
      }

      return strdup(buf);
   }
};

#endif


///////////////////////////////////////////////////////////////////////
// $Log: CheckBoxType.hpp,v $
// Revision 1.5  2003/04/06 12:58:48  endo
// gcc 3.1.1
//
// Revision 1.4  1996/05/05  21:45:27  doug
// fixing compile warnings
//
// Revision 1.4  1996/05/05  21:45:27  doug
// fixing compile warnings
//
// Revision 1.3  1996/04/11  04:07:16  doug
// *** empty log message ***
//
// Revision 1.2  1996/04/06  23:54:20  doug
// *** empty log message ***
//
// Revision 1.1  1996/04/06  20:34:08  doug
// Initial revision
//
///////////////////////////////////////////////////////////////////////
