/**********************************************************************
 **                                                                  **
 **                         RadioButtonType.hpp                      **
 **                                                                  **
 **                                                                  **
 **  Written by:  Douglas C. MacKenzie                               **
 **                                                                  **
 **  Copyright 1995, Georgia Tech Research Corporation               **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **********************************************************************/

/* $Id: RadioButtonType.hpp,v 1.3 2003/04/06 12:58:48 endo Exp $ */


#ifndef RadioButtonType_HPP
#define RadioButtonType_HPP

#include <Xm/Xm.h>
#include <Xm/LabelG.h>
#include <Xm/RowColumn.h>
#include <Xm/ToggleB.h>

#include "utilities.h"
#include "TypeInfo.hpp"

/*-----------------------------------------------------------------------*/
// Define the RadioButtonType class
class RadioButtonType : public TypeInfo
{
   char  **values;
   int    num_values;

   struct toggle_info {
      int  our_value;
      int *value;
   };

   static void
   toggled(Widget w, toggle_info *info)
   {
      // Set the current value to our position
      *info->value = info->our_value;
   }

public:
   RadioButtonType(const char *TypeName, char *valuelist[], int value_cnt) : 
	TypeInfo(TypeName)
   {
      num_values = value_cnt;
      if( num_values > 0 )
      {
         values = new (char *)[num_values];
         for(int i=0; i<num_values; i++)
	    values[i] = strdup(valuelist[i]);
      }
      else
      {
	 values = NULL;
      }
   }

   ~RadioButtonType() 
   { 
      if( values )
      {
	 delete values;
	 values = NULL;
      }
   }

   Widget 
   CreateDisplayWidget(Widget parent, Symbol *parm)
   {
      // Create name 
      XmString str = XmStringCreateSimple("Select one of the following");

      Widget w = XtVaCreateManagedWidget("",
	       		xmLabelGadgetClass, parent,
	       		XmNlabelString, str,
	       		NULL);
      XmStringFree(str);

      char *label = "";
      if( parm->name )
      {
         if( parm->name[0] == '%' )
            label = &parm->name[1];
         else
            label = parm->name;
      }

      int *current_selection = new int;

      w = XtVaCreateWidget("", xmRowColumnWidgetClass, parent,
		XtVaTypedArg, XmNtitleString, XmRString, label, strlen(label)+1,
		XmNpacking,	XmPACK_COLUMN,
		XmNnumColumns,	1,
		XmNradioBehavior, true,
		XmNradioAlwaysOne, true,
                XmNuserData, (int)current_selection,
		NULL);

      // Default is not selected
      *current_selection = -1;

      // Find out which item is currently selected
      if( parm->input_generator && parm->input_generator->name )
      {
         char *val = parm->input_generator->name;

         // Check each of our values in turn to see which is specified.
         for(int i=0; i<num_values; i++)
	 {
	    // Found a match
	    if( strcmp(val, values[i]) == 0)
	    {
	       *current_selection = i;
	       break;
	    }
         }
      }

      for(int i=0; i<num_values; i++)
      {
	 Widget btn = XtVaCreateManagedWidget(values[i],
	  	xmToggleButtonWidgetClass, w, 
	  	XmNset,		i == *current_selection,
	  	NULL);

         // Make a record to handle the toggling actions
         toggle_info *info = new toggle_info;
	 info->our_value = i;
	 info->value = current_selection;
	 XtAddCallback(btn, XmNvalueChangedCallback, 
		    (XtCallbackProc)toggled, (XtPointer)info);
      }
      XtManageChild(w);
      return w;
   }

   void
   UpdateFromDisplayWidget(Widget w, char **value)
   {
      if( value == NULL )
      {
	 fprintf(stderr,"INTERNAL ERROR: Null value to UpdateFromDisplayWidget\nUnable to save modifed values\n");
	 return;
      }

      int *current_selection;
      XtVaGetValues(w, 
		XmNuserData, (int *)&current_selection,
		NULL);

      char *val;

      // Add the specified value to the string 
      if( *current_selection >= 0 && *current_selection < num_values )
         val = strdup(values[*current_selection]);
      else
	 val = strdup("");

      // Return to sender
      free(*value);
      *value = val;
   }


   char *
   PrintValue(struct Symbol *parm, const char *separator)
   {
      char buf[2048];
      buf[0] = '\0';  

      // Get the currently selected item
      if( parm && parm->input_generator && parm->input_generator->name )
      {
         strcpy(buf, parm->input_generator->name);

         // If empty, then return "None" instead of leaving it blank
         if( buf[0] == '\0' )
	 {
	    strcpy(buf, "None");
	 }
      }
      else
      {
	 // Didn't get a valid parameter list
	 strcpy(buf,"???");
      }

      return strdup(buf);
   }
};

#endif


///////////////////////////////////////////////////////////////////////
// $Log: RadioButtonType.hpp,v $
// Revision 1.3  2003/04/06 12:58:48  endo
// gcc 3.1.1
//
// Revision 1.2  1996/05/05  21:45:27  doug
// fixing compile warnings
//
// Revision 1.2  1996/05/05  21:45:27  doug
// fixing compile warnings
//
// Revision 1.1  1996/04/08  16:04:06  doug
// Initial revision
//
//
///////////////////////////////////////////////////////////////////////
