/**********************************************************************
 **                                                                  **
 **                            3d_sphigs_display.c                   **
 **                                                                  **
 **         interface to sphigs routines                             **
 **                                                                  **
 **  Written by:  Zhong Chen                                         **
 **                                                                  **
 **  Copyright 1995, 1996, 1997 Georgia Tech Research Corporation    **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **********************************************************************/

#define DEBUG 1
extern "C" {
#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <math.h>
}
#include "console.h"
#include "3d_sphigs_display.h"

float convert_meters_to_feet_x ( float x);
float convert_meters_to_feet_y ( float x);
#define XSCALE (30.0/mission_area_width_meters)
#define YSCALE (20.0/mission_area_height_meters)

/****************** Global variables for making the room outline.*/
static int  number_room_vertices;
static int  number_facets;
static point (*room_vertices);
static vertex_index (*room_facets);

/* This defines how high to make a 3d boundary from a 2d boundary      */
#define height_to_make_3d_from_2d 150


/******************* Global variables for colors .****************/
int             TEXT_COLOR;
int             GOAL_COLOR;
int             BASE_COLOR;
int             ROOM_COLOR;
int             CUBE_COLOR;
int             OBSTACLE_COLOR;
int             MANTIS_BASE_COLOR;


/******************* declarations for view mapping ***************/

matrix          rtv_vo_matrix;
matrix          rtv_vm_matrix;
vector          rtv_vpn;
vector          rtv_vup;
vector          rtv_vrp;
vector          rtv_prp;

matrix          rsv_vo_matrix;
matrix          rsv_vm_matrix;
vector          rsv_vpn;
vector          rsv_vup;
vector          rsv_vrp;
vector          rsv_prp;

matrix          rfv_vo_matrix;
matrix          rfv_vm_matrix;
vector          rfv_vpn;
vector          rfv_vup;
vector          rfv_vrp;
vector          rfv_prp;


/******************** Some Constants *****************************/
#define PRP    			1000.0
#define CUBE_SIZE               2.0
#define MOVE_DELTA 0.0001   /* how far robot moves before update*/
#define ROBOT_UPDATE_RATE 5  /* how often to update robot */

/******************* Default Values from overlay file ************/
static point origin = {0,0,0};


/******************* view labels *********************************/
#define TOP_ROOM_VIEW   	1
#define SIDE_ROOM_VIEW  	2
#define FRONT_ROOM_VIEW 	3
#define FRONT_ROBOT_VIEW 	5
#define SIDE_ROBOT_VIEW 	6
#define TOP_ROBOT_VIEW 		7


/******************* Structure Labels ****************************/
#define NULL_OBJECT		0
#define ROOM_VIEW	        9
#define ROBOT_BASE	       10
#define TURN_ROBOT_BASE	       11
#define MOVE_ROBOT_BASE	       12

#define MANTIS_BASE            13
#define MANTIS_BODY            14
#define MANTIS_LEG             15
#define MANTIS                 16

#define TOP_RV	               28
#define SIDE_RV	               29
#define FRONT_RV	       30
#define OBSTACLE               32
#define PASSAGE_POINT          33


/******************* COLORS **************************************/
#define WHITE		13
#define BLACK		14
#define LIGHT_YELLOW	2
#define LIGHT_GREY	3
#define LIGHT_BLUE	4
#define LIGHT_RED	5
#define LIGHT_GREEN	6
#define YELLOW		7
#define RED		8
#define GREY		9
#define DARK_GREY	10
#define GREEN	        11
#define BLUE 	        12
#define OBJECT_COLOR_BASE 15

/****************** OBJECTS **************************************/
#define ROOM_START      45
#define ROOM_END        46
#define OBJECT_START    47
#define OBJECT_END      48
#define OBJECTS_BEGIN   49
#define OBJECT_BASE     50

int objectCount=0, maxObjectCount=0;

/******************* Structures **********************************/

/* unit obstacle with radius 0.1 feet */
/* yes I know it is a cube, but let's see you try to draw a sphere */
static point    obstacle_vertices[] =
{
   {-1.0, -1.0, -1.0},
   {-1.0, 1.0, -1.0},
   {1.0, 1.0, -1.0},
   {1.0, -1.0, -1.0},
   {-1.0, -1.0, 1.0},
   {-1.0, 1.0, 1.0},
   {1.0, 1.0, 1.0},
   {1.0, -1.0, 1.0}
};

static vertex_index obstacle_facets[] =
{
   0, 1, 2, 3, -1,
   4, 7, 6, 5, -1,
   0, 3, 7, 4, -1,
   3, 2, 6, 7, -1,
   2, 1, 5, 6, -1,
   1, 0, 4, 5, -1
};



static point passage_point [] =
{
  {-1,-1,0},
  {1,-1,0},
  {1,1,0},
  {-1,1,0}
};

/* vertices for body in 10ths of a foot from center bottom of robot */
#define LINK0_LENGTH (2.0)
static point    body_vertices[] = {
   {13.0, 6.0, 1.0},
   {13.0, -6.0, 1.0},
   {6.0, -13.0, 1.0},
   {-6.0, -13.0, 1.0},
   {-13.0, -6.0, 1.0},
   {-13.0, 6.0, 1.0},
   {-6.0, 13.0, 1.0},
   {6.0, 13.0, 1.0},

   {13.0, 6.0, LINK0_LENGTH * 10},
   {13.0, -6.0, LINK0_LENGTH * 10},
   {6.0, -13.0, LINK0_LENGTH * 10},
   {-6.0, -13.0, LINK0_LENGTH * 10},
   {-13.0, -6.0, LINK0_LENGTH * 10},
   {-13.0, 6.0, LINK0_LENGTH * 10},
   {-6.0, 13.0, LINK0_LENGTH * 10},
   {6.0, 13.0, LINK0_LENGTH * 10}
};

/* facets for body */

static vertex_index body_facets[] = {
   0, 8, 9, 1, -1,
   1, 9, 10, 2, -1,
   2, 10, 11, 3, -1,
   3, 11, 12, 4, -1,
   4, 12, 13, 5, -1,
   5, 13, 14, 6, -1,
   6, 14, 15, 7, -1,
   7, 15, 8, 0, -1,
   15, 14, 13, 12, 11, 10, 9, 8, -1,
   0, 1, 2, 3, 4, 5, 6, 7, -1
};

/* vertices for cube in 10ths of a foot from center */

static point    cube_vertices[] =
{
   {-CUBE_SIZE, -CUBE_SIZE, -CUBE_SIZE},
   {-CUBE_SIZE, CUBE_SIZE, -CUBE_SIZE},
   {CUBE_SIZE, CUBE_SIZE, -CUBE_SIZE},
   {CUBE_SIZE, -CUBE_SIZE, -CUBE_SIZE},
   {-CUBE_SIZE, -CUBE_SIZE, CUBE_SIZE},
   {-CUBE_SIZE, CUBE_SIZE, CUBE_SIZE},
   {CUBE_SIZE, CUBE_SIZE, CUBE_SIZE},
   {CUBE_SIZE, -CUBE_SIZE, CUBE_SIZE}
};

static vertex_index cube_facets[] =
{
   0, 1, 2, 3, -1,
   4, 7, 6, 5, -1,
   0, 3, 7, 4, -1,
   3, 2, 6, 7, -1,
   2, 1, 5, 6, -1,
   1, 0, 4, 5, -1
};

/* vertices for the praying mantis model */

static point mantis_vertices[] =
{
  {3,1,0},
  {3,-1,0},
  {3,-1,1},
  {3,1,1},
  {3,1,0},
  {3,-1,0},
  {-3,-1,1},
  {-3,1,1}
};

static vertex_index mantis_facets[] =
{
  0,1,2,3,-1,
  3,2,6,7, -1,
  7,6,5,4, -1,
  4,5,1,0,-1,
  0,3,7,4,-1,
  2,1,5,6,-1
};

#define BODY_LENGTH 5

static point mantis_body_vertices[] =
{
  {BODY_LENGTH,1,0},
  {BODY_LENGTH,-1,0},
  {BODY_LENGTH,-1,1},
  {BODY_LENGTH,1,1},
  {-BODY_LENGTH,1,0},
  {-BODY_LENGTH,-1,0},
  {-BODY_LENGTH,-1,1},
  {-BODY_LENGTH,1,1}
};

static vertex_index mantis_body_facets[] =
{
  0,1,2,3,-1,
  3,2,6,7, -1,
  7,6,5,4, -1,
  4,5,1,0,-1,
  0,3,7,4,-1,
  2,1,5,6,-1
};
#define LEG_LENGTH 2.5
#define LEGD  0.5

static point mantis_leg_vertices[] =
{
  {LEGD,LEG_LENGTH,0},
  {LEGD,-LEG_LENGTH,0},
  {LEGD,-LEG_LENGTH,LEGD},
  {LEGD,LEG_LENGTH,LEGD},
  {-LEGD,LEG_LENGTH,0},
  {-LEGD,-LEG_LENGTH,0},
  {-LEGD,-LEG_LENGTH,LEGD},
  {-LEGD,LEG_LENGTH,LEGD},

};

static vertex_index mantis_leg_facets[] =
{
  0,1,2,3,-1,
  3,2,6,7, -1,
  7,6,5,4, -1,
  4,5,1,0,-1,
  0,3,7,4,-1,
  2,1,5,6,-1
};



/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
float convert_meters_to_feet_x ( float x) 
{
return( x * 39.37*XSCALE / 12);
}

float convert_meters_to_feet_y ( float x) 
{
return( x * 39.37*YSCALE / 12);
}

/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void add_3d_obstacle_sphigs(double x, double y, double radius)
{
  /* radius, x, and y are coming in meters need to convert to feet */
  matrix          temp_matrix1;
  matrix          temp_matrix2;
  matrix          temp_matrix;
  double           x_scale, y_scale, z_scale;
  x_scale = radius * 39.37*XSCALE / 12;
  y_scale =  radius * 39.37*YSCALE /12;
  z_scale = 1;				 
  x =   x * 39.37*XSCALE / 12;
  y =  y * 39.37*YSCALE / 12;
//  printf("obstacle added at x: %f  y: %f\n", x, y);


  SPH_openStructure(ROOM_VIEW);
  SPH_offsetElementPointer(-1);
  SPH_translate(x * 10, y * 10, 10.0, temp_matrix2); 
  SPH_scale(x_scale * 10, y_scale * 10, 10.0, temp_matrix1); 
  SPH_compose(temp_matrix1, temp_matrix2, temp_matrix);
  SPH_setModelingTransformation(temp_matrix,ASSIGN);
  SPH_executeStructure(OBSTACLE); 
  SPH_closeStructure();

}

#define MAXOBJ 200
double objectx[MAXOBJ], objecty[MAXOBJ];
int objectRemoved[MAXOBJ];

void add_3d_object_sphigs(double x, double y, double radius, char *color)
{
  /* radius, x, and y are coming in meters need to convert to feet */
  matrix          temp_matrix1;
  matrix          temp_matrix2;
  matrix          temp_matrix;
  double           x_scale, y_scale, z_scale;
  int              index=objectCount;

  objectCount++;
  objectx[index]=x; objecty[index]=y;
  objectRemoved[index]=0;
  x_scale = radius * 39.37*XSCALE / 12;
  y_scale =  radius * 39.37*YSCALE /12;
  z_scale = 1;				 
  x =   x * 39.37*XSCALE / 12;
  y =  y * 39.37*YSCALE / 12;
//  printf("object added at x: %f  y: %f %s\n", x, y, color);


  define_object(index,color);
  SPH_openStructure(ROOM_VIEW);
  SPH_offsetElementPointer(-1);
  if(!index) SPH_label(OBJECTS_BEGIN);
  SPH_translate(x * 10, y * 10, 10.0, temp_matrix2); 
  SPH_scale(x_scale * 10, y_scale * 10, 10.0, temp_matrix1); 
  SPH_compose(temp_matrix1, temp_matrix2, temp_matrix);
  SPH_setModelingTransformation(temp_matrix,ASSIGN);
  SPH_executeStructure(index+OBJECT_BASE); 
  SPH_closeStructure();
}

void remove_3d_object_sphigs(double x, double y)
{
  int i;
  int index=0;

  for(i=0;i<objectCount;i++) {
    if((!objectRemoved[i]) && fabs(objectx[i]-x)<0.00001 && fabs(objecty[i]-y)<0.00001) {
      SPH_openStructure(ROOM_VIEW);
      SPH_setElementPointer(0);
      SPH_moveElementPointerToLabel(OBJECTS_BEGIN);
      SPH_offsetElementPointer(2*index+1);
      SPH_deleteElement();
      SPH_deleteElement();
      SPH_closeStructure();
//      printf("remove_3d_object_sphigs %f %f\n",x,y);
      objectRemoved[i]=1;
      break;
    }
    if(!objectRemoved[i]) index++;
  }
}


/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void add_3d_passage_point_sphigs(char *name, float x, float y, float diameter)
{

double          radius;
matrix          temp_matrix;
matrix          temp_matrix1;
matrix          temp_matrix2;
point           origin;
char            str[80];


x = convert_meters_to_feet_x(x);
y = convert_meters_to_feet_y(y);
radius =  convert_meters_to_feet_x(diameter) ;
//  printf("passage added at x: %f  y: %f with radius: %f\n", x, y, radius);
SPH_openStructure(ROOM_VIEW);
SPH_offsetElementPointer(-1);
SPH_scale(radius * 10, radius * 10, 0.0, temp_matrix1); 
SPH_translate(x*10, y*10,0.5, temp_matrix2);
SPH_compose(temp_matrix1, temp_matrix2, temp_matrix);
SPH_setModelingTransformation(temp_matrix, ASSIGN);
SPH_executeStructure(PASSAGE_POINT);

/*
 labels are too confusing to be shown in 3d views
origin[0] = 0;
origin[1] = 0;
origin[2] = 0;
SPH_text(origin,name);
*/

SPH_closeStructure();
}



/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void draw_room()
{

  /* Double buffered so have to regen. twice */
  SPH_regenerateScreen();
//  SPH_regenerateScreen();
}


/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void
exit_display()
{
   SPH_end();
}


/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void
init_front_room_view(void)
{
   matrix          vo_matrix;
   matrix          vm_matrix;
   vector          vpn;
   vector          vup;
   vector          vrp;
   vector          prp;
   double          fplane;
   double          bplane;

   vrp[0] = 180.0;
   vrp[1] = -120.0;
   vrp[2] = 40.0;
   vpn[0] = 0.0;
   vpn[1] = -20;
   vpn[2] = 1.0;
   vup[0] = 0.0;
   vup[1] = 0.0;
   vup[2] = 1.0;
   prp[0] = 0.0;
   prp[1] = 0.0;
   prp[2] = PRP;

   fplane = 0.0;
   bplane = -800.0;

   SPH_evaluateViewOrientationMatrix
      (vrp,
       vpn,
       vup,
       vo_matrix);

   SPH_evaluateViewMappingMatrix
      (-200, 800, -50, 200, PERSPECTIVE,
       prp,
       fplane, bplane,
       0.417, 1.25, 0, 0.42, 0.0, 1.0,
       vm_matrix);

   SPH_setViewRepresentation
      (FRONT_ROOM_VIEW,
       vo_matrix, vm_matrix,
       0.417, 1.25, 0, 0.42, 0.0, 1.0);

   SPH_setViewBackgroundColor(FRONT_ROOM_VIEW, LIGHT_GREY);

}


/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void
init_top_room_view(void)
{
   matrix          vo_matrix;
   matrix          vm_matrix;
   vector          vpn;
   vector          vup;
   vector          vrp;
   vector          prp;
   double          fplane;
   double          bplane;

   vrp[0] = 180.0;
   vrp[1] = 90;
   vrp[2] = 100.0; /* 160 */

   vpn[0] = 0.0;
   vpn[1] = 0.0;
   vpn[2] = 1.0;

   vup[0] = 0.0;
   vup[1] = 1.0;
   vup[2] = 0.0;

   prp[0] = 0.0;
   prp[1] = 0.0;
   prp[2] = PRP;
   fplane = 0;
   bplane = -400;

   SPH_evaluateViewOrientationMatrix
      (vrp,
       vpn,
       vup,
       vo_matrix);

   SPH_evaluateViewMappingMatrix
      (-200, 800, -115, 460, PERSPECTIVE,
       prp,
       fplane, bplane,
       0.417, 1.25, 0.44, 1.0, 0.0, 1.0,
       vm_matrix);

   SPH_setViewRepresentation
      (TOP_ROOM_VIEW,
       vo_matrix, vm_matrix,
       0.417, 1.25, 0.44, 1.0, 0.0, 1.0); 


   SPH_setViewBackgroundColor(TOP_ROOM_VIEW, LIGHT_GREY);
}

/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/

void
init_side_room_view(void)
{
   matrix          vo_matrix;
   matrix          vm_matrix;
   vector          vpn;
   vector          vup;
   vector          vrp;
   vector          prp;
   double          fplane;
   double          bplane;

   vrp[0] = -220.0;
   vrp[1] = 100.0;
   vrp[2] = 40.0;

   vpn[0] = -20.0;
   vpn[1] = 0.0;
   vpn[2] = 1.0;


   vup[0] = 0.0;
   vup[1] = 0.0;
   vup[2] = 1.0;
   prp[0] = 0.0;
   prp[1] = 0.0;
   prp[2] = PRP;
   fplane = 0;
   bplane = -800;

   SPH_evaluateViewOrientationMatrix
      (vrp,
       vpn,
       vup,
       vo_matrix);

   SPH_evaluateViewMappingMatrix
      (-460, 250, -50, 200, PERSPECTIVE,
       prp,
       fplane, bplane,
       0.0, 0.417, 0, 0.42, 0.0, 1.0,
       vm_matrix);

   SPH_setViewRepresentation
      (SIDE_ROOM_VIEW,
       vo_matrix, vm_matrix,
       0.0, 0.417, 0, 0.42, 0.0, 1.0);

   SPH_setViewBackgroundColor(SIDE_ROOM_VIEW, LIGHT_GREY);
}



/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void
init_front_robot_view(void)
{
   rfv_vrp[0] = -100.0;
   rfv_vrp[1] = 0.0;
   rfv_vrp[2] = 55.0;
   rfv_vpn[0] = -100.0;
   rfv_vpn[1] = 0.0;
   rfv_vpn[2] = 15.0;
   rfv_vup[0] = 0.0;
   rfv_vup[1] = 0.0;
   rfv_vup[2] = 1.0;
   rfv_prp[0] = 0.0;
   rfv_prp[1] = 0.0;
   rfv_prp[2] = PRP;

   SPH_evaluateViewOrientationMatrix
      (rfv_vrp,
       rfv_vpn,
       rfv_vup,
       rfv_vo_matrix);

   SPH_evaluateViewMappingMatrix
      (-40, 40, -30, 30, PERSPECTIVE,
       rfv_prp,
       0, -400,
       0.01, 0.29, 0.01, 0.29, 0.0, 1.0,
       rfv_vm_matrix);

   SPH_setViewRepresentation
      (FRONT_ROBOT_VIEW,
       rfv_vo_matrix, rfv_vm_matrix,
       0.01, 0.29, 0.01, 0.29, 0.0, 1.0);


}

/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void
init_side_robot_view(void)
{
   rsv_vrp[0] = 0.0;
   rsv_vrp[1] = -100.0;
   rsv_vrp[2] = 55.0;
   rsv_vpn[0] = 0.0;
   rsv_vpn[1] = -200.0;
   rsv_vpn[2] = 30.0;
   rsv_vup[0] = 0.0;
   rsv_vup[1] = 0.0;
   rsv_vup[2] = 1.0;
   rsv_prp[0] = 0.0;
   rsv_prp[1] = 0.0;
   rsv_prp[2] = PRP;

   SPH_evaluateViewOrientationMatrix
      (rsv_vrp,
       rsv_vpn,
       rsv_vup,
       rsv_vo_matrix);

   SPH_evaluateViewMappingMatrix
      (-40, 40, -30, 30, PERSPECTIVE,
       rsv_prp,
       0, -400,
       0.31, 0.59, 0.01, 0.29, 0.0, 1.0,
       rsv_vm_matrix);

   SPH_setViewRepresentation
      (SIDE_ROBOT_VIEW,
       rsv_vo_matrix, rsv_vm_matrix,
       0.31, 0.59, 0.01, 0.29, 0.0, 1.0);


}

/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void
init_top_robot_view(void)
{

   rtv_vrp[0] = 0.0;
   rtv_vrp[1] = 0.0;
   rtv_vrp[2] = 80.0;
   rtv_vpn[0] = 0.0;
   rtv_vpn[1] = 0.0;
   rtv_vpn[2] = 1.0;
   rtv_vup[0] = 0.0;
   rtv_vup[1] = 1.0;
   rtv_vup[2] = 0.0;
   rtv_prp[0] = 0.0;
   rtv_prp[1] = 0.0;
   rtv_prp[2] = PRP;

   SPH_evaluateViewOrientationMatrix
      (rtv_vrp,
       rtv_vpn,
       rtv_vup,
       rtv_vo_matrix);

   SPH_evaluateViewMappingMatrix
      (-40, 40, -40, 40, PERSPECTIVE,
       rtv_prp,
       0, -400,
       0.61, 0.89, 0.01, 0.29, 0.0, 1.0,
       rtv_vm_matrix);

   SPH_setViewRepresentation
      (TOP_ROBOT_VIEW,
       rtv_vo_matrix, rtv_vm_matrix,
       0.61, 0.89, 0.01, 0.29, 0.0, 1.0);


}


/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void
update_robot_views(float robot_x, float robot_y)
{
     /* robot top view */
     rtv_vrp[0] = robot_x * 10;
     rtv_vrp[1] = robot_y * 10;
     SPH_evaluateViewOrientationMatrix
       (rtv_vrp,
	rtv_vpn,
	rtv_vup,
	rtv_vo_matrix);
     SPH_setViewRepresentation
       (TOP_ROBOT_VIEW,
	rtv_vo_matrix, rtv_vm_matrix,
	0.61, 0.89, 0.01, 0.29, 0.0, 1.0);

     /* robot side view */
     rsv_vrp[0] = robot_x * 10;
     rsv_vrp[1] = robot_y * 10 - 100;
     SPH_evaluateViewOrientationMatrix
       (rsv_vrp,
	rsv_vpn,
	rsv_vup,
	rsv_vo_matrix);
     SPH_setViewRepresentation
       (SIDE_ROBOT_VIEW,
	rsv_vo_matrix, rsv_vm_matrix,
	0.31, 0.59, 0.01, 0.29, 0.0, 1.0);

     /* robot front view */
     rfv_vrp[0] = robot_x * 10 - 100;
     rfv_vrp[1] = robot_y * 10;
     SPH_evaluateViewOrientationMatrix
       (rfv_vrp,
	rfv_vpn,
	rfv_vup,
	rfv_vo_matrix);
     SPH_setViewRepresentation
       (FRONT_ROBOT_VIEW,
	rfv_vo_matrix, rfv_vm_matrix,
	0.01, 0.29, 0.01, 0.29, 0.0, 1.0);
}



/*****************************************************************************
*                                                                            *
* void moveto( float robot_x, float robot_y, float robot_theta )             *
*                                                                            *
*    Move the robot to the desired x,y location with the desired heading.    *
*                                                                            *
*    NOTE: X,Y are in feet, referenced from 0,0 at north east corner of room.*
*          X increases along the long wall towards the window.  	     *
*          Y increases along the short wall, towards the door.               *
*          robot_theta is in DEGREES clockwise from the robot's home position*
*                                                                            *
*****************************************************************************/

int firstTime = 0;
int call_count = 0;     

/** GRG:  BUG: is_up is always passed as 1, but should be 0 the first time */
void
moveto(float robot_x, float robot_y, float robot_theta, int color, int is_up)
{
   matrix          temp_matrix1, temp_matrix2, temp_matrix;
   static float last_x= 0, last_y=0;
   
   robot_x = convert_meters_to_feet_x(robot_x);
   robot_y = convert_meters_to_feet_y(robot_y);

#if 0
   printf("moveto called: x: %f y: %f ", robot_x, robot_y);
   printf("moveto called: x: %f y: %f\n", last_x, last_y);
#endif
   
   if ((fabs(robot_x - last_x ) >MOVE_DELTA) || (fabs(robot_y - last_y) > MOVE_DELTA )) {
     if (firstTime == 0) {
       firstTime = 1;
       /* add the robot to the room */
       SPH_openStructure(ROOM_VIEW);
       SPH_offsetElementPointer(-1);
       SPH_label(MOVE_ROBOT_BASE);
       SPH_translate(robot_x * 10, robot_y *10, 0.0, temp_matrix2);
       SPH_scale(4*XSCALE, 4*YSCALE, 1, temp_matrix1); 
       SPH_compose(temp_matrix1, temp_matrix2, temp_matrix);
       SPH_setModXform(temp_matrix,ASSIGN);
/*       SPH_executeStructure(ROBOT_BASE); */
//       SPH_executeStructure(MANTIS_BASE); 

       SPH_setInteriorColor(CUBE_COLOR);
       SPH_label(TURN_ROBOT_BASE);
       SPH_setModXform(SPH_rotateZ(0, temp_matrix), POSTCONCATENATE);
       SPH_setModXform(SPH_translate(0, 0, 0, 
				     temp_matrix), POSTCONCATENATE);

       SPH_executeStructure(MANTIS); 
//       SPH_polyhedron(8, 6, cube_vertices, cube_facets);
       SPH_closeStructure();
//       update_robot_views(robot_x,robot_y);
       draw_room();
     }
     else if ( (call_count++ % ROBOT_UPDATE_RATE) == 0) {
       SPH_openStructure(ROOM_VIEW);
       SPH_setElementPointer(0);
       SPH_moveElementPointerToLabel(MOVE_ROBOT_BASE);
       SPH_offsetElementPointer(1);
       SPH_deleteElement();
       SPH_translate(robot_x * 10, robot_y *10, 0.0, temp_matrix2);
       SPH_scale(4, 4, 1, temp_matrix1); 
       SPH_compose(temp_matrix1, temp_matrix2, temp_matrix);
       SPH_setModXform(temp_matrix,ASSIGN);
       SPH_setElementPointer(0);
       SPH_moveElementPointerToLabel(TURN_ROBOT_BASE);
       SPH_offsetElementPointer(1);
       SPH_deleteElement();
       SPH_setModXform(SPH_rotateZ(robot_theta, temp_matrix), POSTCONCATENATE);
       SPH_closeStructure();
//       update_robot_views(robot_x,robot_y);
       draw_room();
     }
     last_x = robot_x;
     last_y = robot_y;
   }
}

/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
point * convert_points_2d_to_3d( NDC_point points[],int size)

{
  int i;
  static point (*p);

  if ((  p = (point *) malloc(sizeof( point) *size *2)) == NULL)
      fprintf(stderr, "Error: Ran of out memory creating %s multipoint measure.");

  for (i = 0; i < size; i++) {
    p[i][0] = points[i].x;
    p[i][1] = points[i].y;
    p[i][2] = 0;
    p[i+size][0] = points[i].x;
    p[i+size][1] = points[i].y;
    p[i+size][2] = height_to_make_3d_from_2d;
  }
  return p;
}

/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void  convert_points_to_feet(NDC_point twod_points[],int size)
{
int i;
for (i = 0; i < size; i++) {
    twod_points[i].x = (twod_points[i].x * 39.37*XSCALE ) / 1.2;
    twod_points[i].y = (twod_points[i].y * 39.37*YSCALE ) / 1.2;}
}

/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
vertex_index * make_facets (point p[],int size)
{
  vertex_index (*q);
  int i, index;

  /* the array q needs (size + 1) + size * ( 4 + 1) places */
  if ((  q = (vertex_index *)malloc (sizeof(vertex_index) *
		      ( (size + 1) + size * (4 +1 ))))  == NULL) 
      fprintf(stderr, "Error: Ran of out memory creating %s multipoint measure.");

  /* make the floor */
  for (i = 0 ; i< size ; i++)
    q[i] = i;
  q[size] = -1;

  /* make all the facets of the room */
  /* there will be one for each vertex */
  index = size + 1;
  for (i = 0 ;  i < (size - 1) ; i++) {
    q[index++] = i; 
    q[index++] = i + size;  
    q[index++] = i+1+size;
    q[index++] = i+1;
    q[index++] = -1;
  }
  q[index++] = i; 
  q[index++] = i + size;  
  q[index++] = size;
  q[index++] = 0;
  q[index++] = -1;
  return q;
}
  

/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void add_room_structure()
{  int i;
int size;
point origin;
/* Draw The Room */

   SPH_openStructure(ROOM_VIEW);
   SPH_offsetElementPointer(-1);

SPH_setInteriorColor(ROOM_COLOR);
#if DEBUG
for (i = 0; i <number_room_vertices   ; i++) 
  printf("(%d) x:%f y: %f z: %f\n", i,
	 room_vertices[i][0], room_vertices[i][1], room_vertices[i][2]);
size = number_room_vertices / 2;
for (i= 0; i< ((size + 1) + size * 5) ;i++) {
  printf(" %d ",room_facets[i]);
  if (room_facets[i] == -1) 
    printf("\n");
}
#endif


SPH_polyhedron(number_room_vertices, number_facets, room_vertices, room_facets);

/*SPH_polyhedron(12, 7, room_vertices2, room_facets2); */
SPH_closeStructure();
}

char colors[100][20];
char colorCount=0;
int getColor(char *color)
{
  int i;
  for(i=0;i<colorCount;i++) {
    if(!strcmp(color, colors[i])) return i+OBJECT_COLOR_BASE;
  }
  SPH_loadCommonColor(colorCount+OBJECT_COLOR_BASE, color);
  strcpy(colors[colorCount],color);
  colorCount++;
  return colorCount-1+OBJECT_COLOR_BASE;
}

/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void define_object(int index, char *color)
{
  int colori=getColor(color);
// printf("color %s, index %d\n",color,colori);
  SPH_openStructure(index+OBJECT_BASE);
  if(index>maxObjectCount-1) {
    maxObjectCount=index+1;
    SPH_label(OBJECT_START);
    SPH_setInteriorColor(colori);
    SPH_polyhedron(8, 6, obstacle_vertices, obstacle_facets);
    SPH_label(OBJECT_END);
  } else {
    SPH_setElementPointer(0);
    SPH_deleteElementsBetweenLabels(OBJECT_START,OBJECT_END);
    SPH_setElementPointer(0);
    SPH_moveElementPointerToLabel(OBJECT_START);
    SPH_offsetElementPointer(1);
    SPH_setInteriorColor(colori);
    SPH_polyhedron(8, 6, obstacle_vertices, obstacle_facets);
  }
  SPH_closeStructure();
}

void
define_structures()
{
   matrix          temp_matrix, temp_matrix1, temp_matrix2, temp_matrix3;
   point           origin;

   /* Create NULL_OBJECT */
   SPH_openStructure(NULL_OBJECT);
   SPH_closeStructure();

   /* create obstacle object */
   SPH_openStructure(OBSTACLE);
   SPH_setInteriorColor(OBSTACLE_COLOR);
   SPH_polyhedron(8, 6, obstacle_vertices, obstacle_facets);
   SPH_closeStructure();

   /* Create a passage_pooint */
   SPH_openStructure(PASSAGE_POINT);
   SPH_setInteriorColor(BLACK);
   SPH_fillArea(4, passage_point);
   SPH_closeStructure();

   /* Create ROBOT_BASE */
   SPH_openStructure(ROBOT_BASE);
   SPH_setInteriorColor(BASE_COLOR);
   SPH_polyhedron(16, 10, body_vertices, body_facets);
   SPH_closeStructure();

   /* Create MANTIS_BASE */
   SPH_openStructure(MANTIS_BASE);
   SPH_setInteriorColor(MANTIS_BASE_COLOR);
   SPH_polyhedron(8,6, mantis_vertices, mantis_facets);
   SPH_closeStructure();

   SPH_openStructure(MANTIS_BODY);
   SPH_setInteriorColor(MANTIS_BASE_COLOR);
   SPH_polyhedron(8,6, mantis_body_vertices, mantis_body_facets);
   SPH_closeStructure();

   SPH_openStructure(MANTIS_LEG);
   SPH_setInteriorColor(MANTIS_BASE_COLOR);
   SPH_polyhedron(8,6, mantis_leg_vertices, mantis_leg_facets);
   SPH_closeStructure();

   SPH_openStructure(MANTIS);
   SPH_executeStructure(MANTIS_BODY);

/* six legs */
   SPH_translate(BODY_LENGTH-1, -LEG_LENGTH+1, 0, temp_matrix1);
   SPH_rotateX(45.0,temp_matrix2);
   SPH_compose(temp_matrix1, temp_matrix2, temp_matrix);
   SPH_setModXform(temp_matrix, ASSIGN);
   SPH_executeStructure(MANTIS_LEG);

   SPH_translate(0, -LEG_LENGTH+1, 0, temp_matrix1);
   SPH_rotateX(45.0,temp_matrix2);
   SPH_compose(temp_matrix1, temp_matrix2, temp_matrix);
   SPH_setModXform(temp_matrix, ASSIGN);
   SPH_executeStructure(MANTIS_LEG);
   
   SPH_translate(-(BODY_LENGTH-1), -LEG_LENGTH+1, 0, temp_matrix1);
   SPH_rotateX(45.0,temp_matrix2);
   SPH_compose(temp_matrix1, temp_matrix2, temp_matrix);
   SPH_setModXform(temp_matrix, ASSIGN);
   SPH_executeStructure(MANTIS_LEG);

   SPH_translate(BODY_LENGTH-1, LEG_LENGTH-1, 0, temp_matrix1);
   SPH_rotateX(-45.0,temp_matrix2);
   SPH_compose(temp_matrix1, temp_matrix2, temp_matrix);
   SPH_setModXform(temp_matrix, ASSIGN);
   SPH_executeStructure(MANTIS_LEG);

   SPH_translate(0, LEG_LENGTH-1, 0, temp_matrix1);
   SPH_rotateX(-45.0,temp_matrix2);
   SPH_compose(temp_matrix1, temp_matrix2, temp_matrix);
   SPH_setModXform(temp_matrix, ASSIGN);
   SPH_executeStructure(MANTIS_LEG);
   
   SPH_translate(-(BODY_LENGTH-1), LEG_LENGTH-1, 0, temp_matrix1);
   SPH_rotateX(-45.0,temp_matrix2);
   SPH_compose(temp_matrix1, temp_matrix2, temp_matrix);
   SPH_setModXform(temp_matrix, ASSIGN);
   SPH_executeStructure(MANTIS_LEG);

/* head */

   SPH_rotateZ(90.0,temp_matrix1);
   SPH_rotateY(-45.0,temp_matrix2);
   SPH_compose(temp_matrix1, temp_matrix2, temp_matrix3);
   SPH_translate(BODY_LENGTH+LEG_LENGTH-1, 0, 0, temp_matrix2);
   SPH_compose(temp_matrix3, temp_matrix2, temp_matrix);
   SPH_setModXform(temp_matrix, ASSIGN);
   SPH_executeStructure(MANTIS_LEG);

   SPH_closeStructure();

}


/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void
init_views(int render_mode)
{
   init_top_room_view();
   init_side_room_view();
   init_front_room_view();
//   init_front_robot_view();
//   init_side_robot_view();
//   init_top_robot_view();

   SPH_setRenderingMode(TOP_ROOM_VIEW, render_mode);
   SPH_setRenderingMode(SIDE_ROOM_VIEW, render_mode);
   SPH_setRenderingMode(FRONT_ROOM_VIEW, render_mode);
//   SPH_setRenderingMode(FRONT_ROBOT_VIEW, render_mode);
//   SPH_setRenderingMode(SIDE_ROBOT_VIEW, render_mode);
//   SPH_setRenderingMode(TOP_ROBOT_VIEW, render_mode);
   
   SPH_setViewPointLightSource(TOP_ROOM_VIEW, 0, 500, 10);
   SPH_setViewPointLightSource(SIDE_ROOM_VIEW, 0, 500, 10);
   SPH_setViewPointLightSource(FRONT_ROOM_VIEW, 0, 500, 10);
//   SPH_setViewPointLightSource(TOP_ROBOT_VIEW, 0, 500, 10);
//   SPH_setViewPointLightSource(FRONT_ROBOT_VIEW, 0, 500, 10);
//   SPH_setViewPointLightSource(SIDE_ROBOT_VIEW, 0, 500, 10);
   
   SPH_enableView(TOP_ROOM_VIEW);
   SPH_enableView(SIDE_ROOM_VIEW);
   SPH_enableView(FRONT_ROOM_VIEW);
//   SPH_enableView(FRONT_ROBOT_VIEW);
//   SPH_enableView(SIDE_ROBOT_VIEW);
//   SPH_enableView(TOP_ROBOT_VIEW);
}


/***********************************************************************/
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void 
init_display(int for_overlay,int debug,int solid,int color)
{
  int             i;
  int             render_mode;
  extern int      DCM_KLUDGE;
  extern int      FORCE_BIT_MAPPED;
   
  render_mode = LIT_FLAT;
  SPH_setMaxStructureID(200);
  SPH_setMaxViewIndex(25);
  SPH_begin(800, 576, 7, 4,debug);

  SPH_setDoubleBufferingFlag(TRUE);
  SPH_setImplicitRegenerationMode(SUPPRESSED);

  SPH_loadCommonColor(BLACK, "black");
  SPH_loadCommonColor(WHITE, "white");

  SPH_loadCommonColor(LIGHT_GREY, "grey94");
  SPH_loadCommonColor(GREY, "grey79");
  SPH_loadCommonColor(DARK_GREY, "grey70");
  SPH_loadCommonColor(LIGHT_YELLOW, "light yellow");
  SPH_loadCommonColor(LIGHT_BLUE, "light blue");
  SPH_loadCommonColor(LIGHT_RED, "light pink");
  SPH_loadCommonColor(LIGHT_GREEN, "PaleGreen1");
  SPH_loadCommonColor(YELLOW, "yellow");
  SPH_loadCommonColor(RED, "red");
  SPH_loadCommonColor(GREEN, "green");
  SPH_loadCommonColor(BLUE, "blue");
  TEXT_COLOR = BLACK;
  GOAL_COLOR = RED;
  BASE_COLOR = LIGHT_GREEN;
  ROOM_COLOR = LIGHT_YELLOW;
  CUBE_COLOR = WHITE;
  OBSTACLE_COLOR = BLUE;
  MANTIS_BASE_COLOR = BLUE;
  define_structures();   
  init_views(render_mode);
 
  /* Create Front_Rv */
  SPH_openStructure(FRONT_RV);
  SPH_executeStructure(ROOM_VIEW);
  SPH_setTextColor(TEXT_COLOR);
  origin[0] = 450;
  origin[1] = -50;
  origin[2] = 0;
  SPH_text(origin, "Front Room View"); 
  SPH_closeStructure();

  /* Create TOP_RV */
  SPH_openStructure(TOP_RV);
  SPH_executeStructure(ROOM_VIEW);
  SPH_setTextColor(TEXT_COLOR);
  origin[0] = 450;
  origin[1] = -25;
  origin[2] = 0;
  SPH_text(origin, "Top room view");
  SPH_closeStructure();

  /* Create SIDE_RV */
  SPH_openStructure(SIDE_RV);
  SPH_executeStructure(ROOM_VIEW);
  SPH_setTextColor(TEXT_COLOR);
  origin[0] = 0;
  origin[1] = 300;
  origin[2] = -10;
  SPH_text(origin, "Side room view");
  SPH_closeStructure();

  /* post the roots */

  SPH_postRoot(TOP_RV, TOP_ROOM_VIEW);
  SPH_postRoot(FRONT_RV, FRONT_ROOM_VIEW);
  SPH_postRoot(SIDE_RV, SIDE_ROOM_VIEW);
//  SPH_postRoot(ROOM_VIEW, TOP_ROBOT_VIEW);
//  SPH_postRoot(ROOM_VIEW, FRONT_ROBOT_VIEW);
//  SPH_postRoot(ROOM_VIEW, SIDE_ROBOT_VIEW);

  SPH_openStructure(ROOM_VIEW);
  SPH_label(ROOM_START);
  SPH_label(ROOM_END);
  SPH_closeStructure();
};


/***********************************************************************/
/* sets 4 globals: Number_Room_Vertices, Room_Vertices, Number_Facets, */
/*                 Room_Facets.                                        */
/***********************************************************************/
void make_outline_of_room (char *name, gt_Point_list *pt)
{
  NDC_point twod_points[100];
  int size = 0;
  /* horrible hack to convert from numbers in mlab to these numbers */
  /* probably should run through the numbers to get size then malloc twod_points */
  while (pt != NULL) {
    twod_points[size].x = pt->x;
    twod_points[size].y = pt->y;
    size ++;
    pt = pt->next;
  };
  size = size - 1;    
  
  number_room_vertices = size * 2;
  convert_points_to_feet(twod_points,size);
  room_vertices = convert_points_2d_to_3d(twod_points,size);
  number_facets = size + 1;
  room_facets  =  make_facets (room_vertices,size);
  add_room_structure();
}

void restart_display()
{
  objectCount=0;
  firstTime=0; call_count=0;
  SPH_openStructure(ROOM_VIEW);
  SPH_setElementPointer(0);
  SPH_deleteElementsBetweenLabels(ROOM_START,ROOM_END);
  SPH_closeStructure();
}












