/**********************************************************************
 **                                                                  **
 **                          mission_design.c                        **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  Copyright 2002 Georgia Tech Research Corporation                **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/

/* $Id: mission_design.c,v 1.4 2003/04/06 12:44:19 endo Exp $ */

//-----------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------
#include <stdio.h>
#include <math.h>
#include <signal.h>
#include <memory.h>
#include <malloc.h>
#include <time.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <string>
#include <vector>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>

#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>
#include <Xm/ArrowBG.h>
#include <Xm/BulletinB.h>
#include <Xm/CascadeB.h>   
#include <Xm/CascadeBG.h>   
#include <Xm/DialogS.h>
#include <Xm/DrawingA.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/List.h>
#include <Xm/MainW.h>
#include <Xm/MessageB.h>
#include <Xm/MenuShell.h>
#include <Xm/PanedW.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/ScrollBar.h>
#include <Xm/ScrolledW.h>
#include <Xm/Separator.h>
#include <Xm/SeparatoG.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>

#include "mission_design.h"
#include "mission_expert.h"
#include "gt_console_windows.h"
#include "gt_console_db.h"
#include "gt_sim.h"
#include "draw.h"
#include "console.h"
#include "file_utils.h"
#include "version.h"
#include "EventLogging.h"
#include "bitmaps/Biohazard.bit"
#include "bitmaps/BiohazardBtn.bit"
#include "bitmaps/Waypoints.bit"
#include "bitmaps/WaypointsBtn.bit"
#include "bitmaps/Mine.bit"
#include "bitmaps/MineBtn.bit"
#include "bitmaps/Other.bit"
#include "bitmaps/OtherBtn.bit"

//-----------------------------------------------------------------------
// Macros
//-----------------------------------------------------------------------
#define MMD_CREATE_PIXMAP(name,fg,bg,parent) XCreatePixmapFromBitmapData(XtDisplay(parent),\
	RootWindowOfScreen(XtScreen(parent)),\
	(char *)name##_bits, name##_width, name##_height, fg, bg, \
	DefaultDepthOfScreen(XtScreen(parent)))

//-----------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------
const int MlabMissionDesign::DEFAULT_TASKPOINT_RADIUS_ = 15;
const int MlabMissionDesign::DEFAULT_TOOLBOX_WIDTH_ = 165;
const int MlabMissionDesign::DEFAULT_TOOLBOX_XPOS_ = 50;
const int MlabMissionDesign::DEFAULT_TOOLBOX_YPOS_ = 150;
const int MlabMissionDesign::DEFAULT_TOOLBOX_ACTIONAREA_HEIGHT_ = 70;
const int MlabMissionDesign::DEFAULT_TOOLBOX_EXTRASPACE_HEIGHT_ = 40;
const int MlabMissionDesign::DEFAULT_TOOLBOX_BUTTON_HEIGHT_ = 70;
const int MlabMissionDesign::DEFAULT_TOOLBOX_BUTTON_WIDTH_ = 70;
const int MlabMissionDesign::DEFAULT_TOOLBOX_BUTTON_EXTRASPACE_HEIGHT_ = 5;
const int MlabMissionDesign::DEFAULT_TOOLBOX_NUM_COLUMNS_ = 2;
const int MlabMissionDesign::DEFAULT_RIGHTBTN_WIDTH_ = 350;
const int MlabMissionDesign::DEFAULT_RIGHTBTN_LINEHEIGHT_ = 32;
const int MlabMissionDesign::DEFAULT_RIGHTBTN_EDGEOFFSET_ = 20;
const int MlabMissionDesign::DEFAULT_GLOBALFEATURE_WIDTH_ = 350;
const int MlabMissionDesign::DEFAULT_GLOBALFEATURE_LINEHEIGHT_ = 32;
const int MlabMissionDesign::DEFAULT_GLOBALFEATURE_XEDGEOFFSET_ = 50;
const int MlabMissionDesign::DEFAULT_GLOBALFEATURE_YPOS_ = 100;
const string MlabMissionDesign::TOOLBOX_OK_BUTTON_LABEL_ = "Finish";
const string MlabMissionDesign::TOOLBOX_CANCEL_BUTTON_LABEL_ = "Cancel & Return";
const string MlabMissionDesign::TOOLBOX_GLOBALSETTING_BUTTON_LABEL_ = "Global Settings";
const string MlabMissionDesign::EMPTY_STRING_ = "";

//-----------------------------------------------------------------------
// Constructor for the MlabMissionDesign class.
//-----------------------------------------------------------------------
MlabMissionDesign::MlabMissionDesign(
    Display *display,
    Widget parent,
    XtAppContext app,
    Widget drawing_area,
    char *featureFileString,
    char **defaultOverlayName)
{
    int i;
    bool featuresLoaded = false;

    gEventLogging->start("MMD");

    MMDDisplay_ = display;
    MMDParent_ = parent;
    MMDAppContext_ = app;
    MMDDrawingArea_ = drawing_area;

    XtVaGetValues
	(MMDParent_,
	 XmNforeground, &MMDParentFg_,
	 XmNbackground, &MMDParentBg_,
	 NULL);

    MMDMissionAreaHeightPixels_ = drawing_area_height_pixels;

    defaultTaskName_ = EMPTY_STRING_;
    numTaskTypes_ = 0;
    numTasks_ = 0;
    taskPointRadius_ = DEFAULT_TASKPOINT_RADIUS_;

    featuresLoaded = loadFeaturesFromRCFile_();

    if (!featuresLoaded)
    {
	confirmNoRC_();
	quit_mlab();
    }

    // Set the current mode based on the default task ID.
    MMDMode_ = getDefaultMode_();

    // Create a list of global preference with default values.
    globalFeatureMenu_w_ = NULL;

    // Set the color for the toolbox buttons.
    MMDToolboxBtnBgColorSet_ = gColorPixel.white;
    MMDToolboxBtnBgColorUnset_ = gColorPixel.orange;

    // Right-Button-Menu setup
    rightBtnMenuIsUp_ = false;

    // Some misc setup
    generalYesNoWindowDone_ = false;
    returnConfigureMission_ = MMD_EXIT_CANCEL_RETURN;
    windowsCreated_ = false;

    // Name the files to be used.
    featureFilename_ = featureFileString;
    featureFilename_ += ".";
    featureFilename_ += MEXP_EXTENTION_FEATUREFILENAME;

    featureStatusFilename_ = featureFileString;
    featureStatusFilename_ += ".";
    featureStatusFilename_ += MEXP_EXTENTION_FEATURESTATUSFILENAME;

    featureFile_ = NULL;
    featureStatusFile_ = NULL;
    overlayFilename_ = EMPTY_STRING_;
    defaultTaskWeight_ = 1.0;

    // Load the data in the feature file if there is any.
    loadData_();

    // Save the overlay file name.
    if (overlayFilename_ != EMPTY_STRING_)
    {
	if (*defaultOverlayName != NULL)
	{
	    free(*defaultOverlayName);
	}

	*defaultOverlayName = strdup(overlayFilename_.c_str());
    }
}

//-----------------------------------------------------------------------
// Destructor for the MlabMissionDesign class.
//-----------------------------------------------------------------------
MlabMissionDesign::~MlabMissionDesign(void)
{
    switch (returnConfigureMission_) {

    case MMD_EXIT_CONFIGURE:
	saveData_();
	gEventLogging->end("MMD");
	break;

    case MMD_EXIT_CANCEL_RETURN:
    default:
	gEventLogging->cancel("MMD");
	break;
    }

    setFeatureFileStatus_();

    if (windowsCreated_)
    {
	XtDestroyWidget(toolbox_w_);
	XtDestroyWidget(globalFeatureMenu_w_);
    }
    windowsCreated_ = false;

    if (numTasks_)
    {
	specifiedTaskList_.clear();
    }
}

//-----------------------------------------------------------------------
// This function saves the specified features to the feature file.
//-----------------------------------------------------------------------
void MlabMissionDesign::saveData_(void)
{
    int i, j, k;
    MExpFeature_t globalFeature;
    MExpMissionTask_t task;
    MExpFeature_t localFeature;
    MExpFeatureOption_t option;

    featureFile_ = fopen(featureFilename_.c_str(), "w+");

    if (featureFile_ == NULL)
    {
	fprintf(stderr, "Warning (mlab): MlabMissionDesign::saveData_().\n");
	fprintf(stderr, "The feature file could not be opened.\n");
	returnConfigureMission_ = MMD_EXIT_CANCEL_RETURN;
	return;
    }
    
    // Save the overlay file name.
    fprintf(featureFile_, "%s %s\n",
	    MEXP_STRING_OVERLAY.c_str(),
	    overlayFilename_.c_str());
    fprintf(featureFile_, "\n");

    for (i = 0; i < ((int)(globalFeatures_.size())); i++)
    {
	// Save the global preference.
	globalFeature = globalFeatures_[i];
	fprintf(featureFile_, "%s %d %.2f %d %d \"%s\"\n",
		MEXP_STRING_GLOBALFEATURE.c_str(),
		i,
		globalFeature.selectedOption,
		(int)(globalFeature.options.size()),
		globalFeature.optionType,
		globalFeature.name.c_str());

	for (j = 0; j < ((int)(globalFeature.options.size())); j++)
	{
	    option = globalFeature.options[j];
	    fprintf(featureFile_, "%s %d \"%s\"\n",
		    MEXP_STRING_OPTION.c_str(),
		    j,
		    option.value.c_str());
	}
    }
    fprintf(featureFile_, "\n");

    // Save the data specific to each task.
    for (i = 0; i<numTasks_; i++)
    {
	//task = tasksAry_[i];
	task = specifiedTaskList_[i];

	// Save the task name.
	fprintf(featureFile_, "%s %d %d \"%s\"\n",
		MEXP_STRING_TASK.c_str(),
		i,
		task.id,
		task.name.c_str());

	// Save the position of the task.
	fprintf(featureFile_, "%s %5.2f %5.2f\n",
		MEXP_STRING_COORDINATE.c_str(),
		task.position.x,
		task.position.y);

	// Save the data specific to each feature.
	for (j = 0; j < ((int)(task.localFeatures.size())); j++)
	{
	    //feature = task.featurees[j];
	    localFeature = task.localFeatures[j];
	    fprintf(featureFile_, "%s %d %.2f %d %d \"%s\"\n",
		    MEXP_STRING_FEATURE.c_str(),
		    j,
		    localFeature.selectedOption,
		    (int)(localFeature.options.size()),
		    localFeature.optionType,
		    localFeature.name.c_str());

	    for (k = 0; k < ((int)(localFeature.options.size())); k++)
	    {
		option = localFeature.options[k];
		fprintf(featureFile_, "%s %d \"%s\"\n",
			MEXP_STRING_OPTION.c_str(),
			k,
			option.value.c_str());
	    }
	}
	fprintf(featureFile_, "\n");
    }

    fclose(featureFile_);
}

//-----------------------------------------------------------------------
// This function loads the task data from the feature file.
//-----------------------------------------------------------------------
int MlabMissionDesign::loadData_(void)
{
    FILE *featureFile;
    MExpMissionTask_t addedTask;
    char buf1[1024], buf2[1024], buf3[1024];
    string bufString1, bufString2, bufString3;
    float x, y;
    int taskNum, taskId;
    int globalFeatureId, featureId;
    float globalFeatureSelectedOption, featureSelectedOption;
    bool taskCreated = false, hadError = false;
    
    featureFile = fopen(featureFilename_.c_str(), "r");

    if (featureFile == NULL)
    {
	return false;
    }

    gEventLogging->start("MMD Loading the feature data from %s", featureFilename_.c_str());

    while(!feof(featureFile))
    {
	if (fscanf(featureFile, "%s", buf1) == EOF)
	{
	    break;
	}

	bufString1 = buf1;

	if (bufString1 == MEXP_STRING_OVERLAY)
	{
	    // Save the overlay name.
	    fscanf(featureFile, "%s", buf2);
	    overlayFilename_ = buf2;
	}
	else if (bufString1 == MEXP_STRING_GLOBALFEATURE)
	{
	    // Get the global preference ID.
	    if (fscanf(featureFile, "%s", buf2) == EOF)
	    {
		hadError = true;
		break;
	    }
	    globalFeatureId = atoi(buf2);

	    // Get the selected option for the global preference.
	    if (fscanf(featureFile, "%s", buf3) == EOF)
	    {
		hadError = true;
		break;
	    }
	    globalFeatureSelectedOption = atof(buf3);

	    globalFeatures_[globalFeatureId].selectedOption = globalFeatureSelectedOption;
	}
	else if (bufString1 == MEXP_STRING_TASK)
	{
	    // Get the task number (i.e., how many tasks?).
	    if (fscanf(featureFile, "%s", buf2) == EOF)
	    {
		hadError = true;
		break;
	    }
	    taskNum = atoi(buf2);

	    // Get the task ID (e.g., Biohazrd, EOD, etc.)
	    if (fscanf(featureFile, "%s", buf3) == EOF)
	    {
		hadError = true;
		break;
	    }
	    taskId = atoi(buf3);

	    // Add a new task at <0, 0> for now.
	    addNewTask_(0, 0, taskId, &addedTask);
	    taskCreated = true;
	}
	else if ((bufString1 == MEXP_STRING_COORDINATE) && (taskCreated))
	{
	    // Get the X coordinate.
	    if (fscanf(featureFile, "%f", &x) == EOF)
	    {
		hadError = true;
		break;
	    }
	    specifiedTaskList_[numTasks_-1].position.x = (double)x;

	    // Get the Y coordinate.
	    if (fscanf(featureFile, "%f", &y) == EOF)
	    {
		hadError = true;
		break;
	    }
	    specifiedTaskList_[numTasks_-1].position.y = (double)y;
	}
	else if ((bufString1 == MEXP_STRING_FEATURE) && (taskCreated))
	{
	    // Get the feature ID.
	    if (fscanf(featureFile, "%s", buf2) == EOF)
	    {
		hadError = true;
		break;
	    }
	    featureId = atoi(buf2);

	    // Get the selected option for the feature.
	    if (fscanf(featureFile, "%s", buf3) == EOF)
	    {
		hadError = true;
		break;
	    }
	    featureSelectedOption = atof(buf3);

	    specifiedTaskList_[numTasks_-1].localFeatures[featureId].selectedOption = featureSelectedOption;
	}
    }

    fclose(featureFile);

    gEventLogging->end("MMD Loading the feature data");

    if (hadError)
    {
	fprintf(stderr, "Error(mlab): MlabMissionDesign::loadData_(). Feature file contains errror.\n\a");
    }

    if (!taskCreated) return false;

    return true;
}

//-----------------------------------------------------------------------
// This function saves the specified features to the output file.
//-----------------------------------------------------------------------
void MlabMissionDesign::setFeatureFileStatus_(void)
{
    featureStatusFile_ = fopen(featureStatusFilename_.c_str(), "w+");

    if (featureStatusFile_ == NULL)
    {
	fprintf(stderr, "Warning (mlab): MlabMissionDesign::setFeatureFileStatus_().\n");
	fprintf(stderr,	"The feature status file %s could not be opened.\n",
		featureStatusFilename_.c_str());
	return;
    }

    switch (returnConfigureMission_) {

    case MMD_EXIT_CONFIGURE:
	fprintf(featureStatusFile_, "%s %s",
		MEXP_STRING_CONFIGURE.c_str(),
		MEXP_STRING_YES.c_str());
	break;

    case MMD_EXIT_CANCEL_RETURN:
    default:
	fprintf(featureStatusFile_, "%s %s",
		MEXP_STRING_CONFIGURE.c_str(),
		MEXP_STRING_NO.c_str());
	break;
    }

    fclose(featureStatusFile_);
}

//-----------------------------------------------------------------------
// This function creates the toolbox window.
//-----------------------------------------------------------------------
void MlabMissionDesign::createToolbox_(void)
{
    Widget toolbox_frame_w, toolbox_main_w, toolbox_menu_w, toolbox_action_w;
    Widget btnForm, okBtn, cancelBtn, globalSettingBtn;
    MMDToolBoxSetModeCallbackData_t *callbackData;
    MExpMissionTask_t task;
    vector<Widget> stackedWidget;
    Pixmap pixmap, insensitivePixmap;
    int i, numStackedWidget, numTaskBtnRows;
    int buttonHeight, buttonWidth;
    int toolboxHeight, toolboxWidth, toolboxXPos, toolboxYPos;
    int toolboxActionAreaHeight;

    // Get the task button dimensions.
    buttonHeight = DEFAULT_TOOLBOX_BUTTON_HEIGHT_;
    buttonWidth = DEFAULT_TOOLBOX_BUTTON_WIDTH_;

    // Get the action area height.
    toolboxActionAreaHeight = DEFAULT_TOOLBOX_ACTIONAREA_HEIGHT_;

    // Calculate the number of rows for the task buttons.
    numTaskBtnRows = (int)(numTaskTypes_/DEFAULT_TOOLBOX_NUM_COLUMNS_);
    if ((numTaskTypes_ % DEFAULT_TOOLBOX_NUM_COLUMNS_) > 0)
    {
	numTaskBtnRows++;
    }

    // Decide the height of the toolbox based on the number of rows of
    // the button.
    toolboxHeight = (numTaskBtnRows*(buttonHeight+DEFAULT_TOOLBOX_BUTTON_EXTRASPACE_HEIGHT_)) +
	toolboxActionAreaHeight + DEFAULT_TOOLBOX_EXTRASPACE_HEIGHT_;

    // Also, other dimensions of the toolbox.
    toolboxWidth = DEFAULT_TOOLBOX_WIDTH_;
    toolboxXPos = DEFAULT_TOOLBOX_XPOS_;
    toolboxYPos = DEFAULT_TOOLBOX_YPOS_;
    
    // Popup shell
    toolbox_w_ = XtVaCreatePopupShell
 	(" ",
	 xmDialogShellWidgetClass, MMDParent_, 
	 XmNdeleteResponse, XmUNMAP,
	 XmNx, toolboxXPos,
	 XmNy, toolboxYPos,
	 XmNheight, toolboxHeight,
	 XmNwidth, toolboxWidth,
	 NULL);

    toolbox_frame_w = XtVaCreateWidget
	("toolbox_frame_w",
	 xmFrameWidgetClass, toolbox_w_,
	 XmNshadowType, XmSHADOW_ETCHED_OUT,
	 XmNheight, toolboxHeight,
	 XmNwidth, toolboxWidth,
	 NULL);

    stackedWidget.push_back(toolbox_frame_w);

    toolbox_main_w = XtVaCreateWidget
	("toolbox_main_w",
	 xmRowColumnWidgetClass, toolbox_frame_w,
	 XmNpacking, XmPACK_TIGHT,
	 XmNorientation, XmVERTICAL,
	 XmNnumColumns, 1,
	 XmNisAligned, True,
	 XmNentryAlignment, XmALIGNMENT_CENTER,
	 XmNheight, toolboxHeight,
	 XmNwidth, toolboxWidth,
	 NULL);

    stackedWidget.push_back(toolbox_main_w);

    toolbox_menu_w = XtVaCreateWidget
	("toolbox_menu_w",
	 xmRowColumnWidgetClass, toolbox_main_w,
	 XmNpacking, XmPACK_COLUMN,
	 XmNorientation, XmVERTICAL,
	 XmNnumColumns, DEFAULT_TOOLBOX_NUM_COLUMNS_,
	 XmNisAligned, true,
	 XmNentryAlignment, XmALIGNMENT_CENTER,
	 XmNheight, toolboxHeight-toolboxActionAreaHeight,
	 XmNwidth, toolboxWidth,
	 NULL);

    stackedWidget.push_back(toolbox_menu_w);

    // Task buttons
    task_btn_ = new Widget[numTaskTypes_];

    // Create buttons
    for (i = 0; i < numTaskTypes_; i++)
    {
	// The for for the button.
	btnForm = XtVaCreateManagedWidget
	    ("btnForm",
	     xmFormWidgetClass, toolbox_menu_w,
	     XmNalignment, XmALIGNMENT_CENTER,
	     XmNheight, buttonHeight,
	     XmNwidth, buttonWidth,
	     NULL);

	stackedWidget.push_back(btnForm);

	task = defaultTaskList_[i];
	pixmap = getPixmapFromList_(task.buttonName, false);
	insensitivePixmap = getPixmapFromList_(task.buttonName, true);

	task_btn_[i] = XtVaCreateWidget
	(task.buttonName.c_str(),
	 xmPushButtonWidgetClass, btnForm,
	 XmNlabelType, XmPIXMAP,
	 XmNlabelPixmap, pixmap,
	 XmNlabelInsensitivePixmap, insensitivePixmap,
	 XmNbackground, MMDToolboxBtnBgColorUnset_,
	 XmNhighlightColor, MMDParentBg_,
	 XmNborderColor, MMDToolboxBtnBgColorUnset_,
	 XmNarmColor, MMDToolboxBtnBgColorUnset_,
	 NULL);

	stackedWidget.push_back(task_btn_[i]);

	callbackData = new MMDToolBoxSetModeCallbackData_t;
	callbackData->mode = task.id;
	callbackData->mlabMissionDesignInstance = this;
	XtAddCallback
	    (task_btn_[i],
	     XmNactivateCallback,
	     (XtCallbackProc)cbSetmode_, 
	     (void *)callbackData);
    }

    toolbox_action_w = XtVaCreateWidget
	("toolbox_action_w",
	 xmRowColumnWidgetClass, toolbox_main_w,
	 XmNpacking, XmPACK_COLUMN,
	 XmNorientation, XmVERTICAL,
	 XmNnumColumns, 1,
	 XmNisAligned, True,
	 XmNentryAlignment, XmALIGNMENT_CENTER,
	 XmNtopAttachment, XmATTACH_WIDGET,
	 XmNtopWidget, toolbox_menu_w,
	 XmNheight, toolboxActionAreaHeight,
	 XmNwidth, toolboxWidth,
	 NULL);

    stackedWidget.push_back(toolbox_action_w);

    // Global Setting button
    globalSettingBtn = XtVaCreateWidget
	(TOOLBOX_GLOBALSETTING_BUTTON_LABEL_.c_str(),
	 xmPushButtonWidgetClass, toolbox_action_w,
	 XmNwidth, toolboxWidth,
	 XmNforeground, gColorPixel.white,
	 XmNbackground, gColorPixel.blue,
	 NULL);

    stackedWidget.push_back(globalSettingBtn);

    XtAddCallback
	(globalSettingBtn,
	 XmNactivateCallback,
	 (XtCallbackProc)cbToolboxGlobalSettingBtn_,
	 (void *)this);

    // OK button
    okBtn = XtVaCreateWidget
	(TOOLBOX_OK_BUTTON_LABEL_.c_str(),
	 xmPushButtonWidgetClass, toolbox_action_w,
	 XmNwidth, toolboxWidth,
	 XmNforeground, gColorPixel.white,
	 XmNbackground, gColorPixel.red,
	 NULL);

    stackedWidget.push_back(okBtn);

    callbackData = new MMDToolBoxSetModeCallbackData_t;
    callbackData->mode = true;
    callbackData->mlabMissionDesignInstance = this;
    XtAddCallback
	(okBtn,
	 XmNactivateCallback,
	 (XtCallbackProc)cbExit_,
	 (void *)callbackData);

    // Cancel button
    cancelBtn = XtVaCreateWidget
	(TOOLBOX_CANCEL_BUTTON_LABEL_.c_str(),
	 xmPushButtonWidgetClass, toolbox_action_w,
	 XmNwidth, toolboxWidth,
	 XmNforeground, gColorPixel.white,
	 XmNbackground, gColorPixel.green4,
	 NULL);

    stackedWidget.push_back(cancelBtn);

    callbackData = new MMDToolBoxSetModeCallbackData_t;
    callbackData->mode = false;
    callbackData->mlabMissionDesignInstance = this;
    XtAddCallback
	(cancelBtn,
	 XmNactivateCallback,
	 (XtCallbackProc)cbExit_,
	 (void *)callbackData);

    numStackedWidget = stackedWidget.size();

    for (i = 0; i < numStackedWidget; i++)
    {
	XtManageChild(stackedWidget.back());
	stackedWidget.pop_back();
    }

    XtPopup (toolbox_w_, XtGrabNone);
    updateToolboxBtn_();
}

//-----------------------------------------------------------------------
// This function updates the backgroud of the buttons.
//-----------------------------------------------------------------------
void MlabMissionDesign::updateToolboxBtn_(void)
{
    int i;

    for (i = 0; i < numTaskTypes_; i++)
    {
	XtVaSetValues(task_btn_[i], XmNbackground, toolboxBtnBgColor_(i), NULL);
    }
}

//-----------------------------------------------------------------------
// This function changes the backgroud of the button based on the mode.
//-----------------------------------------------------------------------
Pixel MlabMissionDesign::toolboxBtnBgColor_(int button)
{
    if (button == MMDMode_)
    {
	return MMDToolboxBtnBgColorSet_;
    }
    else
    {
	return MMDToolboxBtnBgColorUnset_;
    }
}

//-----------------------------------------------------------------------
// This function creates the toolbox window and its related widgets.
//-----------------------------------------------------------------------
void MlabMissionDesign::createWindows(void)
{
    createPixmapList_(MMDParentFg_, MMDParentBg_, MMDParent_);
    createToolbox_();
    createGlobalFeatureMenu_();

    windowsCreated_ = true;
}

//-----------------------------------------------------------------------
// This function creates a new task and adds it in the array.
//-----------------------------------------------------------------------
void MlabMissionDesign::addNewTask_(double taskPosX, double taskPosY, int taskType, MExpMissionTask_t *addedTask) 
{
    int i, j;
    string taskName;
    MExpFeature_t localFeature;
    vector<MExpFeature_t> localFeatures;
    MExpMissionTask_t task, defaultTask;

    // At first, create a list of local features with default values.
    localFeatures = defaultLocalFeatures_;

    // Next, find a task in the list that matches with this ID.
    defaultTask = defaultTaskList_[0];
    for (i = 0; i < numTaskTypes_; i++)
    {
	if (defaultTaskList_[i].id == taskType)
	{
	    defaultTask = defaultTaskList_[i];
	    break;
	}
    }

    // Use the specified option for the default task if it is specified.
    for (i = 0; i < ((int)(defaultTask.localFeatures.size())); i++)
    {
	for (j = 0; j < ((int)(localFeatures.size())); j++)
	{
	    if (defaultTask.localFeatures[i].name == localFeatures[j].name)
	    {
		localFeatures[j].selectedOption = defaultTask.localFeatures[i].selectedOption;
	    }
	}
    }

    task.localFeatures = localFeatures;
    task.number = numTasks_;
    task.id = defaultTask.id;
    task.name = defaultTask.name;
    task.position.x = taskPosX;
    task.position.y = taskPosY;
    specifiedTaskList_.push_back(task);

    *addedTask = task;

    numTasks_++;
}

//-----------------------------------------------------------------------
// This function places the specified task mark to the specified place
// in the overlay.
//-----------------------------------------------------------------------
void MlabMissionDesign::placeTask_(double taskPosX, double taskPosY, int taskType) 
{
    char entryKey[1024];
    MExpMissionTask_t addedTask;

    addNewTask_(taskPosX, taskPosY, taskType, &addedTask);

    sprintf(entryKey, "Task #%d [%s]", addedTask.number, addedTask.name.c_str());

    gt_add_passage_point(entryKey, taskPosX, taskPosY, taskPointRadius_*drawingUnitLength_);

    gEventLogging->log("MMD Task #%d [%s] added on <%.2f %.2f>",
		       addedTask.number,
		       addedTask.name.c_str(),
		       addedTask.position.x,
		       addedTask.position.y);
}

//-----------------------------------------------------------------------
// This function removes a task from the list.
//-----------------------------------------------------------------------
void MlabMissionDesign::removeTask_(double taskPosX, double taskPosY) 
{
    char entryKey[1024];
    int i, taskNum;
    MExpMissionTask_t taskToDelete;
    
    if (!isInsideTask_(taskPosX, taskPosY, drawingUnitLength_, &taskToDelete)) return;

    // Delete the task from the entry.
    sprintf(entryKey, "Task #%d [%s]", taskToDelete.number, taskToDelete.name.c_str());
    gt_delete_db_entry(entryKey);

    gEventLogging->log("MMD Task #%d [%s] deleted", taskToDelete.number, taskToDelete.name.c_str());

    numTasks_--;


    if (numTasks_)
    {
	taskNum = taskToDelete.number;
	for (i = taskNum; i < numTasks_; i++)
	{
	    // Copy the one above to this one.
	    specifiedTaskList_[i] = specifiedTaskList_[i+1];
	    specifiedTaskList_[i].number = i;

	    // Make a mark on the overlay.
	    sprintf(entryKey, "Task #%d [%s]", specifiedTaskList_[i].number, specifiedTaskList_[i].name.c_str());
	    gt_add_passage_point(entryKey, specifiedTaskList_[i].position.x, specifiedTaskList_[i].position.y,
				 taskPointRadius_*drawingUnitLength_);

	    // Delete the one above.
	    sprintf(entryKey, "Task #%d [%s]", specifiedTaskList_[i+1].number, specifiedTaskList_[i+1].name.c_str());
	    gt_delete_db_entry(entryKey);

	    gEventLogging->log("MMD Renumbering Task #%d to #%d [%s]",
			       specifiedTaskList_[i+1].number,
			       specifiedTaskList_[i].number,
			       specifiedTaskList_[i].name.c_str());
	}
    }
  
    // Remove the last task in the array (vector).
    specifiedTaskList_.pop_back();

    // Clear the screen and redraw
    clear_map();
    draw_world();
}

//-----------------------------------------------------------------------
// This function redraw the task entry by clear them at first, then
// draw them again.
//-----------------------------------------------------------------------
void MlabMissionDesign::refreshTaskEntry(double unitLength) 
{
    char entryKey[1024];
    int i;
    double taskPosX, taskPosY;

    drawingUnitLength_ = unitLength;

    clear_map();

    for (i = 0; i < numTasks_; i++)
    {
	// Delete the task from the entry.
	sprintf(entryKey, "Task #%d [%s]", i, specifiedTaskList_[i].name.c_str());
	gt_delete_db_entry(entryKey);
    }

    for (i = 0; i < numTasks_; i++)
    {
	sprintf(entryKey, "Task #%d [%s]", i, specifiedTaskList_[i].name.c_str());
	taskPosX = specifiedTaskList_[i].position.x;
	taskPosY = specifiedTaskList_[i].position.y;
	gt_add_passage_point(entryKey, taskPosX, taskPosY, taskPointRadius_*drawingUnitLength_);
    } 

    draw_world();
}

//-----------------------------------------------------------------------
// This function creates the global preference menu window.
//-----------------------------------------------------------------------
void MlabMissionDesign::createGlobalFeatureMenu_(void)
{
    Widget main_frame, main, label_frame, label_box, label;
    Widget feature_frame, feature_box, button_frame;
    Widget apply_frame, apply, cancel_frame, cancel;
    vector<Widget> stackedWidget;
    MMDGlobalFeatureMenuCallbackData_t *optionData;
    XmString sliderTitle;
    char preference_title[1024];
    float value, conversion;
    int i, j, *j_ptr, options, line = 0, s = 0;
    int screenNumber;
    int posX, posY, displayWidth, displayHeight;
    int globalFeatureMenuHeight, globalFeatureMenuWidth, globalFeatureMenuLineHeight;
    int numStackedWidget, numGlobalFeatures;
    int maxValue, decimalPoints;

    // Setting the height of the window.
    numGlobalFeatures = globalFeatures_.size();
    globalFeatureMenuLineHeight = DEFAULT_GLOBALFEATURE_LINEHEIGHT_;
    globalFeatureMenuHeight = (1+2*numGlobalFeatures)*(globalFeatureMenuLineHeight+3);
    globalFeatureMenuWidth = DEFAULT_GLOBALFEATURE_WIDTH_;

    // Make sure the window does go outside the desktop.
    screenNumber = DefaultScreen(MMDDisplay_);
    displayWidth = DisplayWidth(MMDDisplay_, screenNumber);
    displayHeight = DisplayHeight(MMDDisplay_, screenNumber);
    posX = displayWidth - globalFeatureMenuWidth - DEFAULT_GLOBALFEATURE_XEDGEOFFSET_;
    posY = DEFAULT_GLOBALFEATURE_YPOS_;

    // Popup shell.
    globalFeatureMenu_w_ = XtVaCreatePopupShell
	("",
	 xmDialogShellWidgetClass, MMDParent_, 
	 XmNdeleteResponse, XmUNMAP,
	 XmNallowShellResize, true,
	 XmNx, posX,
	 XmNy, posY,
	 XmNheight, globalFeatureMenuHeight,
	 XmNwidth, globalFeatureMenuWidth,
	 NULL);

    // The biggest frame.
    main_frame = XtVaCreateWidget
	("main_frame",
	 xmFrameWidgetClass, globalFeatureMenu_w_,
	 XmNalignment, XmALIGNMENT_BEGINNING,
	 XmNheight, globalFeatureMenuHeight,
	 XmNwidth, globalFeatureMenuWidth,
	 XmNtopAttachment, XmATTACH_WIDGET,
	 XmNtopWidget, globalFeatureMenu_w_,
	 XmNleftAttachment, XmATTACH_WIDGET,
	 XmNleftWidget, globalFeatureMenu_w_,
	 XmNrightAttachment, XmATTACH_WIDGET,
	 XmNrightWidget, globalFeatureMenu_w_,
	 XmNbottomAttachment, XmATTACH_WIDGET,
	 XmNbottomWidget, globalFeatureMenu_w_,
	 NULL);

    stackedWidget.push_back(main_frame);

    main = XtVaCreateWidget
	("main",
	 xmRowColumnWidgetClass, main_frame,
	 XmNpacking, XmPACK_TIGHT,
	 XmNorientation, XmVERTICAL,
	 XmNnumColumns, 1,
	 XmNisAligned, True,
	 XmNentryAlignment, XmALIGNMENT_CENTER,
	 XmNheight, globalFeatureMenuHeight,
	 XmNwidth, globalFeatureMenuWidth,
	 NULL);

    stackedWidget.push_back(main);

    // The widget for the main label.
    label_frame = XtVaCreateWidget
	("label_frame",
	 xmFrameWidgetClass, main,
	 XmNalignment, XmALIGNMENT_BEGINNING,
	 XmNheight, globalFeatureMenuLineHeight,
	 XmNwidth, globalFeatureMenuWidth,
	 NULL);

    stackedWidget.push_back(label_frame);

    // The widget for the main label.
    label_box = XtVaCreateWidget
	("label_box",
	 xmRowColumnWidgetClass, label_frame,
	 XmNpacking, XmPACK_COLUMN,
	 XmNorientation, XmVERTICAL,
	 XmNalignment, XmALIGNMENT_BEGINNING,
	 XmNnumColumns, 1,
	 XmNheight, globalFeatureMenuLineHeight,
	 XmNwidth, globalFeatureMenuWidth,
	 XmNtopAttachment, XmATTACH_FORM,
	 XmNrightAttachment, XmATTACH_FORM,
	 XmNleftAttachment, XmATTACH_FORM,
	 XmNbottomAttachment, XmATTACH_FORM,
	 NULL);

    stackedWidget.push_back(label_box);

    // The main label.
    sprintf(preference_title, "Mission Preference  [Global Settings]");
    label = XtVaCreateWidget
	(preference_title,
	 xmLabelGadgetClass, label_box,
	 XmNalignment, XmALIGNMENT_BEGINNING,
	 XmNheight, globalFeatureMenuLineHeight,
	 XmNwidth, globalFeatureMenuWidth,
	 XmNtopAttachment, XmATTACH_WIDGET,
	 XmNtopWidget, label_box,
	 XmNleftAttachment, XmATTACH_WIDGET,
	 XmNleftWidget, label_box,
	 XmNrightAttachment, XmATTACH_WIDGET,
	 XmNrightWidget, label_box,
	 NULL);
    line++;

    stackedWidget.push_back(label);

    // It is a 2D array. Create the 1D part (feature) first.
    globalFeatureMenuOption_w_ = new Widget*[numGlobalFeatures];

    //  Create the preference raws sequentially.
    for (i = 0; i < numGlobalFeatures; i++)
    {
	if (globalFeatures_[i].id != i)
	{
	    // Global features was not stored in order of their IDs.
	    // Something went wroing. Warn it.
	    fprintf(stderr, "Warning (mlab): MlabMissionDesign::createGlobalFeatureMenu_().\n");
	    fprintf(stderr, "The order of global features corrupted.\n");
	}

	options = globalFeatures_[i].options.size();

	// The widget for the main label.
	feature_frame = XtVaCreateWidget
	    ("feature_frame",
	     xmFrameWidgetClass, main,
	     XmNalignment, XmALIGNMENT_BEGINNING,
	     XmNheight, 2*globalFeatureMenuLineHeight,
	     XmNwidth, globalFeatureMenuWidth,
	     NULL);

	stackedWidget.push_back(feature_frame);

	// Two-colum widget for the feature.
	feature_box = XtVaCreateWidget
	    ("globalFeature-feature_box",
	     xmRowColumnWidgetClass, feature_frame,
	     XmNpacking, XmPACK_COLUMN,
	     XmNorientation, XmHORIZONTAL,
	     XmNalignment, XmALIGNMENT_BEGINNING,
	     XmNnumColumns, options,
	     XmNheight, 2*globalFeatureMenuLineHeight,
	     XmNwidth, globalFeatureMenuWidth,
	     XmNtopAttachment, XmATTACH_FORM,
	     XmNrightAttachment, XmATTACH_FORM,
	     XmNleftAttachment, XmATTACH_FORM,
	     XmNbottomAttachment, XmATTACH_FORM,
	     NULL);

	stackedWidget.push_back(feature_box);

	// Create the second dimention (option) of this array.
	globalFeatureMenuOption_w_[i] = new Widget[options];

	switch (globalFeatures_[i].optionType) {

	case MEXP_FEATURE_OPTION_TOGGLE:

	    // The label column.
	    label = XtVaCreateWidget
		(globalFeatures_[i].name.c_str(),
		 xmLabelGadgetClass, feature_box,
		 XmNalignment, XmALIGNMENT_BEGINNING,
		 XmNheight, globalFeatureMenuLineHeight,
		 XmNwidth, globalFeatureMenuWidth/options,
		 XmNtopAttachment, XmATTACH_WIDGET,
		 XmNtopWidget, feature_box,
		 XmNleftAttachment, XmATTACH_WIDGET,
		 XmNleftWidget, feature_box,
		 NULL);

	    stackedWidget.push_back(label);

	    // Fill the blanks in the label row.
	    for (j = 0; j < (options-1); j++)
	    {
		label = XtVaCreateWidget
		    (" ",
		     xmLabelGadgetClass, feature_box,
		     XmNalignment, XmALIGNMENT_BEGINNING,
		     XmNheight, globalFeatureMenuLineHeight,
		     XmNwidth, globalFeatureMenuWidth/options,
		     XmNtopAttachment, XmATTACH_WIDGET,
		     XmNtopWidget, feature_box,
		     NULL);

		stackedWidget.push_back(label);
	    }
	    
	    for (j = 0; j < options; j++)
	    {
		optionData = new MMDGlobalFeatureMenuCallbackData_t;
		optionData->globalFeatureId = i;
		optionData->optionValue = (float)j;
		optionData->mlabMissionDesignInstance = this;
		optionData->optionType = globalFeatures_[i].optionType;

		if ((int)(globalFeatures_[i].selectedOption) == j)
		{
		    value = 1.0;
		}
		else
		{
		    value = 0.0;
		}

		button_frame = XtVaCreateManagedWidget
		    ("button_frame",
		     xmFormWidgetClass, feature_box,
		     XmNshadowThickness, 0,
		     XmNalignment, XmALIGNMENT_BEGINNING,
		     XmNheight, globalFeatureMenuLineHeight,
		     XmNwidth, globalFeatureMenuWidth/options-10,
		     NULL);

		stackedWidget.push_back(button_frame);

		// Toggle features.
		globalFeatureMenuOption_w_[i][j] = XtVaCreateWidget
		    (globalFeatures_[i].options[j].value.c_str(),
		     xmToggleButtonGadgetClass, button_frame,
		     XmNradioBehavior, true,
		     XmNradioAlwaysOne, true,
		     XmNheight, globalFeatureMenuLineHeight,
		     XmNwidth, globalFeatureMenuWidth/options-10,
		     XmNset, (int)value,
		     XmNalignment, XmALIGNMENT_BEGINNING,
		     XmNhighlightColor, MMDParentBg_,
		     NULL);

		stackedWidget.push_back(globalFeatureMenuOption_w_[i][j]);

		XtAddCallback
		    (globalFeatureMenuOption_w_[i][j],
		     XmNvalueChangedCallback,
		     (XtCallbackProc)cbChangeGlobalFeatureMenuValue_,
		     (void *)optionData);
	    }

	    break;

	case MEXP_FEATURE_OPTION_SLIDER10:
	case MEXP_FEATURE_OPTION_SLIDER1:

	    switch (globalFeatures_[i].optionType) {

	    case MEXP_FEATURE_OPTION_SLIDER10:
		sliderTitle = XmStringCreateLtoR
		    ((String)MEXP_DEFAULT_SLIDER10_RANAGE_TITLE.c_str(),
		     XmSTRING_DEFAULT_CHARSET);
		maxValue = 10;
		decimalPoints = 0;
		conversion = 1.0;
		break;

	    case MEXP_FEATURE_OPTION_SLIDER1:
	    default:
		sliderTitle = XmStringCreateLtoR
		    ((String)MEXP_DEFAULT_SLIDER1_RANAGE_TITLE.c_str(),
		     XmSTRING_DEFAULT_CHARSET);
		maxValue = 100;
		decimalPoints = 2;
		conversion = 100.0;
		break;
	    }


	    // The label column.
	    label = XtVaCreateWidget
		(globalFeatures_[i].name.c_str(),
		 xmLabelGadgetClass, feature_box,
		 XmNalignment, XmALIGNMENT_BEGINNING,
		 XmNheight, 2*globalFeatureMenuLineHeight,
		 XmNwidth, globalFeatureMenuWidth/2,
		 XmNtopAttachment, XmATTACH_WIDGET,
		 XmNtopWidget, feature_box,
		 XmNleftAttachment, XmATTACH_WIDGET,
		 XmNleftWidget, feature_box,
		 NULL);

	    stackedWidget.push_back(label);

	    optionData = new MMDGlobalFeatureMenuCallbackData_t;
	    optionData->globalFeatureId = i;
	    optionData->optionValue = globalFeatures_[i].selectedOption;
	    optionData->mlabMissionDesignInstance = this;
	    optionData->optionType = globalFeatures_[i].optionType;

	    value = optionData->optionValue*conversion;

	    // Slider bar.
	    globalFeatureMenuOption_w_[i][0] = XtVaCreateWidget
		("",
		 xmScaleWidgetClass, feature_box,
		 XmNmaximum, maxValue,
		 XmNminimum, 1,
		 XmNvalue, (int)value,
		 XmNorientation, XmHORIZONTAL,
		 XmNprocessingDirection, XmMAX_ON_RIGHT,
		 XmNshowValue, true,
		 XmNdecimalPoints, decimalPoints,
		 XmNheight, 2*globalFeatureMenuLineHeight,
		 XmNwidth, 3*globalFeatureMenuWidth/7,
		 XmNtopAttachment, XmATTACH_WIDGET,
		 XmNtopWidget, feature_box,
		 XmNrightAttachment, XmATTACH_WIDGET,
		 XmNrightWidget, feature_box,
		 XmNbottomAttachment, XmATTACH_WIDGET,
		 XmNbottomWidget, feature_box,
		 XmNtitleString, sliderTitle,
		 XmNhighlightColor, MMDParentBg_,
		 NULL);

	    stackedWidget.push_back(globalFeatureMenuOption_w_[i][0]);

	    XtAddCallback
		(globalFeatureMenuOption_w_[i][0],
		 XmNvalueChangedCallback,
		 (XtCallbackProc)cbChangeGlobalFeatureMenuValue_,
		 (void *)optionData);

	    XmStringFree(sliderTitle);

	    break;
	}
	line += 2;
    }

    numStackedWidget = stackedWidget.size();

    for (i = 0; i < numStackedWidget; i++)
    {
	XtManageChild(stackedWidget.back());
	stackedWidget.pop_back();
    }

    XtPopup (globalFeatureMenu_w_, XtGrabNone);
}

//-----------------------------------------------------------------------
// This function changes the global preference menu value.
//-----------------------------------------------------------------------
void MlabMissionDesign::changeGlobalFeatureMenuValue_(XtPointer client_data)
{
    int i, j, id;
    bool selected;
    int sliderBarValue;
    float optionValue;
    char optionValueStr[256];

    MMDGlobalFeatureMenuCallbackData_t *data = (MMDGlobalFeatureMenuCallbackData_t *)client_data;

    i = data->globalFeatureId;

    if (globalFeatures_[i].id != i)
    {
	// Global features was not stored in order of their IDs.
	// Something went wrong. Warn it.
	fprintf(stderr, "Warning (mlab): MlabMissionDesign::changeGlobalFeatureMenuValue_().\n");
	fprintf(stderr, "The order of global features corrupted.\n");
	return;
    }

    switch (data->optionType) {
	
    case MEXP_FEATURE_OPTION_SLIDER10:
    case MEXP_FEATURE_OPTION_SLIDER1:

	XtVaGetValues(
	    globalFeatureMenuOption_w_[i][0],
	    XmNvalue, &sliderBarValue,
	    NULL);

	switch (data->optionType) {

	case MEXP_FEATURE_OPTION_SLIDER10:
	    optionValue = (float)sliderBarValue;
	    sprintf(optionValueStr, "%d", sliderBarValue);
	    break;

	case MEXP_FEATURE_OPTION_SLIDER1:
	default:
	    optionValue = (float)sliderBarValue/100.0;
	    sprintf(optionValueStr, "%.2f", optionValue);
	    break;
	}

	globalFeatures_[i].selectedOption = optionValue;
	globalFeatures_[i].options[0].value = optionValueStr;

	gEventLogging->log("MMD GlobalFeatureMenu %s %s toggled",
			   globalFeatures_[i].name.c_str(),
			   globalFeatures_[i].options[0].value.c_str());

	break;

    case MEXP_FEATURE_OPTION_TOGGLE:

	globalFeatures_[i].selectedOption = data->optionValue;

	for (j = 0; j < ((int)(globalFeatures_[i].options.size())); j++)
	{
	    if ((int)(globalFeatures_[i].selectedOption) == j)
	    {
		selected = true;

		gEventLogging->log("MMD GlobalFeatureMenu %s %s toggled",
				   globalFeatures_[i].name.c_str(),
				   globalFeatures_[i].options[j].value.c_str());
	    }
	    else
	    {
		selected = false;
	    }

	    XtVaSetValues(
		globalFeatureMenuOption_w_[i][j],
		XmNset, selected,
		NULL);
	}

	break;
    }
}

//-----------------------------------------------------------------------
// This routine creates a widget that contains list of features for the
// specified task. The user are allowed to change the values of the
// features through this widget.
//-----------------------------------------------------------------------
void MlabMissionDesign::createRightBtnMenu_(MExpMissionTask_t task)
{
    XmString sliderTitle;
    Widget main_frame, main, label_frame, label_box, label;
    Widget localFeature_frame, localFeature_box, button_frame;
    Widget apply_frame, apply, cancel_frame, cancel;
    Widget action_frame, action;
    vector<Widget> stackedWidget;
    MMDRightBtnMenuCallbackData_t *optionData;
    char preference_title[1024];
    float value, conversion;
    int i, j, *j_ptr, numOptions, line = 0, s = 0;
    int screenNumber;
    int posX, posY, displayWidth, maxPos, displayHeight, offset;
    int rightBtnMenuHeight, rightBtnMenuWidth, rightBtnMenuLineHeight;
    int numStackedWidget, numLocalFeatures;
    int maxValue, decimalPoints;

    // Count the number of local features.
    numLocalFeatures = task.localFeatures.size();
    
    // Setting the height of the window.
    rightBtnMenuLineHeight = DEFAULT_RIGHTBTN_LINEHEIGHT_;
    rightBtnMenuHeight = 2*(2+numLocalFeatures)*(rightBtnMenuLineHeight+2);
    rightBtnMenuWidth = DEFAULT_RIGHTBTN_WIDTH_;
    offset = DEFAULT_RIGHTBTN_EDGEOFFSET_;

    // Make sure the window does go outside the desktop.
    screenNumber = DefaultScreen(MMDDisplay_);
    displayWidth = DisplayWidth(MMDDisplay_, screenNumber);
    displayHeight = DisplayHeight(MMDDisplay_, screenNumber);
    maxPos = displayWidth - rightBtnMenuWidth - offset;
    posX = (lastMouseXRoot_ < maxPos)? lastMouseXRoot_ : maxPos;
    maxPos = displayHeight - rightBtnMenuHeight - offset;
    posY = (lastMouseYRoot_ < maxPos)? lastMouseYRoot_ : maxPos;
    
    // Popup shell.
    rightBtnMenu_w_ = XtVaCreatePopupShell
	("",
	 xmDialogShellWidgetClass, MMDParent_, 
	 XmNdeleteResponse, XmUNMAP,
	 XmNallowShellResize, false,
	 XmNx, posX,
	 XmNy, posY,
	 XmNheight, rightBtnMenuHeight,
	 XmNwidth, rightBtnMenuWidth,
	 NULL);

    // The biggest frame.
    main_frame = XtVaCreateManagedWidget
	("main_frame",
	 xmFrameWidgetClass, rightBtnMenu_w_,
	 XmNalignment, XmALIGNMENT_BEGINNING,
	 XmNheight, rightBtnMenuHeight,
	 XmNwidth, rightBtnMenuWidth,
	 XmNtopAttachment, XmATTACH_WIDGET,
	 XmNtopWidget, rightBtnMenu_w_,
	 XmNleftAttachment, XmATTACH_WIDGET,
	 XmNleftWidget, rightBtnMenu_w_,
	 XmNrightAttachment, XmATTACH_WIDGET,
	 XmNrightWidget, rightBtnMenu_w_,
	 XmNbottomAttachment, XmATTACH_WIDGET,
	 XmNbottomWidget, rightBtnMenu_w_,
	 NULL);

    stackedWidget.push_back(main_frame);

    main = XtVaCreateWidget
	("main",
	 xmRowColumnWidgetClass, main_frame,
	 XmNpacking, XmPACK_TIGHT,
	 XmNorientation, XmVERTICAL,
	 XmNnumColumns, 1,
	 XmNisAligned, True,
	 XmNentryAlignment, XmALIGNMENT_CENTER,
	 XmNheight, rightBtnMenuHeight,
	 XmNwidth, rightBtnMenuWidth,
	 NULL);

    stackedWidget.push_back(main);

    // The frame for the label.
    label_frame = XtVaCreateManagedWidget
	("label_frame",
	 xmFrameWidgetClass, main,
	 XmNalignment, XmALIGNMENT_BEGINNING,
	 XmNheight, 2*rightBtnMenuLineHeight-2,
	 XmNwidth, rightBtnMenuWidth,
	 NULL);

    stackedWidget.push_back(label_frame);

    // The widget for the main label.
    label_box = XtVaCreateManagedWidget
	("label_box",
	 xmRowColumnWidgetClass, label_frame,
	 XmNpacking, XmPACK_COLUMN,
	 XmNorientation, XmVERTICAL,
	 XmNalignment, XmALIGNMENT_BEGINNING,
	 XmNnumColumns, 1,
	 XmNheight, 2*rightBtnMenuLineHeight,
	 XmNwidth, rightBtnMenuWidth,
	 NULL);

    stackedWidget.push_back(label_box);

    // The main label.
    sprintf(preference_title, "Task Preference");
    label = XtVaCreateManagedWidget
	(preference_title,
	 xmLabelGadgetClass, label_box,
	 XmNalignment, XmALIGNMENT_BEGINNING,
	 XmNheight, rightBtnMenuLineHeight,
	 XmNwidth, rightBtnMenuWidth,
	 XmNtopAttachment, XmATTACH_WIDGET,
	 XmNtopWidget, label_box,
	 XmNleftAttachment, XmATTACH_WIDGET,
	 XmNleftWidget, label_box,
	 XmNrightAttachment, XmATTACH_WIDGET,
	 XmNrightWidget, label_box,
	 NULL);

    stackedWidget.push_back(label);

    line++;

    sprintf(preference_title, "Task #%d  [%s]", task.number, task.name.c_str());
    label = XtVaCreateManagedWidget
	(preference_title,
	 xmLabelGadgetClass, label_box,
	 XmNalignment, XmALIGNMENT_BEGINNING,
	 XmNheight, rightBtnMenuLineHeight,
	 XmNwidth, rightBtnMenuWidth,
	 XmNtopAttachment, XmATTACH_WIDGET,
	 XmNtopWidget, label,
	 XmNleftAttachment, XmATTACH_WIDGET,
	 XmNleftWidget, label_box,
	 XmNrightAttachment, XmATTACH_WIDGET,
	 XmNrightWidget, label_box,
	 NULL);

    stackedWidget.push_back(label);

    line++;

    // It is a 2D array. Create the 1D part (feature) first.
    rightBtnMenuOption_w_ = new Widget*[numLocalFeatures];

    //  Create the feature raws sequentially.
    for (i = 0; i < numLocalFeatures; i++)
    {
	numOptions = task.localFeatures[i].options.size();

	// The frame for the feature.
	localFeature_frame = XtVaCreateManagedWidget
	    ("localFeature_frame",
	     xmFrameWidgetClass, main,
	     XmNalignment, XmALIGNMENT_BEGINNING,
	     XmNheight, 2*rightBtnMenuLineHeight,
	     XmNwidth, rightBtnMenuWidth,
	     NULL);

	stackedWidget.push_back(localFeature_frame);

	// Two-colum widget for the feature.
	localFeature_box = XtVaCreateManagedWidget
	    ("rightBtnMenu-localFeature_box",
	     xmRowColumnWidgetClass, localFeature_frame,
	     XmNpacking, XmPACK_COLUMN,
	     XmNorientation, XmHORIZONTAL,
	     XmNalignment, XmALIGNMENT_CENTER,
	     XmNnumColumns, numOptions,
	     XmNheight, 2*rightBtnMenuLineHeight,
	     XmNwidth, rightBtnMenuWidth,
	     XmNtopAttachment, XmATTACH_FORM,
	     XmNrightAttachment, XmATTACH_FORM,
	     XmNleftAttachment, XmATTACH_FORM,
	     XmNbottomAttachment, XmATTACH_FORM,
	     NULL);

	stackedWidget.push_back(localFeature_box);

	// Create the second dimention (option) of this array.
	rightBtnMenuOption_w_[i] = new Widget[numOptions];

	switch (task.localFeatures[i].optionType) {

	case MEXP_FEATURE_OPTION_TOGGLE:

	    // The label column.
	    label = XtVaCreateManagedWidget
		(task.localFeatures[i].name.c_str(),
		 xmLabelGadgetClass, localFeature_box,
		 XmNalignment, XmALIGNMENT_BEGINNING,
		 XmNheight, rightBtnMenuLineHeight,
		 XmNwidth, rightBtnMenuWidth/numOptions,
		 XmNtopAttachment, XmATTACH_WIDGET,
		 XmNtopWidget, localFeature_box,
		 XmNleftAttachment, XmATTACH_WIDGET,
		 XmNleftWidget, localFeature_box,
		 NULL);

	    stackedWidget.push_back(label);

	    // Fill the blanks in the label row.
	    for (j = 0; j < (numOptions-1); j++)
	    {
		label = XtVaCreateManagedWidget
		    (" ",
		     xmLabelGadgetClass, localFeature_box,
		     XmNalignment, XmALIGNMENT_BEGINNING,
		     XmNheight, rightBtnMenuLineHeight,
		     XmNwidth, rightBtnMenuWidth/numOptions,
		     XmNtopAttachment, XmATTACH_WIDGET,
		     XmNtopWidget, localFeature_box,
		     NULL);

		stackedWidget.push_back(label);
	    }
	
	    for (j = 0; j < numOptions; j++)
	    {
		optionData = new MMDRightBtnMenuCallbackData_t;
		optionData->featureId = i;
		optionData->optionValue = (float)j;
		optionData->mlabMissionDesignInstance = this;
		optionData->optionType = task.localFeatures[i].optionType;

		if ((int)(task.localFeatures[i].selectedOption) == j)
		{
		    value = 1.0;
		}
		else
		{
		    value = 0.0;
		}

		button_frame = XtVaCreateManagedWidget
		    ("button_frame",
		     xmFormWidgetClass, localFeature_box,
		     XmNshadowThickness, 0,
		     XmNalignment, XmALIGNMENT_BEGINNING,
		     XmNheight, rightBtnMenuLineHeight,
		     XmNwidth, rightBtnMenuWidth/numOptions-10,
		     NULL);

		stackedWidget.push_back(button_frame);

		// Toggle features.
		rightBtnMenuOption_w_[i][j] = XtVaCreateManagedWidget
		    (task.localFeatures[i].options[j].value.c_str(),
		     xmToggleButtonGadgetClass, button_frame,
		     XmNradioBehavior, true,
		     XmNradioAlwaysOne, true,
		     XmNheight, rightBtnMenuLineHeight,
		     XmNwidth, rightBtnMenuWidth/numOptions-10,
		     XmNset, (int)value,
		     XmNalignment, XmALIGNMENT_BEGINNING,
		     XmNhighlightColor, MMDParentBg_,
		     NULL);

		stackedWidget.push_back(rightBtnMenuOption_w_[i][j]);

		XtAddCallback
		    (rightBtnMenuOption_w_[i][j],
		     XmNvalueChangedCallback,
		     (XtCallbackProc)cbChangeRightBtnMenuValue_,
		     (void *)optionData);
	    }

	    break;

	case MEXP_FEATURE_OPTION_SLIDER10:
	case MEXP_FEATURE_OPTION_SLIDER1:

	    switch (task.localFeatures[i].optionType) {

	    case MEXP_FEATURE_OPTION_SLIDER10:
		sliderTitle = XmStringCreateLtoR(
		    (String)MEXP_DEFAULT_SLIDER10_RANAGE_TITLE.c_str(),
		    XmSTRING_DEFAULT_CHARSET);
		maxValue = 10;
		decimalPoints = 0;
		conversion = 1.0;
		break;

	    case MEXP_FEATURE_OPTION_SLIDER1:
	    default:
		sliderTitle = XmStringCreateLtoR(
		    (String)MEXP_DEFAULT_SLIDER1_RANAGE_TITLE.c_str(),
		    XmSTRING_DEFAULT_CHARSET);
		maxValue = 100;
		decimalPoints = 2;
		conversion = 100.0;
		break;
	    }

	    // The label column.
	    label = XtVaCreateManagedWidget
		 (task.localFeatures[i].name.c_str(),
		 xmLabelGadgetClass, localFeature_box,
		 XmNalignment, XmALIGNMENT_BEGINNING,
		 XmNheight, 2*rightBtnMenuLineHeight,
		 XmNwidth, rightBtnMenuWidth/2,
		 XmNtopAttachment, XmATTACH_WIDGET,
		 XmNtopWidget, localFeature_box,
		 XmNleftAttachment, XmATTACH_WIDGET,
		 XmNleftWidget, localFeature_box,
		 NULL);

	    stackedWidget.push_back(label);

	    optionData = new MMDRightBtnMenuCallbackData_t;
	    optionData->featureId = i;
	    optionData->optionValue = task.localFeatures[i].selectedOption;
	    optionData->mlabMissionDesignInstance = this;
	    optionData->optionType = task.localFeatures[i].optionType;

	    value = optionData->optionValue*conversion;

	    rightBtnMenuOption_w_[i][0] = XtVaCreateManagedWidget
		("",
		 xmScaleWidgetClass, localFeature_box,
		 XmNmaximum, maxValue,
		 XmNminimum, 1,
		 XmNvalue, (int)value,
		 XmNorientation, XmHORIZONTAL,
		 XmNprocessingDirection, XmMAX_ON_RIGHT,
		 XmNshowValue, true,
		 XmNdecimalPoints, decimalPoints,
		 XmNheight, 2*rightBtnMenuWidth,
		 XmNwidth, 3*rightBtnMenuWidth/7,
		 XmNtopAttachment, XmATTACH_WIDGET,
		 XmNtopWidget, label,
		 XmNrightAttachment, XmATTACH_FORM,
		 XmNtitleString, sliderTitle,
		 XmNhighlightColor, MMDParentBg_,
		 NULL);

	    stackedWidget.push_back(rightBtnMenuOption_w_[i][0]);

	    XtAddCallback
		(rightBtnMenuOption_w_[i][0],
		 XmNvalueChangedCallback,
		 (XtCallbackProc)cbChangeRightBtnMenuValue_,
		 (void *)optionData);

	    XmStringFree(sliderTitle);

	    break;
	}
	line+=2;

    }

    // The frame for the action field.
    action_frame = XtVaCreateManagedWidget
	("label_frame",
	 xmFormWidgetClass, main,
	 XmNalignment, XmALIGNMENT_BEGINNING,
	 XmNheight, 2*rightBtnMenuLineHeight-5,
	 XmNwidth, rightBtnMenuWidth-5,
	 XmNbottomAttachment, XmATTACH_WIDGET,
	 NULL);

    stackedWidget.push_back(action_frame);

    action = XtVaCreateWidget
	("action",
	 xmRowColumnWidgetClass, action_frame,
	 XmNpacking, XmPACK_COLUMN,
	 XmNorientation, XmVERTICAL,
	 XmNnumColumns, 2,
	 XmNisAligned, true,
	 XmNentryAlignment, XmALIGNMENT_CENTER,
	 XmNheight, 2*rightBtnMenuLineHeight-5,
	 XmNwidth, rightBtnMenuWidth-5,
	 XmNrightAttachment, XmATTACH_FORM,
	 XmNleftAttachment, XmATTACH_FORM,
	 XmNtopAttachment, XmATTACH_FORM,
	 XmNbottomAttachment, XmATTACH_FORM,
	 NULL);

    stackedWidget.push_back(action);

    // The frame for the "Apply" buttons.
    apply_frame = XtVaCreateManagedWidget
	("apply_frame",
	 xmFormWidgetClass, action,
	 XmNalignment, XmALIGNMENT_CENTER,
	 XmNheight, 2*rightBtnMenuLineHeight-10,
	 XmNwidth, rightBtnMenuWidth/2-10,
	 XmNtopAttachment, XmATTACH_WIDGET,
	 XmNtopWidget, action,
	 XmNleftAttachment, XmATTACH_WIDGET,
	 XmNleftWidget, action,
	 XmNrightAttachment, XmATTACH_FORM,
	 XmNbottomAttachment, XmATTACH_WIDGET,
	 XmNbottomWidget, action,
	 NULL);

    stackedWidget.push_back(apply_frame);

    // The "Apply" button.
    apply = XtVaCreateManagedWidget
	("Apply",
	 xmPushButtonGadgetClass, apply_frame,
	 XmNalignment, XmALIGNMENT_CENTER,
	 XmNheight, 2*rightBtnMenuLineHeight-10,
	 XmNwidth, rightBtnMenuWidth/2-10,
	 XmNtopAttachment, XmATTACH_FORM,
	 XmNrightAttachment, XmATTACH_FORM,
	 XmNleftAttachment, XmATTACH_FORM,
	 XmNbottomAttachment, XmATTACH_FORM,
	 NULL);

    stackedWidget.push_back(apply);

    XtAddCallback
	(apply,
	 XmNactivateCallback,
	 (XtCallbackProc)cbApplyRightBtnMenu_,
	 (void *)this);

    // The frame for the "Cancel" buttons.
    cancel_frame = XtVaCreateManagedWidget
	("cancel_frame",
	 xmFormWidgetClass, action,
	 XmNalignment, XmALIGNMENT_BEGINNING,
	 XmNx, rightBtnMenuWidth/2,
	 XmNheight, 2*rightBtnMenuLineHeight-10,
	 XmNwidth, rightBtnMenuWidth/2-10,
	 XmNtopAttachment, XmATTACH_WIDGET,
	 XmNtopWidget, action,
	 XmNrightAttachment, XmATTACH_WIDGET,
	 XmNrightWidget, action,
	 XmNbottomAttachment, XmATTACH_WIDGET,
	 XmNbottomWidget, action,
	 NULL);

    stackedWidget.push_back(cancel_frame);

    // The "Cancel" button
    cancel = XtVaCreateManagedWidget
	("Cancel",
	 xmPushButtonGadgetClass, cancel_frame,
	 XmNalignment, XmALIGNMENT_CENTER,
	 XmNheight, 2*rightBtnMenuLineHeight-10,
	 XmNwidth, rightBtnMenuWidth/2-10,
	 XmNtopAttachment, XmATTACH_FORM,
	 XmNrightAttachment, XmATTACH_FORM,
	 XmNleftAttachment, XmATTACH_FORM,
	 XmNbottomAttachment, XmATTACH_FORM,
	 NULL);

    stackedWidget.push_back(cancel);

    XtAddCallback
	(cancel,
	 XmNactivateCallback,
	 (XtCallbackProc)cbCancelRightBtnMenu_,
	 (void *)this);

    rightBtnMenuUpTask_ = task;

    numStackedWidget = stackedWidget.size();

    for (i = 0; i < numStackedWidget; i++)
    {
	XtManageChild(stackedWidget.back());
	stackedWidget.pop_back();
	XtManageChild(stackedWidget[i]);
    }

    XtPopup (rightBtnMenu_w_, XtGrabNone);

    rightBtnMenuIsUp_ = true;
}

//-----------------------------------------------------------------------
// This function pops up the right-button menu.
//-----------------------------------------------------------------------
void MlabMissionDesign::popUpRightBtnMenu_(double taskPosX, double taskPosY) 
{
    MExpMissionTask_t task;
    
    if (!isInsideTask_(taskPosX, taskPosY, drawingUnitLength_, &task))
    {
	return;
    }

    createRightBtnMenu_(task);

    gEventLogging->start("MMD RightBtnMenu Task #%d [%s]", task.number, task.name.c_str());
}

//-----------------------------------------------------------------------
// This function chages the values of the right-button menu.
//-----------------------------------------------------------------------
void MlabMissionDesign::changeRightBtnMenuValue_(XtPointer client_data)
{
    int i, j;
    bool selected;
    int sliderBarValue;
    float optionValue;
    char optionValueStr[256];

    MMDRightBtnMenuCallbackData_t *data = (MMDRightBtnMenuCallbackData_t *)client_data;
    i = data->featureId;

    switch (data->optionType) {

    case MEXP_FEATURE_OPTION_TOGGLE:
	
	rightBtnMenuUpTask_.localFeatures[i].selectedOption = data->optionValue;

	for (j = 0; j < ((int)(rightBtnMenuUpTask_.localFeatures[i].options.size())); j++)
	{
	    if ((int)(rightBtnMenuUpTask_.localFeatures[i].selectedOption) == j)
	    {
		selected = true;

		gEventLogging->log
		    ("MMD RightBtnMenu Task #%d [%s] %s %s toggled",
		     rightBtnMenuUpTask_.number,
		     rightBtnMenuUpTask_.name.c_str(),
		     rightBtnMenuUpTask_.localFeatures[i].name.c_str(),
		     rightBtnMenuUpTask_.localFeatures[i].options[j].value.c_str());
	    }
	    else
	    {
		selected = false;
	    }

	    XtVaSetValues(
		rightBtnMenuOption_w_[i][j],
		XmNset, selected,
		NULL);
	}

	break;

    case MEXP_FEATURE_OPTION_SLIDER10:
    case MEXP_FEATURE_OPTION_SLIDER1:

	XtVaGetValues(
	    rightBtnMenuOption_w_[i][0],
	    XmNvalue, &sliderBarValue,
	    NULL);

	switch (data->optionType) {

	case MEXP_FEATURE_OPTION_SLIDER10:
	    optionValue = (float)sliderBarValue;
	    sprintf(optionValueStr, "%d", sliderBarValue);
	    break;

	case MEXP_FEATURE_OPTION_SLIDER1:
	default:
	    optionValue = (float)sliderBarValue/100.0;
	    sprintf(optionValueStr, "%.2f", optionValue);
	    break;
	}

	rightBtnMenuUpTask_.localFeatures[i].selectedOption = optionValue;
	rightBtnMenuUpTask_.localFeatures[i].options[0].value = optionValueStr;

	gEventLogging->log
	    ("MMD RightBtnMenu Task #%d [%s] %s adjusted to %s",
	     rightBtnMenuUpTask_.number,
	     rightBtnMenuUpTask_.name.c_str(),
	     rightBtnMenuUpTask_.localFeatures[i].name.c_str(),
	     rightBtnMenuUpTask_.localFeatures[i].options[0].value.c_str());

	break;
    }
}

//-----------------------------------------------------------------------
// This function stores the "Up" task in the task array.
//-----------------------------------------------------------------------
void MlabMissionDesign::applyRightBtnMenu_(void) 
{
    int i;

    i = rightBtnMenuUpTask_.number;
    specifiedTaskList_[i] = rightBtnMenuUpTask_;

    gEventLogging->log("MMD RightBtnMenu Apply Change");

    closeRightBtnMenu_();
}

//-----------------------------------------------------------------------
// This function stores the "Up" task in the task array.
//-----------------------------------------------------------------------
void MlabMissionDesign::cancelRightBtnMenu_(void) 
{
    gEventLogging->log("MMD RightBtnMenu Cancel Change");

    closeRightBtnMenu_();
}

//-----------------------------------------------------------------------
// This function closes the right-button menu window.
//-----------------------------------------------------------------------
void MlabMissionDesign::closeRightBtnMenu_(void) 
{
    delete [] rightBtnMenuOption_w_;
    XtDestroyWidget(rightBtnMenu_w_);
    rightBtnMenuIsUp_ = false;
    gEventLogging->end("MMD RightBtnMenu");
}

//-----------------------------------------------------------------------
// This function checks whether the particular point is inside the already-
// placed task.
//-----------------------------------------------------------------------
bool MlabMissionDesign::isInsideTask_(double x, double y, double unitLength, MExpMissionTask_t *task)
{
    int i;
    double dist, closestDist;
    MExpMissionTask_t closestTask;

    if (numTasks_ == 0)
    {
	return false;
    }

    // Find the closest waypoint
    closestTask = specifiedTaskList_[0];
    closestDist = pow((specifiedTaskList_[0].position.x - x),2) + pow((specifiedTaskList_[0].position.y - y),2);
    
    for (i = 1; i < numTasks_; i++)
    {
	dist = pow((specifiedTaskList_[i].position.x - x),2) + pow((specifiedTaskList_[i].position.y - y),2);

	if (dist < closestDist)
	{
	    closestTask = specifiedTaskList_[i];
	    closestDist = dist;
	}
    }

    // Too far from any waypoint
    closestDist = (sqrt(closestDist)) / unitLength;

    if (closestDist > taskPointRadius_)
    {
	return false;
    }

    *task = closestTask;
    return true;
}

//-----------------------------------------------------------------------
// This routine function is called when a mouse is clicked on the
// overlay.
//-----------------------------------------------------------------------
void MlabMissionDesign::mouseClick(double x, double y, XEvent *event, double unitLength) 
{
    lastMouseXRoot_ = event->xbutton.x_root;
    lastMouseYRoot_ = event->xbutton.y_root;
    drawingUnitLength_ = unitLength;

    if (rightBtnMenuIsUp_) closeRightBtnMenu_();

    switch (event->xbutton.button) {

    case LEFT_BTN:
	placeTask_(x, y, MMDMode_);
	break;

    case MIDDLE_BTN:
	removeTask_(x, y);
	break;

    case RIGHT_BTN:
	popUpRightBtnMenu_(x, y);
	break;
    }
}

//-----------------------------------------------------------------------
// This function sets the task mode.
//-----------------------------------------------------------------------
void  MlabMissionDesign::setMode_(int mode)
{
    //MMDMode_ = (MExpTaskType_t)mode;
    MMDMode_ = mode;
    updateToolboxBtn_();

    gEventLogging->log("MMD Mode set to %d", mode);
}

//-----------------------------------------------------------------------
// This function stores the overlay name in the memory.
//-----------------------------------------------------------------------
void MlabMissionDesign::saveOverlayName(char *ovlFilename)
{
    char simpleFilename[1024];

    if (filename_has_directory(ovlFilename))
    {
	remove_directory(ovlFilename, simpleFilename);
	overlayFilename_ = simpleFilename;
    }
    else
    {
	overlayFilename_ = ovlFilename;
    }
}

//-----------------------------------------------------------------------
// This function asks the user if it is OK with just quit this session
// since no task was placed.
//-----------------------------------------------------------------------
bool MlabMissionDesign::confirmEmptyMission_(void)
{
    Widget w;

    w = createGeneralYesNoWindow_("This mission is empty.\n\nCancel this session and return to CfgEdit?");
    popupGeneralYesNoWindow_(w);

    // Wait for the answer.
    while(!generalYesNoWindowDone_)
    {
	XtAppProcessEvent(MMDAppContext_, XtIMAll);
    }

    return generalYesNoWindowResult_;
}

//-----------------------------------------------------------------------
// This function notifies the users that the RC file could not be located.
//-----------------------------------------------------------------------
void MlabMissionDesign::confirmNoRC_(void)
{
    Widget w;
    XmString ok_str;

    ok_str = XmStringCreateSimple("  OK  ");

    w = createGeneralYesNoWindow_("Error: RC file could not be loaded. Returning to CfgEdit.");
    XtVaSetValues(w, XmNokLabelString, ok_str, NULL);
    XtUnmanageChild(XmMessageBoxGetChild(w, XmDIALOG_HELP_BUTTON));

    popupGeneralYesNoWindow_(w);

    // Wait for the answer.
    while(!generalYesNoWindowDone_)
    {
	XtAppProcessEvent(MMDAppContext_, XtIMAll);
    }

    return;
}

//-----------------------------------------------------------------------
// This function creates the window that simply asks a "Yes" / "No"
// question. The question statement is the input.
//-----------------------------------------------------------------------
Widget MlabMissionDesign::createGeneralYesNoWindow_(char *question)
{
    XmString yes_str, no_str, msg_str, title_str;
    Arg wargs[16];
    int n=0;
    GeneralYesNoWindowCallbackData_t *callBackData;
    Widget generalYesNoWindow;

    msg_str = XmStringCreateLtoR(question, XmSTRING_DEFAULT_CHARSET);
    yes_str = XmStringCreateSimple("  Yes  ");
    no_str = XmStringCreateSimple("  No  ");
    title_str = XmStringCreateSimple(" ");

    XtSetArg(wargs[n], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL); n++;
    XtSetArg(wargs[n], XmNmessageString, msg_str); n++; 
    XtSetArg(wargs[n], XmNokLabelString, yes_str); n++;
    XtSetArg(wargs[n], XmNhelpLabelString, no_str); n++;
    XtSetArg(wargs[n], XmNdialogTitle, title_str); n++;
    generalYesNoWindow = XmCreateMessageDialog(MMDParent_, "", wargs, n);

    // "Yes" button
    callBackData = new GeneralYesNoWindowCallbackData_t;
    callBackData->mlabMissionDesignInstance = this;
    callBackData->result = true;
    XtAddCallback
	(generalYesNoWindow,
	 XmNokCallback,
	 (XtCallbackProc)cbGeneralYesNoWindow_,
	 (void *)callBackData);

    // "No" button
    callBackData = new GeneralYesNoWindowCallbackData_t;
    callBackData->mlabMissionDesignInstance = this;
    callBackData->result = false;
    XtAddCallback
	(generalYesNoWindow,
	 XmNhelpCallback,
	 (XtCallbackProc)cbGeneralYesNoWindow_,
	 (void *)callBackData);

    XtUnmanageChild(XmMessageBoxGetChild(generalYesNoWindow, XmDIALOG_CANCEL_BUTTON));

    XmStringFree(msg_str);
    XmStringFree(yes_str);
    XmStringFree(no_str);

    return generalYesNoWindow;
}

//-----------------------------------------------------------------------
// This function popups the window above.
//-----------------------------------------------------------------------
void MlabMissionDesign::popupGeneralYesNoWindow_(Widget w)
{
    generalYesNoWindowResult_ = false;
    generalYesNoWindowDone_ = false;
    XtManageChild(w);
}

//-----------------------------------------------------------------------
// This function sets the "Yes" / "No" variable according to the answer
// given by the user.
//-----------------------------------------------------------------------
void MlabMissionDesign::sendGeneralYesNoWindowResult_(bool result)
{
    generalYesNoWindowResult_ = result;
    generalYesNoWindowDone_ = true;
}

//-----------------------------------------------------------------------
// This function loads the features from the RC file.
//-----------------------------------------------------------------------
bool MlabMissionDesign::loadFeaturesFromRCFile_(void)
{
    MExpFeature_t feature;
    MExpFeatureOption_t option;
    MExpMissionTask_t task;
    const MExpFeature_t EMPTY_FEATURE;
    const MExpFeatureOption_t EMPTY_FEATURE_OPTION;
    const MExpMissionTask_t EMPTY_TASK;
    FILE *rcFile = NULL;
    string rcFilename;
    string bufString1, bufString2;
    char buf1[1024], buf2[1024], buf3[1024];

    // First, find the RC file.
    rcFilename = find_file_in_envpath(MEXP_RC_FILENAME);

    if (rcFilename == EMPTY_STRING_)
    {
	fprintf(stderr, "Error(mlab): RC file (%s) could not be loaded.\n", rcFilename.c_str());
	return false;
    }

    // Load the RC file.
    rcFile = fopen(rcFilename.c_str(), "r");

    while (fscanf(rcFile, "%s", buf1) != EOF)
    {
	bufString1 = buf1;

	if (bufString1[0] == '#')
	{
	    // It's a comment.
	    while ((fgetc(rcFile) != '\n') && (!feof(rcFile)))
	    {
		// Keep reading until next line.
	    }
	}
	else if (bufString1 == MEXP_RC_STRING_FEATURE_KEY)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    bufString2 = buf2;

	    if (bufString2 == MEXP_RC_STRING_FEATURE_BEGIN)
	    {
		feature = EMPTY_FEATURE;
		option = EMPTY_FEATURE_OPTION;
	    }
	    else if (bufString2 == MEXP_RC_STRING_FEATURE_END)
	    {
		switch (feature.type) {
		    
		case MEXP_FEATURE_GLOBAL:
		    feature.id = globalFeatures_.size();
		    globalFeatures_.push_back(feature);
		    break;

		case MEXP_FEATURE_LOCAL:
		    feature.id = defaultLocalFeatures_.size();
		    defaultLocalFeatures_.push_back(feature);
		    break;
		}
	    }
	}
	else if (bufString1 == MEXP_RC_STRING_FEATURE_TYPE)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    feature.type = atoi(buf2);
	}
	else if (bufString1 == MEXP_RC_STRING_FEATURE_NAME)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    feature.name = buf2;
	}
	else if (bufString1 == MEXP_RC_STRING_FEATURE_NON_INDEX)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    feature.nonIndex = atoi(buf2);
	}
	else if (bufString1 == MEXP_RC_STRING_FEATURE_WEIGHT)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    feature.weight = atof(buf2);
	}
	else if (bufString1 == MEXP_RC_STRING_FEATURE_OPTION_TYPE)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    feature.optionType = atoi(buf2);
	}
	else if (bufString1 == MEXP_RC_STRING_FEATURE_SELECTED_OPTION)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    feature.selectedOption = atof(buf2);
	}
	else if (bufString1 == MEXP_RC_STRING_FEATURE_OPTION)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    option.value = buf2;
	    feature.options.push_back(option);
	}
	else if (bufString1 == MEXP_RC_STRING_TASK_KEY)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    bufString2 = buf2;

	    if (bufString2 == MEXP_RC_STRING_TASK_BEGIN)
	    {
		task = EMPTY_TASK;
	    }
	    else if (bufString2 == MEXP_RC_STRING_TASK_END)
	    {
		task.id = defaultTaskList_.size();
		defaultTaskList_.push_back(task);
		numTaskTypes_ = defaultTaskList_.size();
	    }
	}
	else if (bufString1 == MEXP_RC_STRING_TASK_NAME)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    task.name = buf2;
	}
	else if (bufString1 == MEXP_RC_STRING_TASK_BUTTON_NAME)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    task.buttonName = buf2;
	}
	else if (bufString1 == MEXP_RC_STRING_TASK_SPECIFIC_SELECTED_OPTION)
	{
	    if (fscanf(rcFile, "%s %s", buf2, buf3) == EOF)
	    {
		break;
	    }
	    
	    feature.name = buf2;
	    feature.selectedOption = atof(buf3);
	    task.localFeatures.push_back(feature);
	}
	else if (bufString1 == MEXP_RC_STRING_DEFAULT_TASK)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    defaultTaskName_ = buf2;
	}
	else if (bufString1 == MEXP_RC_STRING_DEFAULT_TASK_WEIGHT)
	{
	    if (fscanf(rcFile, "%s", buf2) == EOF)
	    {
		break;
	    }
	    
	    defaultTaskWeight_ = atof(buf2);
	}
    }

    return true;
}

//-----------------------------------------------------------------------
// This function creates pixmaps to be used, and put them in a list.
//-----------------------------------------------------------------------
void MlabMissionDesign::createPixmapList_(Pixel fg, Pixel bg, Widget parent)
{
    MMDPixmap_t p;

    // Biohazard Pixmap
    p.name = "BiohazardBtn";
    p.pixmap = MMD_CREATE_PIXMAP(BiohazardBtn, gColorPixel.black, gColorPixel.orange, parent);
    p.insensitivePixmap = MMD_CREATE_PIXMAP(BiohazardBtn, gColorPixel.gray80, gColorPixel.orange, parent);
    pixmapList_.push_back(p);

    // EOD Pixmap
    p.name = "MineBtn";
    p.pixmap = MMD_CREATE_PIXMAP(MineBtn, gColorPixel.black, gColorPixel.orange, parent);
    p.insensitivePixmap = MMD_CREATE_PIXMAP(MineBtn, gColorPixel.gray80, gColorPixel.orange, parent);
    pixmapList_.push_back(p);

    // Waypoint Pixmap
    p.name = "WaypointsBtn";
    p.pixmap = MMD_CREATE_PIXMAP(WaypointsBtn, gColorPixel.black, gColorPixel.orange, parent);
    p.insensitivePixmap = MMD_CREATE_PIXMAP(WaypointsBtn, gColorPixel.gray80, gColorPixel.orange, parent);
    pixmapList_.push_back(p);

    // Other Pixmap
    p.name = "OtherBtn";
    p.pixmap = MMD_CREATE_PIXMAP(OtherBtn, gColorPixel.black, gColorPixel.orange, parent);
    p.insensitivePixmap = MMD_CREATE_PIXMAP(OtherBtn, gColorPixel.gray80, gColorPixel.orange, parent);
    pixmapList_.push_back(p);
}

//-----------------------------------------------------------------------
// This function returns pixmap (or insensitive pixmap) from the list.
//-----------------------------------------------------------------------
Pixmap MlabMissionDesign::getPixmapFromList_(string name, bool insensitive)
{
    Pixmap p;
    int i;

    for (i = 0; i < ((int)(pixmapList_.size())); i++)
    {
	if (pixmapList_[i].name == name)
	{
	    if (insensitive)
	    {
		p = pixmapList_[i].insensitivePixmap;
	    }
	    else
	    {
		p = pixmapList_[i].pixmap;
	    }
	    
	    break;
	}
    }

    return p;
}

//-----------------------------------------------------------------------
// This function assigns the default mode to be the ID for the default
// task.
//-----------------------------------------------------------------------
int MlabMissionDesign::getDefaultMode_(void)
{
    int i, mode;

    mode = 0;

    for (i = 0; i < ((int)(defaultTaskList_.size())); i++)
    {
	if (defaultTaskList_[i].name == defaultTaskName_)
	{
	    mode = defaultTaskList_[i].id;
	    break;
	}
    }

    return mode;
}

//-----------------------------------------------------------------------
// This callback function sets the current mode.
//-----------------------------------------------------------------------
void MlabMissionDesign::cbSetmode_(Widget w, XtPointer client_data, XtPointer callback_data)
{
    MlabMissionDesign *mlabMissionDesignInstance = NULL;
    int mode;

    mlabMissionDesignInstance = ((MMDToolBoxSetModeCallbackData_t *)client_data)->mlabMissionDesignInstance;
    mode = ((MMDToolBoxSetModeCallbackData_t *)client_data)->mode;

    mlabMissionDesignInstance->setMode_(mode);
}

//-----------------------------------------------------------------------
// This callback function changes global preference menu value.
//-----------------------------------------------------------------------
void MlabMissionDesign::cbChangeGlobalFeatureMenuValue_(Widget w, XtPointer client_data, XtPointer callback_data)
{
    MlabMissionDesign *mlabMissionDesignInstance = NULL;

    mlabMissionDesignInstance = ((MMDGlobalFeatureMenuCallbackData_t *)client_data)->mlabMissionDesignInstance;
    mlabMissionDesignInstance->changeGlobalFeatureMenuValue_(client_data);
}

//-----------------------------------------------------------------------
// This callback function gets called when "Apply" button in the Right-
// Button Menu is pressed.
//-----------------------------------------------------------------------
void MlabMissionDesign::cbApplyRightBtnMenu_(Widget w, XtPointer client_data, XtPointer callback_data)
{
    MlabMissionDesign *mlabMissionDesignInstance = NULL;

    mlabMissionDesignInstance = (MlabMissionDesign *)client_data;
    mlabMissionDesignInstance->applyRightBtnMenu_();
}

//-----------------------------------------------------------------------
// This callback function gets called when "Cancel" button in the Right-
// Button Menu is pressed.
//-----------------------------------------------------------------------
void MlabMissionDesign::cbCancelRightBtnMenu_(Widget w, XtPointer client_data, XtPointer callback_data)
{
    MlabMissionDesign *mlabMissionDesignInstance = NULL;

    mlabMissionDesignInstance = (MlabMissionDesign *)client_data;
    mlabMissionDesignInstance->cancelRightBtnMenu_();
}

//-----------------------------------------------------------------------
// This callback function gets called when a toggle button in the Right-
// Button Menu is pressed.
//-----------------------------------------------------------------------
void MlabMissionDesign::cbChangeRightBtnMenuValue_(Widget w, XtPointer client_data, XtPointer callback_data)
{
    MlabMissionDesign *mlabMissionDesignInstance = NULL;

    mlabMissionDesignInstance = ((MMDRightBtnMenuCallbackData_t *)client_data)->mlabMissionDesignInstance;
    mlabMissionDesignInstance->changeRightBtnMenuValue_(client_data);
}

//-----------------------------------------------------------------------
// This callback function gets called when either "Global Setting"
// button in the toolbar is pressed.
//-----------------------------------------------------------------------
void MlabMissionDesign::cbToolboxGlobalSettingBtn_(Widget w, XtPointer client_data, XtPointer callback_data)
{
    MlabMissionDesign *mlabMissionDesignInstance = NULL;

    mlabMissionDesignInstance = (MlabMissionDesign *)client_data;
    XtDestroyWidget(mlabMissionDesignInstance->globalFeatureMenu_w_);
    mlabMissionDesignInstance->createGlobalFeatureMenu_();
}

//-----------------------------------------------------------------------
// This callback function gets called when either "Finish" or "Cancel &
// Return" button in the toolbar is pressed.
//-----------------------------------------------------------------------
void MlabMissionDesign::cbExit_(Widget w, XtPointer client_data, XtPointer callback_data)
{
    MlabMissionDesign *mlabMissionDesignInstance = NULL;
    int exitMode;

    mlabMissionDesignInstance = ((MMDToolBoxSetModeCallbackData_t *)client_data)->mlabMissionDesignInstance;
    exitMode = ((MMDToolBoxSetModeCallbackData_t *)client_data)->mode;
    mlabMissionDesignInstance->returnConfigureMission_ = exitMode;

    // If no task was placed, and the user is trying to configure it, verify if
    // the user is OK with just quit.
    if ((exitMode == MMD_EXIT_CONFIGURE) && (mlabMissionDesignInstance->numTasks_ == 0))
    {
	if (mlabMissionDesignInstance->confirmEmptyMission_() == false)
	{
	    return;
	}
    }

    quit_mlab();
}

//-----------------------------------------------------------------------
// This callback function sets the "Yes" / "No" variable according to the
// answer given by the user.
//-----------------------------------------------------------------------
void MlabMissionDesign::cbGeneralYesNoWindow_(Widget w, XtPointer client_data, XtPointer callback_data)
{
    MlabMissionDesign *mlabMissionDesignInstance = NULL;
    bool result;

    mlabMissionDesignInstance = ((GeneralYesNoWindowCallbackData_t *)client_data)->mlabMissionDesignInstance;
    result = ((GeneralYesNoWindowCallbackData_t *)client_data)->result;

    mlabMissionDesignInstance->sendGeneralYesNoWindowResult_(result);

    XtDestroyWidget(w);
}

/**********************************************************************
 * $Log: mission_design.c,v $
 * Revision 1.4  2003/04/06 12:44:19  endo
 * gcc 3.1.1
 *
 * Revision 1.3  2003/04/06 09:03:48  endo
 * Updated for CBR Wizard Prototype II.
 *
 * Revision 1.2  2002/01/31 10:51:28  endo
 * Parameters chaged due to the change of max_vel and base_vel.
 *
 * Revision 1.1  2002/01/12 22:54:39  endo
 * Initial revision
 *
 **********************************************************************/
